using System;
using System.Collections.Generic;
using System.Text;
using System.Collections.ObjectModel;
using System.Globalization;
using System.IO;
using System.Xml.Serialization;
using System.Threading;

namespace PublicDomain
{
    /// <summary>
    /// Represents a Time Zone from the Olson tz database.
    /// 
    /// From http://www.twinsun.com/tz/tz-link.htm
    /// "The public-domain time zone database contains code
    /// and data that represent the history of local time
    /// for many representative locations around the globe."
    /// </summary>
    [Serializable]
    [XmlType(Namespace = GlobalConstants.PublicDomainNamespace)]
    [SoapType(Namespace = GlobalConstants.PublicDomainNamespace)]
    public class TzTimeZone : TimeZone
    {
        /// <summary>
        /// 
        /// </summary>
        public static bool TreatUnspecifiedKindAsLocal = true;

        /// <summary>
        /// 
        /// </summary>
        public static readonly TzTimeZone ZoneUTC;

        /// <summary>
        /// 
        /// </summary>
        public static readonly TzTimeZone ZoneUsEastern;

        /// <summary>
        /// 
        /// </summary>
        public static readonly TzTimeZone ZoneUsCentral;

        /// <summary>
        /// 
        /// </summary>
        public static readonly TzTimeZone ZoneUsMountain;

        /// <summary>
        /// 
        /// </summary>
        public static readonly TzTimeZone ZoneUsPacific;

        private static Dictionary<double, string> s_mainTimeZones = new Dictionary<double, string>();
        private static Dictionary<string, TzZoneInfo> s_zones = new Dictionary<string, TzZoneInfo>();
        private static ReadOnlyCollection<TzZoneInfo> s_zoneList;
        private static string[] s_allZoneNames;
        private static TzTimeZone s_currentTimeZone;
        private static ReaderWriterLock s_zonesLock = new ReaderWriterLock();
        private static object s_allZonesLock = new object();
        private static object s_zoneListLock = new object();

        private TzZoneInfo m_info;

        static TzTimeZone()
        {
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-11:00")] = TzConstants.TimezonePacificMidway;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-10:00")] = TzConstants.TimezonePacificHonolulu;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-9:30")] = TzConstants.TimezonePacificMarquesas;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-9:00")] = TzConstants.TimezoneAmericaAnchorage;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-8:00")] = TzConstants.TimezoneAmericaLosAngeles;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-7:00")] = TzConstants.TimezoneAmericaDenver;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-6:00")] = TzConstants.TimezoneAmericaChicago;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-5:00")] = TzConstants.TimezoneAmericaNewYork;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-4:00")] = TzConstants.TimezoneAmericaLaPaz;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-3:30")] = TzConstants.TimezoneAmericaStJohns;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-3:00")] = TzConstants.TimezoneAmericaArgentinaBuenosAires;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-2:00")] = TzConstants.TimezoneAmericaNoronha;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("-1:00")] = TzConstants.TimezoneAtlanticAzores;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("0:00")] = TzConstants.TimezoneUtc;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("1:00")] = TzConstants.TimezoneEuropeParis;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("2:00")] = TzConstants.TimezoneEuropeAthens;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("3:00")] = TzConstants.TimezoneEuropeMoscow;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("3:30")] = TzConstants.TimezoneAsiaTehran;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("4:00")] = TzConstants.TimezoneAsiaDubai;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("4:30")] = TzConstants.TimezoneAsiaKabul;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("5:00")] = TzConstants.TimezoneAsiaKarachi;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("5:30")] = TzConstants.TimezoneAsiaCalcutta;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("6:00")] = TzConstants.TimezoneAsiaOmsk;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("6:30")] = TzConstants.TimezoneIndianCocos;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("7:00")] = TzConstants.TimezoneAsiaJakarta;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("8:00")] = TzConstants.TimezoneAsiaShanghai;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("9:00")] = TzConstants.TimezoneAsiaTokyo;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("9:30")] = TzConstants.TimezoneAustraliaDarwin;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("10:00")] = TzConstants.TimezonePacificGuam;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("10:30")] = TzConstants.TimezoneAustraliaLordHowe;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("11:00")] = TzConstants.TimezonePacificGuadalcanal;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("11:30")] = TzConstants.TimezonePacificNorfolk;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("12:00")] = TzConstants.TimezonePacificFiji;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("13:00")] = TzConstants.TimezonePacificEnderbury;
            s_mainTimeZones[DateTimeUtlities.ConvertTimeSpanToDouble("14:00")] = TzConstants.TimezonePacificKiritimati;

            // Initialize "common" time zones
            ZoneUTC = TzTimeZone.GetTimeZone(TzConstants.TimezoneUtc);
            ZoneUsEastern = TzTimeZone.GetTimeZone(TzConstants.TimezoneUsEastern);
            ZoneUsCentral = TzTimeZone.GetTimeZone(TzConstants.TimezoneUsCentral);
            ZoneUsMountain = TzTimeZone.GetTimeZone(TzConstants.TimezoneUsMountain);
            ZoneUsPacific = TzTimeZone.GetTimeZone(TzConstants.TimezoneUsPacific);

            // Get the current time zone
            TimeZone cur = TimeZone.CurrentTimeZone;
            TimeSpan utcOffset = cur.GetUtcOffset(DateTime.MinValue);
            s_currentTimeZone = GetTimeZoneByOffset(utcOffset);
        }

        /// <summary>
        /// 
        /// </summary>
        public static ReadOnlyDictionary<string, TzZoneInfo> Zones
        {
            get
            {
                return new ReadOnlyDictionary<string,TzZoneInfo>(s_zones);
            }
        }

        /// <summary>
        /// 
        /// </summary>
        public static ReadOnlyCollection<TzZoneInfo> ZoneList
        {
            get
            {
                lock (s_zoneListLock)
                {
                    string[] allZonesNames = AllZoneNames;

                    List<TzZoneInfo> list = new List<TzZoneInfo>();
                    foreach (string zoneName in allZonesNames)
                    {
                        list.Add(s_zones[zoneName]);
                    }
                    s_zoneList = new ReadOnlyCollection<TzZoneInfo>(list);
                }
                return s_zoneList;
            }
        }

        /// <summary>
        /// Gets all zone names.
        /// </summary>
        /// <value>All zone names.</value>
        public static string[] AllZoneNames
        {
            get
            {
                lock (s_allZonesLock)
                {
                    InitializeAllZones();
                    List<string> zoneNames = new List<string>(s_zones.Keys);
                    zoneNames.Sort();
                    s_allZoneNames = zoneNames.ToArray();
                }
                return s_allZoneNames;
            }
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TzTimeZone"/> class.
        /// </summary>
        public TzTimeZone()
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="TzTimeZone"/> class.
        /// </summary>
        /// <param name="info">The info.</param>
        public TzTimeZone(TzZoneInfo info)
        {
            if (info == null)
            {
                throw new ArgumentNullException("info");
            }
            m_info = info;
        }

        /// <summary>
        /// Gets the info.
        /// </summary>
        /// <value>The info.</value>
        protected TzZoneInfo Info
        {
            get
            {
                return m_info;
            }
        }

        /// <summary>
        /// This will always return the unique zone name
        /// for this zone as specified in the Olson time zone database.
        /// This zone name does not contain spaces. To retrieve the
        /// abbreviated form which is sensitive to daylight savings time,
        /// use the GetAbbreviation method. The
        /// GetAbbreviation method will return an abbreviation such,
        /// for example, EDT or EST, in the Eastern time zone in the United
        /// States, depending on the point in time given to it.
        /// </summary>
        /// <value></value>
        /// <returns>The standard time zone name.</returns>
        /// <exception cref="T:System.ArgumentNullException">Attempted to set this property to null. </exception>
        public override string StandardName
        {
            get
            {
                return m_info.ZoneName;
            }
        }

        /// <summary>
        /// Gets the historical information, including zones and rules
        /// for this time zone.
        /// </summary>
        /// <value>The historical info.</value>
        public virtual TzZoneInfo HistoricalData
        {
            get
            {
                return m_info;
            }
        }

        /// <summary>
        /// This will always return the <see cref="StandardName"/> name
        /// for this zone as specified in the Olson time zone database.
        /// This zone name does not contain spaces. To retrieve the
        /// abbreviated form which is sensitive to daylight savings time,
        /// use the GetAbbreviation method. The
        /// GetAbbreviation method will return an abbreviation such,
        /// for example, EDT or EST, in the Eastern time zone in the United
        /// States, depending on the point in time given to it.
        /// </summary>
        /// <value></value>
        /// <returns>The daylight saving time zone name.</returns>
        public override string DaylightName
        {
            get
            {
                return StandardName;
            }
        }

        /// <summary>
        /// Returns the daylight saving time period for a particular year. By default,
        /// returns the start and end DateTimes in Local format. Use the overload
        /// of GetDaylightChanges to provide the inflection point Kind (such as Utc)
        /// to do the proper conversion.
        /// </summary>
        /// <param name="year">The year to which the daylight saving time period applies.</param>
        /// <returns>
        /// A <see cref="T:System.Globalization.DaylightTime"></see> instance containing the start and end date for daylight saving time in year.
        /// </returns>
        /// <exception cref="T:System.ArgumentOutOfRangeException">year is less than 1 or greater than 9999. </exception>
        public override DaylightTime GetDaylightChanges(int year)
        {
            return GetDaylightChanges(year, DateTimeKind.Local);
        }

        /// <summary>
        /// Returns the daylight saving time period for a particular year.
        /// </summary>
        /// <param name="year">The year to which the daylight saving time period applies.</param>
        /// <param name="inflectionKind">Kind of the inflection.</param>
        /// <returns>
        /// A <see cref="T:System.Globalization.DaylightTime"></see> instance containing the start and end date for daylight saving time in year.
        /// </returns>
        /// <exception cref="T:System.ArgumentOutOfRangeException">year is less than 1 or greater than 9999. </exception>
        public virtual DaylightTime GetDaylightChanges(int year, DateTimeKind inflectionKind)
        {
            if (year < 1 || year > 9999)
            {
                throw new ArgumentOutOfRangeException("year");
            }

            DaylightTime result = null;

            DateTime point = new DateTime(year, 1, 1);
            TzDatabase.TzZone targetZone = FindZone(point);

            PublicDomain.TzDatabase.TzRule one;
            PublicDomain.TzDatabase.TzRule two;
            GetDaylightChangeRules(ref point, targetZone, inflectionKind, out one, out two, out result);

            return result;
        }

        private void GetDaylightChangeRules(ref DateTime time, TzDatabase.TzZone targetZone, DateTimeKind inflectionKind, out TzDatabase.TzRule one, out TzDatabase.TzRule two, out DaylightTime daylightTime)
        {
            one = two = null;
            daylightTime = null;

            // First, figure out which zone we're in

            if (targetZone != null)
            {
                // The zone might not have any daylight savings time rules
                if (targetZone.RuleName != TzDatabase.NotApplicableValue)
                {
                    int index = FindRuleIndex(targetZone, time);
                    if (index != -1)
                    {
                        one = m_info.Rules[index];

                        two = GetCompanionRule(targetZone, one, index, time);

                        if (one != null && two != null)
                        {
                            DateTime from = one.GetDateTime(time.Year, targetZone.UtcOffset, two);
                            DateTime to = two.GetDateTime(time.Year, targetZone.UtcOffset, one);

                            if (from > to && Info.IsLatitudeNorth)
                            {
                                DateTime swap = to;
                                to = from;
                                from = swap;

                                TzDatabase.TzRule swapRule = one;
                                one = two;
                                two = swapRule;
                            }
                            else if (from < to && !Info.IsLatitudeNorth)
                            {
                                DateTime swap = to;
                                to = from;
                                from = swap;

                                TzDatabase.TzRule swapRule = one;
                                one = two;
                                two = swapRule;
                            }

                            TimeSpan delta = one.SaveTime;
                            if (delta.Ticks == 0)
                            {
                                delta = two.SaveTime;
                            }

                            daylightTime = new DaylightTime(from, to, delta);
                        }
                    }
                }
            }
        }

        private PublicDomain.TzDatabase.TzRule GetCompanionRule(TzDatabase.TzZone zone, PublicDomain.TzDatabase.TzRule rule, int ruleIndex, DateTime point)
        {
            TzDatabase.TzRule result = null;
            int companionIndex = FindRuleIndex(zone, point, rule.Modifier, rule.Modifier == TzDatabase.NotApplicableValue ? ruleIndex : -1, false);
            if (companionIndex != -1)
            {
                result = m_info.Rules[companionIndex];
            }
            return result;
        }

        /// <summary>
        /// Finds the zone.
        /// </summary>
        /// <param name="point">The point.</param>
        /// <returns></returns>
        public TzDatabase.TzZone FindZone(DateTime point)
        {
            // Find the zone for the span of the specified DateTime
            // We iterate backwards since it is more likely that a date
            // is closer to the present
            int length = m_info.Zones.Count;

            if (length == 0)
            {
                throw new Exception("No zones found");
            }

            // We go down to the second element
            int i;
            bool found = false;
            for (i = length - 1; i >= 1; i--)
            {
                TzDatabase.TzZone curZone = m_info.Zones[i];
                TzDatabase.TzZone prevZone = m_info.Zones[i - 1];
                DateTime untilDateTime = curZone.GetUntilDateTime();
                DateTime prevUntilDateTime = prevZone.GetUntilDateTime();

                // untilDateTime states that this zone was active from
                // the end of the previous zone until this specific instant.
                if (point < untilDateTime && point >= prevUntilDateTime)
                {
                    found = true;
                    break;
                }
            }

            if (!found)
            {
                // It is either in the first zone or the last zone
                PublicDomain.TzDatabase.TzZone firstZone = m_info.Zones[0];
                DateTime untilDateTime = firstZone.GetUntilDateTime();
                if (point < untilDateTime)
                {
                    // It is the first zone
                    i = 0;
                }
                else
                {
                    i = length - 1;
                }
            }

            return m_info.Zones[i];
        }

        /// <summary>
        /// Finds the rule.
        /// </summary>
        /// <param name="zone">The zone.</param>
        /// <param name="point">The point.</param>
        /// <returns></returns>
        public TzDatabase.TzRule FindRule(TzDatabase.TzZone zone, DateTime point)
        {
            int ruleIndex = FindRuleIndex(zone, point);
            return ruleIndex == -1 ? null : m_info.Rules[ruleIndex];
        }

        /// <summary>
        /// Finds the index of the rule.
        /// </summary>
        /// <param name="zone">The zone.</param>
        /// <param name="point">The point.</param>
        /// <returns></returns>
        public int FindRuleIndex(TzDatabase.TzZone zone, DateTime point)
        {
            return FindRuleIndex(zone, point, null, -1, true);
        }

        /// <summary>
        /// Finds the rule.
        /// </summary>
        /// <param name="zone">The zone.</param>
        /// <param name="point">The point.</param>
        /// <param name="avoidModifier">The avoid modifier.</param>
        /// <param name="avoidIndex">Index of the avoid.</param>
        /// <param name="exactComparison">if set to <c>true</c> [exact comparison].</param>
        /// <returns></returns>
        public int FindRuleIndex(TzDatabase.TzZone zone, DateTime point, string avoidModifier, int avoidIndex, bool exactComparison)
        {
            // Now, we have the zone, we can start to figure out the amount
            // of time to add to UTC to get standard time

            // Get the standard time from the UTC Offset for the zone

            // Use the rules to see if we need to apply daylight savings time
            // "If this field is - then standard time
            // always applies in the time zone."
            string ruleName = zone.RuleName;

            int curRuleIndex = -1;

            if (ruleName != TzDatabase.NotApplicableValue)
            {
                // Iterate over the rules to find the
                // applicable rule
                int rulesLength = m_info.Rules.Count;
                if (rulesLength == 0 && !string.IsNullOrEmpty(ruleName))
                {
                    throw new BaseException("No rules found for rule name {0}", ruleName);
                }

                for (int j = 0; j < rulesLength; j++)
                {
                    PublicDomain.TzDatabase.TzRule r = m_info.Rules[j];

                    if (r.RuleName == ruleName && (avoidIndex == -1 || (avoidIndex != -1 && avoidIndex != j)))
                    {
                        curRuleIndex = j;
                        if ((exactComparison && point >= r.GetFromDateTime(zone.UtcOffset) && point < r.GetToDateTime(zone.UtcOffset)) ||
                            (!exactComparison && point.Year >= r.FromYear && point.Year <= r.ToYear))
                        {
                            if (avoidModifier == null || avoidModifier != r.Modifier || avoidIndex != -1)
                            {
                                break;
                            }
                        }
                    }
                }

                if (curRuleIndex == -1)
                {
                    throw new Exception("Could not find rule");
                }
            }
            return curRuleIndex;
        }

        /// <summary>
        /// Returns the coordinated universal time (UTC) offset for the specified local time.
        /// </summary>
        /// <param name="time">The local date and time.</param>
        /// <returns>
        /// The UTC offset from time, measured in ticks.
        /// </returns>
        public override TimeSpan GetUtcOffset(DateTime time)
        {
            if (time.Kind == DateTimeKind.Unspecified && TreatUnspecifiedKindAsLocal)
            {
                time = new DateTime(time.Ticks, DateTimeKind.Local);
            }

            TimeSpan result = TimeSpan.Zero;
            PublicDomain.TzDatabase.TzZone zone = FindZone(time);
            if (zone != null)
            {
                result += zone.UtcOffset;

                if (time.Kind == DateTimeKind.Utc)
                {
                    time += zone.UtcOffset;
                }

                PrepOffset(ref time, ref result, zone);
            }

            return result;
        }

        private void PrepOffset(ref DateTime time, ref TimeSpan result, PublicDomain.TzDatabase.TzZone zone)
        {
            // First, figure out which zone we're in
            DateTime point = new DateTime(time.Year, 1, 1);

            PublicDomain.TzDatabase.TzRule one;
            PublicDomain.TzDatabase.TzRule two;
            DaylightTime daylightTime;

            GetDaylightChangeRules(ref point, zone, DateTimeKind.Local, out one, out two, out daylightTime);

            if (daylightTime != null && daylightTime.Start < daylightTime.End && time >= daylightTime.Start && time < daylightTime.End)
            {
                result += one.SaveTime;
            }
            else if (daylightTime != null && daylightTime.Start > daylightTime.End && (time <= daylightTime.Start || time > daylightTime.End))
            {
                result += one.SaveTime;
            }
        }

        /// <summary>
        /// Gets the abbreviation for this time zone at
        /// the current date and time. This is sensitive
        /// to daylight savings time. For example, in the United States,
        /// in the Eastern time zone,
        /// this will return either EDT or EST, depending
        /// on whether or not the current time point is
        /// in the daylight savings time period or not.
        /// </summary>
        /// <returns></returns>
        public virtual string GetAbbreviation()
        {
            return GetAbbreviation(DateTime.Now);
        }

        /// <summary>
        /// Gets the abbreviation for this time zone at
        /// the point of <paramref name="time"/>. This is sensitive
        /// to daylight savings time. For example, in the United States,
        /// in the Eastern time zone,
        /// this will return either EDT or EST, depending
        /// on whether or not the point of <paramref name="time"/> is
        /// in the daylight savings time period or not.
        /// </summary>
        /// <param name="time">The time.</param>
        /// <returns></returns>
        public virtual string GetAbbreviation(DateTime time)
        {
            PublicDomain.TzDatabase.TzZone zone = FindZone(time);
            int rule1Index = FindRuleIndex(zone, time);
            PublicDomain.TzDatabase.TzRule rule1 = (rule1Index == -1 ? null : m_info.Rules[rule1Index]);
            if (rule1 != null)
            {
                PublicDomain.TzDatabase.TzRule modrule = rule1;
                PublicDomain.TzDatabase.TzRule rule2 = GetCompanionRule(zone, rule1, rule1Index, time);
                if (rule2 != null)
                {
                    DateTime x = rule1.GetDateTime(time.Year, zone.UtcOffset);
                    DateTime y = rule2.GetDateTime(time.Year, zone.UtcOffset);

                    if (x > y)
                    {
                        DateTime swap = y;
                        y = x;
                        x = swap;

                        PublicDomain.TzDatabase.TzRule swapRule = rule1;
                        rule1 = rule2;
                        rule2 = swapRule;
                    }

                    if (time < x)
                    {
                        modrule = rule2;
                    }
                    else if (time >= x && time < y)
                    {
                    }
                    else
                    {
                        modrule = rule2;
                    }
                }

                return zone.FormatModifier(modrule);
            }
            else if (zone != null && !string.IsNullOrEmpty(zone.Format) && zone.Format != TzDatabase.NotApplicableValue)
            {
                return zone.Format;
            }
            return StandardName;
        }

        /// <summary>
        /// Returns a value indicating whether the specified date and time is within a daylight saving time period.
        /// </summary>
        /// <param name="time">A date and time.</param>
        /// <returns>
        /// true if time is in a daylight saving time period; false otherwise, or if time is null.
        /// </returns>
        public override bool IsDaylightSavingTime(DateTime time)
        {
            DaylightTime daylightTime = GetDaylightChanges(time.Year);
            if (daylightTime != null)
            {
                if (daylightTime.Start <= daylightTime.End && time < daylightTime.Start)
                {
                    return false;
                }
                else if (time >= daylightTime.Start && time < daylightTime.End)
                {
                    return true;
                }
                else if (daylightTime.Start >= daylightTime.End && (time < daylightTime.End || time >= daylightTime.Start))
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            return false;
        }

        /// <summary>
        /// Returns the local time that corresponds to a specified coordinated universal time (UTC).
        /// The <paramref name="time"/> parameters must have its Kind specifically set to UTC,
        /// otherwise this code will treat the time as a local DateTime and simply return the same
        /// value. It is important to note that normal usage of DateTime's, such as DateTime.Parse
        /// will create a DateTime with Kind Unspecified, which is treated as Local, therefore,
        /// if you are manually creating UTC DateTimes to pass to this function, you must explicitly
        /// create one with the UTC Kind. If you already have a DateTime (from a method like Parse)
        /// with a Local or Unspecified Kind, you can create one with UTC, simply with
        /// new DateTime(oldDateTime.Ticks, DateTimeKind.Utc)
        /// </summary>
        /// <param name="time">A UTC time. See summary of method.</param>
        /// <returns>
        /// A <see cref="T:System.DateTime"></see> instance whose value is the local time that corresponds to time.
        /// </returns>
        public override DateTime ToLocalTime(DateTime time)
        {
            switch (time.Kind)
            {
                case DateTimeKind.Local:
                    // Nothing necessary to change
                    return time;
                case DateTimeKind.Unspecified:
                    if (TreatUnspecifiedKindAsLocal)
                    {
                        // Nothing necessary to change
                        return time;
                    }
                    else
                    {
                        throw new ArgumentException("unspecified kind");
                    }
                case DateTimeKind.Utc:
                    return GetLocalTimeFromUniversalTime(time);
                default:
                    throw new NotImplementedException(time.Kind.ToString());
            }
        }

        /// <summary>
        /// Returns the coordinated universal time (UTC) that corresponds to a specified local time.
        /// </summary>
        /// <param name="time">The local date and time.</param>
        /// <returns>
        /// A <see cref="T:System.DateTime"></see> instance whose value is the UTC time that corresponds to time.
        /// </returns>
        public override DateTime ToUniversalTime(DateTime time)
        {
            switch (time.Kind)
            {
                case DateTimeKind.Local:
                    return GetUniversalTimeFromLocalTime(time);
                case DateTimeKind.Unspecified:
                    if (TreatUnspecifiedKindAsLocal)
                    {
                        return GetUniversalTimeFromLocalTime(time);
                    }
                    else
                    {
                        throw new ArgumentException("unspecified kind");
                    }
                case DateTimeKind.Utc:
                    // It is already UTC
                    return time;
                default:
                    throw new NotImplementedException(time.Kind.ToString());
            }
        }

        private DateTime GetLocalTimeFromUniversalTime(DateTime time)
        {
            time += GetUtcOffset(time);

            return new DateTime(time.Ticks, DateTimeKind.Local);
        }

        private DateTime GetUniversalTimeFromLocalTime(DateTime time)
        {
            if (time.Kind == DateTimeKind.Unspecified && TreatUnspecifiedKindAsLocal)
            {
                time = new DateTime(time.Ticks, DateTimeKind.Local);
            }

            time -= GetUtcOffset(time);

            return new DateTime(time.Ticks, DateTimeKind.Utc);
        }

        /// <summary>
        /// Gets the current time in this time zone.
        /// </summary>
        /// <value>The now.</value>
        public virtual TzDateTime Now
        {
            get
            {
                return new TzDateTime(DateTime.UtcNow, this);
            }
        }

        /// <summary>
        /// Gets the name of the zone. Setting the zone is
        /// the same as finding a new zone.
        /// </summary>
        /// <value>The name of the zone.</value>
        internal string ZoneName
        {
            set
            {
                if (string.IsNullOrEmpty(value))
                {
                    throw new ArgumentNullException("ZoneName");
                }
                TzTimeZone newTimeZone = GetTimeZone(value);
                if (newTimeZone != null)
                {
                    m_info = newTimeZone.m_info;
                }
                else
                {
                    throw new TzDatabase.TzException("Invalid time zone {0}", value);
                }
            }
        }

        /// <summary>
        /// Returns a <see cref="T:System.String"></see> that represents the current <see cref="T:System.Object"></see>.
        /// </summary>
        /// <returns>
        /// A <see cref="T:System.String"></see> that represents the current <see cref="T:System.Object"></see>.
        /// </returns>
        public override string ToString()
        {
            return StandardName;
        }

        /// <summary>
        /// Gets a new object representing the time zone with the
        /// name <paramref name="tzName"/> or null if none can be found.
        /// These <see cref="PublicDomain.TzTimeZone"/> instances are
        /// not cached, so constantly calling this method will continuously
        /// create new <see cref="PublicDomain.TzTimeZone"/> objects.
        /// </summary>
        /// <param name="tzName">Name of the time zone.</param>
        /// <returns>null if no time zone can by found or a TzTimeZone object</returns>
        public static TzTimeZone GetTimeZone(string tzName)
        {
            TzTimeZone result = null;
            TzZoneInfo zoneInfo;

            if (!string.IsNullOrEmpty(tzName))
            {
                try
                {
                    s_zonesLock.AcquireReaderLock(-1);
                    
                    if (Zones.TryGetValue(tzName, out zoneInfo))
                    {
                        result = new TzTimeZone(zoneInfo);
                    }
                    else
                    {
                        LockCookie writerCookie = s_zonesLock.UpgradeToWriterLock(-1);

                        try
                        {
                            if (Zones.TryGetValue(tzName, out zoneInfo))
                            {
                                result = new TzTimeZone(zoneInfo);
                            }
                            else
                            {
                                // No zone found in the cache, lazily load it
                                switch (tzName)
                                {
                                    #region Time Zone lookup

                                    case "Africa/Abidjan":
                                        zoneInfo = InitializeZonesAfricaAbidjan_0(); break;
                                    case "Africa/Accra":
                                        zoneInfo = InitializeZonesAfricaAccra_1(); break;
                                    case "Africa/Addis_Ababa":
                                        zoneInfo = InitializeZonesAfricaAddis_Ababa_2(); break;
                                    case "Africa/Algiers":
                                        zoneInfo = InitializeZonesAfricaAlgiers_3(); break;
                                    case "Africa/Asmara":
                                        zoneInfo = InitializeZonesAfricaAsmara_4(); break;
                                    case "Africa/Asmera":
                                        zoneInfo = InitializeZonesAfricaAsmera_5(); break;
                                    case "Africa/Bamako":
                                        zoneInfo = InitializeZonesAfricaBamako_6(); break;
                                    case "Africa/Bangui":
                                        zoneInfo = InitializeZonesAfricaBangui_7(); break;
                                    case "Africa/Banjul":
                                        zoneInfo = InitializeZonesAfricaBanjul_8(); break;
                                    case "Africa/Bissau":
                                        zoneInfo = InitializeZonesAfricaBissau_9(); break;
                                    case "Africa/Blantyre":
                                        zoneInfo = InitializeZonesAfricaBlantyre_10(); break;
                                    case "Africa/Brazzaville":
                                        zoneInfo = InitializeZonesAfricaBrazzaville_11(); break;
                                    case "Africa/Bujumbura":
                                        zoneInfo = InitializeZonesAfricaBujumbura_12(); break;
                                    case "Africa/Cairo":
                                        zoneInfo = InitializeZonesAfricaCairo_13(); break;
                                    case "Africa/Casablanca":
                                        zoneInfo = InitializeZonesAfricaCasablanca_14(); break;
                                    case "Africa/Ceuta":
                                        zoneInfo = InitializeZonesAfricaCeuta_15(); break;
                                    case "Africa/Conakry":
                                        zoneInfo = InitializeZonesAfricaConakry_16(); break;
                                    case "Africa/Dakar":
                                        zoneInfo = InitializeZonesAfricaDakar_17(); break;
                                    case "Africa/Dar_es_Salaam":
                                        zoneInfo = InitializeZonesAfricaDar_es_Salaam_18(); break;
                                    case "Africa/Djibouti":
                                        zoneInfo = InitializeZonesAfricaDjibouti_19(); break;
                                    case "Africa/Douala":
                                        zoneInfo = InitializeZonesAfricaDouala_20(); break;
                                    case "Africa/El_Aaiun":
                                        zoneInfo = InitializeZonesAfricaEl_Aaiun_21(); break;
                                    case "Africa/Freetown":
                                        zoneInfo = InitializeZonesAfricaFreetown_22(); break;
                                    case "Africa/Gaborone":
                                        zoneInfo = InitializeZonesAfricaGaborone_23(); break;
                                    case "Africa/Harare":
                                        zoneInfo = InitializeZonesAfricaHarare_24(); break;
                                    case "Africa/Johannesburg":
                                        zoneInfo = InitializeZonesAfricaJohannesburg_25(); break;
                                    case "Africa/Kampala":
                                        zoneInfo = InitializeZonesAfricaKampala_26(); break;
                                    case "Africa/Khartoum":
                                        zoneInfo = InitializeZonesAfricaKhartoum_27(); break;
                                    case "Africa/Kigali":
                                        zoneInfo = InitializeZonesAfricaKigali_28(); break;
                                    case "Africa/Kinshasa":
                                        zoneInfo = InitializeZonesAfricaKinshasa_29(); break;
                                    case "Africa/Lagos":
                                        zoneInfo = InitializeZonesAfricaLagos_30(); break;
                                    case "Africa/Libreville":
                                        zoneInfo = InitializeZonesAfricaLibreville_31(); break;
                                    case "Africa/Lome":
                                        zoneInfo = InitializeZonesAfricaLome_32(); break;
                                    case "Africa/Luanda":
                                        zoneInfo = InitializeZonesAfricaLuanda_33(); break;
                                    case "Africa/Lubumbashi":
                                        zoneInfo = InitializeZonesAfricaLubumbashi_34(); break;
                                    case "Africa/Lusaka":
                                        zoneInfo = InitializeZonesAfricaLusaka_35(); break;
                                    case "Africa/Malabo":
                                        zoneInfo = InitializeZonesAfricaMalabo_36(); break;
                                    case "Africa/Maputo":
                                        zoneInfo = InitializeZonesAfricaMaputo_37(); break;
                                    case "Africa/Maseru":
                                        zoneInfo = InitializeZonesAfricaMaseru_38(); break;
                                    case "Africa/Mbabane":
                                        zoneInfo = InitializeZonesAfricaMbabane_39(); break;
                                    case "Africa/Mogadishu":
                                        zoneInfo = InitializeZonesAfricaMogadishu_40(); break;
                                    case "Africa/Monrovia":
                                        zoneInfo = InitializeZonesAfricaMonrovia_41(); break;
                                    case "Africa/Nairobi":
                                        zoneInfo = InitializeZonesAfricaNairobi_42(); break;
                                    case "Africa/Ndjamena":
                                        zoneInfo = InitializeZonesAfricaNdjamena_43(); break;
                                    case "Africa/Niamey":
                                        zoneInfo = InitializeZonesAfricaNiamey_44(); break;
                                    case "Africa/Nouakchott":
                                        zoneInfo = InitializeZonesAfricaNouakchott_45(); break;
                                    case "Africa/Ouagadougou":
                                        zoneInfo = InitializeZonesAfricaOuagadougou_46(); break;
                                    case "Africa/Porto-Novo":
                                        zoneInfo = InitializeZonesAfricaPortoNovo_47(); break;
                                    case "Africa/Sao_Tome":
                                        zoneInfo = InitializeZonesAfricaSao_Tome_48(); break;
                                    case "Africa/Timbuktu":
                                        zoneInfo = InitializeZonesAfricaTimbuktu_49(); break;
                                    case "Africa/Tripoli":
                                        zoneInfo = InitializeZonesAfricaTripoli_50(); break;
                                    case "Africa/Tunis":
                                        zoneInfo = InitializeZonesAfricaTunis_51(); break;
                                    case "Africa/Windhoek":
                                        zoneInfo = InitializeZonesAfricaWindhoek_52(); break;
                                    case "America/Adak":
                                        zoneInfo = InitializeZonesAmericaAdak_53(); break;
                                    case "America/Anchorage":
                                        zoneInfo = InitializeZonesAmericaAnchorage_54(); break;
                                    case "America/Anguilla":
                                        zoneInfo = InitializeZonesAmericaAnguilla_55(); break;
                                    case "America/Antigua":
                                        zoneInfo = InitializeZonesAmericaAntigua_56(); break;
                                    case "America/Araguaina":
                                        zoneInfo = InitializeZonesAmericaAraguaina_57(); break;
                                    case "America/Argentina/Buenos_Aires":
                                        zoneInfo = InitializeZonesAmericaArgentinaBuenos_Aires_58(); break;
                                    case "America/Argentina/Catamarca":
                                        zoneInfo = InitializeZonesAmericaArgentinaCatamarca_59(); break;
                                    case "America/Argentina/ComodRivadavia":
                                        zoneInfo = InitializeZonesAmericaArgentinaComodRivadavia_60(); break;
                                    case "America/Argentina/Cordoba":
                                        zoneInfo = InitializeZonesAmericaArgentinaCordoba_61(); break;
                                    case "America/Argentina/Jujuy":
                                        zoneInfo = InitializeZonesAmericaArgentinaJujuy_62(); break;
                                    case "America/Argentina/La_Rioja":
                                        zoneInfo = InitializeZonesAmericaArgentinaLa_Rioja_63(); break;
                                    case "America/Argentina/Mendoza":
                                        zoneInfo = InitializeZonesAmericaArgentinaMendoza_64(); break;
                                    case "America/Argentina/Rio_Gallegos":
                                        zoneInfo = InitializeZonesAmericaArgentinaRio_Gallegos_65(); break;
                                    case "America/Argentina/San_Juan":
                                        zoneInfo = InitializeZonesAmericaArgentinaSan_Juan_66(); break;
                                    case "America/Argentina/San_Luis":
                                        zoneInfo = InitializeZonesAmericaArgentinaSan_Luis_67(); break;
                                    case "America/Argentina/Tucuman":
                                        zoneInfo = InitializeZonesAmericaArgentinaTucuman_68(); break;
                                    case "America/Argentina/Ushuaia":
                                        zoneInfo = InitializeZonesAmericaArgentinaUshuaia_69(); break;
                                    case "America/Aruba":
                                        zoneInfo = InitializeZonesAmericaAruba_70(); break;
                                    case "America/Asuncion":
                                        zoneInfo = InitializeZonesAmericaAsuncion_71(); break;
                                    case "America/Atikokan":
                                        zoneInfo = InitializeZonesAmericaAtikokan_72(); break;
                                    case "America/Atka":
                                        zoneInfo = InitializeZonesAmericaAtka_73(); break;
                                    case "America/Bahia":
                                        zoneInfo = InitializeZonesAmericaBahia_74(); break;
                                    case "America/Barbados":
                                        zoneInfo = InitializeZonesAmericaBarbados_75(); break;
                                    case "America/Belem":
                                        zoneInfo = InitializeZonesAmericaBelem_76(); break;
                                    case "America/Belize":
                                        zoneInfo = InitializeZonesAmericaBelize_77(); break;
                                    case "America/Blanc-Sablon":
                                        zoneInfo = InitializeZonesAmericaBlancSablon_78(); break;
                                    case "America/Boa_Vista":
                                        zoneInfo = InitializeZonesAmericaBoa_Vista_79(); break;
                                    case "America/Bogota":
                                        zoneInfo = InitializeZonesAmericaBogota_80(); break;
                                    case "America/Boise":
                                        zoneInfo = InitializeZonesAmericaBoise_81(); break;
                                    case "America/Buenos_Aires":
                                        zoneInfo = InitializeZonesAmericaBuenos_Aires_82(); break;
                                    case "America/Cambridge_Bay":
                                        zoneInfo = InitializeZonesAmericaCambridge_Bay_83(); break;
                                    case "America/Campo_Grande":
                                        zoneInfo = InitializeZonesAmericaCampo_Grande_84(); break;
                                    case "America/Cancun":
                                        zoneInfo = InitializeZonesAmericaCancun_85(); break;
                                    case "America/Caracas":
                                        zoneInfo = InitializeZonesAmericaCaracas_86(); break;
                                    case "America/Catamarca":
                                        zoneInfo = InitializeZonesAmericaCatamarca_87(); break;
                                    case "America/Cayenne":
                                        zoneInfo = InitializeZonesAmericaCayenne_88(); break;
                                    case "America/Cayman":
                                        zoneInfo = InitializeZonesAmericaCayman_89(); break;
                                    case "America/Chicago":
                                        zoneInfo = InitializeZonesAmericaChicago_90(); break;
                                    case "America/Chihuahua":
                                        zoneInfo = InitializeZonesAmericaChihuahua_91(); break;
                                    case "America/Coral_Harbour":
                                        zoneInfo = InitializeZonesAmericaCoral_Harbour_92(); break;
                                    case "America/Cordoba":
                                        zoneInfo = InitializeZonesAmericaCordoba_93(); break;
                                    case "America/Costa_Rica":
                                        zoneInfo = InitializeZonesAmericaCosta_Rica_94(); break;
                                    case "America/Cuiaba":
                                        zoneInfo = InitializeZonesAmericaCuiaba_95(); break;
                                    case "America/Curacao":
                                        zoneInfo = InitializeZonesAmericaCuracao_96(); break;
                                    case "America/Danmarkshavn":
                                        zoneInfo = InitializeZonesAmericaDanmarkshavn_97(); break;
                                    case "America/Dawson":
                                        zoneInfo = InitializeZonesAmericaDawson_98(); break;
                                    case "America/Dawson_Creek":
                                        zoneInfo = InitializeZonesAmericaDawson_Creek_99(); break;
                                    case "America/Denver":
                                        zoneInfo = InitializeZonesAmericaDenver_100(); break;
                                    case "America/Detroit":
                                        zoneInfo = InitializeZonesAmericaDetroit_101(); break;
                                    case "America/Dominica":
                                        zoneInfo = InitializeZonesAmericaDominica_102(); break;
                                    case "America/Edmonton":
                                        zoneInfo = InitializeZonesAmericaEdmonton_103(); break;
                                    case "America/Eirunepe":
                                        zoneInfo = InitializeZonesAmericaEirunepe_104(); break;
                                    case "America/El_Salvador":
                                        zoneInfo = InitializeZonesAmericaEl_Salvador_105(); break;
                                    case "America/Ensenada":
                                        zoneInfo = InitializeZonesAmericaEnsenada_106(); break;
                                    case "America/Fort_Wayne":
                                        zoneInfo = InitializeZonesAmericaFort_Wayne_107(); break;
                                    case "America/Fortaleza":
                                        zoneInfo = InitializeZonesAmericaFortaleza_108(); break;
                                    case "America/Glace_Bay":
                                        zoneInfo = InitializeZonesAmericaGlace_Bay_109(); break;
                                    case "America/Godthab":
                                        zoneInfo = InitializeZonesAmericaGodthab_110(); break;
                                    case "America/Goose_Bay":
                                        zoneInfo = InitializeZonesAmericaGoose_Bay_111(); break;
                                    case "America/Grand_Turk":
                                        zoneInfo = InitializeZonesAmericaGrand_Turk_112(); break;
                                    case "America/Grenada":
                                        zoneInfo = InitializeZonesAmericaGrenada_113(); break;
                                    case "America/Guadeloupe":
                                        zoneInfo = InitializeZonesAmericaGuadeloupe_114(); break;
                                    case "America/Guatemala":
                                        zoneInfo = InitializeZonesAmericaGuatemala_115(); break;
                                    case "America/Guayaquil":
                                        zoneInfo = InitializeZonesAmericaGuayaquil_116(); break;
                                    case "America/Guyana":
                                        zoneInfo = InitializeZonesAmericaGuyana_117(); break;
                                    case "America/Halifax":
                                        zoneInfo = InitializeZonesAmericaHalifax_118(); break;
                                    case "America/Havana":
                                        zoneInfo = InitializeZonesAmericaHavana_119(); break;
                                    case "America/Hermosillo":
                                        zoneInfo = InitializeZonesAmericaHermosillo_120(); break;
                                    case "America/Indiana/Indianapolis":
                                        zoneInfo = InitializeZonesAmericaIndianaIndianapolis_121(); break;
                                    case "America/Indiana/Knox":
                                        zoneInfo = InitializeZonesAmericaIndianaKnox_122(); break;
                                    case "America/Indiana/Marengo":
                                        zoneInfo = InitializeZonesAmericaIndianaMarengo_123(); break;
                                    case "America/Indiana/Petersburg":
                                        zoneInfo = InitializeZonesAmericaIndianaPetersburg_124(); break;
                                    case "America/Indiana/Tell_City":
                                        zoneInfo = InitializeZonesAmericaIndianaTell_City_125(); break;
                                    case "America/Indiana/Vevay":
                                        zoneInfo = InitializeZonesAmericaIndianaVevay_126(); break;
                                    case "America/Indiana/Vincennes":
                                        zoneInfo = InitializeZonesAmericaIndianaVincennes_127(); break;
                                    case "America/Indiana/Winamac":
                                        zoneInfo = InitializeZonesAmericaIndianaWinamac_128(); break;
                                    case "America/Indianapolis":
                                        zoneInfo = InitializeZonesAmericaIndianapolis_129(); break;
                                    case "America/Inuvik":
                                        zoneInfo = InitializeZonesAmericaInuvik_130(); break;
                                    case "America/Iqaluit":
                                        zoneInfo = InitializeZonesAmericaIqaluit_131(); break;
                                    case "America/Jamaica":
                                        zoneInfo = InitializeZonesAmericaJamaica_132(); break;
                                    case "America/Jujuy":
                                        zoneInfo = InitializeZonesAmericaJujuy_133(); break;
                                    case "America/Juneau":
                                        zoneInfo = InitializeZonesAmericaJuneau_134(); break;
                                    case "America/Kentucky/Louisville":
                                        zoneInfo = InitializeZonesAmericaKentuckyLouisville_135(); break;
                                    case "America/Kentucky/Monticello":
                                        zoneInfo = InitializeZonesAmericaKentuckyMonticello_136(); break;
                                    case "America/Knox_IN":
                                        zoneInfo = InitializeZonesAmericaKnox_IN_137(); break;
                                    case "America/La_Paz":
                                        zoneInfo = InitializeZonesAmericaLa_Paz_138(); break;
                                    case "America/Lima":
                                        zoneInfo = InitializeZonesAmericaLima_139(); break;
                                    case "America/Los_Angeles":
                                        zoneInfo = InitializeZonesAmericaLos_Angeles_140(); break;
                                    case "America/Louisville":
                                        zoneInfo = InitializeZonesAmericaLouisville_141(); break;
                                    case "America/Maceio":
                                        zoneInfo = InitializeZonesAmericaMaceio_142(); break;
                                    case "America/Managua":
                                        zoneInfo = InitializeZonesAmericaManagua_143(); break;
                                    case "America/Manaus":
                                        zoneInfo = InitializeZonesAmericaManaus_144(); break;
                                    case "America/Marigot":
                                        zoneInfo = InitializeZonesAmericaMarigot_145(); break;
                                    case "America/Martinique":
                                        zoneInfo = InitializeZonesAmericaMartinique_146(); break;
                                    case "America/Mazatlan":
                                        zoneInfo = InitializeZonesAmericaMazatlan_147(); break;
                                    case "America/Mendoza":
                                        zoneInfo = InitializeZonesAmericaMendoza_148(); break;
                                    case "America/Menominee":
                                        zoneInfo = InitializeZonesAmericaMenominee_149(); break;
                                    case "America/Merida":
                                        zoneInfo = InitializeZonesAmericaMerida_150(); break;
                                    case "America/Mexico_City":
                                        zoneInfo = InitializeZonesAmericaMexico_City_151(); break;
                                    case "America/Miquelon":
                                        zoneInfo = InitializeZonesAmericaMiquelon_152(); break;
                                    case "America/Moncton":
                                        zoneInfo = InitializeZonesAmericaMoncton_153(); break;
                                    case "America/Monterrey":
                                        zoneInfo = InitializeZonesAmericaMonterrey_154(); break;
                                    case "America/Montevideo":
                                        zoneInfo = InitializeZonesAmericaMontevideo_155(); break;
                                    case "America/Montreal":
                                        zoneInfo = InitializeZonesAmericaMontreal_156(); break;
                                    case "America/Montserrat":
                                        zoneInfo = InitializeZonesAmericaMontserrat_157(); break;
                                    case "America/Nassau":
                                        zoneInfo = InitializeZonesAmericaNassau_158(); break;
                                    case "America/New_York":
                                        zoneInfo = InitializeZonesAmericaNew_York_159(); break;
                                    case "America/Nipigon":
                                        zoneInfo = InitializeZonesAmericaNipigon_160(); break;
                                    case "America/Nome":
                                        zoneInfo = InitializeZonesAmericaNome_161(); break;
                                    case "America/Noronha":
                                        zoneInfo = InitializeZonesAmericaNoronha_162(); break;
                                    case "America/North_Dakota/Center":
                                        zoneInfo = InitializeZonesAmericaNorth_DakotaCenter_163(); break;
                                    case "America/North_Dakota/New_Salem":
                                        zoneInfo = InitializeZonesAmericaNorth_DakotaNew_Salem_164(); break;
                                    case "America/Panama":
                                        zoneInfo = InitializeZonesAmericaPanama_165(); break;
                                    case "America/Pangnirtung":
                                        zoneInfo = InitializeZonesAmericaPangnirtung_166(); break;
                                    case "America/Paramaribo":
                                        zoneInfo = InitializeZonesAmericaParamaribo_167(); break;
                                    case "America/Phoenix":
                                        zoneInfo = InitializeZonesAmericaPhoenix_168(); break;
                                    case "America/Port_of_Spain":
                                        zoneInfo = InitializeZonesAmericaPort_of_Spain_169(); break;
                                    case "America/Port-au-Prince":
                                        zoneInfo = InitializeZonesAmericaPortauPrince_170(); break;
                                    case "America/Porto_Acre":
                                        zoneInfo = InitializeZonesAmericaPorto_Acre_171(); break;
                                    case "America/Porto_Velho":
                                        zoneInfo = InitializeZonesAmericaPorto_Velho_172(); break;
                                    case "America/Puerto_Rico":
                                        zoneInfo = InitializeZonesAmericaPuerto_Rico_173(); break;
                                    case "America/Rainy_River":
                                        zoneInfo = InitializeZonesAmericaRainy_River_174(); break;
                                    case "America/Rankin_Inlet":
                                        zoneInfo = InitializeZonesAmericaRankin_Inlet_175(); break;
                                    case "America/Recife":
                                        zoneInfo = InitializeZonesAmericaRecife_176(); break;
                                    case "America/Regina":
                                        zoneInfo = InitializeZonesAmericaRegina_177(); break;
                                    case "America/Resolute":
                                        zoneInfo = InitializeZonesAmericaResolute_178(); break;
                                    case "America/Rio_Branco":
                                        zoneInfo = InitializeZonesAmericaRio_Branco_179(); break;
                                    case "America/Rosario":
                                        zoneInfo = InitializeZonesAmericaRosario_180(); break;
                                    case "America/Santarem":
                                        zoneInfo = InitializeZonesAmericaSantarem_181(); break;
                                    case "America/Santiago":
                                        zoneInfo = InitializeZonesAmericaSantiago_182(); break;
                                    case "America/Santo_Domingo":
                                        zoneInfo = InitializeZonesAmericaSanto_Domingo_183(); break;
                                    case "America/Sao_Paulo":
                                        zoneInfo = InitializeZonesAmericaSao_Paulo_184(); break;
                                    case "America/Scoresbysund":
                                        zoneInfo = InitializeZonesAmericaScoresbysund_185(); break;
                                    case "America/Shiprock":
                                        zoneInfo = InitializeZonesAmericaShiprock_186(); break;
                                    case "America/St_Barthelemy":
                                        zoneInfo = InitializeZonesAmericaSt_Barthelemy_187(); break;
                                    case "America/St_Johns":
                                        zoneInfo = InitializeZonesAmericaSt_Johns_188(); break;
                                    case "America/St_Kitts":
                                        zoneInfo = InitializeZonesAmericaSt_Kitts_189(); break;
                                    case "America/St_Lucia":
                                        zoneInfo = InitializeZonesAmericaSt_Lucia_190(); break;
                                    case "America/St_Thomas":
                                        zoneInfo = InitializeZonesAmericaSt_Thomas_191(); break;
                                    case "America/St_Vincent":
                                        zoneInfo = InitializeZonesAmericaSt_Vincent_192(); break;
                                    case "America/Swift_Current":
                                        zoneInfo = InitializeZonesAmericaSwift_Current_193(); break;
                                    case "America/Tegucigalpa":
                                        zoneInfo = InitializeZonesAmericaTegucigalpa_194(); break;
                                    case "America/Thule":
                                        zoneInfo = InitializeZonesAmericaThule_195(); break;
                                    case "America/Thunder_Bay":
                                        zoneInfo = InitializeZonesAmericaThunder_Bay_196(); break;
                                    case "America/Tijuana":
                                        zoneInfo = InitializeZonesAmericaTijuana_197(); break;
                                    case "America/Toronto":
                                        zoneInfo = InitializeZonesAmericaToronto_198(); break;
                                    case "America/Tortola":
                                        zoneInfo = InitializeZonesAmericaTortola_199(); break;
                                    case "America/Vancouver":
                                        zoneInfo = InitializeZonesAmericaVancouver_200(); break;
                                    case "America/Virgin":
                                        zoneInfo = InitializeZonesAmericaVirgin_201(); break;
                                    case "America/Whitehorse":
                                        zoneInfo = InitializeZonesAmericaWhitehorse_202(); break;
                                    case "America/Winnipeg":
                                        zoneInfo = InitializeZonesAmericaWinnipeg_203(); break;
                                    case "America/Yakutat":
                                        zoneInfo = InitializeZonesAmericaYakutat_204(); break;
                                    case "America/Yellowknife":
                                        zoneInfo = InitializeZonesAmericaYellowknife_205(); break;
                                    case "Antarctica/Casey":
                                        zoneInfo = InitializeZonesAntarcticaCasey_206(); break;
                                    case "Antarctica/Davis":
                                        zoneInfo = InitializeZonesAntarcticaDavis_207(); break;
                                    case "Antarctica/DumontDUrville":
                                        zoneInfo = InitializeZonesAntarcticaDumontDUrville_208(); break;
                                    case "Antarctica/Mawson":
                                        zoneInfo = InitializeZonesAntarcticaMawson_209(); break;
                                    case "Antarctica/McMurdo":
                                        zoneInfo = InitializeZonesAntarcticaMcMurdo_210(); break;
                                    case "Antarctica/Palmer":
                                        zoneInfo = InitializeZonesAntarcticaPalmer_211(); break;
                                    case "Antarctica/Rothera":
                                        zoneInfo = InitializeZonesAntarcticaRothera_212(); break;
                                    case "Antarctica/South_Pole":
                                        zoneInfo = InitializeZonesAntarcticaSouth_Pole_213(); break;
                                    case "Antarctica/Syowa":
                                        zoneInfo = InitializeZonesAntarcticaSyowa_214(); break;
                                    case "Antarctica/Vostok":
                                        zoneInfo = InitializeZonesAntarcticaVostok_215(); break;
                                    case "Arctic/Longyearbyen":
                                        zoneInfo = InitializeZonesArcticLongyearbyen_216(); break;
                                    case "Asia/Aden":
                                        zoneInfo = InitializeZonesAsiaAden_217(); break;
                                    case "Asia/Almaty":
                                        zoneInfo = InitializeZonesAsiaAlmaty_218(); break;
                                    case "Asia/Amman":
                                        zoneInfo = InitializeZonesAsiaAmman_219(); break;
                                    case "Asia/Anadyr":
                                        zoneInfo = InitializeZonesAsiaAnadyr_220(); break;
                                    case "Asia/Aqtau":
                                        zoneInfo = InitializeZonesAsiaAqtau_221(); break;
                                    case "Asia/Aqtobe":
                                        zoneInfo = InitializeZonesAsiaAqtobe_222(); break;
                                    case "Asia/Ashgabat":
                                        zoneInfo = InitializeZonesAsiaAshgabat_223(); break;
                                    case "Asia/Ashkhabad":
                                        zoneInfo = InitializeZonesAsiaAshkhabad_224(); break;
                                    case "Asia/Baghdad":
                                        zoneInfo = InitializeZonesAsiaBaghdad_225(); break;
                                    case "Asia/Bahrain":
                                        zoneInfo = InitializeZonesAsiaBahrain_226(); break;
                                    case "Asia/Baku":
                                        zoneInfo = InitializeZonesAsiaBaku_227(); break;
                                    case "Asia/Bangkok":
                                        zoneInfo = InitializeZonesAsiaBangkok_228(); break;
                                    case "Asia/Beirut":
                                        zoneInfo = InitializeZonesAsiaBeirut_229(); break;
                                    case "Asia/Bishkek":
                                        zoneInfo = InitializeZonesAsiaBishkek_230(); break;
                                    case "Asia/Brunei":
                                        zoneInfo = InitializeZonesAsiaBrunei_231(); break;
                                    case "Asia/Calcutta":
                                        zoneInfo = InitializeZonesAsiaCalcutta_232(); break;
                                    case "Asia/Choibalsan":
                                        zoneInfo = InitializeZonesAsiaChoibalsan_233(); break;
                                    case "Asia/Chongqing":
                                        zoneInfo = InitializeZonesAsiaChongqing_234(); break;
                                    case "Asia/Chungking":
                                        zoneInfo = InitializeZonesAsiaChungking_235(); break;
                                    case "Asia/Colombo":
                                        zoneInfo = InitializeZonesAsiaColombo_236(); break;
                                    case "Asia/Dacca":
                                        zoneInfo = InitializeZonesAsiaDacca_237(); break;
                                    case "Asia/Damascus":
                                        zoneInfo = InitializeZonesAsiaDamascus_238(); break;
                                    case "Asia/Dhaka":
                                        zoneInfo = InitializeZonesAsiaDhaka_239(); break;
                                    case "Asia/Dili":
                                        zoneInfo = InitializeZonesAsiaDili_240(); break;
                                    case "Asia/Dubai":
                                        zoneInfo = InitializeZonesAsiaDubai_241(); break;
                                    case "Asia/Dushanbe":
                                        zoneInfo = InitializeZonesAsiaDushanbe_242(); break;
                                    case "Asia/Gaza":
                                        zoneInfo = InitializeZonesAsiaGaza_243(); break;
                                    case "Asia/Harbin":
                                        zoneInfo = InitializeZonesAsiaHarbin_244(); break;
                                    case "Asia/Ho_Chi_Minh":
                                        zoneInfo = InitializeZonesAsiaHo_Chi_Minh_245(); break;
                                    case "Asia/Hong_Kong":
                                        zoneInfo = InitializeZonesAsiaHong_Kong_246(); break;
                                    case "Asia/Hovd":
                                        zoneInfo = InitializeZonesAsiaHovd_247(); break;
                                    case "Asia/Irkutsk":
                                        zoneInfo = InitializeZonesAsiaIrkutsk_248(); break;
                                    case "Asia/Istanbul":
                                        zoneInfo = InitializeZonesAsiaIstanbul_249(); break;
                                    case "Asia/Jakarta":
                                        zoneInfo = InitializeZonesAsiaJakarta_250(); break;
                                    case "Asia/Jayapura":
                                        zoneInfo = InitializeZonesAsiaJayapura_251(); break;
                                    case "Asia/Jerusalem":
                                        zoneInfo = InitializeZonesAsiaJerusalem_252(); break;
                                    case "Asia/Kabul":
                                        zoneInfo = InitializeZonesAsiaKabul_253(); break;
                                    case "Asia/Kamchatka":
                                        zoneInfo = InitializeZonesAsiaKamchatka_254(); break;
                                    case "Asia/Karachi":
                                        zoneInfo = InitializeZonesAsiaKarachi_255(); break;
                                    case "Asia/Kashgar":
                                        zoneInfo = InitializeZonesAsiaKashgar_256(); break;
                                    case "Asia/Katmandu":
                                        zoneInfo = InitializeZonesAsiaKatmandu_257(); break;
                                    case "Asia/Kolkata":
                                        zoneInfo = InitializeZonesAsiaKolkata_258(); break;
                                    case "Asia/Krasnoyarsk":
                                        zoneInfo = InitializeZonesAsiaKrasnoyarsk_259(); break;
                                    case "Asia/Kuala_Lumpur":
                                        zoneInfo = InitializeZonesAsiaKuala_Lumpur_260(); break;
                                    case "Asia/Kuching":
                                        zoneInfo = InitializeZonesAsiaKuching_261(); break;
                                    case "Asia/Kuwait":
                                        zoneInfo = InitializeZonesAsiaKuwait_262(); break;
                                    case "Asia/Macao":
                                        zoneInfo = InitializeZonesAsiaMacao_263(); break;
                                    case "Asia/Macau":
                                        zoneInfo = InitializeZonesAsiaMacau_264(); break;
                                    case "Asia/Magadan":
                                        zoneInfo = InitializeZonesAsiaMagadan_265(); break;
                                    case "Asia/Makassar":
                                        zoneInfo = InitializeZonesAsiaMakassar_266(); break;
                                    case "Asia/Manila":
                                        zoneInfo = InitializeZonesAsiaManila_267(); break;
                                    case "Asia/Muscat":
                                        zoneInfo = InitializeZonesAsiaMuscat_268(); break;
                                    case "Asia/Nicosia":
                                        zoneInfo = InitializeZonesAsiaNicosia_269(); break;
                                    case "Asia/Novosibirsk":
                                        zoneInfo = InitializeZonesAsiaNovosibirsk_270(); break;
                                    case "Asia/Omsk":
                                        zoneInfo = InitializeZonesAsiaOmsk_271(); break;
                                    case "Asia/Oral":
                                        zoneInfo = InitializeZonesAsiaOral_272(); break;
                                    case "Asia/Phnom_Penh":
                                        zoneInfo = InitializeZonesAsiaPhnom_Penh_273(); break;
                                    case "Asia/Pontianak":
                                        zoneInfo = InitializeZonesAsiaPontianak_274(); break;
                                    case "Asia/Pyongyang":
                                        zoneInfo = InitializeZonesAsiaPyongyang_275(); break;
                                    case "Asia/Qatar":
                                        zoneInfo = InitializeZonesAsiaQatar_276(); break;
                                    case "Asia/Qyzylorda":
                                        zoneInfo = InitializeZonesAsiaQyzylorda_277(); break;
                                    case "Asia/Rangoon":
                                        zoneInfo = InitializeZonesAsiaRangoon_278(); break;
                                    case "Asia/Riyadh":
                                        zoneInfo = InitializeZonesAsiaRiyadh_279(); break;
                                    case "Asia/Riyadh87":
                                        zoneInfo = InitializeZonesAsiaRiyadh87_280(); break;
                                    case "Asia/Riyadh88":
                                        zoneInfo = InitializeZonesAsiaRiyadh88_281(); break;
                                    case "Asia/Riyadh89":
                                        zoneInfo = InitializeZonesAsiaRiyadh89_282(); break;
                                    case "Asia/Saigon":
                                        zoneInfo = InitializeZonesAsiaSaigon_283(); break;
                                    case "Asia/Sakhalin":
                                        zoneInfo = InitializeZonesAsiaSakhalin_284(); break;
                                    case "Asia/Samarkand":
                                        zoneInfo = InitializeZonesAsiaSamarkand_285(); break;
                                    case "Asia/Seoul":
                                        zoneInfo = InitializeZonesAsiaSeoul_286(); break;
                                    case "Asia/Shanghai":
                                        zoneInfo = InitializeZonesAsiaShanghai_287(); break;
                                    case "Asia/Singapore":
                                        zoneInfo = InitializeZonesAsiaSingapore_288(); break;
                                    case "Asia/Taipei":
                                        zoneInfo = InitializeZonesAsiaTaipei_289(); break;
                                    case "Asia/Tashkent":
                                        zoneInfo = InitializeZonesAsiaTashkent_290(); break;
                                    case "Asia/Tbilisi":
                                        zoneInfo = InitializeZonesAsiaTbilisi_291(); break;
                                    case "Asia/Tehran":
                                        zoneInfo = InitializeZonesAsiaTehran_292(); break;
                                    case "Asia/Tel_Aviv":
                                        zoneInfo = InitializeZonesAsiaTel_Aviv_293(); break;
                                    case "Asia/Thimbu":
                                        zoneInfo = InitializeZonesAsiaThimbu_294(); break;
                                    case "Asia/Thimphu":
                                        zoneInfo = InitializeZonesAsiaThimphu_295(); break;
                                    case "Asia/Tokyo":
                                        zoneInfo = InitializeZonesAsiaTokyo_296(); break;
                                    case "Asia/Ujung_Pandang":
                                        zoneInfo = InitializeZonesAsiaUjung_Pandang_297(); break;
                                    case "Asia/Ulaanbaatar":
                                        zoneInfo = InitializeZonesAsiaUlaanbaatar_298(); break;
                                    case "Asia/Ulan_Bator":
                                        zoneInfo = InitializeZonesAsiaUlan_Bator_299(); break;
                                    case "Asia/Urumqi":
                                        zoneInfo = InitializeZonesAsiaUrumqi_300(); break;
                                    case "Asia/Vientiane":
                                        zoneInfo = InitializeZonesAsiaVientiane_301(); break;
                                    case "Asia/Vladivostok":
                                        zoneInfo = InitializeZonesAsiaVladivostok_302(); break;
                                    case "Asia/Yakutsk":
                                        zoneInfo = InitializeZonesAsiaYakutsk_303(); break;
                                    case "Asia/Yekaterinburg":
                                        zoneInfo = InitializeZonesAsiaYekaterinburg_304(); break;
                                    case "Asia/Yerevan":
                                        zoneInfo = InitializeZonesAsiaYerevan_305(); break;
                                    case "Atlantic/Azores":
                                        zoneInfo = InitializeZonesAtlanticAzores_306(); break;
                                    case "Atlantic/Bermuda":
                                        zoneInfo = InitializeZonesAtlanticBermuda_307(); break;
                                    case "Atlantic/Canary":
                                        zoneInfo = InitializeZonesAtlanticCanary_308(); break;
                                    case "Atlantic/Cape_Verde":
                                        zoneInfo = InitializeZonesAtlanticCape_Verde_309(); break;
                                    case "Atlantic/Faeroe":
                                        zoneInfo = InitializeZonesAtlanticFaeroe_310(); break;
                                    case "Atlantic/Faroe":
                                        zoneInfo = InitializeZonesAtlanticFaroe_311(); break;
                                    case "Atlantic/Jan_Mayen":
                                        zoneInfo = InitializeZonesAtlanticJan_Mayen_312(); break;
                                    case "Atlantic/Madeira":
                                        zoneInfo = InitializeZonesAtlanticMadeira_313(); break;
                                    case "Atlantic/Reykjavik":
                                        zoneInfo = InitializeZonesAtlanticReykjavik_314(); break;
                                    case "Atlantic/South_Georgia":
                                        zoneInfo = InitializeZonesAtlanticSouth_Georgia_315(); break;
                                    case "Atlantic/St_Helena":
                                        zoneInfo = InitializeZonesAtlanticSt_Helena_316(); break;
                                    case "Atlantic/Stanley":
                                        zoneInfo = InitializeZonesAtlanticStanley_317(); break;
                                    case "Australia/ACT":
                                        zoneInfo = InitializeZonesAustraliaACT_318(); break;
                                    case "Australia/Adelaide":
                                        zoneInfo = InitializeZonesAustraliaAdelaide_319(); break;
                                    case "Australia/Brisbane":
                                        zoneInfo = InitializeZonesAustraliaBrisbane_320(); break;
                                    case "Australia/Broken_Hill":
                                        zoneInfo = InitializeZonesAustraliaBroken_Hill_321(); break;
                                    case "Australia/Canberra":
                                        zoneInfo = InitializeZonesAustraliaCanberra_322(); break;
                                    case "Australia/Currie":
                                        zoneInfo = InitializeZonesAustraliaCurrie_323(); break;
                                    case "Australia/Darwin":
                                        zoneInfo = InitializeZonesAustraliaDarwin_324(); break;
                                    case "Australia/Eucla":
                                        zoneInfo = InitializeZonesAustraliaEucla_325(); break;
                                    case "Australia/Hobart":
                                        zoneInfo = InitializeZonesAustraliaHobart_326(); break;
                                    case "Australia/LHI":
                                        zoneInfo = InitializeZonesAustraliaLHI_327(); break;
                                    case "Australia/Lindeman":
                                        zoneInfo = InitializeZonesAustraliaLindeman_328(); break;
                                    case "Australia/Lord_Howe":
                                        zoneInfo = InitializeZonesAustraliaLord_Howe_329(); break;
                                    case "Australia/Melbourne":
                                        zoneInfo = InitializeZonesAustraliaMelbourne_330(); break;
                                    case "Australia/North":
                                        zoneInfo = InitializeZonesAustraliaNorth_331(); break;
                                    case "Australia/NSW":
                                        zoneInfo = InitializeZonesAustraliaNSW_332(); break;
                                    case "Australia/Perth":
                                        zoneInfo = InitializeZonesAustraliaPerth_333(); break;
                                    case "Australia/Queensland":
                                        zoneInfo = InitializeZonesAustraliaQueensland_334(); break;
                                    case "Australia/South":
                                        zoneInfo = InitializeZonesAustraliaSouth_335(); break;
                                    case "Australia/Sydney":
                                        zoneInfo = InitializeZonesAustraliaSydney_336(); break;
                                    case "Australia/Tasmania":
                                        zoneInfo = InitializeZonesAustraliaTasmania_337(); break;
                                    case "Australia/Victoria":
                                        zoneInfo = InitializeZonesAustraliaVictoria_338(); break;
                                    case "Australia/West":
                                        zoneInfo = InitializeZonesAustraliaWest_339(); break;
                                    case "Australia/Yancowinna":
                                        zoneInfo = InitializeZonesAustraliaYancowinna_340(); break;
                                    case "Brazil/Acre":
                                        zoneInfo = InitializeZonesBrazilAcre_341(); break;
                                    case "Brazil/DeNoronha":
                                        zoneInfo = InitializeZonesBrazilDeNoronha_342(); break;
                                    case "Brazil/East":
                                        zoneInfo = InitializeZonesBrazilEast_343(); break;
                                    case "Brazil/West":
                                        zoneInfo = InitializeZonesBrazilWest_344(); break;
                                    case "Canada/Atlantic":
                                        zoneInfo = InitializeZonesCanadaAtlantic_345(); break;
                                    case "Canada/Central":
                                        zoneInfo = InitializeZonesCanadaCentral_346(); break;
                                    case "Canada/Eastern":
                                        zoneInfo = InitializeZonesCanadaEastern_347(); break;
                                    case "Canada/East-Saskatchewan":
                                        zoneInfo = InitializeZonesCanadaEastSaskatchewan_348(); break;
                                    case "Canada/Mountain":
                                        zoneInfo = InitializeZonesCanadaMountain_349(); break;
                                    case "Canada/Newfoundland":
                                        zoneInfo = InitializeZonesCanadaNewfoundland_350(); break;
                                    case "Canada/Pacific":
                                        zoneInfo = InitializeZonesCanadaPacific_351(); break;
                                    case "Canada/Saskatchewan":
                                        zoneInfo = InitializeZonesCanadaSaskatchewan_352(); break;
                                    case "Canada/Yukon":
                                        zoneInfo = InitializeZonesCanadaYukon_353(); break;
                                    case "CET":
                                        zoneInfo = InitializeZonesCET_354(); break;
                                    case "Chile/Continental":
                                        zoneInfo = InitializeZonesChileContinental_355(); break;
                                    case "Chile/EasterIsland":
                                        zoneInfo = InitializeZonesChileEasterIsland_356(); break;
                                    case "CST6CDT":
                                        zoneInfo = InitializeZonesCST6CDT_357(); break;
                                    case "Cuba":
                                        zoneInfo = InitializeZonesCuba_358(); break;
                                    case "EET":
                                        zoneInfo = InitializeZonesEET_359(); break;
                                    case "Egypt":
                                        zoneInfo = InitializeZonesEgypt_360(); break;
                                    case "Eire":
                                        zoneInfo = InitializeZonesEire_361(); break;
                                    case "EST":
                                        zoneInfo = InitializeZonesEST_362(); break;
                                    case "EST5EDT":
                                        zoneInfo = InitializeZonesEST5EDT_363(); break;
                                    case "Etc/GMT":
                                        zoneInfo = InitializeZonesEtcGMT_364(); break;
                                    case "Etc/GMT+0":
                                        zoneInfo = InitializeZonesEtcGMT0_365(); break;
                                    case "Etc/GMT+1":
                                        zoneInfo = InitializeZonesEtcGMT1_366(); break;
                                    case "Etc/GMT+10":
                                        zoneInfo = InitializeZonesEtcGMT10_367(); break;
                                    case "Etc/GMT+11":
                                        zoneInfo = InitializeZonesEtcGMT11_368(); break;
                                    case "Etc/GMT+12":
                                        zoneInfo = InitializeZonesEtcGMT12_369(); break;
                                    case "Etc/GMT+2":
                                        zoneInfo = InitializeZonesEtcGMT2_370(); break;
                                    case "Etc/GMT+3":
                                        zoneInfo = InitializeZonesEtcGMT3_371(); break;
                                    case "Etc/GMT+4":
                                        zoneInfo = InitializeZonesEtcGMT4_372(); break;
                                    case "Etc/GMT+5":
                                        zoneInfo = InitializeZonesEtcGMT5_373(); break;
                                    case "Etc/GMT+6":
                                        zoneInfo = InitializeZonesEtcGMT6_374(); break;
                                    case "Etc/GMT+7":
                                        zoneInfo = InitializeZonesEtcGMT7_375(); break;
                                    case "Etc/GMT+8":
                                        zoneInfo = InitializeZonesEtcGMT8_376(); break;
                                    case "Etc/GMT+9":
                                        zoneInfo = InitializeZonesEtcGMT9_377(); break;
                                    case "Etc/GMT0":
                                        zoneInfo = InitializeZonesEtcGMT0_378(); break;
                                    case "Etc/GMT-0":
                                        zoneInfo = InitializeZonesEtcGMT0_379(); break;
                                    case "Etc/GMT-1":
                                        zoneInfo = InitializeZonesEtcGMT1_380(); break;
                                    case "Etc/GMT-10":
                                        zoneInfo = InitializeZonesEtcGMT10_381(); break;
                                    case "Etc/GMT-11":
                                        zoneInfo = InitializeZonesEtcGMT11_382(); break;
                                    case "Etc/GMT-12":
                                        zoneInfo = InitializeZonesEtcGMT12_383(); break;
                                    case "Etc/GMT-13":
                                        zoneInfo = InitializeZonesEtcGMT13_384(); break;
                                    case "Etc/GMT-14":
                                        zoneInfo = InitializeZonesEtcGMT14_385(); break;
                                    case "Etc/GMT-2":
                                        zoneInfo = InitializeZonesEtcGMT2_386(); break;
                                    case "Etc/GMT-3":
                                        zoneInfo = InitializeZonesEtcGMT3_387(); break;
                                    case "Etc/GMT-4":
                                        zoneInfo = InitializeZonesEtcGMT4_388(); break;
                                    case "Etc/GMT-5":
                                        zoneInfo = InitializeZonesEtcGMT5_389(); break;
                                    case "Etc/GMT-6":
                                        zoneInfo = InitializeZonesEtcGMT6_390(); break;
                                    case "Etc/GMT-7":
                                        zoneInfo = InitializeZonesEtcGMT7_391(); break;
                                    case "Etc/GMT-8":
                                        zoneInfo = InitializeZonesEtcGMT8_392(); break;
                                    case "Etc/GMT-9":
                                        zoneInfo = InitializeZonesEtcGMT9_393(); break;
                                    case "Etc/Greenwich":
                                        zoneInfo = InitializeZonesEtcGreenwich_394(); break;
                                    case "Etc/UCT":
                                        zoneInfo = InitializeZonesEtcUCT_395(); break;
                                    case "Etc/Universal":
                                        zoneInfo = InitializeZonesEtcUniversal_396(); break;
                                    case "Etc/UTC":
                                        zoneInfo = InitializeZonesEtcUTC_397(); break;
                                    case "Etc/Zulu":
                                        zoneInfo = InitializeZonesEtcZulu_398(); break;
                                    case "Europe/Amsterdam":
                                        zoneInfo = InitializeZonesEuropeAmsterdam_399(); break;
                                    case "Europe/Andorra":
                                        zoneInfo = InitializeZonesEuropeAndorra_400(); break;
                                    case "Europe/Athens":
                                        zoneInfo = InitializeZonesEuropeAthens_401(); break;
                                    case "Europe/Belfast":
                                        zoneInfo = InitializeZonesEuropeBelfast_402(); break;
                                    case "Europe/Belgrade":
                                        zoneInfo = InitializeZonesEuropeBelgrade_403(); break;
                                    case "Europe/Berlin":
                                        zoneInfo = InitializeZonesEuropeBerlin_404(); break;
                                    case "Europe/Bratislava":
                                        zoneInfo = InitializeZonesEuropeBratislava_405(); break;
                                    case "Europe/Brussels":
                                        zoneInfo = InitializeZonesEuropeBrussels_406(); break;
                                    case "Europe/Bucharest":
                                        zoneInfo = InitializeZonesEuropeBucharest_407(); break;
                                    case "Europe/Budapest":
                                        zoneInfo = InitializeZonesEuropeBudapest_408(); break;
                                    case "Europe/Chisinau":
                                        zoneInfo = InitializeZonesEuropeChisinau_409(); break;
                                    case "Europe/Copenhagen":
                                        zoneInfo = InitializeZonesEuropeCopenhagen_410(); break;
                                    case "Europe/Dublin":
                                        zoneInfo = InitializeZonesEuropeDublin_411(); break;
                                    case "Europe/Gibraltar":
                                        zoneInfo = InitializeZonesEuropeGibraltar_412(); break;
                                    case "Europe/Guernsey":
                                        zoneInfo = InitializeZonesEuropeGuernsey_413(); break;
                                    case "Europe/Helsinki":
                                        zoneInfo = InitializeZonesEuropeHelsinki_414(); break;
                                    case "Europe/Isle_of_Man":
                                        zoneInfo = InitializeZonesEuropeIsle_of_Man_415(); break;
                                    case "Europe/Istanbul":
                                        zoneInfo = InitializeZonesEuropeIstanbul_416(); break;
                                    case "Europe/Jersey":
                                        zoneInfo = InitializeZonesEuropeJersey_417(); break;
                                    case "Europe/Kaliningrad":
                                        zoneInfo = InitializeZonesEuropeKaliningrad_418(); break;
                                    case "Europe/Kiev":
                                        zoneInfo = InitializeZonesEuropeKiev_419(); break;
                                    case "Europe/Lisbon":
                                        zoneInfo = InitializeZonesEuropeLisbon_420(); break;
                                    case "Europe/Ljubljana":
                                        zoneInfo = InitializeZonesEuropeLjubljana_421(); break;
                                    case "Europe/London":
                                        zoneInfo = InitializeZonesEuropeLondon_422(); break;
                                    case "Europe/Luxembourg":
                                        zoneInfo = InitializeZonesEuropeLuxembourg_423(); break;
                                    case "Europe/Madrid":
                                        zoneInfo = InitializeZonesEuropeMadrid_424(); break;
                                    case "Europe/Malta":
                                        zoneInfo = InitializeZonesEuropeMalta_425(); break;
                                    case "Europe/Mariehamn":
                                        zoneInfo = InitializeZonesEuropeMariehamn_426(); break;
                                    case "Europe/Minsk":
                                        zoneInfo = InitializeZonesEuropeMinsk_427(); break;
                                    case "Europe/Monaco":
                                        zoneInfo = InitializeZonesEuropeMonaco_428(); break;
                                    case "Europe/Moscow":
                                        zoneInfo = InitializeZonesEuropeMoscow_429(); break;
                                    case "Europe/Nicosia":
                                        zoneInfo = InitializeZonesEuropeNicosia_430(); break;
                                    case "Europe/Oslo":
                                        zoneInfo = InitializeZonesEuropeOslo_431(); break;
                                    case "Europe/Paris":
                                        zoneInfo = InitializeZonesEuropeParis_432(); break;
                                    case "Europe/Podgorica":
                                        zoneInfo = InitializeZonesEuropePodgorica_433(); break;
                                    case "Europe/Prague":
                                        zoneInfo = InitializeZonesEuropePrague_434(); break;
                                    case "Europe/Riga":
                                        zoneInfo = InitializeZonesEuropeRiga_435(); break;
                                    case "Europe/Rome":
                                        zoneInfo = InitializeZonesEuropeRome_436(); break;
                                    case "Europe/Samara":
                                        zoneInfo = InitializeZonesEuropeSamara_437(); break;
                                    case "Europe/San_Marino":
                                        zoneInfo = InitializeZonesEuropeSan_Marino_438(); break;
                                    case "Europe/Sarajevo":
                                        zoneInfo = InitializeZonesEuropeSarajevo_439(); break;
                                    case "Europe/Simferopol":
                                        zoneInfo = InitializeZonesEuropeSimferopol_440(); break;
                                    case "Europe/Skopje":
                                        zoneInfo = InitializeZonesEuropeSkopje_441(); break;
                                    case "Europe/Sofia":
                                        zoneInfo = InitializeZonesEuropeSofia_442(); break;
                                    case "Europe/Stockholm":
                                        zoneInfo = InitializeZonesEuropeStockholm_443(); break;
                                    case "Europe/Tallinn":
                                        zoneInfo = InitializeZonesEuropeTallinn_444(); break;
                                    case "Europe/Tirane":
                                        zoneInfo = InitializeZonesEuropeTirane_445(); break;
                                    case "Europe/Tiraspol":
                                        zoneInfo = InitializeZonesEuropeTiraspol_446(); break;
                                    case "Europe/Uzhgorod":
                                        zoneInfo = InitializeZonesEuropeUzhgorod_447(); break;
                                    case "Europe/Vaduz":
                                        zoneInfo = InitializeZonesEuropeVaduz_448(); break;
                                    case "Europe/Vatican":
                                        zoneInfo = InitializeZonesEuropeVatican_449(); break;
                                    case "Europe/Vienna":
                                        zoneInfo = InitializeZonesEuropeVienna_450(); break;
                                    case "Europe/Vilnius":
                                        zoneInfo = InitializeZonesEuropeVilnius_451(); break;
                                    case "Europe/Volgograd":
                                        zoneInfo = InitializeZonesEuropeVolgograd_452(); break;
                                    case "Europe/Warsaw":
                                        zoneInfo = InitializeZonesEuropeWarsaw_453(); break;
                                    case "Europe/Zagreb":
                                        zoneInfo = InitializeZonesEuropeZagreb_454(); break;
                                    case "Europe/Zaporozhye":
                                        zoneInfo = InitializeZonesEuropeZaporozhye_455(); break;
                                    case "Europe/Zurich":
                                        zoneInfo = InitializeZonesEuropeZurich_456(); break;
                                    case "GB":
                                        zoneInfo = InitializeZonesGB_457(); break;
                                    case "GB-Eire":
                                        zoneInfo = InitializeZonesGBEire_458(); break;
                                    case "GMT":
                                        zoneInfo = InitializeZonesGMT_459(); break;
                                    case "GMT+0":
                                        zoneInfo = InitializeZonesGMT0_460(); break;
                                    case "GMT0":
                                        zoneInfo = InitializeZonesGMT0_461(); break;
                                    case "GMT-0":
                                        zoneInfo = InitializeZonesGMT0_462(); break;
                                    case "Greenwich":
                                        zoneInfo = InitializeZonesGreenwich_463(); break;
                                    case "Hongkong":
                                        zoneInfo = InitializeZonesHongkong_464(); break;
                                    case "HST":
                                        zoneInfo = InitializeZonesHST_465(); break;
                                    case "Iceland":
                                        zoneInfo = InitializeZonesIceland_466(); break;
                                    case "Indian/Antananarivo":
                                        zoneInfo = InitializeZonesIndianAntananarivo_467(); break;
                                    case "Indian/Chagos":
                                        zoneInfo = InitializeZonesIndianChagos_468(); break;
                                    case "Indian/Christmas":
                                        zoneInfo = InitializeZonesIndianChristmas_469(); break;
                                    case "Indian/Cocos":
                                        zoneInfo = InitializeZonesIndianCocos_470(); break;
                                    case "Indian/Comoro":
                                        zoneInfo = InitializeZonesIndianComoro_471(); break;
                                    case "Indian/Kerguelen":
                                        zoneInfo = InitializeZonesIndianKerguelen_472(); break;
                                    case "Indian/Mahe":
                                        zoneInfo = InitializeZonesIndianMahe_473(); break;
                                    case "Indian/Maldives":
                                        zoneInfo = InitializeZonesIndianMaldives_474(); break;
                                    case "Indian/Mauritius":
                                        zoneInfo = InitializeZonesIndianMauritius_475(); break;
                                    case "Indian/Mayotte":
                                        zoneInfo = InitializeZonesIndianMayotte_476(); break;
                                    case "Indian/Reunion":
                                        zoneInfo = InitializeZonesIndianReunion_477(); break;
                                    case "Iran":
                                        zoneInfo = InitializeZonesIran_478(); break;
                                    case "Israel":
                                        zoneInfo = InitializeZonesIsrael_479(); break;
                                    case "Jamaica":
                                        zoneInfo = InitializeZonesJamaica_480(); break;
                                    case "Japan":
                                        zoneInfo = InitializeZonesJapan_481(); break;
                                    case "Kwajalein":
                                        zoneInfo = InitializeZonesKwajalein_482(); break;
                                    case "Libya":
                                        zoneInfo = InitializeZonesLibya_483(); break;
                                    case "MET":
                                        zoneInfo = InitializeZonesMET_484(); break;
                                    case "Mexico/BajaNorte":
                                        zoneInfo = InitializeZonesMexicoBajaNorte_485(); break;
                                    case "Mexico/BajaSur":
                                        zoneInfo = InitializeZonesMexicoBajaSur_486(); break;
                                    case "Mexico/General":
                                        zoneInfo = InitializeZonesMexicoGeneral_487(); break;
                                    case "Mideast/Riyadh87":
                                        zoneInfo = InitializeZonesMideastRiyadh87_488(); break;
                                    case "Mideast/Riyadh88":
                                        zoneInfo = InitializeZonesMideastRiyadh88_489(); break;
                                    case "Mideast/Riyadh89":
                                        zoneInfo = InitializeZonesMideastRiyadh89_490(); break;
                                    case "MST":
                                        zoneInfo = InitializeZonesMST_491(); break;
                                    case "MST7MDT":
                                        zoneInfo = InitializeZonesMST7MDT_492(); break;
                                    case "Navajo":
                                        zoneInfo = InitializeZonesNavajo_493(); break;
                                    case "NZ":
                                        zoneInfo = InitializeZonesNZ_494(); break;
                                    case "NZ-CHAT":
                                        zoneInfo = InitializeZonesNZCHAT_495(); break;
                                    case "Pacific/Apia":
                                        zoneInfo = InitializeZonesPacificApia_496(); break;
                                    case "Pacific/Auckland":
                                        zoneInfo = InitializeZonesPacificAuckland_497(); break;
                                    case "Pacific/Chatham":
                                        zoneInfo = InitializeZonesPacificChatham_498(); break;
                                    case "Pacific/Easter":
                                        zoneInfo = InitializeZonesPacificEaster_499(); break;
                                    case "Pacific/Efate":
                                        zoneInfo = InitializeZonesPacificEfate_500(); break;
                                    case "Pacific/Enderbury":
                                        zoneInfo = InitializeZonesPacificEnderbury_501(); break;
                                    case "Pacific/Fakaofo":
                                        zoneInfo = InitializeZonesPacificFakaofo_502(); break;
                                    case "Pacific/Fiji":
                                        zoneInfo = InitializeZonesPacificFiji_503(); break;
                                    case "Pacific/Funafuti":
                                        zoneInfo = InitializeZonesPacificFunafuti_504(); break;
                                    case "Pacific/Galapagos":
                                        zoneInfo = InitializeZonesPacificGalapagos_505(); break;
                                    case "Pacific/Gambier":
                                        zoneInfo = InitializeZonesPacificGambier_506(); break;
                                    case "Pacific/Guadalcanal":
                                        zoneInfo = InitializeZonesPacificGuadalcanal_507(); break;
                                    case "Pacific/Guam":
                                        zoneInfo = InitializeZonesPacificGuam_508(); break;
                                    case "Pacific/Honolulu":
                                        zoneInfo = InitializeZonesPacificHonolulu_509(); break;
                                    case "Pacific/Johnston":
                                        zoneInfo = InitializeZonesPacificJohnston_510(); break;
                                    case "Pacific/Kiritimati":
                                        zoneInfo = InitializeZonesPacificKiritimati_511(); break;
                                    case "Pacific/Kosrae":
                                        zoneInfo = InitializeZonesPacificKosrae_512(); break;
                                    case "Pacific/Kwajalein":
                                        zoneInfo = InitializeZonesPacificKwajalein_513(); break;
                                    case "Pacific/Majuro":
                                        zoneInfo = InitializeZonesPacificMajuro_514(); break;
                                    case "Pacific/Marquesas":
                                        zoneInfo = InitializeZonesPacificMarquesas_515(); break;
                                    case "Pacific/Midway":
                                        zoneInfo = InitializeZonesPacificMidway_516(); break;
                                    case "Pacific/Nauru":
                                        zoneInfo = InitializeZonesPacificNauru_517(); break;
                                    case "Pacific/Niue":
                                        zoneInfo = InitializeZonesPacificNiue_518(); break;
                                    case "Pacific/Norfolk":
                                        zoneInfo = InitializeZonesPacificNorfolk_519(); break;
                                    case "Pacific/Noumea":
                                        zoneInfo = InitializeZonesPacificNoumea_520(); break;
                                    case "Pacific/Pago_Pago":
                                        zoneInfo = InitializeZonesPacificPago_Pago_521(); break;
                                    case "Pacific/Palau":
                                        zoneInfo = InitializeZonesPacificPalau_522(); break;
                                    case "Pacific/Pitcairn":
                                        zoneInfo = InitializeZonesPacificPitcairn_523(); break;
                                    case "Pacific/Ponape":
                                        zoneInfo = InitializeZonesPacificPonape_524(); break;
                                    case "Pacific/Port_Moresby":
                                        zoneInfo = InitializeZonesPacificPort_Moresby_525(); break;
                                    case "Pacific/Rarotonga":
                                        zoneInfo = InitializeZonesPacificRarotonga_526(); break;
                                    case "Pacific/Saipan":
                                        zoneInfo = InitializeZonesPacificSaipan_527(); break;
                                    case "Pacific/Samoa":
                                        zoneInfo = InitializeZonesPacificSamoa_528(); break;
                                    case "Pacific/Tahiti":
                                        zoneInfo = InitializeZonesPacificTahiti_529(); break;
                                    case "Pacific/Tarawa":
                                        zoneInfo = InitializeZonesPacificTarawa_530(); break;
                                    case "Pacific/Tongatapu":
                                        zoneInfo = InitializeZonesPacificTongatapu_531(); break;
                                    case "Pacific/Truk":
                                        zoneInfo = InitializeZonesPacificTruk_532(); break;
                                    case "Pacific/Wake":
                                        zoneInfo = InitializeZonesPacificWake_533(); break;
                                    case "Pacific/Wallis":
                                        zoneInfo = InitializeZonesPacificWallis_534(); break;
                                    case "Pacific/Yap":
                                        zoneInfo = InitializeZonesPacificYap_535(); break;
                                    case "Poland":
                                        zoneInfo = InitializeZonesPoland_536(); break;
                                    case "Portugal":
                                        zoneInfo = InitializeZonesPortugal_537(); break;
                                    case "PRC":
                                        zoneInfo = InitializeZonesPRC_538(); break;
                                    case "PST8PDT":
                                        zoneInfo = InitializeZonesPST8PDT_539(); break;
                                    case "ROC":
                                        zoneInfo = InitializeZonesROC_540(); break;
                                    case "ROK":
                                        zoneInfo = InitializeZonesROK_541(); break;
                                    case "Singapore":
                                        zoneInfo = InitializeZonesSingapore_542(); break;
                                    case "Turkey":
                                        zoneInfo = InitializeZonesTurkey_543(); break;
                                    case "UCT":
                                        zoneInfo = InitializeZonesUCT_544(); break;
                                    case "Universal":
                                        zoneInfo = InitializeZonesUniversal_545(); break;
                                    case "US/Alaska":
                                        zoneInfo = InitializeZonesUSAlaska_546(); break;
                                    case "US/Aleutian":
                                        zoneInfo = InitializeZonesUSAleutian_547(); break;
                                    case "US/Arizona":
                                        zoneInfo = InitializeZonesUSArizona_548(); break;
                                    case "US/Central":
                                        zoneInfo = InitializeZonesUSCentral_549(); break;
                                    case "US/Eastern":
                                        zoneInfo = InitializeZonesUSEastern_550(); break;
                                    case "US/East-Indiana":
                                        zoneInfo = InitializeZonesUSEastIndiana_551(); break;
                                    case "US/Hawaii":
                                        zoneInfo = InitializeZonesUSHawaii_552(); break;
                                    case "US/Indiana-Starke":
                                        zoneInfo = InitializeZonesUSIndianaStarke_553(); break;
                                    case "US/Michigan":
                                        zoneInfo = InitializeZonesUSMichigan_554(); break;
                                    case "US/Mountain":
                                        zoneInfo = InitializeZonesUSMountain_555(); break;
                                    case "US/Pacific":
                                        zoneInfo = InitializeZonesUSPacific_556(); break;
                                    case "US/Pacific-New":
                                        zoneInfo = InitializeZonesUSPacificNew_557(); break;
                                    case "US/Samoa":
                                        zoneInfo = InitializeZonesUSSamoa_558(); break;
                                    case "UTC":
                                        zoneInfo = InitializeZonesUTC_559(); break;
                                    case "WET":
                                        zoneInfo = InitializeZonesWET_560(); break;
                                    case "W-SU":
                                        zoneInfo = InitializeZonesWSU_561(); break;
                                    case "Zulu":
                                        zoneInfo = InitializeZonesZulu_562(); break;

                                    #endregion
                                }

                                if (zoneInfo != null)
                                {
                                    // Found the time zone
                                    result = new TzTimeZone(zoneInfo);
                                    s_zones[tzName] = zoneInfo;
                                }
                            }
                        }
                        finally
                        {
                            s_zonesLock.DowngradeFromWriterLock(ref writerCookie);
                        }
                    }
                }
                finally
                {
                    s_zonesLock.ReleaseReaderLock();
                }
            }

            return result;
        }

        /// <summary>
        /// Gets the time zone by offset.
        /// </summary>
        /// <param name="utcOffsetTime">The utc offset time.</param>
        /// <returns></returns>
        public static TzTimeZone GetTimeZoneByOffset(string utcOffsetTime)
        {
            return GetTimeZoneByOffset(DateTimeUtlities.ParseTimeSpan(utcOffsetTime, DateTimeUtlities.TimeSpanAssumption.Hours));
        }

        /// <summary>
        /// Gets the time zone by offset.
        /// </summary>
        /// <param name="utcOffsetTime">The utc offset time.</param>
        /// <returns></returns>
        public static TzTimeZone GetTimeZoneByOffset(TimeSpan utcOffsetTime)
        {
            if (utcOffsetTime.Equals(TimeSpan.Zero))
            {
                return TzTimeZone.ZoneUTC;
            }
            else
            {
                // We pick the most "well-known" time zone in the set for each offset
                string zoneName;
                if (s_mainTimeZones.TryGetValue(DateTimeUtlities.ConvertTimeSpanToDouble(utcOffsetTime), out zoneName))
                {
                    return GetTimeZone(zoneName);
                }
                else
                {
                    // TODO do a manual search
                }
            }
            throw new TzDatabase.TzException("Cannot find time zone with UTC offset {0}.", utcOffsetTime);
        }

        /// <summary>
        /// Gets the time zone of the current computer system.
        /// </summary>
        /// <value>A System.TimeZone instance representing the current, local time zone.</value>
        public static new TzTimeZone CurrentTimeZone
        {
            get
            {
                return s_currentTimeZone;
            }
        }

        /// <summary>
        ///     Returns a value indicating whether the specified date and time is within
        ///     the specified daylight saving time period.
        /// </summary>
        /// <param name="time">A date and time.</param>
        /// <param name="daylightTimes">A daylight saving time period.</param>
        /// <returns>
        /// 	true if time is in daylightTimes; otherwise, false.
        /// </returns>
        public static new bool IsDaylightSavingTime(DateTime time, DaylightTime daylightTimes)
        {
            return TimeZone.IsDaylightSavingTime(time, daylightTimes);
        }

        /// <summary>
        /// Determines whether the specified <see cref="T:System.Object"></see> is equal to the current <see cref="T:System.Object"></see>.
        /// </summary>
        /// <param name="obj">The <see cref="T:System.Object"></see> to compare with the current <see cref="T:System.Object"></see>.</param>
        /// <returns>
        /// true if the specified <see cref="T:System.Object"></see> is equal to the current <see cref="T:System.Object"></see>; otherwise, false.
        /// </returns>
        public override bool Equals(object obj)
        {
            TzTimeZone cmp = obj as TzTimeZone;
            if (cmp != null)
            {
                return StandardName.Equals(cmp.StandardName);
            }
            return base.Equals(obj);
        }

        /// <summary>
        /// Serves as a hash function for a particular type. <see cref="M:System.Object.GetHashCode"></see> is suitable for use in hashing algorithms and data structures like a hash table.
        /// </summary>
        /// <returns>
        /// A hash code for the current <see cref="T:System.Object"></see>.
        /// </returns>
        public override int GetHashCode()
        {
            return StandardName.GetHashCode();
        }

        #region All time zones

        private static void InitializeAllZones()
        {
            string[] allZones = new string[] { "Africa/Abidjan", "Africa/Accra", "Africa/Addis_Ababa", "Africa/Algiers", "Africa/Asmara", "Africa/Asmera", "Africa/Bamako", "Africa/Bangui", "Africa/Banjul", "Africa/Bissau", "Africa/Blantyre", "Africa/Brazzaville", "Africa/Bujumbura", "Africa/Cairo", "Africa/Casablanca", "Africa/Ceuta", "Africa/Conakry", "Africa/Dakar", "Africa/Dar_es_Salaam", "Africa/Djibouti", "Africa/Douala", "Africa/El_Aaiun", "Africa/Freetown", "Africa/Gaborone", "Africa/Harare", "Africa/Johannesburg", "Africa/Kampala", "Africa/Khartoum", "Africa/Kigali", "Africa/Kinshasa", "Africa/Lagos", "Africa/Libreville", "Africa/Lome", "Africa/Luanda", "Africa/Lubumbashi", "Africa/Lusaka", "Africa/Malabo", "Africa/Maputo", "Africa/Maseru", "Africa/Mbabane", "Africa/Mogadishu", "Africa/Monrovia", "Africa/Nairobi", "Africa/Ndjamena", "Africa/Niamey", "Africa/Nouakchott", "Africa/Ouagadougou", "Africa/Porto-Novo", "Africa/Sao_Tome", "Africa/Timbuktu", "Africa/Tripoli", "Africa/Tunis", "Africa/Windhoek", "America/Adak", "America/Anchorage", "America/Anguilla", "America/Antigua", "America/Araguaina", "America/Argentina/Buenos_Aires", "America/Argentina/Catamarca", "America/Argentina/ComodRivadavia", "America/Argentina/Cordoba", "America/Argentina/Jujuy", "America/Argentina/La_Rioja", "America/Argentina/Mendoza", "America/Argentina/Rio_Gallegos", "America/Argentina/San_Juan", "America/Argentina/San_Luis", "America/Argentina/Tucuman", "America/Argentina/Ushuaia", "America/Aruba", "America/Asuncion", "America/Atikokan", "America/Atka", "America/Bahia", "America/Barbados", "America/Belem", "America/Belize", "America/Blanc-Sablon", "America/Boa_Vista", "America/Bogota", "America/Boise", "America/Buenos_Aires", "America/Cambridge_Bay", "America/Campo_Grande", "America/Cancun", "America/Caracas", "America/Catamarca", "America/Cayenne", "America/Cayman", "America/Chicago", "America/Chihuahua", "America/Coral_Harbour", "America/Cordoba", "America/Costa_Rica", "America/Cuiaba", "America/Curacao", "America/Danmarkshavn", "America/Dawson", "America/Dawson_Creek", "America/Denver", "America/Detroit", "America/Dominica", "America/Edmonton", "America/Eirunepe", "America/El_Salvador", "America/Ensenada", "America/Fort_Wayne", "America/Fortaleza", "America/Glace_Bay", "America/Godthab", "America/Goose_Bay", "America/Grand_Turk", "America/Grenada", "America/Guadeloupe", "America/Guatemala", "America/Guayaquil", "America/Guyana", "America/Halifax", "America/Havana", "America/Hermosillo", "America/Indiana/Indianapolis", "America/Indiana/Knox", "America/Indiana/Marengo", "America/Indiana/Petersburg", "America/Indiana/Tell_City", "America/Indiana/Vevay", "America/Indiana/Vincennes", "America/Indiana/Winamac", "America/Indianapolis", "America/Inuvik", "America/Iqaluit", "America/Jamaica", "America/Jujuy", "America/Juneau", "America/Kentucky/Louisville", "America/Kentucky/Monticello", "America/Knox_IN", "America/La_Paz", "America/Lima", "America/Los_Angeles", "America/Louisville", "America/Maceio", "America/Managua", "America/Manaus", "America/Marigot", "America/Martinique", "America/Mazatlan", "America/Mendoza", "America/Menominee", "America/Merida", "America/Mexico_City", "America/Miquelon", "America/Moncton", "America/Monterrey", "America/Montevideo", "America/Montreal", "America/Montserrat", "America/Nassau", "America/New_York", "America/Nipigon", "America/Nome", "America/Noronha", "America/North_Dakota/Center", "America/North_Dakota/New_Salem", "America/Panama", "America/Pangnirtung", "America/Paramaribo", "America/Phoenix", "America/Port_of_Spain", "America/Port-au-Prince", "America/Porto_Acre", "America/Porto_Velho", "America/Puerto_Rico", "America/Rainy_River", "America/Rankin_Inlet", "America/Recife", "America/Regina", "America/Resolute", "America/Rio_Branco", "America/Rosario", "America/Santarem", "America/Santiago", "America/Santo_Domingo", "America/Sao_Paulo", "America/Scoresbysund", "America/Shiprock", "America/St_Barthelemy", "America/St_Johns", "America/St_Kitts", "America/St_Lucia", "America/St_Thomas", "America/St_Vincent", "America/Swift_Current", "America/Tegucigalpa", "America/Thule", "America/Thunder_Bay", "America/Tijuana", "America/Toronto", "America/Tortola", "America/Vancouver", "America/Virgin", "America/Whitehorse", "America/Winnipeg", "America/Yakutat", "America/Yellowknife", "Antarctica/Casey", "Antarctica/Davis", "Antarctica/DumontDUrville", "Antarctica/Mawson", "Antarctica/McMurdo", "Antarctica/Palmer", "Antarctica/Rothera", "Antarctica/South_Pole", "Antarctica/Syowa", "Antarctica/Vostok", "Arctic/Longyearbyen", "Asia/Aden", "Asia/Almaty", "Asia/Amman", "Asia/Anadyr", "Asia/Aqtau", "Asia/Aqtobe", "Asia/Ashgabat", "Asia/Ashkhabad", "Asia/Baghdad", "Asia/Bahrain", "Asia/Baku", "Asia/Bangkok", "Asia/Beirut", "Asia/Bishkek", "Asia/Brunei", "Asia/Calcutta", "Asia/Choibalsan", "Asia/Chongqing", "Asia/Chungking", "Asia/Colombo", "Asia/Dacca", "Asia/Damascus", "Asia/Dhaka", "Asia/Dili", "Asia/Dubai", "Asia/Dushanbe", "Asia/Gaza", "Asia/Harbin", "Asia/Ho_Chi_Minh", "Asia/Hong_Kong", "Asia/Hovd", "Asia/Irkutsk", "Asia/Istanbul", "Asia/Jakarta", "Asia/Jayapura", "Asia/Jerusalem", "Asia/Kabul", "Asia/Kamchatka", "Asia/Karachi", "Asia/Kashgar", "Asia/Katmandu", "Asia/Kolkata", "Asia/Krasnoyarsk", "Asia/Kuala_Lumpur", "Asia/Kuching", "Asia/Kuwait", "Asia/Macao", "Asia/Macau", "Asia/Magadan", "Asia/Makassar", "Asia/Manila", "Asia/Muscat", "Asia/Nicosia", "Asia/Novosibirsk", "Asia/Omsk", "Asia/Oral", "Asia/Phnom_Penh", "Asia/Pontianak", "Asia/Pyongyang", "Asia/Qatar", "Asia/Qyzylorda", "Asia/Rangoon", "Asia/Riyadh", "Asia/Riyadh87", "Asia/Riyadh88", "Asia/Riyadh89", "Asia/Saigon", "Asia/Sakhalin", "Asia/Samarkand", "Asia/Seoul", "Asia/Shanghai", "Asia/Singapore", "Asia/Taipei", "Asia/Tashkent", "Asia/Tbilisi", "Asia/Tehran", "Asia/Tel_Aviv", "Asia/Thimbu", "Asia/Thimphu", "Asia/Tokyo", "Asia/Ujung_Pandang", "Asia/Ulaanbaatar", "Asia/Ulan_Bator", "Asia/Urumqi", "Asia/Vientiane", "Asia/Vladivostok", "Asia/Yakutsk", "Asia/Yekaterinburg", "Asia/Yerevan", "Atlantic/Azores", "Atlantic/Bermuda", "Atlantic/Canary", "Atlantic/Cape_Verde", "Atlantic/Faeroe", "Atlantic/Faroe", "Atlantic/Jan_Mayen", "Atlantic/Madeira", "Atlantic/Reykjavik", "Atlantic/South_Georgia", "Atlantic/St_Helena", "Atlantic/Stanley", "Australia/ACT", "Australia/Adelaide", "Australia/Brisbane", "Australia/Broken_Hill", "Australia/Canberra", "Australia/Currie", "Australia/Darwin", "Australia/Eucla", "Australia/Hobart", "Australia/LHI", "Australia/Lindeman", "Australia/Lord_Howe", "Australia/Melbourne", "Australia/North", "Australia/NSW", "Australia/Perth", "Australia/Queensland", "Australia/South", "Australia/Sydney", "Australia/Tasmania", "Australia/Victoria", "Australia/West", "Australia/Yancowinna", "Brazil/Acre", "Brazil/DeNoronha", "Brazil/East", "Brazil/West", "Canada/Atlantic", "Canada/Central", "Canada/Eastern", "Canada/East-Saskatchewan", "Canada/Mountain", "Canada/Newfoundland", "Canada/Pacific", "Canada/Saskatchewan", "Canada/Yukon", "CET", "Chile/Continental", "Chile/EasterIsland", "CST6CDT", "Cuba", "EET", "Egypt", "Eire", "EST", "EST5EDT", "Etc/GMT", "Etc/GMT+0", "Etc/GMT+1", "Etc/GMT+10", "Etc/GMT+11", "Etc/GMT+12", "Etc/GMT+2", "Etc/GMT+3", "Etc/GMT+4", "Etc/GMT+5", "Etc/GMT+6", "Etc/GMT+7", "Etc/GMT+8", "Etc/GMT+9", "Etc/GMT0", "Etc/GMT-0", "Etc/GMT-1", "Etc/GMT-10", "Etc/GMT-11", "Etc/GMT-12", "Etc/GMT-13", "Etc/GMT-14", "Etc/GMT-2", "Etc/GMT-3", "Etc/GMT-4", "Etc/GMT-5", "Etc/GMT-6", "Etc/GMT-7", "Etc/GMT-8", "Etc/GMT-9", "Etc/Greenwich", "Etc/UCT", "Etc/Universal", "Etc/UTC", "Etc/Zulu", "Europe/Amsterdam", "Europe/Andorra", "Europe/Athens", "Europe/Belfast", "Europe/Belgrade", "Europe/Berlin", "Europe/Bratislava", "Europe/Brussels", "Europe/Bucharest", "Europe/Budapest", "Europe/Chisinau", "Europe/Copenhagen", "Europe/Dublin", "Europe/Gibraltar", "Europe/Guernsey", "Europe/Helsinki", "Europe/Isle_of_Man", "Europe/Istanbul", "Europe/Jersey", "Europe/Kaliningrad", "Europe/Kiev", "Europe/Lisbon", "Europe/Ljubljana", "Europe/London", "Europe/Luxembourg", "Europe/Madrid", "Europe/Malta", "Europe/Mariehamn", "Europe/Minsk", "Europe/Monaco", "Europe/Moscow", "Europe/Nicosia", "Europe/Oslo", "Europe/Paris", "Europe/Podgorica", "Europe/Prague", "Europe/Riga", "Europe/Rome", "Europe/Samara", "Europe/San_Marino", "Europe/Sarajevo", "Europe/Simferopol", "Europe/Skopje", "Europe/Sofia", "Europe/Stockholm", "Europe/Tallinn", "Europe/Tirane", "Europe/Tiraspol", "Europe/Uzhgorod", "Europe/Vaduz", "Europe/Vatican", "Europe/Vienna", "Europe/Vilnius", "Europe/Volgograd", "Europe/Warsaw", "Europe/Zagreb", "Europe/Zaporozhye", "Europe/Zurich", "GB", "GB-Eire", "GMT", "GMT+0", "GMT0", "GMT-0", "Greenwich", "Hongkong", "HST", "Iceland", "Indian/Antananarivo", "Indian/Chagos", "Indian/Christmas", "Indian/Cocos", "Indian/Comoro", "Indian/Kerguelen", "Indian/Mahe", "Indian/Maldives", "Indian/Mauritius", "Indian/Mayotte", "Indian/Reunion", "Iran", "Israel", "Jamaica", "Japan", "Kwajalein", "Libya", "MET", "Mexico/BajaNorte", "Mexico/BajaSur", "Mexico/General", "Mideast/Riyadh87", "Mideast/Riyadh88", "Mideast/Riyadh89", "MST", "MST7MDT", "Navajo", "NZ", "NZ-CHAT", "Pacific/Apia", "Pacific/Auckland", "Pacific/Chatham", "Pacific/Easter", "Pacific/Efate", "Pacific/Enderbury", "Pacific/Fakaofo", "Pacific/Fiji", "Pacific/Funafuti", "Pacific/Galapagos", "Pacific/Gambier", "Pacific/Guadalcanal", "Pacific/Guam", "Pacific/Honolulu", "Pacific/Johnston", "Pacific/Kiritimati", "Pacific/Kosrae", "Pacific/Kwajalein", "Pacific/Majuro", "Pacific/Marquesas", "Pacific/Midway", "Pacific/Nauru", "Pacific/Niue", "Pacific/Norfolk", "Pacific/Noumea", "Pacific/Pago_Pago", "Pacific/Palau", "Pacific/Pitcairn", "Pacific/Ponape", "Pacific/Port_Moresby", "Pacific/Rarotonga", "Pacific/Saipan", "Pacific/Samoa", "Pacific/Tahiti", "Pacific/Tarawa", "Pacific/Tongatapu", "Pacific/Truk", "Pacific/Wake", "Pacific/Wallis", "Pacific/Yap", "Poland", "Portugal", "PRC", "PST8PDT", "ROC", "ROK", "Singapore", "Turkey", "UCT", "Universal", "US/Alaska", "US/Aleutian", "US/Arizona", "US/Central", "US/Eastern", "US/East-Indiana", "US/Hawaii", "US/Indiana-Starke", "US/Michigan", "US/Mountain", "US/Pacific", "US/Pacific-New", "US/Samoa", "UTC", "WET", "W-SU", "Zulu" };
            foreach (string zone in allZones)
            {
                GetTimeZone(zone);
            }
        }

        #endregion

        #region Time Zone initializations
        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaAbidjan_0()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Abidjan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Abidjan", new TimeSpan(-9680000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Abidjan", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaAccra_1()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Accra", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Accra", new TimeSpan(-520000000), "-", "LMT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Accra", TimeSpan.Zero, "Ghana", "%s", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Ghana", 1936, 1942, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(12000000000), "GHST", null), new PublicDomain.TzDatabase.TzRule("Ghana", 1936, 1942, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "GMT", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaAddis_Ababa_2()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Addis_Ababa", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Addis_Ababa", new TimeSpan(92880000000), "-", "LMT", 1870, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Addis_Ababa", new TimeSpan(93200000000), "-", "ADMT", 1936, PublicDomain.Month.May, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Adis Dera MT"), new PublicDomain.TzDatabase.TzZone("Africa/Addis_Ababa", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaAlgiers_3()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Algiers", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Algiers", new TimeSpan(7320000000), "-", "LMT", 1891, PublicDomain.Month.March, 15, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Algiers", new TimeSpan(5610000000), "-", "PMT", 1911, PublicDomain.Month.March, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Paris Mean Time"), new PublicDomain.TzDatabase.TzZone("Africa/Algiers", TimeSpan.Zero, "Algeria", "WE%sT", 1940, PublicDomain.Month.February, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Algiers", new TimeSpan(36000000000), "Algeria", "CE%sT", 1946, PublicDomain.Month.October, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Algiers", TimeSpan.Zero, "-", "WET", 1956, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Algiers", new TimeSpan(36000000000), "-", "CET", 1963, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Algiers", TimeSpan.Zero, "Algeria", "WE%sT", 1977, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Algiers", new TimeSpan(36000000000), "Algeria", "CE%sT", 1979, PublicDomain.Month.October, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Algiers", TimeSpan.Zero, "Algeria", "WE%sT", 1981, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Algiers", new TimeSpan(36000000000), "-", "CET", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Algeria", 1916, 1916, PublicDomain.Month.June, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1916, 1919, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1917, 1917, PublicDomain.Month.March, 24, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1918, 1918, PublicDomain.Month.March, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1919, 1919, PublicDomain.Month.March, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1920, 1920, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1920, 1920, PublicDomain.Month.October, 23, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1921, 1921, PublicDomain.Month.March, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1921, 1921, PublicDomain.Month.June, 21, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1939, 1939, PublicDomain.Month.September, 11, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1939, 1939, PublicDomain.Month.November, 19, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1944, 1944, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1971, 1971, PublicDomain.Month.April, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1971, 1971, PublicDomain.Month.September, 26, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1977, 1977, PublicDomain.Month.May, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1977, 1977, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1978, 1978, PublicDomain.Month.March, 24, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1978, 1978, PublicDomain.Month.September, 22, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1980, 1980, PublicDomain.Month.April, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Algeria", 1980, 1980, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaAsmara_4()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Asmara", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Asmara", new TimeSpan(93320000000), "-", "LMT", 1870, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Asmara", new TimeSpan(93320000000), "-", "AMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Asmara Mean Time"), new PublicDomain.TzDatabase.TzZone("Africa/Asmara", new TimeSpan(93200000000), "-", "ADMT", 1936, PublicDomain.Month.May, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Adis Dera MT"), new PublicDomain.TzDatabase.TzZone("Africa/Asmara", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaAsmera_5()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Asmera", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Asmara", new TimeSpan(93320000000), "-", "LMT", 1870, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Asmara", new TimeSpan(93320000000), "-", "AMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Asmara Mean Time"), new PublicDomain.TzDatabase.TzZone("Africa/Asmara", new TimeSpan(93200000000), "-", "ADMT", 1936, PublicDomain.Month.May, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Adis Dera MT"), new PublicDomain.TzDatabase.TzZone("Africa/Asmara", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaBamako_6()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Bamako", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Bamako", new TimeSpan(-19200000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Bamako", TimeSpan.Zero, "-", "GMT", 1934, PublicDomain.Month.February, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Bamako", new TimeSpan(-36000000000), "-", "WAT", 1960, PublicDomain.Month.June, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Bamako", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaBangui_7()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Bangui", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Bangui", new TimeSpan(44600000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Bangui", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaBanjul_8()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Banjul", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Banjul", new TimeSpan(-39960000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Banjul", new TimeSpan(-39960000000), "-", "BMT", 1935, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Banjul Mean Time"), new PublicDomain.TzDatabase.TzZone("Africa/Banjul", new TimeSpan(-36000000000), "-", "WAT", 1964, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Banjul", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaBissau_9()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Bissau", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Bissau", new TimeSpan(-37400000000), "-", "LMT", 1911, PublicDomain.Month.May, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Bissau", new TimeSpan(-36000000000), "-", "WAT", 1975, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Bissau", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaBlantyre_10()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Blantyre", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Blantyre", new TimeSpan(84000000000), "-", "LMT", 1903, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Blantyre", new TimeSpan(72000000000), "-", "CAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaBrazzaville_11()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Brazzaville", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Brazzaville", new TimeSpan(36680000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Brazzaville", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaBujumbura_12()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Bujumbura", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Bujumbura", new TimeSpan(70480000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Bujumbura", new TimeSpan(72000000000), "-", "CAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaCairo_13()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Cairo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Cairo", new TimeSpan(75000000000), "-", "LMT", 1900, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Cairo", new TimeSpan(72000000000), "Egypt", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Egypt", 1940, 1940, PublicDomain.Month.July, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1940, 1940, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1941, 1941, PublicDomain.Month.April, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1941, 1941, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1942, 1944, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1942, 1942, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1943, 1945, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1945, 1945, PublicDomain.Month.April, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1957, 1957, PublicDomain.Month.May, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1957, 1958, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1958, 1958, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1959, 1981, PublicDomain.Month.May, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1959, 1965, PublicDomain.Month.September, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1966, 1994, PublicDomain.Month.October, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1982, 1982, PublicDomain.Month.July, 25, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1983, 1983, PublicDomain.Month.July, 12, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1984, 1988, PublicDomain.Month.May, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1989, 1989, PublicDomain.Month.May, 6, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1990, 1994, PublicDomain.Month.May, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1995, 2147483647, PublicDomain.Month.April, -1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1995, 2005, PublicDomain.Month.September, -1, DayOfWeek.Thursday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 2006, 2006, PublicDomain.Month.September, 21, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 2007, 2007, PublicDomain.Month.September, 1, DayOfWeek.Thursday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 2008, 2147483647, PublicDomain.Month.August, -1, DayOfWeek.Thursday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaCasablanca_14()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Casablanca", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Casablanca", new TimeSpan(-18200000000), "-", "LMT", 1913, PublicDomain.Month.October, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Casablanca", TimeSpan.Zero, "Morocco", "WE%sT", 1984, PublicDomain.Month.March, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Casablanca", new TimeSpan(36000000000), "-", "CET", 1986, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Casablanca", TimeSpan.Zero, "Morocco", "WE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Morocco", 1939, 1939, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1939, 1939, PublicDomain.Month.November, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1940, 1940, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1945, 1945, PublicDomain.Month.November, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1950, 1950, PublicDomain.Month.June, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1950, 1950, PublicDomain.Month.October, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1967, 1967, PublicDomain.Month.June, 3, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1967, 1967, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1974, 1974, PublicDomain.Month.June, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1974, 1974, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1976, 1977, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1976, 1976, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1977, 1977, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1978, 1978, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Morocco", 1978, 1978, PublicDomain.Month.August, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Morocco", 2008, 2008, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Morocco", 2008, 2008, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaCeuta_15()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Ceuta", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Ceuta", new TimeSpan(-12760000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ceuta", TimeSpan.Zero, "-", "WET", 1918, PublicDomain.Month.May, 6, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ceuta", TimeSpan.Zero, "1:00", "WEST", 1918, PublicDomain.Month.October, 7, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ceuta", TimeSpan.Zero, "-", "WET", 1924, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ceuta", TimeSpan.Zero, "Spain", "WE%sT", 1929, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ceuta", TimeSpan.Zero, "SpainAfrica", "WE%sT", 1984, PublicDomain.Month.March, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ceuta", new TimeSpan(36000000000), "-", "CET", 1986, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ceuta", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Spain", 1917, 1917, PublicDomain.Month.May, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1917, 1919, PublicDomain.Month.October, 6, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1918, 1918, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1919, 1919, PublicDomain.Month.April, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1924, 1924, PublicDomain.Month.April, 16, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1924, 1924, PublicDomain.Month.October, 4, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1926, 1929, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1929, 1929, PublicDomain.Month.April, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1937, 1937, PublicDomain.Month.May, 22, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1937, 1939, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1938, 1938, PublicDomain.Month.March, 22, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1939, 1939, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1940, 1940, PublicDomain.Month.March, 16, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1942, 1942, PublicDomain.Month.May, 2, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Spain", 1942, 1942, PublicDomain.Month.September, 1, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1943, 1946, PublicDomain.Month.April, 13, DayOfWeek.Saturday, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Spain", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1944, 1944, PublicDomain.Month.October, 10, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1946, 1946, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1949, 1949, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1949, 1949, PublicDomain.Month.September, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("SpainAfrica", 1967, 1967, PublicDomain.Month.June, 3, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("SpainAfrica", 1967, 1967, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1974, 1975, PublicDomain.Month.April, 13, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("SpainAfrica", 1974, 1974, PublicDomain.Month.June, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("SpainAfrica", 1974, 1974, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1974, 1975, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1976, 1976, PublicDomain.Month.March, 27, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("SpainAfrica", 1976, 1977, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("SpainAfrica", 1976, 1976, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1976, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1977, 1978, PublicDomain.Month.April, 2, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("SpainAfrica", 1977, 1977, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("SpainAfrica", 1978, 1978, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("SpainAfrica", 1978, 1978, PublicDomain.Month.August, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaConakry_16()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Conakry", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Conakry", new TimeSpan(-32920000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Conakry", TimeSpan.Zero, "-", "GMT", 1934, PublicDomain.Month.February, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Conakry", new TimeSpan(-36000000000), "-", "WAT", 1960, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Conakry", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaDakar_17()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Dakar", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Dakar", new TimeSpan(-41840000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Dakar", new TimeSpan(-36000000000), "-", "WAT", 1941, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Dakar", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaDar_es_Salaam_18()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Dar_es_Salaam", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Dar_es_Salaam", new TimeSpan(94280000000), "-", "LMT", 1931, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Dar_es_Salaam", new TimeSpan(108000000000), "-", "EAT", 1948, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Dar_es_Salaam", new TimeSpan(98850000000), "-", "BEAUT", 1961, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Dar_es_Salaam", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaDjibouti_19()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Djibouti", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Djibouti", new TimeSpan(103560000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Djibouti", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaDouala_20()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Douala", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Douala", new TimeSpan(23280000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Douala", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaEl_Aaiun_21()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/El_Aaiun", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/El_Aaiun", new TimeSpan(-31680000000), "-", "LMT", 1934, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/El_Aaiun", new TimeSpan(-36000000000), "-", "WAT", 1976, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/El_Aaiun", TimeSpan.Zero, "-", "WET", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaFreetown_22()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Freetown", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Freetown", new TimeSpan(-31800000000), "-", "LMT", 1882, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Freetown", new TimeSpan(-31800000000), "-", "FMT", 1913, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Freetown Mean Time"), new PublicDomain.TzDatabase.TzZone("Africa/Freetown", new TimeSpan(-36000000000), "SL", "%s", 1957, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Freetown", TimeSpan.Zero, "SL", "%s", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("SL", 1935, 1942, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(24000000000), "SLST", null), new PublicDomain.TzDatabase.TzRule("SL", 1935, 1942, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "WAT", null), new PublicDomain.TzDatabase.TzRule("SL", 1957, 1962, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "SLST", null), new PublicDomain.TzDatabase.TzRule("SL", 1957, 1962, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "GMT", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaGaborone_23()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Gaborone", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Gaborone", new TimeSpan(62200000000), "-", "LMT", 1885, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Gaborone", new TimeSpan(72000000000), "-", "CAT", 1943, PublicDomain.Month.September, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Gaborone", new TimeSpan(72000000000), "1:00", "CAST", 1944, PublicDomain.Month.March, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Gaborone", new TimeSpan(72000000000), "-", "CAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaHarare_24()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Harare", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Harare", new TimeSpan(74520000000), "-", "LMT", 1903, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Harare", new TimeSpan(72000000000), "-", "CAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaJohannesburg_25()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Johannesburg", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Johannesburg", new TimeSpan(67200000000), "-", "LMT", 1892, PublicDomain.Month.February, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Johannesburg", new TimeSpan(54000000000), "-", "SAST", 1903, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Johannesburg", new TimeSpan(72000000000), "SA", "SAST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("SA", 1942, 1943, PublicDomain.Month.September, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("SA", 1943, 1944, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaKampala_26()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Kampala", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Kampala", new TimeSpan(77800000000), "-", "LMT", 1928, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Kampala", new TimeSpan(108000000000), "-", "EAT", 1930, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Kampala", new TimeSpan(90000000000), "-", "BEAT", 1948, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Kampala", new TimeSpan(98850000000), "-", "BEAUT", 1957, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Kampala", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaKhartoum_27()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Khartoum", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Khartoum", new TimeSpan(78080000000), "-", "LMT", 1931, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Khartoum", new TimeSpan(72000000000), "Sudan", "CA%sT", 2000, PublicDomain.Month.January, 15, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Khartoum", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Sudan", 1970, 1970, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Sudan", 1970, 1985, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Sudan", 1971, 1971, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Sudan", 1972, 1985, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaKigali_28()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Kigali", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Kigali", new TimeSpan(72160000000), "-", "LMT", 1935, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Kigali", new TimeSpan(72000000000), "-", "CAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaKinshasa_29()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Kinshasa", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Kinshasa", new TimeSpan(36720000000), "-", "LMT", 1897, PublicDomain.Month.November, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Kinshasa", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaLagos_30()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Lagos", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Lagos", new TimeSpan(8160000000), "-", "LMT", 1919, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Lagos", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaLibreville_31()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Libreville", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Libreville", new TimeSpan(22680000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Libreville", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaLome_32()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Lome", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Lome", new TimeSpan(2920000000), "-", "LMT", 1893, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Lome", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaLuanda_33()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Luanda", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Luanda", new TimeSpan(31760000000), "-", "LMT", 1892, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Luanda", new TimeSpan(31240000000), "-", "AOT", 1911, PublicDomain.Month.May, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Angola Time"), new PublicDomain.TzDatabase.TzZone("Africa/Luanda", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaLubumbashi_34()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Lubumbashi", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Lubumbashi", new TimeSpan(65920000000), "-", "LMT", 1897, PublicDomain.Month.November, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Lubumbashi", new TimeSpan(72000000000), "-", "CAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaLusaka_35()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Lusaka", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Lusaka", new TimeSpan(67880000000), "-", "LMT", 1903, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Lusaka", new TimeSpan(72000000000), "-", "CAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaMalabo_36()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Malabo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Malabo", new TimeSpan(21080000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Malabo", TimeSpan.Zero, "-", "GMT", 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Malabo", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaMaputo_37()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Maputo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Maputo", new TimeSpan(78200000000), "-", "LMT", 1903, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Maputo", new TimeSpan(72000000000), "-", "CAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaMaseru_38()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Maseru", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Maseru", new TimeSpan(66000000000), "-", "LMT", 1903, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Maseru", new TimeSpan(72000000000), "-", "SAST", 1943, PublicDomain.Month.September, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Maseru", new TimeSpan(72000000000), "1:00", "SAST", 1944, PublicDomain.Month.March, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Maseru", new TimeSpan(72000000000), "-", "SAST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaMbabane_39()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Mbabane", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Mbabane", new TimeSpan(74640000000), "-", "LMT", 1903, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Mbabane", new TimeSpan(72000000000), "-", "SAST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaMogadishu_40()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Mogadishu", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Mogadishu", new TimeSpan(108880000000), "-", "LMT", 1893, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Mogadishu", new TimeSpan(108000000000), "-", "EAT", 1931, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Mogadishu", new TimeSpan(90000000000), "-", "BEAT", 1957, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Mogadishu", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaMonrovia_41()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Monrovia", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Monrovia", new TimeSpan(-25880000000), "-", "LMT", 1882, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Monrovia", new TimeSpan(-25880000000), "-", "MMT", 1919, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Monrovia Mean Time"), new PublicDomain.TzDatabase.TzZone("Africa/Monrovia", new TimeSpan(-26700000000), "-", "LRT", 1972, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Liberia Time"), new PublicDomain.TzDatabase.TzZone("Africa/Monrovia", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaNairobi_42()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Nairobi", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Nairobi", new TimeSpan(88360000000), "-", "LMT", 1928, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Nairobi", new TimeSpan(108000000000), "-", "EAT", 1930, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Nairobi", new TimeSpan(90000000000), "-", "BEAT", 1940, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Nairobi", new TimeSpan(98850000000), "-", "BEAUT", 1960, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Nairobi", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaNdjamena_43()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Ndjamena", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Ndjamena", new TimeSpan(36120000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ndjamena", new TimeSpan(36000000000), "-", "WAT", 1979, PublicDomain.Month.October, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ndjamena", new TimeSpan(36000000000), "1:00", "WAST", 1980, PublicDomain.Month.March, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ndjamena", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaNiamey_44()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Niamey", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Niamey", new TimeSpan(5080000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Niamey", new TimeSpan(-36000000000), "-", "WAT", 1934, PublicDomain.Month.February, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Niamey", TimeSpan.Zero, "-", "GMT", 1960, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Niamey", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaNouakchott_45()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Nouakchott", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Nouakchott", new TimeSpan(-38280000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Nouakchott", TimeSpan.Zero, "-", "GMT", 1934, PublicDomain.Month.February, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Nouakchott", new TimeSpan(-36000000000), "-", "WAT", 1960, PublicDomain.Month.November, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Nouakchott", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaOuagadougou_46()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Ouagadougou", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Ouagadougou", new TimeSpan(-3640000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Ouagadougou", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaPortoNovo_47()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Porto-Novo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Porto-Novo", new TimeSpan(6280000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Porto-Novo", TimeSpan.Zero, "-", "GMT", 1934, PublicDomain.Month.February, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Porto-Novo", new TimeSpan(36000000000), "-", "WAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaSao_Tome_48()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Sao_Tome", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Sao_Tome", new TimeSpan(16160000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Sao_Tome", new TimeSpan(-21920000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Lisbon Mean Time"), new PublicDomain.TzDatabase.TzZone("Africa/Sao_Tome", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaTimbuktu_49()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Timbuktu", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Bamako", new TimeSpan(-19200000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Bamako", TimeSpan.Zero, "-", "GMT", 1934, PublicDomain.Month.February, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Bamako", new TimeSpan(-36000000000), "-", "WAT", 1960, PublicDomain.Month.June, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Bamako", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaTripoli_50()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Tripoli", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(31640000000), "-", "LMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(36000000000), "Libya", "CE%sT", 1959, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(72000000000), "-", "EET", 1982, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(36000000000), "Libya", "CE%sT", 1990, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(72000000000), "-", "EET", 1996, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(36000000000), "-", "CET", 1997, PublicDomain.Month.April, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(36000000000), "1:00", "CEST", 1997, PublicDomain.Month.October, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(72000000000), "-", "EET", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Libya", 1951, 1951, PublicDomain.Month.October, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1952, 1952, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Libya", 1953, 1953, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1954, 1954, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Libya", 1955, 1955, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1956, 1956, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Libya", 1982, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1982, 1985, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Libya", 1985, 1985, PublicDomain.Month.April, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1986, 1986, PublicDomain.Month.April, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1986, 1986, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Libya", 1987, 1989, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1987, 1989, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaTunis_51()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Tunis", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Tunis", new TimeSpan(24440000000), "-", "LMT", 1881, PublicDomain.Month.May, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tunis", new TimeSpan(5610000000), "-", "PMT", 1911, PublicDomain.Month.March, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Paris Mean Time"), new PublicDomain.TzDatabase.TzZone("Africa/Tunis", new TimeSpan(36000000000), "Tunisia", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Tunisia", 1939, 1939, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1939, 1939, PublicDomain.Month.November, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1940, 1940, PublicDomain.Month.February, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1941, 1941, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1942, 1942, PublicDomain.Month.March, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1943, 1943, PublicDomain.Month.April, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1943, 1943, PublicDomain.Month.April, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1944, 1944, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1945, 1945, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1977, 1977, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1977, 1977, PublicDomain.Month.September, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1978, 1978, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1988, 1988, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1988, 1990, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1989, 1989, PublicDomain.Month.March, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 1990, 1990, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 2005, 2005, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 2005, 2005, PublicDomain.Month.September, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 2006, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tunisia", 2006, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAfricaWindhoek_52()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Africa/Windhoek", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Windhoek", new TimeSpan(41040000000), "-", "LMT", 1892, PublicDomain.Month.February, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Windhoek", new TimeSpan(54000000000), "-", "SWAT", 1903, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "SW Africa Time"), new PublicDomain.TzDatabase.TzZone("Africa/Windhoek", new TimeSpan(72000000000), "-", "SAST", 1942, PublicDomain.Month.September, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Windhoek", new TimeSpan(72000000000), "1:00", "SAST", 1943, PublicDomain.Month.March, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Windhoek", new TimeSpan(72000000000), "-", "SAST", 1990, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Africa/Windhoek", new TimeSpan(72000000000), "-", "CAT", 1994, PublicDomain.Month.April, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Windhoek", new TimeSpan(36000000000), "Namibia", "WA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Namibia", 1994, 2147483647, PublicDomain.Month.September, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Namibia", 1995, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaAdak_53()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Adak", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(440010000000), "-", "LMT", 1867, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-423980000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "-", "NST", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "US", "N%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "-", "NST", 1967, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "-", "BST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "US", "B%sT", 1983, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-360000000000), "US", "AH%sT", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-360000000000), "US", "HA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaAnchorage_54()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Anchorage", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(504240000000), "-", "LMT", 1867, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-359760000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "-", "CAT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "US", "CAT/CAWT", 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "US", "CAT/CAPT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Peace"), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "-", "CAT", 1967, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "-", "AHST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "US", "AH%sT", 1983, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-324000000000), "US", "Y%sT", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-324000000000), "US", "AK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaAnguilla_55()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Anguilla", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Anguilla", new TimeSpan(-151360000000), "-", "LMT", 1912, PublicDomain.Month.March, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anguilla", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaAntigua_56()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Antigua", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Antigua", new TimeSpan(-148320000000), "-", "LMT", 1912, PublicDomain.Month.March, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Antigua", new TimeSpan(-180000000000), "-", "EST", 1951, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Antigua", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaAraguaina_57()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Araguaina", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Araguaina", new TimeSpan(-115680000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Araguaina", new TimeSpan(-108000000000), "Brazil", "BR%sT", 1990, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Araguaina", new TimeSpan(-108000000000), "-", "BRT", 1995, PublicDomain.Month.September, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Araguaina", new TimeSpan(-108000000000), "Brazil", "BR%sT", 2003, PublicDomain.Month.September, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Araguaina", new TimeSpan(-108000000000), "-", "BRT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaBuenos_Aires_58()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/Buenos_Aires", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-140280000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Cordoba Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaCatamarca_59()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/Catamarca", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-157880000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "Arg", "AR%sT", 1991, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.June, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaComodRivadavia_60()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/ComodRivadavia", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-157880000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "Arg", "AR%sT", 1991, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.June, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaCordoba_61()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/Cordoba", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-154080000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-108000000000), "Arg", "AR%sT", 1991, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaJujuy_62()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/Jujuy", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-156720000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-108000000000), "Arg", "AR%sT", 1990, PublicDomain.Month.March, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "-", "WART", 1990, PublicDomain.Month.October, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "1:00", "WARST", 1991, PublicDomain.Month.March, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-108000000000), "1:00", "ARST", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaLa_Rioja_63()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/La_Rioja", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-160440000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-108000000000), "Arg", "AR%sT", 1991, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.May, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.June, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/La_Rioja", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaMendoza_64()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/Mendoza", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-165160000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-108000000000), "Arg", "AR%sT", 1990, PublicDomain.Month.March, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "-", "WART", 1990, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "1:00", "WARST", 1991, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "1:00", "WARST", 1992, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "-", "WART", 1992, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.May, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.September, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaRio_Gallegos_65()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/Rio_Gallegos", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Rio_Gallegos", new TimeSpan(-166120000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Rio_Gallegos", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Cordoba Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Argentina/Rio_Gallegos", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Rio_Gallegos", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Rio_Gallegos", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Rio_Gallegos", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Rio_Gallegos", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Rio_Gallegos", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.June, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Rio_Gallegos", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaSan_Juan_66()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/San_Juan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-164440000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-108000000000), "Arg", "AR%sT", 1991, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.May, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.May, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.July, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Juan", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaSan_Luis_67()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/San_Luis", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-159240000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-108000000000), "Arg", "AR%sT", 1990, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-108000000000), "1:00", "ARST", 1990, PublicDomain.Month.March, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-144000000000), "-", "WART", 1990, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-144000000000), "1:00", "WARST", 1991, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-108000000000), "-", "ART", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-144000000000), "1:00", "WARST", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.May, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.July, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-108000000000), "Arg", "AR%sT", 2008, PublicDomain.Month.January, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/San_Luis", new TimeSpan(-108000000000), "-", "ART", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaTucuman_68()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/Tucuman", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-156520000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-108000000000), "Arg", "AR%sT", 1991, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.June, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Tucuman", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaArgentinaUshuaia_69()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Argentina/Ushuaia", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Ushuaia", new TimeSpan(-163920000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Ushuaia", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Cordoba Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Argentina/Ushuaia", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Ushuaia", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Ushuaia", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Ushuaia", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Ushuaia", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.May, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Ushuaia", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.June, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Ushuaia", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaAruba_70()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Aruba", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Aruba", new TimeSpan(-168240000000), "-", "LMT", 1912, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Oranjestad"), new PublicDomain.TzDatabase.TzZone("America/Aruba", new TimeSpan(-162000000000), "-", "ANT", 1965, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Netherlands Antilles Time"), new PublicDomain.TzDatabase.TzZone("America/Aruba", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaAsuncion_71()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Asuncion", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Asuncion", new TimeSpan(-138400000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Asuncion", new TimeSpan(-138400000000), "-", "AMT", 1931, PublicDomain.Month.October, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Asuncion Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Asuncion", new TimeSpan(-144000000000), "-", "PYT", 1972, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Paraguay Time"), new PublicDomain.TzDatabase.TzZone("America/Asuncion", new TimeSpan(-108000000000), "-", "PYT", 1974, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Asuncion", new TimeSpan(-144000000000), "Para", "PY%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Para", 1975, 1978, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Para", 1975, 1988, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Para", 1979, 1991, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Para", 1989, 1989, PublicDomain.Month.October, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Para", 1990, 1990, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Para", 1991, 1991, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Para", 1992, 1992, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Para", 1992, 1992, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Para", 1993, 1993, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Para", 1993, 1995, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Para", 1994, 1995, PublicDomain.Month.February, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Para", 1996, 1996, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Para", 1996, 2001, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Para", 1997, 1997, PublicDomain.Month.February, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Para", 1998, 2001, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Para", 2002, 2004, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Para", 2002, 2003, PublicDomain.Month.September, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Para", 2004, 2147483647, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Para", 2005, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaAtikokan_72()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Atikokan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Atikokan", new TimeSpan(-219880000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Atikokan", new TimeSpan(-216000000000), "Canada", "C%sT", 1940, PublicDomain.Month.September, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Atikokan", new TimeSpan(-216000000000), "1:00", "CDT", 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("America/Atikokan", new TimeSpan(-216000000000), "Canada", "C%sT", 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Atikokan", new TimeSpan(-180000000000), "-", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaAtka_73()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Atka", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(440010000000), "-", "LMT", 1867, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-423980000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "-", "NST", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "US", "N%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "-", "NST", 1967, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "-", "BST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "US", "B%sT", 1983, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-360000000000), "US", "AH%sT", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-360000000000), "US", "HA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaBahia_74()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Bahia", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Bahia", new TimeSpan(-92440000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Bahia", new TimeSpan(-108000000000), "Brazil", "BR%sT", 2003, PublicDomain.Month.September, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Bahia", new TimeSpan(-108000000000), "-", "BRT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaBarbados_75()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Barbados", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Barbados", new TimeSpan(-143080000000), "-", "LMT", 1924, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bridgetown"), new PublicDomain.TzDatabase.TzZone("America/Barbados", new TimeSpan(-143080000000), "-", "BMT", 1932, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bridgetown Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Barbados", new TimeSpan(-144000000000), "Barb", "A%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Barb", 1977, 1977, PublicDomain.Month.June, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Barb", 1977, 1978, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Barb", 1978, 1980, PublicDomain.Month.April, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Barb", 1979, 1979, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Barb", 1980, 1980, PublicDomain.Month.September, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaBelem_76()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Belem", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Belem", new TimeSpan(-116360000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Belem", new TimeSpan(-108000000000), "Brazil", "BR%sT", 1988, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Belem", new TimeSpan(-108000000000), "-", "BRT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaBelize_77()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Belize", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Belize", new TimeSpan(-211680000000), "-", "LMT", 1912, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Belize", new TimeSpan(-216000000000), "Belize", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Belize", 1918, 1942, PublicDomain.Month.October, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HD", null), new PublicDomain.TzDatabase.TzRule("Belize", 1919, 1943, PublicDomain.Month.February, 9, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Belize", 1973, 1973, PublicDomain.Month.December, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Belize", 1974, 1974, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Belize", 1982, 1982, PublicDomain.Month.December, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Belize", 1983, 1983, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaBlancSablon_78()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Blanc-Sablon", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Blanc-Sablon", new TimeSpan(-137080000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Blanc-Sablon", new TimeSpan(-144000000000), "Canada", "A%sT", 1970, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Blanc-Sablon", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaBoa_Vista_79()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Boa_Vista", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Boa_Vista", new TimeSpan(-145600000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Boa_Vista", new TimeSpan(-144000000000), "Brazil", "AM%sT", 1988, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Boa_Vista", new TimeSpan(-144000000000), "-", "AMT", 1999, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Boa_Vista", new TimeSpan(-144000000000), "Brazil", "AM%sT", 2000, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Boa_Vista", new TimeSpan(-144000000000), "-", "AMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaBogota_80()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Bogota", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Bogota", new TimeSpan(-177800000000), "-", "LMT", 1884, PublicDomain.Month.March, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Bogota", new TimeSpan(-177800000000), "-", "BMT", 1914, PublicDomain.Month.November, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bogota Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Bogota", new TimeSpan(-180000000000), "CO", "CO%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Colombia Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("CO", 1992, 1992, PublicDomain.Month.May, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("CO", 1993, 1993, PublicDomain.Month.April, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaBoise_81()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Boise", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Boise", new TimeSpan(-278890000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(441110000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Boise", new TimeSpan(-288000000000), "US", "P%sT", 1923, PublicDomain.Month.May, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Boise", new TimeSpan(-252000000000), "US", "M%sT", 1974, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Boise", new TimeSpan(-252000000000), "-", "MST", 1974, PublicDomain.Month.February, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Boise", new TimeSpan(-252000000000), "US", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaBuenos_Aires_82()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Buenos_Aires", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-140280000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Cordoba Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Buenos_Aires", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCambridge_Bay_83()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Cambridge_Bay", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Cambridge_Bay", TimeSpan.Zero, "-", "zzz", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "trading post est.?"), new PublicDomain.TzDatabase.TzZone("America/Cambridge_Bay", new TimeSpan(-252000000000), "NT_YK", "M%sT", 1999, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cambridge_Bay", new TimeSpan(-216000000000), "Canada", "C%sT", 2000, PublicDomain.Month.October, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cambridge_Bay", new TimeSpan(-180000000000), "-", "EST", 2000, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cambridge_Bay", new TimeSpan(-216000000000), "-", "CST", 2001, PublicDomain.Month.April, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cambridge_Bay", new TimeSpan(-252000000000), "Canada", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.May, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCampo_Grande_84()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Campo_Grande", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Campo_Grande", new TimeSpan(-131080000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Campo_Grande", new TimeSpan(-144000000000), "Brazil", "AM%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCancun_85()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Cancun", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Cancun", new TimeSpan(-208240000000), "-", "LMT", 1922, PublicDomain.Month.January, 1, null, new TimeSpan(7760000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cancun", new TimeSpan(-216000000000), "-", "CST", 1981, PublicDomain.Month.December, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cancun", new TimeSpan(-180000000000), "Mexico", "E%sT", 1998, PublicDomain.Month.August, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cancun", new TimeSpan(-216000000000), "Mexico", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCaracas_86()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Caracas", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Caracas", new TimeSpan(-160640000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Caracas", new TimeSpan(-160600000000), "-", "CMT", 1912, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Caracas Mean Time?"), new PublicDomain.TzDatabase.TzZone("America/Caracas", new TimeSpan(-162000000000), "-", "VET", 1965, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Venezuela Time"), new PublicDomain.TzDatabase.TzZone("America/Caracas", new TimeSpan(-144000000000), "-", "VET", 2007, PublicDomain.Month.December, 9, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Caracas", new TimeSpan(-162000000000), "-", "VET", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCatamarca_87()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Catamarca", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-157880000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "Arg", "AR%sT", 1991, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.June, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Catamarca", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCayenne_88()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Cayenne", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Cayenne", new TimeSpan(-125600000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cayenne", new TimeSpan(-144000000000), "-", "GFT", 1967, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "French Guiana Time"), new PublicDomain.TzDatabase.TzZone("America/Cayenne", new TimeSpan(-108000000000), "-", "GFT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCayman_89()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Cayman", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Cayman", new TimeSpan(-195320000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Georgetown"), new PublicDomain.TzDatabase.TzZone("America/Cayman", new TimeSpan(-184320000000), "-", "KMT", 1912, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kingston Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Cayman", new TimeSpan(-180000000000), "-", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaChicago_90()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Chicago", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-210360000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(437640000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "US", "C%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "Chicago", "C%sT", 1936, PublicDomain.Month.March, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-180000000000), "-", "EST", 1936, PublicDomain.Month.November, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "Chicago", "C%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "Chicago", "C%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "US", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1920, 1920, PublicDomain.Month.June, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1920, 1921, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1921, 1921, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1922, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1922, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1955, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaChihuahua_91()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Chihuahua", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Chihuahua", new TimeSpan(-254600000000), "-", "LMT", 1921, PublicDomain.Month.December, 31, null, new TimeSpan(861400000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chihuahua", new TimeSpan(-252000000000), "-", "MST", 1927, PublicDomain.Month.June, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chihuahua", new TimeSpan(-216000000000), "-", "CST", 1930, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chihuahua", new TimeSpan(-252000000000), "-", "MST", 1931, PublicDomain.Month.May, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chihuahua", new TimeSpan(-216000000000), "-", "CST", 1931, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chihuahua", new TimeSpan(-252000000000), "-", "MST", 1932, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chihuahua", new TimeSpan(-216000000000), "-", "CST", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chihuahua", new TimeSpan(-216000000000), "Mexico", "C%sT", 1998, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chihuahua", new TimeSpan(-216000000000), "-", "CST", 1998, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chihuahua", new TimeSpan(-252000000000), "Mexico", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCoral_Harbour_92()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Coral_Harbour", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Atikokan", new TimeSpan(-219880000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Atikokan", new TimeSpan(-216000000000), "Canada", "C%sT", 1940, PublicDomain.Month.September, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Atikokan", new TimeSpan(-216000000000), "1:00", "CDT", 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("America/Atikokan", new TimeSpan(-216000000000), "Canada", "C%sT", 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Atikokan", new TimeSpan(-180000000000), "-", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCordoba_93()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Cordoba", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-154080000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-108000000000), "Arg", "AR%sT", 1991, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCosta_Rica_94()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Costa_Rica", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Costa_Rica", new TimeSpan(-201800000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "San Jose"), new PublicDomain.TzDatabase.TzZone("America/Costa_Rica", new TimeSpan(-201800000000), "-", "SJMT", 1921, PublicDomain.Month.January, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "San Jose Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Costa_Rica", new TimeSpan(-216000000000), "CR", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("CR", 1979, 1980, PublicDomain.Month.February, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CR", 1979, 1980, PublicDomain.Month.June, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CR", 1991, 1992, PublicDomain.Month.January, 15, DayOfWeek.Saturday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CR", 1991, 1991, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CR", 1992, 1992, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCuiaba_95()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Cuiaba", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Cuiaba", new TimeSpan(-134600000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cuiaba", new TimeSpan(-144000000000), "Brazil", "AM%sT", 2003, PublicDomain.Month.September, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cuiaba", new TimeSpan(-144000000000), "-", "AMT", 2004, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Cuiaba", new TimeSpan(-144000000000), "Brazil", "AM%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaCuracao_96()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Curacao", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Curacao", new TimeSpan(-165440000000), "-", "LMT", 1912, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Willemstad"), new PublicDomain.TzDatabase.TzZone("America/Curacao", new TimeSpan(-162000000000), "-", "ANT", 1965, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Netherlands Antilles Time"), new PublicDomain.TzDatabase.TzZone("America/Curacao", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaDanmarkshavn_97()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Danmarkshavn", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Danmarkshavn", new TimeSpan(-44800000000), "-", "LMT", 1916, PublicDomain.Month.July, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Danmarkshavn", new TimeSpan(-108000000000), "-", "WGT", 1980, PublicDomain.Month.April, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Danmarkshavn", new TimeSpan(-108000000000), "EU", "WG%sT", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Danmarkshavn", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaDawson_98()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Dawson", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Dawson", new TimeSpan(-334600000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Dawson", new TimeSpan(-324000000000), "NT_YK", "Y%sT", 1973, PublicDomain.Month.October, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Dawson", new TimeSpan(-288000000000), "NT_YK", "P%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Dawson", new TimeSpan(-288000000000), "Canada", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.May, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaDawson_Creek_99()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Dawson_Creek", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Dawson_Creek", new TimeSpan(-288560000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Dawson_Creek", new TimeSpan(-288000000000), "Canada", "P%sT", 1947, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Dawson_Creek", new TimeSpan(-288000000000), "Vanc", "P%sT", 1972, PublicDomain.Month.August, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Dawson_Creek", new TimeSpan(-252000000000), "-", "MST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Vanc", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Vanc", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Vanc", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1946, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1946, 1946, PublicDomain.Month.October, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1947, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1962, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaDenver_100()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Denver", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-251960000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(432040000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "Denver", "M%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "Denver", "M%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1920, 1921, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Denver", 1920, 1920, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1921, 1921, PublicDomain.Month.May, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1965, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Denver", 1965, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaDetroit_101()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Detroit", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-199310000000), "-", "LMT", 1905, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-216000000000), "-", "CST", 1915, PublicDomain.Month.May, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "-", "EST", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "US", "E%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "Detroit", "E%sT", 1973, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "US", "E%sT", 1975, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "-", "EST", 1975, PublicDomain.Month.April, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Detroit", 1948, 1948, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Detroit", 1948, 1948, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Detroit", 1967, 1967, PublicDomain.Month.June, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Detroit", 1967, 1967, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaDominica_102()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Dominica", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Dominica", new TimeSpan(-147360000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, "Roseau"), new PublicDomain.TzDatabase.TzZone("America/Dominica", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaEdmonton_103()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Edmonton", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Edmonton", new TimeSpan(-272320000000), "-", "LMT", 1906, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Edmonton", new TimeSpan(-252000000000), "Edm", "M%sT", 1987, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Edmonton", new TimeSpan(-252000000000), "Canada", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Edm", 1918, 1919, PublicDomain.Month.April, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1919, 1919, PublicDomain.Month.May, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1920, 1923, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1920, 1920, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1921, 1923, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Edm", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1945, 1945, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1947, 1947, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1947, 1947, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1967, 1967, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1967, 1967, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1969, 1969, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1969, 1969, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1972, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1972, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaEirunepe_104()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Eirunepe", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Eirunepe", new TimeSpan(-167680000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Eirunepe", new TimeSpan(-180000000000), "Brazil", "AC%sT", 1988, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Eirunepe", new TimeSpan(-180000000000), "-", "ACT", 1993, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Eirunepe", new TimeSpan(-180000000000), "Brazil", "AC%sT", 1994, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Eirunepe", new TimeSpan(-180000000000), "-", "ACT", 2008, PublicDomain.Month.June, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Eirunepe", new TimeSpan(-144000000000), "-", "AMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaEl_Salvador_105()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/El_Salvador", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/El_Salvador", new TimeSpan(-214080000000), "-", "LMT", 1921, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "San Salvador"), new PublicDomain.TzDatabase.TzZone("America/El_Salvador", new TimeSpan(-216000000000), "Salv", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Salv", 1987, 1988, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Salv", 1987, 1988, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaEnsenada_106()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Ensenada", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-280840000000), "-", "LMT", 1922, PublicDomain.Month.January, 1, null, new TimeSpan(7160000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-252000000000), "-", "MST", 1924, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1927, PublicDomain.Month.June, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-252000000000), "-", "MST", 1930, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PDT", 1931, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1942, PublicDomain.Month.April, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PWT", 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PPT", 1945, PublicDomain.Month.November, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Peace"), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1948, PublicDomain.Month.April, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PDT", 1949, PublicDomain.Month.January, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1954, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "CA", "P%sT", 1961, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1976, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "US", "P%sT", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "Mexico", "P%sT", 2001, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "US", "P%sT", 2002, PublicDomain.Month.February, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "Mexico", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1949, 1949, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1962, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaFort_Wayne_107()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Fort_Wayne", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-206780000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(441220000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "US", "C%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "Indianapolis", "C%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "Indianapolis", "C%sT", 1955, PublicDomain.Month.April, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 1957, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "-", "CST", 1958, PublicDomain.Month.April, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "US", "E%sT", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1941, 1941, PublicDomain.Month.June, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1941, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1946, 1954, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaFortaleza_108()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Fortaleza", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Fortaleza", new TimeSpan(-92400000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Fortaleza", new TimeSpan(-108000000000), "Brazil", "BR%sT", 1990, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Fortaleza", new TimeSpan(-108000000000), "-", "BRT", 1999, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Fortaleza", new TimeSpan(-108000000000), "Brazil", "BR%sT", 2000, PublicDomain.Month.October, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Fortaleza", new TimeSpan(-108000000000), "-", "BRT", 2001, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Fortaleza", new TimeSpan(-108000000000), "Brazil", "BR%sT", 2002, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Fortaleza", new TimeSpan(-108000000000), "-", "BRT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaGlace_Bay_109()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Glace_Bay", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Glace_Bay", new TimeSpan(-143880000000), "-", "LMT", 1902, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Glace_Bay", new TimeSpan(-144000000000), "Canada", "A%sT", 1953, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Glace_Bay", new TimeSpan(-144000000000), "Halifax", "A%sT", 1954, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Glace_Bay", new TimeSpan(-144000000000), "-", "AST", 1972, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Glace_Bay", new TimeSpan(-144000000000), "Halifax", "A%sT", 1974, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Glace_Bay", new TimeSpan(-144000000000), "Canada", "A%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Halifax", 1916, 1916, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1920, 1920, PublicDomain.Month.May, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1920, 1920, PublicDomain.Month.August, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1921, 1921, PublicDomain.Month.May, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1921, 1922, PublicDomain.Month.September, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1922, 1922, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1923, 1925, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1923, 1923, PublicDomain.Month.September, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1924, 1924, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1925, 1925, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1926, 1926, PublicDomain.Month.May, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1926, 1926, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1927, 1927, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1927, 1927, PublicDomain.Month.September, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1928, 1931, PublicDomain.Month.May, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1928, 1928, PublicDomain.Month.September, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1929, 1929, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1930, 1930, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1931, 1932, PublicDomain.Month.September, 24, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1932, 1932, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1933, 1933, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1933, 1933, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1934, 1934, PublicDomain.Month.May, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1934, 1934, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1935, 1935, PublicDomain.Month.June, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1935, 1935, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1936, 1936, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1936, 1936, PublicDomain.Month.September, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1937, 1938, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1937, 1941, PublicDomain.Month.September, 24, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1939, 1939, PublicDomain.Month.May, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1940, 1941, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1946, 1949, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1946, 1949, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1951, 1954, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1951, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1956, 1959, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1956, 1959, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1962, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1962, 1973, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaGodthab_110()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Godthab", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Godthab", new TimeSpan(-124160000000), "-", "LMT", 1916, PublicDomain.Month.July, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Nuuk"), new PublicDomain.TzDatabase.TzZone("America/Godthab", new TimeSpan(-108000000000), "-", "WGT", 1980, PublicDomain.Month.April, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Godthab", new TimeSpan(-108000000000), "EU", "WG%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaGoose_Bay_111()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Goose_Bay", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Goose_Bay", new TimeSpan(-145000000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Happy Valley-Goose Bay"), new PublicDomain.TzDatabase.TzZone("America/Goose_Bay", new TimeSpan(-126520000000), "-", "NST", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Goose_Bay", new TimeSpan(-126520000000), "Canada", "N%sT", 1919, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Goose_Bay", new TimeSpan(-126520000000), "-", "NST", 1935, PublicDomain.Month.March, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Goose_Bay", new TimeSpan(-126000000000), "-", "NST", 1936, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Goose_Bay", new TimeSpan(-126000000000), "StJohns", "N%sT", 1942, PublicDomain.Month.May, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Goose_Bay", new TimeSpan(-126000000000), "Canada", "N%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Goose_Bay", new TimeSpan(-126000000000), "StJohns", "N%sT", 1966, PublicDomain.Month.March, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Goose_Bay", new TimeSpan(-144000000000), "StJohns", "A%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("StJohns", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1919, 1919, PublicDomain.Month.May, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1919, 1919, PublicDomain.Month.August, 12, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1920, 1935, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1920, 1935, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1936, 1941, PublicDomain.Month.May, 9, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1936, 1941, PublicDomain.Month.October, 2, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1946, 1950, PublicDomain.Month.May, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1946, 1950, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1951, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1951, 1959, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1960, 1986, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1987, 1987, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1987, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1988, 1988, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1989, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaGrand_Turk_112()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Grand_Turk", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Grand_Turk", new TimeSpan(-170720000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Grand_Turk", new TimeSpan(-184320000000), "-", "KMT", 1912, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kingston Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Grand_Turk", new TimeSpan(-180000000000), "TC", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("TC", 1979, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("TC", 1979, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("TC", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("TC", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("TC", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaGrenada_113()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Grenada", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Grenada", new TimeSpan(-148200000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "St George's"), new PublicDomain.TzDatabase.TzZone("America/Grenada", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaGuadeloupe_114()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Guadeloupe", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Guadeloupe", new TimeSpan(-147680000000), "-", "LMT", 1911, PublicDomain.Month.June, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Pointe a Pitre"), new PublicDomain.TzDatabase.TzZone("America/Guadeloupe", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaGuatemala_115()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Guatemala", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Guatemala", new TimeSpan(-217240000000), "-", "LMT", 1918, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Guatemala", new TimeSpan(-216000000000), "Guat", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Guat", 1973, 1973, PublicDomain.Month.November, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Guat", 1974, 1974, PublicDomain.Month.February, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Guat", 1983, 1983, PublicDomain.Month.May, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Guat", 1983, 1983, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Guat", 1991, 1991, PublicDomain.Month.March, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Guat", 1991, 1991, PublicDomain.Month.September, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Guat", 2006, 2006, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Guat", 2006, 2006, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaGuayaquil_116()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Guayaquil", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Guayaquil", new TimeSpan(-191600000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Guayaquil", new TimeSpan(-188400000000), "-", "QMT", 1931, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Quito Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Guayaquil", new TimeSpan(-180000000000), "-", "ECT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Ecuador Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaGuyana_117()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Guyana", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Guyana", new TimeSpan(-139600000000), "-", "LMT", 1915, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Georgetown"), new PublicDomain.TzDatabase.TzZone("America/Guyana", new TimeSpan(-135000000000), "-", "GBGT", 1966, PublicDomain.Month.May, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Br Guiana Time"), new PublicDomain.TzDatabase.TzZone("America/Guyana", new TimeSpan(-135000000000), "-", "GYT", 1975, PublicDomain.Month.July, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Guyana Time"), new PublicDomain.TzDatabase.TzZone("America/Guyana", new TimeSpan(-108000000000), "-", "GYT", 1991, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Guyana", new TimeSpan(-144000000000), "-", "GYT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaHalifax_118()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Halifax", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-152640000000), "-", "LMT", 1902, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Halifax", "A%sT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Canada", "A%sT", 1919, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Halifax", "A%sT", 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Canada", "A%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Halifax", "A%sT", 1974, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Canada", "A%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Halifax", 1916, 1916, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1920, 1920, PublicDomain.Month.May, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1920, 1920, PublicDomain.Month.August, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1921, 1921, PublicDomain.Month.May, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1921, 1922, PublicDomain.Month.September, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1922, 1922, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1923, 1925, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1923, 1923, PublicDomain.Month.September, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1924, 1924, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1925, 1925, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1926, 1926, PublicDomain.Month.May, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1926, 1926, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1927, 1927, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1927, 1927, PublicDomain.Month.September, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1928, 1931, PublicDomain.Month.May, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1928, 1928, PublicDomain.Month.September, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1929, 1929, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1930, 1930, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1931, 1932, PublicDomain.Month.September, 24, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1932, 1932, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1933, 1933, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1933, 1933, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1934, 1934, PublicDomain.Month.May, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1934, 1934, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1935, 1935, PublicDomain.Month.June, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1935, 1935, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1936, 1936, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1936, 1936, PublicDomain.Month.September, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1937, 1938, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1937, 1941, PublicDomain.Month.September, 24, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1939, 1939, PublicDomain.Month.May, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1940, 1941, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1946, 1949, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1946, 1949, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1951, 1954, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1951, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1956, 1959, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1956, 1959, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1962, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1962, 1973, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaHavana_119()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Havana", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Havana", new TimeSpan(-197680000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Havana", new TimeSpan(-197760000000), "-", "HMT", 1925, PublicDomain.Month.July, 19, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, "Havana MT"), new PublicDomain.TzDatabase.TzZone("America/Havana", new TimeSpan(-180000000000), "Cuba", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Cuba", 1928, 1928, PublicDomain.Month.June, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1928, 1928, PublicDomain.Month.October, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1940, 1942, PublicDomain.Month.June, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1940, 1942, PublicDomain.Month.September, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1945, 1946, PublicDomain.Month.June, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1945, 1946, PublicDomain.Month.September, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1965, 1965, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1965, 1965, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1966, 1966, PublicDomain.Month.May, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1966, 1966, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1967, 1967, PublicDomain.Month.April, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1967, 1968, PublicDomain.Month.September, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1968, 1968, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1969, 1977, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1969, 1971, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1972, 1974, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1975, 1977, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1978, 1978, PublicDomain.Month.May, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1978, 1990, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1979, 1980, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1981, 1985, PublicDomain.Month.May, 5, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1986, 1989, PublicDomain.Month.March, 14, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1990, 1997, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1991, 1995, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1997, 1997, PublicDomain.Month.October, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1998, 1999, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1998, 2003, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 2000, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 2006, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 2007, 2007, PublicDomain.Month.March, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaHermosillo_120()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Hermosillo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-266320000000), "-", "LMT", 1921, PublicDomain.Month.December, 31, null, new TimeSpan(849680000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-252000000000), "-", "MST", 1927, PublicDomain.Month.June, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-216000000000), "-", "CST", 1930, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-252000000000), "-", "MST", 1931, PublicDomain.Month.May, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-216000000000), "-", "CST", 1931, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-252000000000), "-", "MST", 1932, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-216000000000), "-", "CST", 1942, PublicDomain.Month.April, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-252000000000), "-", "MST", 1949, PublicDomain.Month.January, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-288000000000), "-", "PST", 1970, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-252000000000), "Mexico", "M%sT", 1999, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Hermosillo", new TimeSpan(-252000000000), "-", "MST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaIndianaIndianapolis_121()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Indiana/Indianapolis", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-206780000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(441220000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "US", "C%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "Indianapolis", "C%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "Indianapolis", "C%sT", 1955, PublicDomain.Month.April, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 1957, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "-", "CST", 1958, PublicDomain.Month.April, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "US", "E%sT", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1941, 1941, PublicDomain.Month.June, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1941, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1946, 1954, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaIndianaKnox_122()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Indiana/Knox", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-207900000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "US", "C%sT", 1947, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "Starke", "C%sT", 1962, PublicDomain.Month.April, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-180000000000), "-", "EST", 1963, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "US", "C%sT", 1991, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "US", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1947, 1961, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Starke", 1947, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1955, 1956, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1957, 1958, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1959, 1961, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaIndianaMarengo_123()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Indiana/Marengo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Marengo", new TimeSpan(-207230000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(440770000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Marengo", new TimeSpan(-216000000000), "US", "C%sT", 1951, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Marengo", new TimeSpan(-216000000000), "Marengo", "C%sT", 1961, PublicDomain.Month.April, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Marengo", new TimeSpan(-180000000000), "-", "EST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Marengo", new TimeSpan(-180000000000), "US", "E%sT", 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Marengo", new TimeSpan(-216000000000), "1:00", "CDT", 1974, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Marengo", new TimeSpan(-180000000000), "US", "E%sT", 1976, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Marengo", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Marengo", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Marengo", 1951, 1951, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Marengo", 1951, 1951, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Marengo", 1954, 1960, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Marengo", 1954, 1960, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaIndianaPetersburg_124()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Indiana/Petersburg", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Petersburg", new TimeSpan(-209470000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(438530000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Petersburg", new TimeSpan(-216000000000), "US", "C%sT", 1955, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Petersburg", new TimeSpan(-216000000000), "Pike", "C%sT", 1965, PublicDomain.Month.April, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Petersburg", new TimeSpan(-180000000000), "-", "EST", 1966, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Petersburg", new TimeSpan(-216000000000), "US", "C%sT", 1977, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Petersburg", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Petersburg", new TimeSpan(-216000000000), "US", "C%sT", 2007, PublicDomain.Month.November, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Petersburg", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Pike", 1955, 1955, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Pike", 1955, 1960, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Pike", 1956, 1964, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Pike", 1961, 1964, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaIndianaTell_City_125()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Indiana/Tell_City", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Tell_City", new TimeSpan(-208230000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(439770000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Tell_City", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Tell_City", new TimeSpan(-216000000000), "Perry", "C%sT", 1964, PublicDomain.Month.April, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Tell_City", new TimeSpan(-180000000000), "-", "EST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Tell_City", new TimeSpan(-180000000000), "US", "E%sT", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Tell_City", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Tell_City", new TimeSpan(-216000000000), "US", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Perry", 1946, 1946, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Perry", 1946, 1946, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Perry", 1953, 1954, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Perry", 1953, 1959, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Perry", 1955, 1955, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Perry", 1956, 1963, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Perry", 1960, 1960, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Perry", 1961, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Perry", 1962, 1963, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaIndianaVevay_126()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Indiana/Vevay", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Vevay", new TimeSpan(-204160000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(443840000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vevay", new TimeSpan(-216000000000), "US", "C%sT", 1954, PublicDomain.Month.April, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vevay", new TimeSpan(-180000000000), "-", "EST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vevay", new TimeSpan(-180000000000), "US", "E%sT", 1973, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vevay", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vevay", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaIndianaVincennes_127()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Indiana/Vincennes", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Vincennes", new TimeSpan(-210070000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(437930000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vincennes", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vincennes", new TimeSpan(-216000000000), "Vincennes", "C%sT", 1964, PublicDomain.Month.April, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vincennes", new TimeSpan(-180000000000), "-", "EST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vincennes", new TimeSpan(-180000000000), "US", "E%sT", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vincennes", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vincennes", new TimeSpan(-216000000000), "US", "C%sT", 2007, PublicDomain.Month.November, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Vincennes", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vincennes", 1946, 1946, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Vincennes", 1946, 1946, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vincennes", 1953, 1954, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Vincennes", 1953, 1959, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vincennes", 1955, 1955, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Vincennes", 1956, 1963, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Vincennes", 1960, 1960, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vincennes", 1961, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vincennes", 1962, 1963, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaIndianaWinamac_128()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Indiana/Winamac", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Winamac", new TimeSpan(-207850000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(440150000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Winamac", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Winamac", new TimeSpan(-216000000000), "Pulaski", "C%sT", 1961, PublicDomain.Month.April, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Winamac", new TimeSpan(-180000000000), "-", "EST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Winamac", new TimeSpan(-180000000000), "US", "E%sT", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Winamac", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Winamac", new TimeSpan(-216000000000), "US", "C%sT", 2007, PublicDomain.Month.March, 11, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Winamac", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Pulaski", 1946, 1960, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Pulaski", 1946, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Pulaski", 1955, 1956, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Pulaski", 1957, 1960, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaIndianapolis_129()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Indianapolis", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-206780000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(441220000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "US", "C%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "Indianapolis", "C%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "Indianapolis", "C%sT", 1955, PublicDomain.Month.April, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 1957, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "-", "CST", 1958, PublicDomain.Month.April, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "US", "E%sT", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1941, 1941, PublicDomain.Month.June, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1941, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1946, 1954, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaInuvik_130()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Inuvik", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Inuvik", TimeSpan.Zero, "-", "zzz", 1953, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Inuvik founded"), new PublicDomain.TzDatabase.TzZone("America/Inuvik", new TimeSpan(-288000000000), "NT_YK", "P%sT", 1979, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Inuvik", new TimeSpan(-252000000000), "NT_YK", "M%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Inuvik", new TimeSpan(-252000000000), "Canada", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.May, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaIqaluit_131()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Iqaluit", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Iqaluit", TimeSpan.Zero, "-", "zzz", 1942, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Frobisher Bay est."), new PublicDomain.TzDatabase.TzZone("America/Iqaluit", new TimeSpan(-180000000000), "NT_YK", "E%sT", 1999, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Iqaluit", new TimeSpan(-216000000000), "Canada", "C%sT", 2000, PublicDomain.Month.October, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Iqaluit", new TimeSpan(-180000000000), "Canada", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.May, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaJamaica_132()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Jamaica", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Jamaica", new TimeSpan(-184320000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kingston"), new PublicDomain.TzDatabase.TzZone("America/Jamaica", new TimeSpan(-184320000000), "-", "KMT", 1912, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kingston Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Jamaica", new TimeSpan(-180000000000), "-", "EST", 1974, PublicDomain.Month.April, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Jamaica", new TimeSpan(-180000000000), "US", "E%sT", 1984, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Jamaica", new TimeSpan(-180000000000), "-", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaJujuy_133()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Jujuy", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-156720000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-108000000000), "Arg", "AR%sT", 1990, PublicDomain.Month.March, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "-", "WART", 1990, PublicDomain.Month.October, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "1:00", "WARST", 1991, PublicDomain.Month.March, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-108000000000), "1:00", "ARST", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Jujuy", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaJuneau_134()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Juneau", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Juneau", new TimeSpan(541390000000), "-", "LMT", 1867, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Juneau", new TimeSpan(-322610000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Juneau", new TimeSpan(-288000000000), "-", "PST", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Juneau", new TimeSpan(-288000000000), "US", "P%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Juneau", new TimeSpan(-288000000000), "-", "PST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Juneau", new TimeSpan(-288000000000), "US", "P%sT", 1983, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Juneau", new TimeSpan(-324000000000), "US", "Y%sT", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Juneau", new TimeSpan(-324000000000), "US", "AK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaKentuckyLouisville_135()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Kentucky/Louisville", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-205820000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(442180000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-216000000000), "US", "C%sT", 1921, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-216000000000), "Louisville", "C%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-216000000000), "Louisville", "C%sT", 1961, PublicDomain.Month.July, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-180000000000), "-", "EST", 1968, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-180000000000), "US", "E%sT", 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-216000000000), "1:00", "CDT", 1974, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1921, 1921, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1921, 1921, PublicDomain.Month.September, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1941, 1961, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1941, 1941, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1946, 1946, PublicDomain.Month.June, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1950, 1955, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1956, 1960, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaKentuckyMonticello_136()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Kentucky/Monticello", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Kentucky/Monticello", new TimeSpan(-203640000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(444360000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Monticello", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Monticello", new TimeSpan(-216000000000), "-", "CST", 1968, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Monticello", new TimeSpan(-216000000000), "US", "C%sT", 2000, PublicDomain.Month.October, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Monticello", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaKnox_IN_137()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Knox_IN", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-207900000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "US", "C%sT", 1947, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "Starke", "C%sT", 1962, PublicDomain.Month.April, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-180000000000), "-", "EST", 1963, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "US", "C%sT", 1991, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "US", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1947, 1961, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Starke", 1947, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1955, 1956, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1957, 1958, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1959, 1961, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaLa_Paz_138()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/La_Paz", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/La_Paz", new TimeSpan(-163560000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/La_Paz", new TimeSpan(-163560000000), "-", "CMT", 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Calamarca MT"), new PublicDomain.TzDatabase.TzZone("America/La_Paz", new TimeSpan(-163560000000), "1:00", "BOST", 1932, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bolivia ST"), new PublicDomain.TzDatabase.TzZone("America/La_Paz", new TimeSpan(-144000000000), "-", "BOT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bolivia Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaLima_139()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Lima", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Lima", new TimeSpan(-184920000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Lima", new TimeSpan(-185160000000), "-", "LMT", 1908, PublicDomain.Month.July, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Lima Mean Time?"), new PublicDomain.TzDatabase.TzZone("America/Lima", new TimeSpan(-180000000000), "Peru", "PE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Peru Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Peru", 1938, 1938, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Peru", 1938, 1938, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Peru", 1938, 1939, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Peru", 1939, 1940, PublicDomain.Month.March, 24, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Peru", 1986, 1987, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Peru", 1986, 1987, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Peru", 1990, 1990, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Peru", 1990, 1990, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Peru", 1994, 1994, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Peru", 1994, 1994, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaLos_Angeles_140()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Los_Angeles", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-283780000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(436220000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-288000000000), "US", "P%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-288000000000), "CA", "P%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-288000000000), "US", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1949, 1949, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1962, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaLouisville_141()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Louisville", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-205820000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(442180000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-216000000000), "US", "C%sT", 1921, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-216000000000), "Louisville", "C%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-216000000000), "Louisville", "C%sT", 1961, PublicDomain.Month.July, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-180000000000), "-", "EST", 1968, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-180000000000), "US", "E%sT", 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-216000000000), "1:00", "CDT", 1974, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Kentucky/Louisville", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1921, 1921, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1921, 1921, PublicDomain.Month.September, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1941, 1961, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1941, 1941, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1946, 1946, PublicDomain.Month.June, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1950, 1955, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Louisville", 1956, 1960, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMaceio_142()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Maceio", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Maceio", new TimeSpan(-85720000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Maceio", new TimeSpan(-108000000000), "Brazil", "BR%sT", 1990, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Maceio", new TimeSpan(-108000000000), "-", "BRT", 1995, PublicDomain.Month.October, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Maceio", new TimeSpan(-108000000000), "Brazil", "BR%sT", 1996, PublicDomain.Month.September, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Maceio", new TimeSpan(-108000000000), "-", "BRT", 1999, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Maceio", new TimeSpan(-108000000000), "Brazil", "BR%sT", 2000, PublicDomain.Month.October, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Maceio", new TimeSpan(-108000000000), "-", "BRT", 2001, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Maceio", new TimeSpan(-108000000000), "Brazil", "BR%sT", 2002, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Maceio", new TimeSpan(-108000000000), "-", "BRT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaManagua_143()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Managua", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Managua", new TimeSpan(-207080000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Managua", new TimeSpan(-207120000000), "-", "MMT", 1934, PublicDomain.Month.June, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Managua Mean Time?"), new PublicDomain.TzDatabase.TzZone("America/Managua", new TimeSpan(-216000000000), "-", "CST", 1973, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Managua", new TimeSpan(-180000000000), "-", "EST", 1975, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Managua", new TimeSpan(-216000000000), "Nic", "C%sT", 1992, PublicDomain.Month.January, 1, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Managua", new TimeSpan(-180000000000), "-", "EST", 1992, PublicDomain.Month.September, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Managua", new TimeSpan(-216000000000), "-", "CST", 1993, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Managua", new TimeSpan(-180000000000), "-", "EST", 1997, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Managua", new TimeSpan(-216000000000), "Nic", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Nic", 1979, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Nic", 1979, 1980, PublicDomain.Month.June, 23, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Nic", 2005, 2005, PublicDomain.Month.April, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Nic", 2005, 2005, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Nic", 2006, 2006, PublicDomain.Month.April, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Nic", 2006, 2006, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaManaus_144()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Manaus", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Manaus", new TimeSpan(-144040000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Manaus", new TimeSpan(-144000000000), "Brazil", "AM%sT", 1988, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Manaus", new TimeSpan(-144000000000), "-", "AMT", 1993, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Manaus", new TimeSpan(-144000000000), "Brazil", "AM%sT", 1994, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Manaus", new TimeSpan(-144000000000), "-", "AMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMarigot_145()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Marigot", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Guadeloupe", new TimeSpan(-147680000000), "-", "LMT", 1911, PublicDomain.Month.June, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Pointe a Pitre"), new PublicDomain.TzDatabase.TzZone("America/Guadeloupe", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMartinique_146()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Martinique", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Martinique", new TimeSpan(-146600000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Fort-de-France"), new PublicDomain.TzDatabase.TzZone("America/Martinique", new TimeSpan(-146600000000), "-", "FFMT", 1911, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Fort-de-France MT"), new PublicDomain.TzDatabase.TzZone("America/Martinique", new TimeSpan(-144000000000), "-", "AST", 1980, PublicDomain.Month.April, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Martinique", new TimeSpan(-144000000000), "1:00", "ADT", 1980, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Martinique", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMazatlan_147()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Mazatlan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-255400000000), "-", "LMT", 1921, PublicDomain.Month.December, 31, null, new TimeSpan(860600000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-252000000000), "-", "MST", 1927, PublicDomain.Month.June, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-216000000000), "-", "CST", 1930, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-252000000000), "-", "MST", 1931, PublicDomain.Month.May, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-216000000000), "-", "CST", 1931, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-252000000000), "-", "MST", 1932, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-216000000000), "-", "CST", 1942, PublicDomain.Month.April, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-252000000000), "-", "MST", 1949, PublicDomain.Month.January, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-288000000000), "-", "PST", 1970, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-252000000000), "Mexico", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMendoza_148()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Mendoza", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-165160000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-108000000000), "Arg", "AR%sT", 1990, PublicDomain.Month.March, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "-", "WART", 1990, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "1:00", "WARST", 1991, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "1:00", "WARST", 1992, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "-", "WART", 1992, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-108000000000), "-", "ART", 2004, PublicDomain.Month.May, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-144000000000), "-", "WART", 2004, PublicDomain.Month.September, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Mendoza", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMenominee_149()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Menominee", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Menominee", new TimeSpan(-210270000000), "-", "LMT", 1885, PublicDomain.Month.September, 18, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Menominee", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Menominee", new TimeSpan(-216000000000), "Menominee", "C%sT", 1969, PublicDomain.Month.April, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Menominee", new TimeSpan(-180000000000), "-", "EST", 1973, PublicDomain.Month.April, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Menominee", new TimeSpan(-216000000000), "US", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Menominee", 1946, 1946, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Menominee", 1946, 1946, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Menominee", 1966, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Menominee", 1966, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMerida_150()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Merida", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Merida", new TimeSpan(-215080000000), "-", "LMT", 1922, PublicDomain.Month.January, 1, null, new TimeSpan(920000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Merida", new TimeSpan(-216000000000), "-", "CST", 1981, PublicDomain.Month.December, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Merida", new TimeSpan(-180000000000), "-", "EST", 1982, PublicDomain.Month.December, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Merida", new TimeSpan(-216000000000), "Mexico", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMexico_City_151()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Mexico_City", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-237960000000), "-", "LMT", 1922, PublicDomain.Month.January, 1, null, new TimeSpan(14040000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-252000000000), "-", "MST", 1927, PublicDomain.Month.June, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-216000000000), "-", "CST", 1930, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-252000000000), "-", "MST", 1931, PublicDomain.Month.May, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-216000000000), "-", "CST", 1931, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-252000000000), "-", "MST", 1932, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-216000000000), "Mexico", "C%sT", 2001, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-216000000000), "-", "CST", 2002, PublicDomain.Month.February, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-216000000000), "Mexico", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMiquelon_152()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Miquelon", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Miquelon", new TimeSpan(-134800000000), "-", "LMT", 1911, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "St Pierre"), new PublicDomain.TzDatabase.TzZone("America/Miquelon", new TimeSpan(-144000000000), "-", "AST", 1980, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Miquelon", new TimeSpan(-108000000000), "-", "PMST", 1987, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Pierre & Miquelon Time"), new PublicDomain.TzDatabase.TzZone("America/Miquelon", new TimeSpan(-108000000000), "Canada", "PM%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMoncton_153()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Moncton", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Moncton", new TimeSpan(-155480000000), "-", "LMT", 1883, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Moncton", new TimeSpan(-180000000000), "-", "EST", 1902, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Moncton", new TimeSpan(-144000000000), "Canada", "A%sT", 1933, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Moncton", new TimeSpan(-144000000000), "Moncton", "A%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Moncton", new TimeSpan(-144000000000), "Canada", "A%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Moncton", new TimeSpan(-144000000000), "Moncton", "A%sT", 1973, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Moncton", new TimeSpan(-144000000000), "Canada", "A%sT", 1993, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Moncton", new TimeSpan(-144000000000), "Moncton", "A%sT", 2007, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Moncton", new TimeSpan(-144000000000), "Canada", "A%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1933, 1935, PublicDomain.Month.June, 8, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1933, 1935, PublicDomain.Month.September, 8, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1936, 1938, PublicDomain.Month.June, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1936, 1938, PublicDomain.Month.September, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1939, 1939, PublicDomain.Month.May, 27, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1939, 1941, PublicDomain.Month.September, 21, DayOfWeek.Saturday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1940, 1940, PublicDomain.Month.May, 19, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1941, 1941, PublicDomain.Month.May, 4, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1946, 1972, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1946, 1956, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1957, 1972, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1993, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Moncton", 1993, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMonterrey_154()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Monterrey", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Monterrey", new TimeSpan(-240760000000), "-", "LMT", 1921, PublicDomain.Month.December, 31, null, new TimeSpan(839240000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Monterrey", new TimeSpan(-216000000000), "-", "CST", 1988, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Monterrey", new TimeSpan(-216000000000), "US", "C%sT", 1989, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Monterrey", new TimeSpan(-216000000000), "Mexico", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMontevideo_155()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Montevideo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Montevideo", new TimeSpan(-134840000000), "-", "LMT", 1898, PublicDomain.Month.June, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Montevideo", new TimeSpan(-134840000000), "-", "MMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Montevideo MT"), new PublicDomain.TzDatabase.TzZone("America/Montevideo", new TimeSpan(-126000000000), "Uruguay", "UY%sT", 1942, PublicDomain.Month.December, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Uruguay Time"), new PublicDomain.TzDatabase.TzZone("America/Montevideo", new TimeSpan(-108000000000), "Uruguay", "UY%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Uruguay", 1923, 1923, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HS", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1924, 1926, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1924, 1925, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HS", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1933, 1935, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HS", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1934, 1936, PublicDomain.Month.March, 25, DayOfWeek.Saturday, new TimeSpan(846000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1936, 1936, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HS", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1937, 1941, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1937, 1940, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HS", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1941, 1941, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HS", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1942, 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1942, 1942, PublicDomain.Month.December, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1943, 1943, PublicDomain.Month.March, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1959, 1959, PublicDomain.Month.May, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1959, 1959, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1960, 1960, PublicDomain.Month.January, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1960, 1960, PublicDomain.Month.March, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1965, 1967, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1965, 1965, PublicDomain.Month.September, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1966, 1967, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1968, 1970, PublicDomain.Month.May, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HS", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1968, 1970, PublicDomain.Month.December, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1972, 1972, PublicDomain.Month.April, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1972, 1972, PublicDomain.Month.August, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1974, 1974, PublicDomain.Month.March, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HS", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1974, 1974, PublicDomain.Month.December, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1976, 1976, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1977, 1977, PublicDomain.Month.December, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1978, 1978, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1979, 1979, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1980, 1980, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1987, 1987, PublicDomain.Month.December, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1988, 1988, PublicDomain.Month.March, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1988, 1988, PublicDomain.Month.December, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1989, 1989, PublicDomain.Month.March, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1989, 1989, PublicDomain.Month.October, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1990, 1992, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1990, 1991, PublicDomain.Month.October, 21, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1992, 1992, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 1993, 1993, PublicDomain.Month.February, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 2004, 2004, PublicDomain.Month.September, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 2005, 2005, PublicDomain.Month.March, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 2005, 2005, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 2006, 2006, PublicDomain.Month.March, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 2006, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Uruguay", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMontreal_156()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Montreal", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Montreal", new TimeSpan(-176560000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Montreal", new TimeSpan(-180000000000), "Mont", "E%sT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Montreal", new TimeSpan(-180000000000), "Canada", "E%sT", 1919, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Montreal", new TimeSpan(-180000000000), "Mont", "E%sT", 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("America/Montreal", new TimeSpan(-180000000000), "Canada", "E%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Montreal", new TimeSpan(-180000000000), "Mont", "E%sT", 1974, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Montreal", new TimeSpan(-180000000000), "Canada", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mont", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1917, 1917, PublicDomain.Month.April, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1919, 1919, PublicDomain.Month.March, 31, null, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1919, 1919, PublicDomain.Month.October, 25, null, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1920, 1920, PublicDomain.Month.May, 2, null, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1920, 1922, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1921, 1921, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1922, 1922, PublicDomain.Month.April, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1924, 1924, PublicDomain.Month.May, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1924, 1926, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1925, 1926, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1927, 1927, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1927, 1932, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1928, 1931, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1932, 1932, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1933, 1940, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1933, 1933, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1934, 1939, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1945, 1948, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1946, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1949, 1950, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1951, 1956, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1957, 1973, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaMontserrat_157()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Montserrat", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Montserrat", new TimeSpan(-149320000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, "Cork Hill"), new PublicDomain.TzDatabase.TzZone("America/Montserrat", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaNassau_158()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Nassau", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Nassau", new TimeSpan(-185640000000), "-", "LMT", 1912, PublicDomain.Month.March, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nassau", new TimeSpan(-180000000000), "Bahamas", "E%sT", 1976, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nassau", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Bahamas", 1964, 1975, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Bahamas", 1964, 1975, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaNew_York_159()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/New_York", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-177620000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(434380000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-180000000000), "US", "E%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-180000000000), "NYC", "E%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-180000000000), "US", "E%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-180000000000), "NYC", "E%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NYC", 1920, 1920, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NYC", 1920, 1920, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NYC", 1921, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NYC", 1921, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NYC", 1955, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaNipigon_160()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Nipigon", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Nipigon", new TimeSpan(-211840000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nipigon", new TimeSpan(-180000000000), "Canada", "E%sT", 1940, PublicDomain.Month.September, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nipigon", new TimeSpan(-180000000000), "1:00", "EDT", 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("America/Nipigon", new TimeSpan(-180000000000), "Canada", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaNome_161()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Nome", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Nome", new TimeSpan(467010000000), "-", "LMT", 1867, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nome", new TimeSpan(-396980000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nome", new TimeSpan(-396000000000), "-", "NST", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nome", new TimeSpan(-396000000000), "US", "N%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nome", new TimeSpan(-396000000000), "-", "NST", 1967, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nome", new TimeSpan(-396000000000), "-", "BST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nome", new TimeSpan(-396000000000), "US", "B%sT", 1983, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nome", new TimeSpan(-324000000000), "US", "Y%sT", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Nome", new TimeSpan(-324000000000), "US", "AK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaNoronha_162()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Noronha", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-77800000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "Brazil", "FN%sT", 1990, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "-", "FNT", 1999, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "Brazil", "FN%sT", 2000, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "-", "FNT", 2001, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "Brazil", "FN%sT", 2002, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "-", "FNT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaNorth_DakotaCenter_163()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/North_Dakota/Center", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/North_Dakota/Center", new TimeSpan(-243120000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(440880000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/North_Dakota/Center", new TimeSpan(-252000000000), "US", "M%sT", 1992, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/North_Dakota/Center", new TimeSpan(-216000000000), "US", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaNorth_DakotaNew_Salem_164()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/North_Dakota/New_Salem", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/North_Dakota/New_Salem", new TimeSpan(-243390000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(440610000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/North_Dakota/New_Salem", new TimeSpan(-252000000000), "US", "M%sT", 2003, PublicDomain.Month.October, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/North_Dakota/New_Salem", new TimeSpan(-216000000000), "US", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaPanama_165()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Panama", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Panama", new TimeSpan(-190880000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Panama", new TimeSpan(-191760000000), "-", "CMT", 1908, PublicDomain.Month.April, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Colon Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Panama", new TimeSpan(-180000000000), "-", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaPangnirtung_166()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Pangnirtung", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Pangnirtung", TimeSpan.Zero, "-", "zzz", 1921, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "trading post est."), new PublicDomain.TzDatabase.TzZone("America/Pangnirtung", new TimeSpan(-144000000000), "NT_YK", "A%sT", 1995, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Pangnirtung", new TimeSpan(-180000000000), "Canada", "E%sT", 1999, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Pangnirtung", new TimeSpan(-216000000000), "Canada", "C%sT", 2000, PublicDomain.Month.October, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Pangnirtung", new TimeSpan(-180000000000), "Canada", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.May, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaParamaribo_167()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Paramaribo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Paramaribo", new TimeSpan(-132400000000), "-", "LMT", 1911, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Paramaribo", new TimeSpan(-132520000000), "-", "PMT", 1935, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Paramaribo Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Paramaribo", new TimeSpan(-132360000000), "-", "PMT", 1945, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "The capital moved?"), new PublicDomain.TzDatabase.TzZone("America/Paramaribo", new TimeSpan(-126000000000), "-", "NEGT", 1975, PublicDomain.Month.November, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Dutch Guiana Time"), new PublicDomain.TzDatabase.TzZone("America/Paramaribo", new TimeSpan(-126000000000), "-", "SRT", 1984, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Suriname Time"), new PublicDomain.TzDatabase.TzZone("America/Paramaribo", new TimeSpan(-108000000000), "-", "SRT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaPhoenix_168()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Phoenix", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-268980000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(415020000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "US", "M%sT", 1944, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "-", "MST", 1944, PublicDomain.Month.April, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "US", "M%sT", 1944, PublicDomain.Month.October, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "-", "MST", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "US", "M%sT", 1968, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "-", "MST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaPort_of_Spain_169()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Port_of_Spain", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Port_of_Spain", new TimeSpan(-147640000000), "-", "LMT", 1912, PublicDomain.Month.March, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Port_of_Spain", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaPortauPrince_170()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Port-au-Prince", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Port-au-Prince", new TimeSpan(-173600000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Port-au-Prince", new TimeSpan(-173400000000), "-", "PPMT", 1917, PublicDomain.Month.January, 24, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, "P-a-P MT"), new PublicDomain.TzDatabase.TzZone("America/Port-au-Prince", new TimeSpan(-180000000000), "Haiti", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Haiti", 1983, 1983, PublicDomain.Month.May, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Haiti", 1983, 1987, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Haiti", 1984, 1987, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Haiti", 1988, 1997, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Haiti", 1988, 1997, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Haiti", 2005, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Haiti", 2005, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaPorto_Acre_171()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Porto_Acre", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-162720000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-180000000000), "Brazil", "AC%sT", 1988, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-180000000000), "-", "ACT", 2008, PublicDomain.Month.June, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-144000000000), "-", "AMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaPorto_Velho_172()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Porto_Velho", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Porto_Velho", new TimeSpan(-153360000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Porto_Velho", new TimeSpan(-144000000000), "Brazil", "AM%sT", 1988, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Porto_Velho", new TimeSpan(-144000000000), "-", "AMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaPuerto_Rico_173()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Puerto_Rico", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Puerto_Rico", new TimeSpan(-158650000000), "-", "LMT", 1899, PublicDomain.Month.March, 28, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, "San Juan"), new PublicDomain.TzDatabase.TzZone("America/Puerto_Rico", new TimeSpan(-144000000000), "-", "AST", 1942, PublicDomain.Month.May, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Puerto_Rico", new TimeSpan(-144000000000), "US", "A%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Puerto_Rico", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaRainy_River_174()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Rainy_River", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Rainy_River", new TimeSpan(-226960000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rainy_River", new TimeSpan(-216000000000), "Canada", "C%sT", 1940, PublicDomain.Month.September, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rainy_River", new TimeSpan(-216000000000), "1:00", "CDT", 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("America/Rainy_River", new TimeSpan(-216000000000), "Canada", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaRankin_Inlet_175()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Rankin_Inlet", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Rankin_Inlet", TimeSpan.Zero, "-", "zzz", 1957, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Rankin Inlet founded"), new PublicDomain.TzDatabase.TzZone("America/Rankin_Inlet", new TimeSpan(-216000000000), "NT_YK", "C%sT", 2000, PublicDomain.Month.October, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rankin_Inlet", new TimeSpan(-180000000000), "-", "EST", 2001, PublicDomain.Month.April, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rankin_Inlet", new TimeSpan(-216000000000), "Canada", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.May, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaRecife_176()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Recife", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Recife", new TimeSpan(-83760000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Recife", new TimeSpan(-108000000000), "Brazil", "BR%sT", 1990, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Recife", new TimeSpan(-108000000000), "-", "BRT", 1999, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Recife", new TimeSpan(-108000000000), "Brazil", "BR%sT", 2000, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Recife", new TimeSpan(-108000000000), "-", "BRT", 2001, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Recife", new TimeSpan(-108000000000), "Brazil", "BR%sT", 2002, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Recife", new TimeSpan(-108000000000), "-", "BRT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaRegina_177()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Regina", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Regina", new TimeSpan(-251160000000), "-", "LMT", 1905, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Regina", new TimeSpan(-252000000000), "Regina", "M%sT", 1960, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Regina", new TimeSpan(-216000000000), "-", "CST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Regina", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1930, 1934, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1930, 1934, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1937, 1941, PublicDomain.Month.April, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1937, 1937, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1938, 1938, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1939, 1941, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Regina", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Regina", 1945, 1945, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1946, 1946, PublicDomain.Month.April, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1946, 1946, PublicDomain.Month.October, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1947, 1957, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1947, 1957, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1959, 1959, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1959, 1959, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaResolute_178()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Resolute", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Resolute", TimeSpan.Zero, "-", "zzz", 1947, PublicDomain.Month.August, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Resolute founded"), new PublicDomain.TzDatabase.TzZone("America/Resolute", new TimeSpan(-216000000000), "NT_YK", "C%sT", 2000, PublicDomain.Month.October, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Resolute", new TimeSpan(-180000000000), "-", "EST", 2001, PublicDomain.Month.April, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Resolute", new TimeSpan(-216000000000), "Canada", "C%sT", 2006, PublicDomain.Month.October, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Resolute", new TimeSpan(-180000000000), "-", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.May, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaRio_Branco_179()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Rio_Branco", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-162720000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-180000000000), "Brazil", "AC%sT", 1988, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-180000000000), "-", "ACT", 2008, PublicDomain.Month.June, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-144000000000), "-", "AMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaRosario_180()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Rosario", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-154080000000), "-", "LMT", 1894, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-154080000000), "-", "CMT", 1920, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "-", "ART", 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "Arg", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-108000000000), "Arg", "AR%sT", 1991, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "-", "WART", 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-108000000000), "Arg", "AR%sT", 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-144000000000), "Arg", "AR%sT", 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Argentina/Cordoba", new TimeSpan(-108000000000), "Arg", "AR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Arg", 1930, 1930, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1931, 1931, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1940, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1932, 1939, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1940, 1940, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1941, 1941, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1943, 1943, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1963, 1963, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1988, 1988, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 1989, 1992, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 1999, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2000, 2000, PublicDomain.Month.March, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2007, 2007, PublicDomain.Month.December, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Arg", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSantarem_181()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Santarem", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Santarem", new TimeSpan(-131280000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Santarem", new TimeSpan(-144000000000), "Brazil", "AM%sT", 1988, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Santarem", new TimeSpan(-144000000000), "-", "AMT", 2008, PublicDomain.Month.June, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Santarem", new TimeSpan(-108000000000), "-", "BRT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSantiago_182()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Santiago", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-169660000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-169660000000), "-", "SMT", 1910, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Santiago Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-180000000000), "-", "CLT", 1916, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Chile Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-169660000000), "-", "SMT", 1918, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Santiago Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-144000000000), "-", "CLT", 1919, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Chile Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-169660000000), "-", "SMT", 1927, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Santiago Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-180000000000), "Chile", "CL%sT", 1947, PublicDomain.Month.May, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Chile Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-144000000000), "Chile", "CL%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Chile", 1927, 1932, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1928, 1932, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1942, 1942, PublicDomain.Month.June, 1, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1942, 1942, PublicDomain.Month.August, 1, null, new TimeSpan(180000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1946, 1946, PublicDomain.Month.July, 15, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1946, 1946, PublicDomain.Month.September, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1947, 1947, PublicDomain.Month.April, 1, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1968, 1968, PublicDomain.Month.November, 3, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1969, 1969, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1969, 1969, PublicDomain.Month.November, 23, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1970, 1970, PublicDomain.Month.March, 29, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1970, 1972, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1971, 1971, PublicDomain.Month.March, 14, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1972, 1986, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1973, 1973, PublicDomain.Month.September, 30, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1974, 1987, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1987, 1987, PublicDomain.Month.April, 12, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1988, 1989, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1988, 1988, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1989, 1989, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1990, 1990, PublicDomain.Month.March, 18, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1990, 1990, PublicDomain.Month.September, 16, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1991, 1996, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1991, 1997, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1997, 1997, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1998, 1998, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1998, 1998, PublicDomain.Month.September, 27, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1999, 1999, PublicDomain.Month.April, 4, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1999, 2147483647, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 2000, 2007, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 2008, 2008, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 2009, 2147483647, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSanto_Domingo_183()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Santo_Domingo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Santo_Domingo", new TimeSpan(-167760000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Santo_Domingo", new TimeSpan(-168000000000), "-", "SDMT", 1933, PublicDomain.Month.April, 1, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, "S. Dom. MT"), new PublicDomain.TzDatabase.TzZone("America/Santo_Domingo", new TimeSpan(-180000000000), "DR", "E%sT", 1974, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Santo_Domingo", new TimeSpan(-144000000000), "-", "AST", 2000, PublicDomain.Month.October, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Santo_Domingo", new TimeSpan(-180000000000), "US", "E%sT", 2000, PublicDomain.Month.December, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Santo_Domingo", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("DR", 1966, 1966, PublicDomain.Month.October, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("DR", 1967, 1967, PublicDomain.Month.February, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("DR", 1969, 1973, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HD", null), new PublicDomain.TzDatabase.TzRule("DR", 1970, 1970, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("DR", 1971, 1971, PublicDomain.Month.January, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("DR", 1972, 1974, PublicDomain.Month.January, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSao_Paulo_184()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Sao_Paulo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Sao_Paulo", new TimeSpan(-111880000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Sao_Paulo", new TimeSpan(-108000000000), "Brazil", "BR%sT", 1963, PublicDomain.Month.October, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Sao_Paulo", new TimeSpan(-108000000000), "1:00", "BRST", 1964, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Sao_Paulo", new TimeSpan(-108000000000), "Brazil", "BR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaScoresbysund_185()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Scoresbysund", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Scoresbysund", new TimeSpan(-52720000000), "-", "LMT", 1916, PublicDomain.Month.July, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Ittoqqortoormiit"), new PublicDomain.TzDatabase.TzZone("America/Scoresbysund", new TimeSpan(-72000000000), "-", "CGT", 1980, PublicDomain.Month.April, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Scoresbysund", new TimeSpan(-72000000000), "C-Eur", "CG%sT", 1981, PublicDomain.Month.March, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Scoresbysund", new TimeSpan(-36000000000), "EU", "EG%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaShiprock_186()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Shiprock", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-251960000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(432040000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "Denver", "M%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "Denver", "M%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1920, 1921, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Denver", 1920, 1920, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1921, 1921, PublicDomain.Month.May, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1965, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Denver", 1965, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSt_Barthelemy_187()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/St_Barthelemy", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Guadeloupe", new TimeSpan(-147680000000), "-", "LMT", 1911, PublicDomain.Month.June, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Pointe a Pitre"), new PublicDomain.TzDatabase.TzZone("America/Guadeloupe", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSt_Johns_188()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/St_Johns", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126520000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126520000000), "StJohns", "N%sT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126520000000), "Canada", "N%sT", 1919, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126520000000), "StJohns", "N%sT", 1935, PublicDomain.Month.March, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126000000000), "StJohns", "N%sT", 1942, PublicDomain.Month.May, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126000000000), "Canada", "N%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126000000000), "StJohns", "N%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("StJohns", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1919, 1919, PublicDomain.Month.May, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1919, 1919, PublicDomain.Month.August, 12, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1920, 1935, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1920, 1935, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1936, 1941, PublicDomain.Month.May, 9, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1936, 1941, PublicDomain.Month.October, 2, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1946, 1950, PublicDomain.Month.May, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1946, 1950, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1951, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1951, 1959, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1960, 1986, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1987, 1987, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1987, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1988, 1988, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1989, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSt_Kitts_189()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/St_Kitts", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/St_Kitts", new TimeSpan(-150520000000), "-", "LMT", 1912, PublicDomain.Month.March, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Basseterre"), new PublicDomain.TzDatabase.TzZone("America/St_Kitts", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSt_Lucia_190()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/St_Lucia", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/St_Lucia", new TimeSpan(-146400000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Castries"), new PublicDomain.TzDatabase.TzZone("America/St_Lucia", new TimeSpan(-146400000000), "-", "CMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Castries Mean Time"), new PublicDomain.TzDatabase.TzZone("America/St_Lucia", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSt_Thomas_191()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/St_Thomas", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/St_Thomas", new TimeSpan(-155840000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Charlotte Amalie"), new PublicDomain.TzDatabase.TzZone("America/St_Thomas", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSt_Vincent_192()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/St_Vincent", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/St_Vincent", new TimeSpan(-146960000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kingstown"), new PublicDomain.TzDatabase.TzZone("America/St_Vincent", new TimeSpan(-146960000000), "-", "KMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kingstown Mean Time"), new PublicDomain.TzDatabase.TzZone("America/St_Vincent", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaSwift_Current_193()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Swift_Current", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Swift_Current", new TimeSpan(-258800000000), "-", "LMT", 1905, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Swift_Current", new TimeSpan(-252000000000), "Canada", "M%sT", 1946, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Swift_Current", new TimeSpan(-252000000000), "Regina", "M%sT", 1950, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Swift_Current", new TimeSpan(-252000000000), "Swift", "M%sT", 1972, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Swift_Current", new TimeSpan(-216000000000), "-", "CST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1930, 1934, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1930, 1934, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1937, 1941, PublicDomain.Month.April, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1937, 1937, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1938, 1938, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1939, 1941, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Regina", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1945, 1945, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1946, 1946, PublicDomain.Month.April, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1946, 1946, PublicDomain.Month.October, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1947, 1957, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1947, 1957, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Swift", 1957, 1957, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Swift", 1957, 1957, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Swift", 1959, 1961, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1959, 1959, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Swift", 1959, 1959, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1959, 1959, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Swift", 1960, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaTegucigalpa_194()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Tegucigalpa", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Tegucigalpa", new TimeSpan(-209320000000), "-", "LMT", 1921, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tegucigalpa", new TimeSpan(-216000000000), "Hond", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Hond", 1987, 1988, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Hond", 1987, 1988, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Hond", 2006, 2006, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Hond", 2006, 2006, PublicDomain.Month.August, 1, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaThule_195()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Thule", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Thule", new TimeSpan(-165080000000), "-", "LMT", 1916, PublicDomain.Month.July, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Pituffik air base"), new PublicDomain.TzDatabase.TzZone("America/Thule", new TimeSpan(-144000000000), "Thule", "A%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Thule", 1991, 1992, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Thule", 1991, 1992, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Thule", 1993, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Thule", 1993, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Thule", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Thule", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaThunder_Bay_196()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Thunder_Bay", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Thunder_Bay", new TimeSpan(-214200000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Thunder_Bay", new TimeSpan(-216000000000), "-", "CST", 1910, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Thunder_Bay", new TimeSpan(-180000000000), "-", "EST", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Thunder_Bay", new TimeSpan(-180000000000), "Canada", "E%sT", 1970, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Thunder_Bay", new TimeSpan(-180000000000), "Mont", "E%sT", 1973, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Thunder_Bay", new TimeSpan(-180000000000), "-", "EST", 1974, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Thunder_Bay", new TimeSpan(-180000000000), "Canada", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mont", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1917, 1917, PublicDomain.Month.April, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1919, 1919, PublicDomain.Month.March, 31, null, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1919, 1919, PublicDomain.Month.October, 25, null, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1920, 1920, PublicDomain.Month.May, 2, null, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1920, 1922, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1921, 1921, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1922, 1922, PublicDomain.Month.April, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1924, 1924, PublicDomain.Month.May, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1924, 1926, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1925, 1926, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1927, 1927, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1927, 1932, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1928, 1931, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1932, 1932, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1933, 1940, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1933, 1933, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1934, 1939, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1945, 1948, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1946, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mont", 1949, 1950, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1951, 1956, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mont", 1957, 1973, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaTijuana_197()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Tijuana", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-280840000000), "-", "LMT", 1922, PublicDomain.Month.January, 1, null, new TimeSpan(7160000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-252000000000), "-", "MST", 1924, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1927, PublicDomain.Month.June, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-252000000000), "-", "MST", 1930, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PDT", 1931, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1942, PublicDomain.Month.April, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PWT", 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PPT", 1945, PublicDomain.Month.November, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Peace"), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1948, PublicDomain.Month.April, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PDT", 1949, PublicDomain.Month.January, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1954, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "CA", "P%sT", 1961, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1976, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "US", "P%sT", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "Mexico", "P%sT", 2001, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "US", "P%sT", 2002, PublicDomain.Month.February, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "Mexico", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1949, 1949, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1962, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaToronto_198()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Toronto", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-190520000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-180000000000), "Canada", "E%sT", 1919, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-180000000000), "Toronto", "E%sT", 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-180000000000), "Canada", "E%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-180000000000), "Toronto", "E%sT", 1974, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-180000000000), "Canada", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(846000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1919, 1919, PublicDomain.Month.October, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1920, 1920, PublicDomain.Month.May, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1920, 1920, PublicDomain.Month.September, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1921, 1921, PublicDomain.Month.May, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1921, 1921, PublicDomain.Month.September, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1922, 1923, PublicDomain.Month.May, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1922, 1926, PublicDomain.Month.September, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1924, 1927, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1927, 1932, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1928, 1931, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1932, 1932, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1933, 1940, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1933, 1933, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1934, 1939, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Toronto", 1945, 1946, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1946, 1946, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1947, 1949, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1947, 1948, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1949, 1949, PublicDomain.Month.November, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1950, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1950, 1950, PublicDomain.Month.November, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1951, 1956, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1957, 1973, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaTortola_199()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Tortola", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Tortola", new TimeSpan(-155080000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Road Town"), new PublicDomain.TzDatabase.TzZone("America/Tortola", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaVancouver_200()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Vancouver", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Vancouver", new TimeSpan(-295480000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Vancouver", new TimeSpan(-288000000000), "Vanc", "P%sT", 1987, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Vancouver", new TimeSpan(-288000000000), "Canada", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Vanc", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Vanc", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1946, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1946, 1946, PublicDomain.Month.October, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1947, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1962, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaVirgin_201()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Virgin", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/St_Thomas", new TimeSpan(-155840000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Charlotte Amalie"), new PublicDomain.TzDatabase.TzZone("America/St_Thomas", new TimeSpan(-144000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaWhitehorse_202()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Whitehorse", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Whitehorse", new TimeSpan(-324120000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Whitehorse", new TimeSpan(-324000000000), "NT_YK", "Y%sT", 1966, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Whitehorse", new TimeSpan(-288000000000), "NT_YK", "P%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Whitehorse", new TimeSpan(-288000000000), "Canada", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.May, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaWinnipeg_203()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Winnipeg", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Winnipeg", new TimeSpan(-233160000000), "-", "LMT", 1887, PublicDomain.Month.July, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Winnipeg", new TimeSpan(-216000000000), "Winn", "C%sT", 2006, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Winnipeg", new TimeSpan(-216000000000), "Canada", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Winn", 1916, 1916, PublicDomain.Month.April, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1916, 1916, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1937, 1937, PublicDomain.Month.May, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1937, 1937, PublicDomain.Month.September, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Winn", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Winn", 1945, 1945, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1946, 1946, PublicDomain.Month.May, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1946, 1946, PublicDomain.Month.October, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1947, 1949, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1947, 1949, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1950, 1950, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1950, 1950, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1951, 1960, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1951, 1958, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1959, 1959, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1960, 1960, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1963, 1963, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1963, 1963, PublicDomain.Month.September, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1966, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1966, 2005, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1987, 2005, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaYakutat_204()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Yakutat", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Yakutat", new TimeSpan(528650000000), "-", "LMT", 1867, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Yakutat", new TimeSpan(-335350000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Yakutat", new TimeSpan(-324000000000), "-", "YST", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Yakutat", new TimeSpan(-324000000000), "US", "Y%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Yakutat", new TimeSpan(-324000000000), "-", "YST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Yakutat", new TimeSpan(-324000000000), "US", "Y%sT", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Yakutat", new TimeSpan(-324000000000), "US", "AK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAmericaYellowknife_205()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("America/Yellowknife", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Yellowknife", TimeSpan.Zero, "-", "zzz", 1935, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Yellowknife founded?"), new PublicDomain.TzDatabase.TzZone("America/Yellowknife", new TimeSpan(-252000000000), "NT_YK", "M%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Yellowknife", new TimeSpan(-252000000000), "Canada", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.May, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAntarcticaCasey_206()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Antarctica/Casey", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Antarctica/Casey", TimeSpan.Zero, "-", "zzz", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/Casey", new TimeSpan(288000000000), "-", "WST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Western (Aus) Standard Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAntarcticaDavis_207()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Antarctica/Davis", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Antarctica/Davis", TimeSpan.Zero, "-", "zzz", 1957, PublicDomain.Month.January, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/Davis", new TimeSpan(252000000000), "-", "DAVT", 1964, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Davis Time"), new PublicDomain.TzDatabase.TzZone("Antarctica/Davis", TimeSpan.Zero, "-", "zzz", 1969, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/Davis", new TimeSpan(252000000000), "-", "DAVT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAntarcticaDumontDUrville_208()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Antarctica/DumontDUrville", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Antarctica/DumontDUrville", TimeSpan.Zero, "-", "zzz", 1947, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/DumontDUrville", new TimeSpan(360000000000), "-", "PMT", 1952, PublicDomain.Month.January, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Port-Martin Time"), new PublicDomain.TzDatabase.TzZone("Antarctica/DumontDUrville", TimeSpan.Zero, "-", "zzz", 1956, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/DumontDUrville", new TimeSpan(360000000000), "-", "DDUT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Dumont-d'Urville Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAntarcticaMawson_209()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Antarctica/Mawson", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Antarctica/Mawson", TimeSpan.Zero, "-", "zzz", 1954, PublicDomain.Month.February, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/Mawson", new TimeSpan(216000000000), "-", "MAWT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Mawson Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAntarcticaMcMurdo_210()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Antarctica/McMurdo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Antarctica/McMurdo", TimeSpan.Zero, "-", "zzz", 1956, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/McMurdo", new TimeSpan(432000000000), "NZAQ", "NZ%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NZAQ", 1974, 1974, PublicDomain.Month.November, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1975, 1988, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1976, 1989, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1989, 1989, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1990, 2007, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1990, 2006, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 2007, 2147483647, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAntarcticaPalmer_211()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Antarctica/Palmer", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Antarctica/Palmer", TimeSpan.Zero, "-", "zzz", 1965, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/Palmer", new TimeSpan(-144000000000), "ArgAQ", "AR%sT", 1969, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/Palmer", new TimeSpan(-108000000000), "ArgAQ", "AR%sT", 1982, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/Palmer", new TimeSpan(-144000000000), "ChileAQ", "CL%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("ArgAQ", 1964, 1966, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ArgAQ", 1964, 1966, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("ArgAQ", 1967, 1967, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ArgAQ", 1967, 1968, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("ArgAQ", 1968, 1969, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1972, 1986, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ArgAQ", 1974, 1974, PublicDomain.Month.January, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("ArgAQ", 1974, 1974, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1974, 1987, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1987, 1987, PublicDomain.Month.April, 12, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1988, 1989, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1988, 1988, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1989, 1989, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1990, 1990, PublicDomain.Month.March, 18, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1990, 1990, PublicDomain.Month.September, 16, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1991, 1996, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1991, 1997, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1997, 1997, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1998, 1998, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1998, 1998, PublicDomain.Month.September, 27, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1999, 1999, PublicDomain.Month.April, 4, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 1999, 2147483647, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("ChileAQ", 2000, 2147483647, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAntarcticaRothera_212()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Antarctica/Rothera", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Antarctica/Rothera", TimeSpan.Zero, "-", "zzz", 1976, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/Rothera", new TimeSpan(-108000000000), "-", "ROTT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Rothera time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAntarcticaSouth_Pole_213()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Antarctica/South_Pole", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Antarctica/McMurdo", TimeSpan.Zero, "-", "zzz", 1956, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/McMurdo", new TimeSpan(432000000000), "NZAQ", "NZ%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NZAQ", 1974, 1974, PublicDomain.Month.November, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1975, 1988, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1976, 1989, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1989, 1989, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1990, 2007, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 1990, 2006, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 2007, 2147483647, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZAQ", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAntarcticaSyowa_214()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Antarctica/Syowa", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Antarctica/Syowa", TimeSpan.Zero, "-", "zzz", 1957, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/Syowa", new TimeSpan(108000000000), "-", "SYOT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Syowa Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAntarcticaVostok_215()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Antarctica/Vostok", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Antarctica/Vostok", TimeSpan.Zero, "-", "zzz", 1957, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Antarctica/Vostok", new TimeSpan(216000000000), "-", "VOST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Vostok time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesArcticLongyearbyen_216()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Arctic/Longyearbyen", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(25800000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "Norway", "CE%sT", 1940, PublicDomain.Month.August, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "Norway", "CE%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Norway", 1916, 1916, PublicDomain.Month.May, 22, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Norway", 1916, 1916, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1945, 1945, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1959, 1964, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Norway", 1959, 1965, PublicDomain.Month.September, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1965, 1965, PublicDomain.Month.April, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaAden_217()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Aden", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Aden", new TimeSpan(108480000000), "-", "LMT", 1950, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aden", new TimeSpan(108000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaAlmaty_218()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Almaty", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Almaty", new TimeSpan(184680000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Alma-Ata"), new PublicDomain.TzDatabase.TzZone("Asia/Almaty", new TimeSpan(180000000000), "-", "ALMT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Alma-Ata Time"), new PublicDomain.TzDatabase.TzZone("Asia/Almaty", new TimeSpan(216000000000), "RussiaAsia", "ALM%sT", 1991, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Almaty", new TimeSpan(216000000000), "-", "ALMT", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Almaty", new TimeSpan(216000000000), "RussiaAsia", "ALM%sT", 2005, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Almaty", new TimeSpan(216000000000), "-", "ALMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaAmman_219()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Amman", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Amman", new TimeSpan(86240000000), "-", "LMT", 1931, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Amman", new TimeSpan(72000000000), "Jordan", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Jordan", 1973, 1973, PublicDomain.Month.June, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1973, 1975, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1974, 1977, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1976, 1976, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1977, 1977, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1978, 1978, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1978, 1978, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1985, 1985, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1985, 1985, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1986, 1988, PublicDomain.Month.April, 1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1986, 1990, PublicDomain.Month.October, 1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1989, 1989, PublicDomain.Month.May, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1990, 1990, PublicDomain.Month.April, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1991, 1991, PublicDomain.Month.April, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1991, 1991, PublicDomain.Month.September, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1992, 1992, PublicDomain.Month.April, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1992, 1993, PublicDomain.Month.October, 1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1993, 1998, PublicDomain.Month.April, 1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1994, 1994, PublicDomain.Month.September, 15, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1995, 1998, PublicDomain.Month.September, 15, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1999, 1999, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1999, 2002, PublicDomain.Month.September, -1, DayOfWeek.Thursday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 2000, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Thursday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 2003, 2003, PublicDomain.Month.October, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 2004, 2004, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 2005, 2005, PublicDomain.Month.September, -1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 2006, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaAnadyr_220()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Anadyr", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Anadyr", new TimeSpan(425960000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Anadyr", new TimeSpan(432000000000), "-", "ANAT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Anadyr Time"), new PublicDomain.TzDatabase.TzZone("Asia/Anadyr", new TimeSpan(468000000000), "Russia", "ANA%sT", 1982, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Anadyr", new TimeSpan(432000000000), "Russia", "ANA%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Anadyr", new TimeSpan(396000000000), "Russia", "ANA%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Anadyr", new TimeSpan(432000000000), "Russia", "ANA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaAqtau_221()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Aqtau", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Aqtau", new TimeSpan(120640000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aqtau", new TimeSpan(144000000000), "-", "FORT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Fort Shevchenko T"), new PublicDomain.TzDatabase.TzZone("Asia/Aqtau", new TimeSpan(180000000000), "-", "FORT", 1963, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aqtau", new TimeSpan(180000000000), "-", "SHET", 1981, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Shevchenko Time"), new PublicDomain.TzDatabase.TzZone("Asia/Aqtau", new TimeSpan(216000000000), "-", "SHET", 1982, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aqtau", new TimeSpan(180000000000), "RussiaAsia", "SHE%sT", 1991, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aqtau", new TimeSpan(180000000000), "-", "SHET", 1991, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Aqtau", new TimeSpan(180000000000), "RussiaAsia", "AQT%sT", 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, "Aqtau Time"), new PublicDomain.TzDatabase.TzZone("Asia/Aqtau", new TimeSpan(144000000000), "RussiaAsia", "AQT%sT", 2005, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aqtau", new TimeSpan(180000000000), "-", "AQTT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaAqtobe_222()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Aqtobe", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Aqtobe", new TimeSpan(137200000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aqtobe", new TimeSpan(144000000000), "-", "AKTT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Aktyubinsk Time"), new PublicDomain.TzDatabase.TzZone("Asia/Aqtobe", new TimeSpan(180000000000), "-", "AKTT", 1981, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aqtobe", new TimeSpan(180000000000), "1:00", "AKTST", 1981, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aqtobe", new TimeSpan(216000000000), "-", "AKTT", 1982, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aqtobe", new TimeSpan(180000000000), "RussiaAsia", "AKT%sT", 1991, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Aqtobe", new TimeSpan(180000000000), "-", "AKTT", 1991, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Aqtobe", new TimeSpan(180000000000), "RussiaAsia", "AQT%sT", 2005, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Aqtobe Time"), new PublicDomain.TzDatabase.TzZone("Asia/Aqtobe", new TimeSpan(180000000000), "-", "AQTT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaAshgabat_223()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Ashgabat", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(140120000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Ashkhabad"), new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(144000000000), "-", "ASHT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Ashkhabad Time"), new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(180000000000), "RussiaAsia", "ASH%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(144000000000), "RussiaAsia", "ASH%sT", 1991, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(144000000000), "RussiaAsia", "TM%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(180000000000), "-", "TMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaAshkhabad_224()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Ashkhabad", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(140120000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Ashkhabad"), new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(144000000000), "-", "ASHT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Ashkhabad Time"), new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(180000000000), "RussiaAsia", "ASH%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(144000000000), "RussiaAsia", "ASH%sT", 1991, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(144000000000), "RussiaAsia", "TM%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ashgabat", new TimeSpan(180000000000), "-", "TMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaBaghdad_225()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Baghdad", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Baghdad", new TimeSpan(106600000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Baghdad", new TimeSpan(106560000000), "-", "BMT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Baghdad Mean Time?"), new PublicDomain.TzDatabase.TzZone("Asia/Baghdad", new TimeSpan(108000000000), "-", "AST", 1982, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Baghdad", new TimeSpan(108000000000), "Iraq", "A%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Iraq", 1982, 1982, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iraq", 1982, 1984, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iraq", 1983, 1983, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iraq", 1984, 1985, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iraq", 1985, 1990, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iraq", 1986, 1990, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iraq", 1991, 2007, PublicDomain.Month.April, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iraq", 1991, 2007, PublicDomain.Month.October, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaBahrain_226()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Bahrain", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Bahrain", new TimeSpan(121400000000), "-", "LMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Al Manamah"), new PublicDomain.TzDatabase.TzZone("Asia/Bahrain", new TimeSpan(144000000000), "-", "GST", 1972, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Bahrain", new TimeSpan(108000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaBaku_227()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Baku", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Baku", new TimeSpan(119640000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Baku", new TimeSpan(108000000000), "-", "BAKT", 1957, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Baku Time"), new PublicDomain.TzDatabase.TzZone("Asia/Baku", new TimeSpan(144000000000), "RussiaAsia", "BAK%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Baku", new TimeSpan(108000000000), "1:00", "BAKST", 1991, PublicDomain.Month.August, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Baku", new TimeSpan(108000000000), "RussiaAsia", "AZ%sT", 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Baku", new TimeSpan(144000000000), "-", "AZT", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Azerbaijan time"), new PublicDomain.TzDatabase.TzZone("Asia/Baku", new TimeSpan(144000000000), "EUAsia", "AZ%sT", 1997, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Baku", new TimeSpan(144000000000), "Azer", "AZ%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("EUAsia", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EUAsia", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EUAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Azer", 1997, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Azer", 1997, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(180000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaBangkok_228()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Bangkok", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Bangkok", new TimeSpan(241240000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Bangkok", new TimeSpan(241240000000), "-", "BMT", 1920, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bangkok Mean Time"), new PublicDomain.TzDatabase.TzZone("Asia/Bangkok", new TimeSpan(252000000000), "-", "ICT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaBeirut_229()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Beirut", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Beirut", new TimeSpan(85200000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Beirut", new TimeSpan(72000000000), "Lebanon", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Lebanon", 1920, 1920, PublicDomain.Month.March, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1920, 1920, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1921, 1921, PublicDomain.Month.April, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1921, 1921, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1922, 1922, PublicDomain.Month.March, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1922, 1922, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1923, 1923, PublicDomain.Month.April, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1923, 1923, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1957, 1961, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1957, 1961, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1972, 1972, PublicDomain.Month.June, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1972, 1977, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1973, 1977, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1978, 1978, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1978, 1978, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1984, 1987, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1984, 1991, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1988, 1988, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1989, 1989, PublicDomain.Month.May, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1990, 1992, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1992, 1992, PublicDomain.Month.October, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1993, 1998, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lebanon", 1999, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaBishkek_230()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Bishkek", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Bishkek", new TimeSpan(179040000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Bishkek", new TimeSpan(180000000000), "-", "FRUT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Frunze Time"), new PublicDomain.TzDatabase.TzZone("Asia/Bishkek", new TimeSpan(216000000000), "RussiaAsia", "FRU%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Bishkek", new TimeSpan(180000000000), "1:00", "FRUST", 1991, PublicDomain.Month.August, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Bishkek", new TimeSpan(180000000000), "Kyrgyz", "KG%sT", 2005, PublicDomain.Month.August, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kyrgyzstan Time"), new PublicDomain.TzDatabase.TzZone("Asia/Bishkek", new TimeSpan(216000000000), "-", "KGT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Kyrgyz", 1992, 1996, PublicDomain.Month.April, 7, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Kyrgyz", 1992, 1996, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Kyrgyz", 1997, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Kyrgyz", 1997, 2004, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(90000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaBrunei_231()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Brunei", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Brunei", new TimeSpan(275800000000), "-", "LMT", 1926, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bandar Seri Begawan"), new PublicDomain.TzDatabase.TzZone("Asia/Brunei", new TimeSpan(270000000000), "-", "BNT", 1933, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Brunei", new TimeSpan(288000000000), "-", "BNT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaCalcutta_232()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Calcutta", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(212080000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kolkata"), new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(212000000000), "-", "HMT", 1941, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Howrah Mean Time?"), new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(234000000000), "-", "BURT", 1942, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Burma Time"), new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(198000000000), "-", "IST", 1942, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(198000000000), "1:00", "IST", 1945, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(198000000000), "-", "IST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaChoibalsan_233()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Choibalsan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Choibalsan", new TimeSpan(274800000000), "-", "LMT", 1905, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Choibalsan", new TimeSpan(252000000000), "-", "ULAT", 1978, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Choibalsan", new TimeSpan(288000000000), "-", "ULAT", 1983, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Choibalsan", new TimeSpan(324000000000), "Mongol", "CHO%sT", 2008, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Choibalsan Time"), new PublicDomain.TzDatabase.TzZone("Asia/Choibalsan", new TimeSpan(288000000000), "Mongol", "CHO%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mongol", 1983, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1983, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1984, 1998, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1985, 1998, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2001, 2001, PublicDomain.Month.April, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2001, 2006, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2002, 2006, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaChongqing_234()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Chongqing", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Chongqing", new TimeSpan(255800000000), "-", "LMT", 1928, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Chungking"), new PublicDomain.TzDatabase.TzZone("Asia/Chongqing", new TimeSpan(252000000000), "-", "LONT", 1980, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Long-shu Time"), new PublicDomain.TzDatabase.TzZone("Asia/Chongqing", new TimeSpan(288000000000), "PRC", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1986, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1991, PublicDomain.Month.September, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("PRC", 1987, 1991, PublicDomain.Month.April, 10, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaChungking_235()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Chungking", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Chongqing", new TimeSpan(255800000000), "-", "LMT", 1928, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Chungking"), new PublicDomain.TzDatabase.TzZone("Asia/Chongqing", new TimeSpan(252000000000), "-", "LONT", 1980, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Long-shu Time"), new PublicDomain.TzDatabase.TzZone("Asia/Chongqing", new TimeSpan(288000000000), "PRC", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1986, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1991, PublicDomain.Month.September, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("PRC", 1987, 1991, PublicDomain.Month.April, 10, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaColombo_236()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Colombo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Colombo", new TimeSpan(191640000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Colombo", new TimeSpan(191720000000), "-", "MMT", 1906, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Moratuwa Mean Time"), new PublicDomain.TzDatabase.TzZone("Asia/Colombo", new TimeSpan(198000000000), "-", "IST", 1942, PublicDomain.Month.January, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Colombo", new TimeSpan(198000000000), "0:30", "IHST", 1942, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Colombo", new TimeSpan(198000000000), "1:00", "IST", 1945, PublicDomain.Month.October, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Colombo", new TimeSpan(198000000000), "-", "IST", 1996, PublicDomain.Month.May, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Colombo", new TimeSpan(234000000000), "-", "LKT", 1996, PublicDomain.Month.October, 26, null, new TimeSpan(18000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Colombo", new TimeSpan(216000000000), "-", "LKT", 2006, PublicDomain.Month.April, 15, null, new TimeSpan(18000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Colombo", new TimeSpan(198000000000), "-", "IST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("0:30", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), null, null), new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaDacca_237()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Dacca", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(217000000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(212000000000), "-", "HMT", 1941, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Howrah Mean Time?"), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(234000000000), "-", "BURT", 1942, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Burma Time"), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(198000000000), "-", "IST", 1942, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(234000000000), "-", "BURT", 1951, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(216000000000), "-", "DACT", 1971, PublicDomain.Month.March, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Dacca Time"), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(216000000000), "-", "BDT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bangladesh Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaDamascus_238()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Damascus", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Damascus", new TimeSpan(87120000000), "-", "LMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Dimashq"), new PublicDomain.TzDatabase.TzZone("Asia/Damascus", new TimeSpan(72000000000), "Syria", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Syria", 1920, 1923, PublicDomain.Month.April, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1920, 1923, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1962, 1962, PublicDomain.Month.April, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1962, 1962, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1963, 1965, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1963, 1963, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1964, 1964, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1965, 1965, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1966, 1966, PublicDomain.Month.April, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1966, 1976, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1967, 1978, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1977, 1978, PublicDomain.Month.September, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1983, 1984, PublicDomain.Month.April, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1983, 1984, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1986, 1986, PublicDomain.Month.February, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1986, 1986, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1987, 1987, PublicDomain.Month.March, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1987, 1988, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1988, 1988, PublicDomain.Month.March, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1989, 1989, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1989, 1989, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1990, 1990, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1990, 1990, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1991, 1991, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1991, 1992, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1992, 1992, PublicDomain.Month.April, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1993, 1993, PublicDomain.Month.March, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1993, 1993, PublicDomain.Month.September, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1994, 1996, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1994, 2005, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 1997, 1998, PublicDomain.Month.March, -1, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 1999, 2006, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 2006, 2006, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 2007, 2007, PublicDomain.Month.November, 1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Syria", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Syria", 2008, 2147483647, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaDhaka_239()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Dhaka", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(217000000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(212000000000), "-", "HMT", 1941, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Howrah Mean Time?"), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(234000000000), "-", "BURT", 1942, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Burma Time"), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(198000000000), "-", "IST", 1942, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(234000000000), "-", "BURT", 1951, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(216000000000), "-", "DACT", 1971, PublicDomain.Month.March, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Dacca Time"), new PublicDomain.TzDatabase.TzZone("Asia/Dhaka", new TimeSpan(216000000000), "-", "BDT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bangladesh Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaDili_240()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Dili", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Dili", new TimeSpan(301400000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dili", new TimeSpan(288000000000), "-", "TLT", 1942, PublicDomain.Month.February, 21, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, "E Timor Time"), new PublicDomain.TzDatabase.TzZone("Asia/Dili", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.September, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dili", new TimeSpan(324000000000), "-", "TLT", 1976, PublicDomain.Month.May, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dili", new TimeSpan(288000000000), "-", "CIT", 2000, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dili", new TimeSpan(324000000000), "-", "TLT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaDubai_241()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Dubai", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Dubai", new TimeSpan(132720000000), "-", "LMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dubai", new TimeSpan(144000000000), "-", "GST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaDushanbe_242()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Dushanbe", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Dushanbe", new TimeSpan(165120000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dushanbe", new TimeSpan(180000000000), "-", "DUST", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Dushanbe Time"), new PublicDomain.TzDatabase.TzZone("Asia/Dushanbe", new TimeSpan(216000000000), "RussiaAsia", "DUS%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dushanbe", new TimeSpan(180000000000), "1:00", "DUSST", 1991, PublicDomain.Month.September, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Dushanbe", new TimeSpan(180000000000), "-", "TJT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tajikistan Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaGaza_243()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Gaza", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Gaza", new TimeSpan(82720000000), "-", "LMT", 1900, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Gaza", new TimeSpan(72000000000), "Zion", "EET", 1948, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Gaza", new TimeSpan(72000000000), "EgyptAsia", "EE%sT", 1967, PublicDomain.Month.June, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Gaza", new TimeSpan(72000000000), "Zion", "I%sT", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Gaza", new TimeSpan(72000000000), "Jordan", "EE%sT", 1999, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Gaza", new TimeSpan(72000000000), "Palestine", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Zion", 1940, 1940, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1942, 1944, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1943, 1943, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1944, 1944, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1945, 1945, PublicDomain.Month.April, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1945, 1945, PublicDomain.Month.November, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1946, 1946, PublicDomain.Month.April, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1946, 1946, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1948, PublicDomain.Month.May, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1948, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1949, PublicDomain.Month.November, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1949, 1949, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1950, 1950, PublicDomain.Month.April, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1950, 1950, PublicDomain.Month.September, 15, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1951, 1951, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1951, 1951, PublicDomain.Month.November, 11, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1952, 1952, PublicDomain.Month.April, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1952, 1952, PublicDomain.Month.October, 19, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1953, 1953, PublicDomain.Month.April, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1953, 1953, PublicDomain.Month.September, 13, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1954, 1954, PublicDomain.Month.June, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1954, 1954, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1955, 1955, PublicDomain.Month.June, 11, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1955, 1955, PublicDomain.Month.September, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1956, 1956, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1956, 1956, PublicDomain.Month.September, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1957, 1957, PublicDomain.Month.April, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("EgyptAsia", 1957, 1957, PublicDomain.Month.May, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1957, 1957, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("EgyptAsia", 1957, 1958, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EgyptAsia", 1958, 1958, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EgyptAsia", 1959, 1967, PublicDomain.Month.May, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EgyptAsia", 1959, 1965, PublicDomain.Month.September, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EgyptAsia", 1966, 1966, PublicDomain.Month.October, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1973, 1973, PublicDomain.Month.June, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1973, 1975, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1974, 1977, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1974, 1974, PublicDomain.Month.July, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1974, 1974, PublicDomain.Month.October, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1975, 1975, PublicDomain.Month.April, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1975, 1975, PublicDomain.Month.August, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1976, 1976, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1977, 1977, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1978, 1978, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1978, 1978, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1985, 1985, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1985, 1985, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1985, 1985, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1985, 1985, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1986, 1988, PublicDomain.Month.April, 1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1986, 1986, PublicDomain.Month.May, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1986, 1986, PublicDomain.Month.September, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1986, 1990, PublicDomain.Month.October, 1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Zion", 1987, 1987, PublicDomain.Month.April, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1987, 1987, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1988, 1988, PublicDomain.Month.April, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1988, 1988, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1989, 1989, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1989, 1989, PublicDomain.Month.May, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1989, 1989, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1990, 1990, PublicDomain.Month.March, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1990, 1990, PublicDomain.Month.April, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1990, 1990, PublicDomain.Month.August, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1991, 1991, PublicDomain.Month.March, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1991, 1991, PublicDomain.Month.April, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1991, 1991, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1991, 1991, PublicDomain.Month.September, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Zion", 1992, 1992, PublicDomain.Month.March, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1992, 1992, PublicDomain.Month.April, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1992, 1992, PublicDomain.Month.September, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1992, 1993, PublicDomain.Month.October, 1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1993, 1998, PublicDomain.Month.April, 1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1993, 1993, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1993, 1993, PublicDomain.Month.September, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1994, 1994, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1994, 1994, PublicDomain.Month.August, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1994, 1994, PublicDomain.Month.September, 15, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Zion", 1995, 1995, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1995, 1995, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1995, 1998, PublicDomain.Month.September, 15, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Zion", 1996, 1996, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1996, 1996, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1997, 1997, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1997, 1997, PublicDomain.Month.September, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1998, 1998, PublicDomain.Month.March, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1998, 1998, PublicDomain.Month.September, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1999, 1999, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Palestine", 1999, 2005, PublicDomain.Month.April, 15, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1999, 1999, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1999, 1999, PublicDomain.Month.September, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 1999, 2002, PublicDomain.Month.September, -1, DayOfWeek.Thursday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Palestine", 1999, 2003, PublicDomain.Month.October, 15, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 2000, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Thursday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2000, 2000, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2000, 2000, PublicDomain.Month.October, 6, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2001, 2001, PublicDomain.Month.April, 9, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2001, 2001, PublicDomain.Month.September, 24, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2002, 2002, PublicDomain.Month.March, 29, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2002, 2002, PublicDomain.Month.October, 7, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2003, 2003, PublicDomain.Month.March, 28, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2003, 2003, PublicDomain.Month.October, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 2003, 2003, PublicDomain.Month.October, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Zion", 2004, 2004, PublicDomain.Month.April, 7, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2004, 2004, PublicDomain.Month.September, 22, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Palestine", 2004, 2004, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Jordan", 2004, 2004, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Zion", 2005, 2005, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Jordan", 2005, 2005, PublicDomain.Month.September, -1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Palestine", 2005, 2005, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Zion", 2005, 2005, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2006, 2010, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Palestine", 2006, 2147483647, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Palestine", 2006, 2006, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Zion", 2006, 2006, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Jordan", 2006, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Palestine", 2007, 2147483647, PublicDomain.Month.September, 8, DayOfWeek.Thursday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Zion", 2007, 2007, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2008, 2008, PublicDomain.Month.October, 5, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2009, 2009, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2010, 2010, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2011, 2011, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2011, 2011, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2012, 2015, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2012, 2012, PublicDomain.Month.September, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2013, 2013, PublicDomain.Month.September, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2014, 2014, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2015, 2015, PublicDomain.Month.September, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2016, 2016, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2016, 2016, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2017, 2021, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2017, 2017, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2018, 2018, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2019, 2019, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2020, 2020, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2021, 2021, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2022, 2022, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2022, 2022, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2023, 2032, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2023, 2023, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2024, 2024, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2025, 2025, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2026, 2026, PublicDomain.Month.September, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2027, 2027, PublicDomain.Month.October, 10, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2028, 2028, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2029, 2029, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2030, 2030, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2031, 2031, PublicDomain.Month.September, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2032, 2032, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2033, 2033, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2033, 2033, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2034, 2037, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2034, 2034, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2035, 2035, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2036, 2036, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2037, 2037, PublicDomain.Month.September, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaHarbin_244()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Harbin", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Harbin", new TimeSpan(304040000000), "-", "LMT", 1928, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Haerbin"), new PublicDomain.TzDatabase.TzZone("Asia/Harbin", new TimeSpan(306000000000), "-", "CHAT", 1932, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Changbai Time"), new PublicDomain.TzDatabase.TzZone("Asia/Harbin", new TimeSpan(288000000000), "-", "CST", 1940, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Harbin", new TimeSpan(324000000000), "-", "CHAT", 1966, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Harbin", new TimeSpan(306000000000), "-", "CHAT", 1980, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Harbin", new TimeSpan(288000000000), "PRC", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1986, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1991, PublicDomain.Month.September, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("PRC", 1987, 1991, PublicDomain.Month.April, 10, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaHo_Chi_Minh_245()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Ho_Chi_Minh", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Ho_Chi_Minh", new TimeSpan(256000000000), "-", "LMT", 1906, PublicDomain.Month.June, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ho_Chi_Minh", new TimeSpan(255800000000), "-", "SMT", 1911, PublicDomain.Month.March, 11, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, "Saigon MT?"), new PublicDomain.TzDatabase.TzZone("Asia/Ho_Chi_Minh", new TimeSpan(252000000000), "-", "ICT", 1912, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ho_Chi_Minh", new TimeSpan(288000000000), "-", "ICT", 1931, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ho_Chi_Minh", new TimeSpan(252000000000), "-", "ICT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaHong_Kong_246()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Hong_Kong", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Hong_Kong", new TimeSpan(273960000000), "-", "LMT", 1904, PublicDomain.Month.October, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Hong_Kong", new TimeSpan(288000000000), "HK", "HK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("HK", 1946, 1946, PublicDomain.Month.April, 20, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1946, 1946, PublicDomain.Month.December, 1, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1947, 1947, PublicDomain.Month.December, 30, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1948, 1948, PublicDomain.Month.May, 2, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1948, 1952, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1949, 1953, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1953, 1953, PublicDomain.Month.November, 1, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1954, 1964, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1954, 1954, PublicDomain.Month.October, 31, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1955, 1964, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1965, 1977, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1965, 1977, PublicDomain.Month.October, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1979, 1980, PublicDomain.Month.May, 8, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1979, 1980, PublicDomain.Month.October, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaHovd_247()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Hovd", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Hovd", new TimeSpan(219960000000), "-", "LMT", 1905, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Hovd", new TimeSpan(216000000000), "-", "HOVT", 1978, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Hovd Time"), new PublicDomain.TzDatabase.TzZone("Asia/Hovd", new TimeSpan(252000000000), "Mongol", "HOV%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mongol", 1983, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1983, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1984, 1998, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1985, 1998, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2001, 2001, PublicDomain.Month.April, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2001, 2006, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2002, 2006, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaIrkutsk_248()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Irkutsk", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Irkutsk", new TimeSpan(250400000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Irkutsk", new TimeSpan(250400000000), "-", "IMT", 1920, PublicDomain.Month.January, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Irkutsk Mean Time"), new PublicDomain.TzDatabase.TzZone("Asia/Irkutsk", new TimeSpan(252000000000), "-", "IRKT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Irkutsk Time"), new PublicDomain.TzDatabase.TzZone("Asia/Irkutsk", new TimeSpan(288000000000), "Russia", "IRK%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Irkutsk", new TimeSpan(252000000000), "Russia", "IRK%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Irkutsk", new TimeSpan(288000000000), "Russia", "IRK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaIstanbul_249()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Istanbul", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(69520000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(70160000000), "-", "IMT", 1910, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Istanbul Mean Time?"), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(72000000000), "Turkey", "EE%sT", 1978, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(108000000000), "Turkey", "TR%sT", 1985, PublicDomain.Month.April, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Turkey Time"), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(72000000000), "Turkey", "EE%sT", 2007, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Turkey", 1916, 1916, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1920, 1920, PublicDomain.Month.March, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1920, 1920, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1921, 1921, PublicDomain.Month.April, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1921, 1921, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1922, 1922, PublicDomain.Month.March, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1922, 1922, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1924, 1924, PublicDomain.Month.May, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1924, 1925, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1925, 1925, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1940, 1940, PublicDomain.Month.June, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1940, 1940, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1940, 1940, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1941, 1941, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1942, 1942, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1942, 1942, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1945, 1945, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1945, 1945, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1946, 1946, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1947, 1948, PublicDomain.Month.April, 16, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1947, 1950, PublicDomain.Month.October, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1949, 1949, PublicDomain.Month.April, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1950, 1950, PublicDomain.Month.April, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1951, 1951, PublicDomain.Month.April, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1951, 1951, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1962, 1962, PublicDomain.Month.July, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1962, 1962, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1964, 1964, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1964, 1964, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1970, 1972, PublicDomain.Month.May, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1970, 1972, PublicDomain.Month.October, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1973, 1973, PublicDomain.Month.June, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1973, 1973, PublicDomain.Month.November, 4, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1974, 1974, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1974, 1974, PublicDomain.Month.November, 3, null, new TimeSpan(180000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1975, 1975, PublicDomain.Month.March, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1975, 1976, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1976, 1976, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1977, 1978, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1977, 1977, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1979, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1979, 1982, PublicDomain.Month.October, 11, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1981, 1982, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1983, 1983, PublicDomain.Month.July, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1983, 1983, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1985, 1985, PublicDomain.Month.April, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1985, 1985, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1986, 1990, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1986, 1990, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1991, 2006, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1991, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1996, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaJakarta_250()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Jakarta", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Jakarta", new TimeSpan(256320000000), "-", "LMT", 1867, PublicDomain.Month.August, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jakarta", new TimeSpan(256320000000), "-", "JMT", 1923, PublicDomain.Month.December, 31, null, new TimeSpan(856320000000), TzDatabase.TimeModifier.LocalWallTime, "Jakarta"), new PublicDomain.TzDatabase.TzZone("Asia/Jakarta", new TimeSpan(264000000000), "-", "JAVT", 1932, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Java Time"), new PublicDomain.TzDatabase.TzZone("Asia/Jakarta", new TimeSpan(270000000000), "-", "WIT", 1942, PublicDomain.Month.March, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jakarta", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.September, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jakarta", new TimeSpan(270000000000), "-", "WIT", 1948, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jakarta", new TimeSpan(288000000000), "-", "WIT", 1950, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jakarta", new TimeSpan(270000000000), "-", "WIT", 1964, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jakarta", new TimeSpan(252000000000), "-", "WIT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaJayapura_251()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Jayapura", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Jayapura", new TimeSpan(337680000000), "-", "LMT", 1932, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jayapura", new TimeSpan(324000000000), "-", "EIT", 1944, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jayapura", new TimeSpan(342000000000), "-", "CST", 1964, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jayapura", new TimeSpan(324000000000), "-", "EIT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaJerusalem_252()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Jerusalem", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Jerusalem", new TimeSpan(84560000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jerusalem", new TimeSpan(84400000000), "-", "JMT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Jerusalem Mean Time?"), new PublicDomain.TzDatabase.TzZone("Asia/Jerusalem", new TimeSpan(72000000000), "Zion", "I%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Zion", 1940, 1940, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1942, 1944, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1943, 1943, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1944, 1944, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1945, 1945, PublicDomain.Month.April, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1945, 1945, PublicDomain.Month.November, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1946, 1946, PublicDomain.Month.April, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1946, 1946, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1948, PublicDomain.Month.May, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1948, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1949, PublicDomain.Month.November, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1949, 1949, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1950, 1950, PublicDomain.Month.April, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1950, 1950, PublicDomain.Month.September, 15, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1951, 1951, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1951, 1951, PublicDomain.Month.November, 11, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1952, 1952, PublicDomain.Month.April, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1952, 1952, PublicDomain.Month.October, 19, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1953, 1953, PublicDomain.Month.April, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1953, 1953, PublicDomain.Month.September, 13, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1954, 1954, PublicDomain.Month.June, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1954, 1954, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1955, 1955, PublicDomain.Month.June, 11, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1955, 1955, PublicDomain.Month.September, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1956, 1956, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1956, 1956, PublicDomain.Month.September, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1957, 1957, PublicDomain.Month.April, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1957, 1957, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1974, 1974, PublicDomain.Month.July, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1974, 1974, PublicDomain.Month.October, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1975, 1975, PublicDomain.Month.April, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1975, 1975, PublicDomain.Month.August, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1985, 1985, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1985, 1985, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1986, 1986, PublicDomain.Month.May, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1986, 1986, PublicDomain.Month.September, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1987, 1987, PublicDomain.Month.April, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1987, 1987, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1988, 1988, PublicDomain.Month.April, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1988, 1988, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1989, 1989, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1989, 1989, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1990, 1990, PublicDomain.Month.March, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1990, 1990, PublicDomain.Month.August, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1991, 1991, PublicDomain.Month.March, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1991, 1991, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1992, 1992, PublicDomain.Month.March, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1992, 1992, PublicDomain.Month.September, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1993, 1993, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1993, 1993, PublicDomain.Month.September, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1994, 1994, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1994, 1994, PublicDomain.Month.August, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1995, 1995, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1995, 1995, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1996, 1996, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1996, 1996, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1997, 1997, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1997, 1997, PublicDomain.Month.September, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1998, 1998, PublicDomain.Month.March, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1998, 1998, PublicDomain.Month.September, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1999, 1999, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1999, 1999, PublicDomain.Month.September, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2000, 2000, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2000, 2000, PublicDomain.Month.October, 6, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2001, 2001, PublicDomain.Month.April, 9, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2001, 2001, PublicDomain.Month.September, 24, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2002, 2002, PublicDomain.Month.March, 29, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2002, 2002, PublicDomain.Month.October, 7, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2003, 2003, PublicDomain.Month.March, 28, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2003, 2003, PublicDomain.Month.October, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2004, 2004, PublicDomain.Month.April, 7, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2004, 2004, PublicDomain.Month.September, 22, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2005, 2005, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2005, 2005, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2006, 2010, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2006, 2006, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2007, 2007, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2008, 2008, PublicDomain.Month.October, 5, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2009, 2009, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2010, 2010, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2011, 2011, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2011, 2011, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2012, 2015, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2012, 2012, PublicDomain.Month.September, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2013, 2013, PublicDomain.Month.September, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2014, 2014, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2015, 2015, PublicDomain.Month.September, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2016, 2016, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2016, 2016, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2017, 2021, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2017, 2017, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2018, 2018, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2019, 2019, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2020, 2020, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2021, 2021, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2022, 2022, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2022, 2022, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2023, 2032, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2023, 2023, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2024, 2024, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2025, 2025, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2026, 2026, PublicDomain.Month.September, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2027, 2027, PublicDomain.Month.October, 10, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2028, 2028, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2029, 2029, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2030, 2030, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2031, 2031, PublicDomain.Month.September, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2032, 2032, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2033, 2033, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2033, 2033, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2034, 2037, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2034, 2034, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2035, 2035, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2036, 2036, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2037, 2037, PublicDomain.Month.September, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaKabul_253()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Kabul", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Kabul", new TimeSpan(166080000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kabul", new TimeSpan(144000000000), "-", "AFT", 1945, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kabul", new TimeSpan(162000000000), "-", "AFT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaKamchatka_254()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Kamchatka", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Kamchatka", new TimeSpan(380760000000), "-", "LMT", 1922, PublicDomain.Month.November, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kamchatka", new TimeSpan(396000000000), "-", "PETT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "P-K Time"), new PublicDomain.TzDatabase.TzZone("Asia/Kamchatka", new TimeSpan(432000000000), "Russia", "PET%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kamchatka", new TimeSpan(396000000000), "Russia", "PET%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kamchatka", new TimeSpan(432000000000), "Russia", "PET%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaKarachi_255()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Karachi", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Karachi", new TimeSpan(160920000000), "-", "LMT", 1907, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Karachi", new TimeSpan(198000000000), "-", "IST", 1942, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Karachi", new TimeSpan(198000000000), "1:00", "IST", 1945, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Karachi", new TimeSpan(198000000000), "-", "IST", 1951, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Karachi", new TimeSpan(180000000000), "-", "KART", 1971, PublicDomain.Month.March, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Karachi Time"), new PublicDomain.TzDatabase.TzZone("Asia/Karachi", new TimeSpan(180000000000), "Pakistan", "PK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Pakistan Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Pakistan", 2002, 2002, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Pakistan", 2002, 2002, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Pakistan", 2008, 2008, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Pakistan", 2008, 2008, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaKashgar_256()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Kashgar", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Kashgar", new TimeSpan(182360000000), "-", "LMT", 1928, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Kashi or Kaxgar"), new PublicDomain.TzDatabase.TzZone("Asia/Kashgar", new TimeSpan(198000000000), "-", "KAST", 1940, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kashgar Time"), new PublicDomain.TzDatabase.TzZone("Asia/Kashgar", new TimeSpan(180000000000), "-", "KAST", 1980, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kashgar", new TimeSpan(288000000000), "PRC", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1986, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1991, PublicDomain.Month.September, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("PRC", 1987, 1991, PublicDomain.Month.April, 10, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaKatmandu_257()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Katmandu", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Katmandu", new TimeSpan(204760000000), "-", "LMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Katmandu", new TimeSpan(198000000000), "-", "IST", 1986, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Katmandu", new TimeSpan(207000000000), "-", "NPT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Nepal Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaKolkata_258()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Kolkata", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(212080000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kolkata"), new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(212000000000), "-", "HMT", 1941, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Howrah Mean Time?"), new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(234000000000), "-", "BURT", 1942, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Burma Time"), new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(198000000000), "-", "IST", 1942, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(198000000000), "1:00", "IST", 1945, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kolkata", new TimeSpan(198000000000), "-", "IST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaKrasnoyarsk_259()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Krasnoyarsk", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Krasnoyarsk", new TimeSpan(222800000000), "-", "LMT", 1920, PublicDomain.Month.January, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Krasnoyarsk", new TimeSpan(216000000000), "-", "KRAT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Krasnoyarsk Time"), new PublicDomain.TzDatabase.TzZone("Asia/Krasnoyarsk", new TimeSpan(252000000000), "Russia", "KRA%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Krasnoyarsk", new TimeSpan(216000000000), "Russia", "KRA%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Krasnoyarsk", new TimeSpan(252000000000), "Russia", "KRA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaKuala_Lumpur_260()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Kuala_Lumpur", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Kuala_Lumpur", new TimeSpan(244060000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuala_Lumpur", new TimeSpan(249250000000), "-", "SMT", 1905, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Singapore M.T."), new PublicDomain.TzDatabase.TzZone("Asia/Kuala_Lumpur", new TimeSpan(252000000000), "-", "MALT", 1933, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Malaya Time"), new PublicDomain.TzDatabase.TzZone("Asia/Kuala_Lumpur", new TimeSpan(252000000000), "0:20", "MALST", 1936, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuala_Lumpur", new TimeSpan(264000000000), "-", "MALT", 1941, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuala_Lumpur", new TimeSpan(270000000000), "-", "MALT", 1942, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuala_Lumpur", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuala_Lumpur", new TimeSpan(270000000000), "-", "MALT", 1982, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuala_Lumpur", new TimeSpan(288000000000), "-", "MYT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Malaysia Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("0:20", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(12000000000), null, null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaKuching_261()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Kuching", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Kuching", new TimeSpan(264800000000), "-", "LMT", 1926, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuching", new TimeSpan(270000000000), "-", "BORT", 1933, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Borneo Time"), new PublicDomain.TzDatabase.TzZone("Asia/Kuching", new TimeSpan(288000000000), "NBorneo", "BOR%sT", 1942, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuching", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuching", new TimeSpan(288000000000), "-", "BORT", 1982, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuching", new TimeSpan(288000000000), "-", "MYT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NBorneo", 1935, 1941, PublicDomain.Month.September, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(12000000000), "TS", "one-Third Summer"), new PublicDomain.TzDatabase.TzRule("NBorneo", 1935, 1941, PublicDomain.Month.December, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaKuwait_262()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Kuwait", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Kuwait", new TimeSpan(115160000000), "-", "LMT", 1950, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Kuwait", new TimeSpan(108000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaMacao_263()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Macao", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Macau", new TimeSpan(272600000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Macau", new TimeSpan(288000000000), "Macau", "MO%sT", 1999, PublicDomain.Month.December, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "return to China"), new PublicDomain.TzDatabase.TzZone("Asia/Macau", new TimeSpan(288000000000), "PRC", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Macau", 1961, 1962, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1961, 1964, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Macau", 1963, 1963, PublicDomain.Month.March, 16, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1964, 1964, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1965, 1965, PublicDomain.Month.March, 16, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1965, 1965, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Macau", 1966, 1971, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1966, 1971, PublicDomain.Month.October, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Macau", 1972, 1974, PublicDomain.Month.April, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1972, 1973, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Macau", 1974, 1977, PublicDomain.Month.October, 15, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Macau", 1975, 1977, PublicDomain.Month.April, 15, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1978, 1980, PublicDomain.Month.April, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1978, 1980, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1986, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1991, PublicDomain.Month.September, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("PRC", 1987, 1991, PublicDomain.Month.April, 10, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaMacau_264()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Macau", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Macau", new TimeSpan(272600000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Macau", new TimeSpan(288000000000), "Macau", "MO%sT", 1999, PublicDomain.Month.December, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "return to China"), new PublicDomain.TzDatabase.TzZone("Asia/Macau", new TimeSpan(288000000000), "PRC", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Macau", 1961, 1962, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1961, 1964, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Macau", 1963, 1963, PublicDomain.Month.March, 16, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1964, 1964, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1965, 1965, PublicDomain.Month.March, 16, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1965, 1965, PublicDomain.Month.October, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Macau", 1966, 1971, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1966, 1971, PublicDomain.Month.October, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Macau", 1972, 1974, PublicDomain.Month.April, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1972, 1973, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Macau", 1974, 1977, PublicDomain.Month.October, 15, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Macau", 1975, 1977, PublicDomain.Month.April, 15, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1978, 1980, PublicDomain.Month.April, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Macau", 1978, 1980, PublicDomain.Month.October, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1986, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1991, PublicDomain.Month.September, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("PRC", 1987, 1991, PublicDomain.Month.April, 10, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaMagadan_265()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Magadan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Magadan", new TimeSpan(361920000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Magadan", new TimeSpan(360000000000), "-", "MAGT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Magadan Time"), new PublicDomain.TzDatabase.TzZone("Asia/Magadan", new TimeSpan(396000000000), "Russia", "MAG%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Magadan", new TimeSpan(360000000000), "Russia", "MAG%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Magadan", new TimeSpan(396000000000), "Russia", "MAG%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaMakassar_266()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Makassar", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Makassar", new TimeSpan(286560000000), "-", "LMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Makassar", new TimeSpan(286560000000), "-", "MMT", 1932, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Macassar MT"), new PublicDomain.TzDatabase.TzZone("Asia/Makassar", new TimeSpan(288000000000), "-", "CIT", 1942, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Makassar", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.September, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Makassar", new TimeSpan(288000000000), "-", "CIT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaManila_267()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Manila", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Manila", new TimeSpan(-573600000000), "-", "LMT", 1844, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Manila", new TimeSpan(290400000000), "-", "LMT", 1899, PublicDomain.Month.May, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Manila", new TimeSpan(288000000000), "Phil", "PH%sT", 1942, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Manila", new TimeSpan(324000000000), "-", "JST", 1944, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Manila", new TimeSpan(288000000000), "Phil", "PH%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Phil", 1936, 1936, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Phil", 1937, 1937, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Phil", 1954, 1954, PublicDomain.Month.April, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Phil", 1954, 1954, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Phil", 1978, 1978, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Phil", 1978, 1978, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaMuscat_268()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Muscat", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Muscat", new TimeSpan(140600000000), "-", "LMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Muscat", new TimeSpan(144000000000), "-", "GST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaNicosia_269()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Nicosia", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Nicosia", new TimeSpan(80080000000), "-", "LMT", 1921, PublicDomain.Month.November, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Nicosia", new TimeSpan(72000000000), "Cyprus", "EE%sT", 1998, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Nicosia", new TimeSpan(72000000000), "EUAsia", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Cyprus", 1975, 1975, PublicDomain.Month.April, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1975, 1975, PublicDomain.Month.October, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1976, 1976, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1976, 1976, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1977, 1977, PublicDomain.Month.September, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1978, 1978, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1979, 1997, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EUAsia", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1981, 1998, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EUAsia", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EUAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaNovosibirsk_270()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Novosibirsk", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Novosibirsk", new TimeSpan(199000000000), "-", "LMT", 1919, PublicDomain.Month.December, 14, null, new TimeSpan(216000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Novosibirsk", new TimeSpan(216000000000), "-", "NOVT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Novosibirsk Time"), new PublicDomain.TzDatabase.TzZone("Asia/Novosibirsk", new TimeSpan(252000000000), "Russia", "NOV%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Novosibirsk", new TimeSpan(216000000000), "Russia", "NOV%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Novosibirsk", new TimeSpan(252000000000), "Russia", "NOV%sT", 1993, PublicDomain.Month.May, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "say Shanks & P."), new PublicDomain.TzDatabase.TzZone("Asia/Novosibirsk", new TimeSpan(216000000000), "Russia", "NOV%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaOmsk_271()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Omsk", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Omsk", new TimeSpan(176160000000), "-", "LMT", 1919, PublicDomain.Month.November, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Omsk", new TimeSpan(180000000000), "-", "OMST", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Omsk TIme"), new PublicDomain.TzDatabase.TzZone("Asia/Omsk", new TimeSpan(216000000000), "Russia", "OMS%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Omsk", new TimeSpan(180000000000), "Russia", "OMS%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Omsk", new TimeSpan(216000000000), "Russia", "OMS%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaOral_272()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Oral", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Oral", new TimeSpan(123240000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Ural'sk"), new PublicDomain.TzDatabase.TzZone("Asia/Oral", new TimeSpan(144000000000), "-", "URAT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Ural'sk time"), new PublicDomain.TzDatabase.TzZone("Asia/Oral", new TimeSpan(180000000000), "-", "URAT", 1981, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Oral", new TimeSpan(180000000000), "1:00", "URAST", 1981, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Oral", new TimeSpan(216000000000), "-", "URAT", 1982, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Oral", new TimeSpan(180000000000), "RussiaAsia", "URA%sT", 1989, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Oral", new TimeSpan(144000000000), "RussiaAsia", "URA%sT", 1991, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Oral", new TimeSpan(144000000000), "-", "URAT", 1991, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Oral", new TimeSpan(144000000000), "RussiaAsia", "ORA%sT", 2005, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Oral Time"), new PublicDomain.TzDatabase.TzZone("Asia/Oral", new TimeSpan(180000000000), "-", "ORAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaPhnom_Penh_273()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Phnom_Penh", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Phnom_Penh", new TimeSpan(251800000000), "-", "LMT", 1906, PublicDomain.Month.June, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Phnom_Penh", new TimeSpan(255800000000), "-", "SMT", 1911, PublicDomain.Month.March, 11, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, "Saigon MT?"), new PublicDomain.TzDatabase.TzZone("Asia/Phnom_Penh", new TimeSpan(252000000000), "-", "ICT", 1912, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Phnom_Penh", new TimeSpan(288000000000), "-", "ICT", 1931, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Phnom_Penh", new TimeSpan(252000000000), "-", "ICT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaPontianak_274()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Pontianak", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Pontianak", new TimeSpan(262400000000), "-", "LMT", 1908, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pontianak", new TimeSpan(262400000000), "-", "PMT", 1932, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Pontianak MT"), new PublicDomain.TzDatabase.TzZone("Asia/Pontianak", new TimeSpan(270000000000), "-", "WIT", 1942, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pontianak", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.September, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pontianak", new TimeSpan(270000000000), "-", "WIT", 1948, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pontianak", new TimeSpan(288000000000), "-", "WIT", 1950, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pontianak", new TimeSpan(270000000000), "-", "WIT", 1964, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pontianak", new TimeSpan(288000000000), "-", "CIT", 1988, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pontianak", new TimeSpan(252000000000), "-", "WIT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaPyongyang_275()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Pyongyang", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Pyongyang", new TimeSpan(301800000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pyongyang", new TimeSpan(306000000000), "-", "KST", 1904, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pyongyang", new TimeSpan(324000000000), "-", "KST", 1928, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pyongyang", new TimeSpan(306000000000), "-", "KST", 1932, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pyongyang", new TimeSpan(324000000000), "-", "KST", 1954, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pyongyang", new TimeSpan(288000000000), "-", "KST", 1961, PublicDomain.Month.August, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Pyongyang", new TimeSpan(324000000000), "-", "KST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaQatar_276()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Qatar", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Qatar", new TimeSpan(123680000000), "-", "LMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Al Dawhah / Doha"), new PublicDomain.TzDatabase.TzZone("Asia/Qatar", new TimeSpan(144000000000), "-", "GST", 1972, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Qatar", new TimeSpan(108000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaQyzylorda_277()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Qyzylorda", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Qyzylorda", new TimeSpan(157120000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Qyzylorda", new TimeSpan(144000000000), "-", "KIZT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kizilorda Time"), new PublicDomain.TzDatabase.TzZone("Asia/Qyzylorda", new TimeSpan(180000000000), "-", "KIZT", 1981, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Qyzylorda", new TimeSpan(180000000000), "1:00", "KIZST", 1981, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Qyzylorda", new TimeSpan(216000000000), "-", "KIZT", 1982, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Qyzylorda", new TimeSpan(180000000000), "RussiaAsia", "KIZ%sT", 1991, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Qyzylorda", new TimeSpan(180000000000), "-", "KIZT", 1991, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Qyzylorda", new TimeSpan(180000000000), "-", "QYZT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Qyzylorda", new TimeSpan(216000000000), "RussiaAsia", "QYZ%sT", 2005, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Qyzylorda", new TimeSpan(216000000000), "-", "QYZT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaRangoon_278()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Rangoon", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Rangoon", new TimeSpan(230800000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Yangon"), new PublicDomain.TzDatabase.TzZone("Asia/Rangoon", new TimeSpan(230760000000), "-", "RMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Rangoon Mean Time?"), new PublicDomain.TzDatabase.TzZone("Asia/Rangoon", new TimeSpan(234000000000), "-", "BURT", 1942, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Burma Time"), new PublicDomain.TzDatabase.TzZone("Asia/Rangoon", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.May, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Rangoon", new TimeSpan(234000000000), "-", "MMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Myanmar Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaRiyadh_279()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Riyadh", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Riyadh", new TimeSpan(112120000000), "-", "LMT", 1950, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh", new TimeSpan(108000000000), "-", "AST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaRiyadh87_280()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Riyadh87", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Riyadh87", new TimeSpan(112240000000), "-", "zzz", 1987, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh87", new TimeSpan(112240000000), "sol87", "zzz", 1988, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh87", new TimeSpan(112240000000), "-", "zzz", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 1, null, new TimeSpan(434000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 2, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 3, null, new TimeSpan(434550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 4, null, new TimeSpan(434850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 5, null, new TimeSpan(435100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 6, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 7, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 8, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 9, null, new TimeSpan(436150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 10, null, new TimeSpan(436400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 11, null, new TimeSpan(436650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 12, null, new TimeSpan(436900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 13, null, new TimeSpan(437100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 14, null, new TimeSpan(437350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 15, null, new TimeSpan(437550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 16, null, new TimeSpan(437750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 17, null, new TimeSpan(437950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 18, null, new TimeSpan(438150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 19, null, new TimeSpan(438350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 20, null, new TimeSpan(438550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 21, null, new TimeSpan(438700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 22, null, new TimeSpan(438900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 23, null, new TimeSpan(439050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 24, null, new TimeSpan(439200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 25, null, new TimeSpan(439350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 26, null, new TimeSpan(439500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 27, null, new TimeSpan(439600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 28, null, new TimeSpan(439750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 29, null, new TimeSpan(439850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 30, null, new TimeSpan(439950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 31, null, new TimeSpan(440050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 1, null, new TimeSpan(440150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 2, null, new TimeSpan(440200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 3, null, new TimeSpan(440300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 4, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 5, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 6, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 7, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 8, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 9, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 10, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 11, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 12, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 13, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 14, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 15, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 16, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 17, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 18, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 19, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 20, null, new TimeSpan(440300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 21, null, new TimeSpan(440250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 22, null, new TimeSpan(440150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 23, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 24, null, new TimeSpan(440000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 25, null, new TimeSpan(439900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 26, null, new TimeSpan(439800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 27, null, new TimeSpan(439700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 28, null, new TimeSpan(439600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 1, null, new TimeSpan(439500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 2, null, new TimeSpan(439400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 3, null, new TimeSpan(439250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 4, null, new TimeSpan(439150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 5, null, new TimeSpan(439000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 6, null, new TimeSpan(438850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 7, null, new TimeSpan(438750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 8, null, new TimeSpan(438600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 9, null, new TimeSpan(438450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 10, null, new TimeSpan(438300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 11, null, new TimeSpan(438150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 12, null, new TimeSpan(437950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 13, null, new TimeSpan(437800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 14, null, new TimeSpan(437650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 15, null, new TimeSpan(437500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 16, null, new TimeSpan(437300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 17, null, new TimeSpan(437150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 18, null, new TimeSpan(436950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 19, null, new TimeSpan(436800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 20, null, new TimeSpan(436600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 21, null, new TimeSpan(436450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 22, null, new TimeSpan(436250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 23, null, new TimeSpan(436100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 24, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 25, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 26, null, new TimeSpan(435550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 27, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 28, null, new TimeSpan(435150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 29, null, new TimeSpan(435000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 30, null, new TimeSpan(434800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 31, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 1, null, new TimeSpan(434450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 2, null, new TimeSpan(434250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 3, null, new TimeSpan(434100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 4, null, new TimeSpan(433900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 5, null, new TimeSpan(433750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 6, null, new TimeSpan(433550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 7, null, new TimeSpan(433400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 8, null, new TimeSpan(433250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 9, null, new TimeSpan(433050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 10, null, new TimeSpan(432900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 11, null, new TimeSpan(432750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 12, null, new TimeSpan(432550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 13, null, new TimeSpan(432400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 14, null, new TimeSpan(432250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 15, null, new TimeSpan(432100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 16, null, new TimeSpan(431950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(50000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 17, null, new TimeSpan(431850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 18, null, new TimeSpan(431700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 19, null, new TimeSpan(431550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 20, null, new TimeSpan(431450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 21, null, new TimeSpan(431300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 22, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 23, null, new TimeSpan(431050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 24, null, new TimeSpan(430950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 25, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 26, null, new TimeSpan(430750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 27, null, new TimeSpan(430650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 28, null, new TimeSpan(430550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 29, null, new TimeSpan(430450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 30, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 1, null, new TimeSpan(430300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 2, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 3, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 4, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 5, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 6, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 7, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 8, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 9, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 10, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 11, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 12, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 13, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 14, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 15, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 16, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 17, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 18, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 19, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 20, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 21, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 22, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 23, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 24, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 25, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 26, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 27, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 28, null, new TimeSpan(430300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 29, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 30, null, new TimeSpan(430450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 31, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 1, null, new TimeSpan(430600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 2, null, new TimeSpan(430700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 3, null, new TimeSpan(430800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 4, null, new TimeSpan(430900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 5, null, new TimeSpan(431000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 6, null, new TimeSpan(431100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 7, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 8, null, new TimeSpan(431300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 9, null, new TimeSpan(431450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 10, null, new TimeSpan(431550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 11, null, new TimeSpan(431700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 12, null, new TimeSpan(431800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 13, null, new TimeSpan(431900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 14, null, new TimeSpan(432050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-50000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 15, null, new TimeSpan(432150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 16, null, new TimeSpan(432300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 17, null, new TimeSpan(432450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 18, null, new TimeSpan(432550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 19, null, new TimeSpan(432700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 20, null, new TimeSpan(432800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 21, null, new TimeSpan(432950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 22, null, new TimeSpan(433100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 23, null, new TimeSpan(433200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 24, null, new TimeSpan(433350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 25, null, new TimeSpan(433450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 26, null, new TimeSpan(433600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 27, null, new TimeSpan(433700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 28, null, new TimeSpan(433850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 29, null, new TimeSpan(433950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 30, null, new TimeSpan(434100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 1, null, new TimeSpan(434200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 2, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 3, null, new TimeSpan(434450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 4, null, new TimeSpan(434550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 5, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 6, null, new TimeSpan(434750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 7, null, new TimeSpan(434850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 8, null, new TimeSpan(434950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 9, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 10, null, new TimeSpan(435150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 11, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 12, null, new TimeSpan(435300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 13, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 14, null, new TimeSpan(435450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 15, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 16, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 17, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 18, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 19, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 20, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 21, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 22, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 23, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 24, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 25, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 26, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 27, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 28, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 29, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 30, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 31, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 1, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 2, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 3, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 4, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 5, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 6, null, new TimeSpan(435550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 7, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 8, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 9, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 10, null, new TimeSpan(435250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 11, null, new TimeSpan(435150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 12, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 13, null, new TimeSpan(434950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 14, null, new TimeSpan(434850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 15, null, new TimeSpan(434750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 16, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 17, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 18, null, new TimeSpan(434400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 19, null, new TimeSpan(434250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 20, null, new TimeSpan(434100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 21, null, new TimeSpan(433950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 22, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 23, null, new TimeSpan(433650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 24, null, new TimeSpan(433500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 25, null, new TimeSpan(433350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 26, null, new TimeSpan(433200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 27, null, new TimeSpan(433000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 28, null, new TimeSpan(432850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 29, null, new TimeSpan(432650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 30, null, new TimeSpan(432500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 31, null, new TimeSpan(432300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 1, null, new TimeSpan(432100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 2, null, new TimeSpan(431900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 3, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 4, null, new TimeSpan(431550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 5, null, new TimeSpan(431350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 6, null, new TimeSpan(431150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 7, null, new TimeSpan(430950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 8, null, new TimeSpan(430750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 9, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 10, null, new TimeSpan(430300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 11, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 12, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 13, null, new TimeSpan(429700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 14, null, new TimeSpan(429450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 15, null, new TimeSpan(429250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 16, null, new TimeSpan(429050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 17, null, new TimeSpan(428850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 18, null, new TimeSpan(428600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 19, null, new TimeSpan(428400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 20, null, new TimeSpan(428200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 21, null, new TimeSpan(427950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 22, null, new TimeSpan(427750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 23, null, new TimeSpan(427550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 24, null, new TimeSpan(427350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 25, null, new TimeSpan(427150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 26, null, new TimeSpan(426950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 27, null, new TimeSpan(426700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 28, null, new TimeSpan(426500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 29, null, new TimeSpan(426300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 30, null, new TimeSpan(426100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 1, null, new TimeSpan(425900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 2, null, new TimeSpan(425750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 3, null, new TimeSpan(425550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 4, null, new TimeSpan(425350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 5, null, new TimeSpan(425150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 6, null, new TimeSpan(425000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 7, null, new TimeSpan(424800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 8, null, new TimeSpan(424650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 9, null, new TimeSpan(424450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 10, null, new TimeSpan(424300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 11, null, new TimeSpan(424150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 12, null, new TimeSpan(424000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 13, null, new TimeSpan(423850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 14, null, new TimeSpan(423700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 15, null, new TimeSpan(423550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 16, null, new TimeSpan(423450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 17, null, new TimeSpan(423300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 18, null, new TimeSpan(423200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 19, null, new TimeSpan(423050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 20, null, new TimeSpan(422950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 21, null, new TimeSpan(422850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 22, null, new TimeSpan(422750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 23, null, new TimeSpan(422650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 24, null, new TimeSpan(422600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 25, null, new TimeSpan(422500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 26, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 27, null, new TimeSpan(422350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 28, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 29, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 30, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 31, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 1, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 2, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 3, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 4, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 5, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 6, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 7, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 8, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 9, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 10, null, new TimeSpan(422350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 11, null, new TimeSpan(422400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 12, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 13, null, new TimeSpan(422550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 14, null, new TimeSpan(422600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 15, null, new TimeSpan(422700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 16, null, new TimeSpan(422800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 17, null, new TimeSpan(422900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 18, null, new TimeSpan(423050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 19, null, new TimeSpan(423150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 20, null, new TimeSpan(423300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 21, null, new TimeSpan(423450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 22, null, new TimeSpan(423600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 23, null, new TimeSpan(423750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 24, null, new TimeSpan(423900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 25, null, new TimeSpan(424100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 26, null, new TimeSpan(424300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 27, null, new TimeSpan(424450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 28, null, new TimeSpan(424650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 29, null, new TimeSpan(424850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 30, null, new TimeSpan(425100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 1, null, new TimeSpan(425300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 2, null, new TimeSpan(425500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 3, null, new TimeSpan(425750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 4, null, new TimeSpan(426000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 5, null, new TimeSpan(426250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 6, null, new TimeSpan(426500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 7, null, new TimeSpan(426750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 8, null, new TimeSpan(427000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 9, null, new TimeSpan(427250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 10, null, new TimeSpan(427500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 11, null, new TimeSpan(427800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 12, null, new TimeSpan(428050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 13, null, new TimeSpan(428350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 14, null, new TimeSpan(428650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 15, null, new TimeSpan(428900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 16, null, new TimeSpan(429200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 17, null, new TimeSpan(429500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 18, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 19, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 20, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 21, null, new TimeSpan(430700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 22, null, new TimeSpan(431000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 23, null, new TimeSpan(431300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 24, null, new TimeSpan(431600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 25, null, new TimeSpan(431900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 26, null, new TimeSpan(432200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 27, null, new TimeSpan(432450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 28, null, new TimeSpan(432750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 29, null, new TimeSpan(433050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 30, null, new TimeSpan(433350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 31, null, new TimeSpan(433650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1650000000), "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaRiyadh88_281()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Riyadh88", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Riyadh88", new TimeSpan(112240000000), "-", "zzz", 1988, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh88", new TimeSpan(112240000000), "sol88", "zzz", 1989, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh88", new TimeSpan(112240000000), "-", "zzz", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 1, null, new TimeSpan(433950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 2, null, new TimeSpan(434200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 3, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 4, null, new TimeSpan(434800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 5, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 6, null, new TimeSpan(435300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 7, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 8, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 9, null, new TimeSpan(436100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 10, null, new TimeSpan(436350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 11, null, new TimeSpan(436600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 12, null, new TimeSpan(436850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 13, null, new TimeSpan(437050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 14, null, new TimeSpan(437300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 15, null, new TimeSpan(437500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 16, null, new TimeSpan(437700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 17, null, new TimeSpan(437900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 18, null, new TimeSpan(438100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 19, null, new TimeSpan(438300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 20, null, new TimeSpan(438500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 21, null, new TimeSpan(438650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 22, null, new TimeSpan(438850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 23, null, new TimeSpan(439000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 24, null, new TimeSpan(439150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 25, null, new TimeSpan(439300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 26, null, new TimeSpan(439450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 27, null, new TimeSpan(439600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 28, null, new TimeSpan(439700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 29, null, new TimeSpan(439800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 30, null, new TimeSpan(439900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 31, null, new TimeSpan(440000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 1, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 2, null, new TimeSpan(440200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 3, null, new TimeSpan(440250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 4, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 5, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 6, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 7, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 8, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 9, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 10, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 11, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 12, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 13, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 14, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 15, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 16, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 17, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 18, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 19, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 20, null, new TimeSpan(440300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 21, null, new TimeSpan(440250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 22, null, new TimeSpan(440200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 23, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 24, null, new TimeSpan(440000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 25, null, new TimeSpan(439950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 26, null, new TimeSpan(439850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 27, null, new TimeSpan(439750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 28, null, new TimeSpan(439650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 29, null, new TimeSpan(439500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 1, null, new TimeSpan(439400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 2, null, new TimeSpan(439300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 3, null, new TimeSpan(439150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 4, null, new TimeSpan(439050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 5, null, new TimeSpan(438900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 6, null, new TimeSpan(438750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 7, null, new TimeSpan(438600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 8, null, new TimeSpan(438450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 9, null, new TimeSpan(438300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 10, null, new TimeSpan(438150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 11, null, new TimeSpan(438000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 12, null, new TimeSpan(437850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 13, null, new TimeSpan(437700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 14, null, new TimeSpan(437500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 15, null, new TimeSpan(437350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 16, null, new TimeSpan(437200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 17, null, new TimeSpan(437000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 18, null, new TimeSpan(436850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 19, null, new TimeSpan(436650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 20, null, new TimeSpan(436500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 21, null, new TimeSpan(436300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 22, null, new TimeSpan(436100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 23, null, new TimeSpan(435950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 24, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 25, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 26, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 27, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 28, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 29, null, new TimeSpan(434850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 30, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 31, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 1, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 2, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 3, null, new TimeSpan(433950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 4, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 5, null, new TimeSpan(433600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 6, null, new TimeSpan(433450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 7, null, new TimeSpan(433250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 8, null, new TimeSpan(433100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 9, null, new TimeSpan(432950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 10, null, new TimeSpan(432750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 11, null, new TimeSpan(432600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 12, null, new TimeSpan(432450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 13, null, new TimeSpan(432300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 14, null, new TimeSpan(432150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 15, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 16, null, new TimeSpan(431850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 17, null, new TimeSpan(431700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 18, null, new TimeSpan(431600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 19, null, new TimeSpan(431450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 20, null, new TimeSpan(431350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 21, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 22, null, new TimeSpan(431100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 23, null, new TimeSpan(430950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 24, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 25, null, new TimeSpan(430750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 26, null, new TimeSpan(430650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 27, null, new TimeSpan(430550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 28, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 29, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 30, null, new TimeSpan(430300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 1, null, new TimeSpan(430250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 2, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 3, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 4, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 5, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 6, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 7, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 8, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 9, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 10, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 11, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 12, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 13, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 14, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 15, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 16, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 17, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 18, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 19, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 20, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 21, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 22, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 23, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 24, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 25, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 26, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 27, null, new TimeSpan(430250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 28, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 29, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 30, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 31, null, new TimeSpan(430600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 1, null, new TimeSpan(430700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 2, null, new TimeSpan(430750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 3, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 4, null, new TimeSpan(430950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 5, null, new TimeSpan(431100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 6, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 7, null, new TimeSpan(431300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 8, null, new TimeSpan(431400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 9, null, new TimeSpan(431550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 10, null, new TimeSpan(431650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 11, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 12, null, new TimeSpan(431900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 13, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 14, null, new TimeSpan(432150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 15, null, new TimeSpan(432250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 16, null, new TimeSpan(432400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 17, null, new TimeSpan(432550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 18, null, new TimeSpan(432650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 19, null, new TimeSpan(432800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 20, null, new TimeSpan(432900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 21, null, new TimeSpan(433050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 22, null, new TimeSpan(433200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 23, null, new TimeSpan(433300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 24, null, new TimeSpan(433450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 25, null, new TimeSpan(433550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 26, null, new TimeSpan(433700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 27, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 28, null, new TimeSpan(433950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 29, null, new TimeSpan(434050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 30, null, new TimeSpan(434200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 1, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 2, null, new TimeSpan(434400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 3, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 4, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 5, null, new TimeSpan(434750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 6, null, new TimeSpan(434850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 7, null, new TimeSpan(434950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 8, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 9, null, new TimeSpan(435100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 10, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 11, null, new TimeSpan(435300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 12, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 13, null, new TimeSpan(435450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 14, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 15, null, new TimeSpan(435550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 16, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 17, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 18, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 19, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 20, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 21, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 22, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 23, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 24, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 25, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 26, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 27, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 28, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 29, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 30, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 31, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 1, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 2, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 3, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 4, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 5, null, new TimeSpan(435550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 6, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 7, null, new TimeSpan(435450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 8, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 9, null, new TimeSpan(435250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 10, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 11, null, new TimeSpan(435100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 12, null, new TimeSpan(435000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 13, null, new TimeSpan(434900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 14, null, new TimeSpan(434750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 15, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 16, null, new TimeSpan(434550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 17, null, new TimeSpan(434400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 18, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 19, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 20, null, new TimeSpan(434000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 21, null, new TimeSpan(433850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 22, null, new TimeSpan(433700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 23, null, new TimeSpan(433550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 24, null, new TimeSpan(433400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 25, null, new TimeSpan(433200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 26, null, new TimeSpan(433050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 27, null, new TimeSpan(432900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 28, null, new TimeSpan(432700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 29, null, new TimeSpan(432500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 30, null, new TimeSpan(432350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 31, null, new TimeSpan(432150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 1, null, new TimeSpan(431950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(50000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 2, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 3, null, new TimeSpan(431600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 4, null, new TimeSpan(431400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 5, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 6, null, new TimeSpan(431000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 7, null, new TimeSpan(430800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 8, null, new TimeSpan(430550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 9, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 10, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 11, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 12, null, new TimeSpan(429750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 13, null, new TimeSpan(429500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 14, null, new TimeSpan(429300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 15, null, new TimeSpan(429100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 16, null, new TimeSpan(428900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 17, null, new TimeSpan(428650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 18, null, new TimeSpan(428450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 19, null, new TimeSpan(428250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 20, null, new TimeSpan(428050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 21, null, new TimeSpan(427800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 22, null, new TimeSpan(427600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 23, null, new TimeSpan(427400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 24, null, new TimeSpan(427200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 25, null, new TimeSpan(427000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 26, null, new TimeSpan(426750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 27, null, new TimeSpan(426550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 28, null, new TimeSpan(426350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 29, null, new TimeSpan(426150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 30, null, new TimeSpan(425950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 1, null, new TimeSpan(425750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 2, null, new TimeSpan(425600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 3, null, new TimeSpan(425400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 4, null, new TimeSpan(425200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 5, null, new TimeSpan(425050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 6, null, new TimeSpan(424850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 7, null, new TimeSpan(424700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 8, null, new TimeSpan(424500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 9, null, new TimeSpan(424350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 10, null, new TimeSpan(424200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 11, null, new TimeSpan(424050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 12, null, new TimeSpan(423900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 13, null, new TimeSpan(423750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 14, null, new TimeSpan(423600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 15, null, new TimeSpan(423450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 16, null, new TimeSpan(423350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 17, null, new TimeSpan(423200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 18, null, new TimeSpan(423100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 19, null, new TimeSpan(423000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 20, null, new TimeSpan(422850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 21, null, new TimeSpan(422800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 22, null, new TimeSpan(422700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 23, null, new TimeSpan(422600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 24, null, new TimeSpan(422500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 25, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 26, null, new TimeSpan(422400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 27, null, new TimeSpan(422350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 28, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 29, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 30, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 31, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 1, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 2, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 3, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 4, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 5, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 6, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 7, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 8, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 9, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 10, null, new TimeSpan(422400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 11, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 12, null, new TimeSpan(422500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 13, null, new TimeSpan(422600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 14, null, new TimeSpan(422700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 15, null, new TimeSpan(422800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 16, null, new TimeSpan(422900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 17, null, new TimeSpan(423000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 18, null, new TimeSpan(423150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 19, null, new TimeSpan(423250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 20, null, new TimeSpan(423400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 21, null, new TimeSpan(423550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 22, null, new TimeSpan(423700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 23, null, new TimeSpan(423900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 24, null, new TimeSpan(424050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 25, null, new TimeSpan(424250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 26, null, new TimeSpan(424400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 27, null, new TimeSpan(424600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 28, null, new TimeSpan(424800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 29, null, new TimeSpan(425050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 30, null, new TimeSpan(425250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 1, null, new TimeSpan(425450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 2, null, new TimeSpan(425700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 3, null, new TimeSpan(425950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 4, null, new TimeSpan(426150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 5, null, new TimeSpan(426400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 6, null, new TimeSpan(426650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 7, null, new TimeSpan(426950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 8, null, new TimeSpan(427200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 9, null, new TimeSpan(427450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 10, null, new TimeSpan(427750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 11, null, new TimeSpan(428000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 12, null, new TimeSpan(428300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 13, null, new TimeSpan(428550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 14, null, new TimeSpan(428850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 15, null, new TimeSpan(429150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 16, null, new TimeSpan(429450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 17, null, new TimeSpan(429750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 18, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 19, null, new TimeSpan(430300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 20, null, new TimeSpan(430600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 21, null, new TimeSpan(430900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 22, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 23, null, new TimeSpan(431500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 24, null, new TimeSpan(431800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 25, null, new TimeSpan(432100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 26, null, new TimeSpan(432400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 27, null, new TimeSpan(432700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 28, null, new TimeSpan(433000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 29, null, new TimeSpan(433300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 30, null, new TimeSpan(433550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 31, null, new TimeSpan(433850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1850000000), "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaRiyadh89_282()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Riyadh89", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Riyadh89", new TimeSpan(112240000000), "-", "zzz", 1989, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh89", new TimeSpan(112240000000), "sol89", "zzz", 1990, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh89", new TimeSpan(112240000000), "-", "zzz", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 1, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 2, null, new TimeSpan(434450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 3, null, new TimeSpan(434700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 4, null, new TimeSpan(435000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 5, null, new TimeSpan(435250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 6, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 7, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 8, null, new TimeSpan(436050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 9, null, new TimeSpan(436300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 10, null, new TimeSpan(436550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 11, null, new TimeSpan(436750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 12, null, new TimeSpan(437000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 13, null, new TimeSpan(437250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 14, null, new TimeSpan(437450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 15, null, new TimeSpan(437650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 16, null, new TimeSpan(437850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 17, null, new TimeSpan(438050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 18, null, new TimeSpan(438250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 19, null, new TimeSpan(438450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 20, null, new TimeSpan(438650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 21, null, new TimeSpan(438800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 22, null, new TimeSpan(438950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 23, null, new TimeSpan(439150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 24, null, new TimeSpan(439300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 25, null, new TimeSpan(439400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 26, null, new TimeSpan(439550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 27, null, new TimeSpan(439700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 28, null, new TimeSpan(439800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 29, null, new TimeSpan(439900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 30, null, new TimeSpan(440000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 31, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 1, null, new TimeSpan(440200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 2, null, new TimeSpan(440250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 3, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 4, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 5, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 6, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 7, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 8, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 9, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 10, null, new TimeSpan(440600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 11, null, new TimeSpan(440600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 12, null, new TimeSpan(440600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 13, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 14, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 15, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 16, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 17, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 18, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 19, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 20, null, new TimeSpan(440300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 21, null, new TimeSpan(440200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 22, null, new TimeSpan(440150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 23, null, new TimeSpan(440050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 24, null, new TimeSpan(439950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 25, null, new TimeSpan(439850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 26, null, new TimeSpan(439750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 27, null, new TimeSpan(439650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 28, null, new TimeSpan(439550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 1, null, new TimeSpan(439450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 2, null, new TimeSpan(439300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 3, null, new TimeSpan(439200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 4, null, new TimeSpan(439050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 5, null, new TimeSpan(438950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 6, null, new TimeSpan(438800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 7, null, new TimeSpan(438650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 8, null, new TimeSpan(438500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 9, null, new TimeSpan(438350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 10, null, new TimeSpan(438200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 11, null, new TimeSpan(438050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 12, null, new TimeSpan(437900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 13, null, new TimeSpan(437700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 14, null, new TimeSpan(437550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 15, null, new TimeSpan(437400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 16, null, new TimeSpan(437200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 17, null, new TimeSpan(437050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 18, null, new TimeSpan(436850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 19, null, new TimeSpan(436700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 20, null, new TimeSpan(436500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 21, null, new TimeSpan(436350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 22, null, new TimeSpan(436150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 23, null, new TimeSpan(435950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 24, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 25, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 26, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 27, null, new TimeSpan(435250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 28, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 29, null, new TimeSpan(434900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 30, null, new TimeSpan(434700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 31, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 1, null, new TimeSpan(434350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 2, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 3, null, new TimeSpan(434000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 4, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 5, null, new TimeSpan(433650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 6, null, new TimeSpan(433450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 7, null, new TimeSpan(433300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 8, null, new TimeSpan(433100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 9, null, new TimeSpan(432950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 10, null, new TimeSpan(432800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 11, null, new TimeSpan(432650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 12, null, new TimeSpan(432500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 13, null, new TimeSpan(432350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 14, null, new TimeSpan(432200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 15, null, new TimeSpan(432050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-50000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 16, null, new TimeSpan(431900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 17, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 18, null, new TimeSpan(431600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 19, null, new TimeSpan(431500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 20, null, new TimeSpan(431350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 21, null, new TimeSpan(431250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 22, null, new TimeSpan(431100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 23, null, new TimeSpan(431000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 24, null, new TimeSpan(430900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 25, null, new TimeSpan(430800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 26, null, new TimeSpan(430700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 27, null, new TimeSpan(430600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 28, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 29, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 30, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 1, null, new TimeSpan(430250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 2, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 3, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 4, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 5, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 6, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 7, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 8, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 9, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 10, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 11, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 12, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 13, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 14, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 15, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 16, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 17, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 18, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 19, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 20, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 21, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 22, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 23, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 24, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 25, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 26, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 27, null, new TimeSpan(430250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 28, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 29, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 30, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 31, null, new TimeSpan(430550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 1, null, new TimeSpan(430650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 2, null, new TimeSpan(430750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 3, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 4, null, new TimeSpan(430950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 5, null, new TimeSpan(431050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 6, null, new TimeSpan(431150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 7, null, new TimeSpan(431250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 8, null, new TimeSpan(431400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 9, null, new TimeSpan(431500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 10, null, new TimeSpan(431600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 11, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 12, null, new TimeSpan(431850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 13, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 14, null, new TimeSpan(432100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 15, null, new TimeSpan(432250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 16, null, new TimeSpan(432350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 17, null, new TimeSpan(432500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 18, null, new TimeSpan(432650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 19, null, new TimeSpan(432750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 20, null, new TimeSpan(432900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 21, null, new TimeSpan(433000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 22, null, new TimeSpan(433150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 23, null, new TimeSpan(433300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 24, null, new TimeSpan(433400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 25, null, new TimeSpan(433550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 26, null, new TimeSpan(433650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 27, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 28, null, new TimeSpan(433900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 29, null, new TimeSpan(434050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 30, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 1, null, new TimeSpan(434250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 2, null, new TimeSpan(434400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 3, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 4, null, new TimeSpan(434600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 5, null, new TimeSpan(434700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 6, null, new TimeSpan(434800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 7, null, new TimeSpan(434900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 8, null, new TimeSpan(435000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 9, null, new TimeSpan(435100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 10, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 11, null, new TimeSpan(435250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 12, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 13, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 14, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 15, null, new TimeSpan(435550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 16, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 17, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 18, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 19, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 20, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 21, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 22, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 23, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 24, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 25, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 26, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 27, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 28, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 29, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 30, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 31, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 1, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 2, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 3, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 4, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 5, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 6, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 7, null, new TimeSpan(435450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 8, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 9, null, new TimeSpan(435300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 10, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 11, null, new TimeSpan(435100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 12, null, new TimeSpan(435000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 13, null, new TimeSpan(434900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 14, null, new TimeSpan(434800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 15, null, new TimeSpan(434700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 16, null, new TimeSpan(434550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 17, null, new TimeSpan(434450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 18, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 19, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 20, null, new TimeSpan(434050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 21, null, new TimeSpan(433900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 22, null, new TimeSpan(433750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 23, null, new TimeSpan(433600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 24, null, new TimeSpan(433400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 25, null, new TimeSpan(433250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 26, null, new TimeSpan(433100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 27, null, new TimeSpan(432900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 28, null, new TimeSpan(432750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 29, null, new TimeSpan(432550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 30, null, new TimeSpan(432400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 31, null, new TimeSpan(432200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 1, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 2, null, new TimeSpan(431850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 3, null, new TimeSpan(431650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 4, null, new TimeSpan(431450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 5, null, new TimeSpan(431250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 6, null, new TimeSpan(431050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 7, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 8, null, new TimeSpan(430650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 9, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 10, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 11, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 12, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 13, null, new TimeSpan(429600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 14, null, new TimeSpan(429350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 15, null, new TimeSpan(429150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 16, null, new TimeSpan(428950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 17, null, new TimeSpan(428750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 18, null, new TimeSpan(428500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 19, null, new TimeSpan(428300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 20, null, new TimeSpan(428100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 21, null, new TimeSpan(427900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 22, null, new TimeSpan(427650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 23, null, new TimeSpan(427450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 24, null, new TimeSpan(427250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 25, null, new TimeSpan(427050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 26, null, new TimeSpan(426850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 27, null, new TimeSpan(426650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 28, null, new TimeSpan(426400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 29, null, new TimeSpan(426200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 30, null, new TimeSpan(426000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 1, null, new TimeSpan(425850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 2, null, new TimeSpan(425650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 3, null, new TimeSpan(425450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 4, null, new TimeSpan(425250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 5, null, new TimeSpan(425100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 6, null, new TimeSpan(424900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 7, null, new TimeSpan(424700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 8, null, new TimeSpan(424550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 9, null, new TimeSpan(424400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 10, null, new TimeSpan(424200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 11, null, new TimeSpan(424050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 12, null, new TimeSpan(423900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 13, null, new TimeSpan(423750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 14, null, new TimeSpan(423600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 15, null, new TimeSpan(423500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 16, null, new TimeSpan(423350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 17, null, new TimeSpan(423200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 18, null, new TimeSpan(423100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 19, null, new TimeSpan(423000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 20, null, new TimeSpan(422900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 21, null, new TimeSpan(422800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 22, null, new TimeSpan(422700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 23, null, new TimeSpan(422600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 24, null, new TimeSpan(422500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 25, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 26, null, new TimeSpan(422400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 27, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 28, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 29, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 30, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 31, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 1, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 2, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 3, null, new TimeSpan(422100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 4, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 5, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 6, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 7, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 8, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 9, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 10, null, new TimeSpan(422350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 11, null, new TimeSpan(422400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 12, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 13, null, new TimeSpan(422550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 14, null, new TimeSpan(422650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 15, null, new TimeSpan(422750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 16, null, new TimeSpan(422850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 17, null, new TimeSpan(422950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 18, null, new TimeSpan(423100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 19, null, new TimeSpan(423200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 20, null, new TimeSpan(423350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 21, null, new TimeSpan(423500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 22, null, new TimeSpan(423650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 23, null, new TimeSpan(423850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 24, null, new TimeSpan(424000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 25, null, new TimeSpan(424200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 26, null, new TimeSpan(424400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 27, null, new TimeSpan(424550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 28, null, new TimeSpan(424750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 29, null, new TimeSpan(425000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 30, null, new TimeSpan(425200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 1, null, new TimeSpan(425400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 2, null, new TimeSpan(425650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 3, null, new TimeSpan(425900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 4, null, new TimeSpan(426150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 5, null, new TimeSpan(426350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 6, null, new TimeSpan(426600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 7, null, new TimeSpan(426900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 8, null, new TimeSpan(427150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 9, null, new TimeSpan(427400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 10, null, new TimeSpan(427700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 11, null, new TimeSpan(427950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 12, null, new TimeSpan(428250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 13, null, new TimeSpan(428500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 14, null, new TimeSpan(428800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 15, null, new TimeSpan(429100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 16, null, new TimeSpan(429400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 17, null, new TimeSpan(429650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 18, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 19, null, new TimeSpan(430250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 20, null, new TimeSpan(430550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 21, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 22, null, new TimeSpan(431150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 23, null, new TimeSpan(431450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 24, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 25, null, new TimeSpan(432050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-50000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 26, null, new TimeSpan(432350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 27, null, new TimeSpan(432650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 28, null, new TimeSpan(432950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 29, null, new TimeSpan(433200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 30, null, new TimeSpan(433500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 31, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaSaigon_283()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Saigon", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Ho_Chi_Minh", new TimeSpan(256000000000), "-", "LMT", 1906, PublicDomain.Month.June, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ho_Chi_Minh", new TimeSpan(255800000000), "-", "SMT", 1911, PublicDomain.Month.March, 11, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, "Saigon MT?"), new PublicDomain.TzDatabase.TzZone("Asia/Ho_Chi_Minh", new TimeSpan(252000000000), "-", "ICT", 1912, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ho_Chi_Minh", new TimeSpan(288000000000), "-", "ICT", 1931, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ho_Chi_Minh", new TimeSpan(252000000000), "-", "ICT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaSakhalin_284()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Sakhalin", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Sakhalin", new TimeSpan(342480000000), "-", "LMT", 1905, PublicDomain.Month.August, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Sakhalin", new TimeSpan(324000000000), "-", "CJT", 1938, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Sakhalin", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.August, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Sakhalin", new TimeSpan(396000000000), "Russia", "SAK%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, "Sakhalin T."), new PublicDomain.TzDatabase.TzZone("Asia/Sakhalin", new TimeSpan(360000000000), "Russia", "SAK%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Sakhalin", new TimeSpan(396000000000), "Russia", "SAK%sT", 1997, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Sakhalin", new TimeSpan(360000000000), "Russia", "SAK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaSamarkand_285()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Samarkand", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Samarkand", new TimeSpan(160320000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Samarkand", new TimeSpan(144000000000), "-", "SAMT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Samarkand Time"), new PublicDomain.TzDatabase.TzZone("Asia/Samarkand", new TimeSpan(180000000000), "-", "SAMT", 1981, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Samarkand", new TimeSpan(180000000000), "1:00", "SAMST", 1981, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Samarkand", new TimeSpan(216000000000), "-", "TAST", 1982, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tashkent Time"), new PublicDomain.TzDatabase.TzZone("Asia/Samarkand", new TimeSpan(180000000000), "RussiaAsia", "SAM%sT", 1991, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Samarkand", new TimeSpan(180000000000), "RussiaAsia", "UZ%sT", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Samarkand", new TimeSpan(180000000000), "-", "UZT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaSeoul_286()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Seoul", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(304720000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(306000000000), "-", "KST", 1904, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(324000000000), "-", "KST", 1928, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(306000000000), "-", "KST", 1932, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(324000000000), "-", "KST", 1954, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(288000000000), "ROK", "K%sT", 1961, PublicDomain.Month.August, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(306000000000), "-", "KST", 1968, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(324000000000), "ROK", "K%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("ROK", 1960, 1960, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("ROK", 1960, 1960, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("ROK", 1987, 1988, PublicDomain.Month.May, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("ROK", 1987, 1988, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaShanghai_287()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Shanghai", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Shanghai", new TimeSpan(291520000000), "-", "LMT", 1928, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Shanghai", new TimeSpan(288000000000), "Shang", "C%sT", 1949, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Shanghai", new TimeSpan(288000000000), "PRC", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Shang", 1940, 1940, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Shang", 1940, 1941, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Shang", 1941, 1941, PublicDomain.Month.March, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1986, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1991, PublicDomain.Month.September, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("PRC", 1987, 1991, PublicDomain.Month.April, 10, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaSingapore_288()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Singapore", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(249250000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(249250000000), "-", "SMT", 1905, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Singapore M.T."), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(252000000000), "-", "MALT", 1933, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Malaya Time"), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(252000000000), "0:20", "MALST", 1936, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(264000000000), "-", "MALT", 1941, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(270000000000), "-", "MALT", 1942, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(270000000000), "-", "MALT", 1965, PublicDomain.Month.August, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(270000000000), "-", "SGT", 1982, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Singapore Time"), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(288000000000), "-", "SGT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("0:20", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(12000000000), null, null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaTaipei_289()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Taipei", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Taipei", new TimeSpan(291600000000), "-", "LMT", 1896, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Taibei or T'ai-pei"), new PublicDomain.TzDatabase.TzZone("Asia/Taipei", new TimeSpan(288000000000), "Taiwan", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Taiwan", 1945, 1951, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1945, 1951, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1952, 1952, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1952, 1954, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1953, 1959, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1955, 1961, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1960, 1961, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1974, 1975, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1974, 1975, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1980, 1980, PublicDomain.Month.June, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1980, 1980, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaTashkent_290()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Tashkent", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Tashkent", new TimeSpan(166320000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tashkent", new TimeSpan(180000000000), "-", "TAST", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tashkent Time"), new PublicDomain.TzDatabase.TzZone("Asia/Tashkent", new TimeSpan(216000000000), "RussiaAsia", "TAS%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tashkent", new TimeSpan(180000000000), "RussiaAsia", "TAS%sT", 1991, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Tashkent", new TimeSpan(180000000000), "RussiaAsia", "UZ%sT", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tashkent", new TimeSpan(180000000000), "-", "UZT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaTbilisi_291()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Tbilisi", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(107560000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(107560000000), "-", "TBMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tbilisi Mean Time"), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(108000000000), "-", "TBIT", 1957, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tbilisi Time"), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(144000000000), "RussiaAsia", "TBI%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(108000000000), "1:00", "TBIST", 1991, PublicDomain.Month.April, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(108000000000), "RussiaAsia", "GE%sT", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Georgia Time"), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(108000000000), "E-EurAsia", "GE%sT", 1994, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(144000000000), "E-EurAsia", "GE%sT", 1996, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(144000000000), "1:00", "GEST", 1997, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(144000000000), "E-EurAsia", "GE%sT", 2004, PublicDomain.Month.June, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(108000000000), "RussiaAsia", "GE%sT", 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tbilisi", new TimeSpan(144000000000), "-", "GET", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("E-EurAsia", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-EurAsia", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-EurAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaTehran_292()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Tehran", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Tehran", new TimeSpan(123440000000), "-", "LMT", 1916, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tehran", new TimeSpan(123440000000), "-", "TMT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tehran Mean Time"), new PublicDomain.TzDatabase.TzZone("Asia/Tehran", new TimeSpan(126000000000), "-", "IRST", 1977, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tehran", new TimeSpan(144000000000), "Iran", "IR%sT", 1979, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tehran", new TimeSpan(126000000000), "Iran", "IR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Iran", 1978, 1980, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 1978, 1978, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 1979, 1979, PublicDomain.Month.September, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 1980, 1980, PublicDomain.Month.September, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 1991, 1991, PublicDomain.Month.May, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 1991, 1995, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 1992, 1995, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 1996, 1996, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 1996, 1996, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 1997, 1999, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 1997, 1999, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2000, 2000, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2000, 2000, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2001, 2003, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2001, 2003, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2004, 2004, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2004, 2004, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2005, 2005, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2005, 2005, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2008, 2008, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2008, 2008, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2009, 2011, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2009, 2011, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2012, 2012, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2012, 2012, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2013, 2015, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2013, 2015, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2016, 2016, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2016, 2016, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2017, 2019, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2017, 2019, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2020, 2020, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2020, 2020, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2021, 2023, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2021, 2023, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2024, 2024, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2024, 2024, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2025, 2027, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2025, 2027, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2028, 2029, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2028, 2029, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2030, 2031, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2030, 2031, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2032, 2033, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2032, 2033, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2034, 2035, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2034, 2035, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2036, 2037, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2036, 2037, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaTel_Aviv_293()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Tel_Aviv", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Jerusalem", new TimeSpan(84560000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jerusalem", new TimeSpan(84400000000), "-", "JMT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Jerusalem Mean Time?"), new PublicDomain.TzDatabase.TzZone("Asia/Jerusalem", new TimeSpan(72000000000), "Zion", "I%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Zion", 1940, 1940, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1942, 1944, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1943, 1943, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1944, 1944, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1945, 1945, PublicDomain.Month.April, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1945, 1945, PublicDomain.Month.November, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1946, 1946, PublicDomain.Month.April, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1946, 1946, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1948, PublicDomain.Month.May, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1948, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1949, PublicDomain.Month.November, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1949, 1949, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1950, 1950, PublicDomain.Month.April, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1950, 1950, PublicDomain.Month.September, 15, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1951, 1951, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1951, 1951, PublicDomain.Month.November, 11, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1952, 1952, PublicDomain.Month.April, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1952, 1952, PublicDomain.Month.October, 19, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1953, 1953, PublicDomain.Month.April, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1953, 1953, PublicDomain.Month.September, 13, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1954, 1954, PublicDomain.Month.June, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1954, 1954, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1955, 1955, PublicDomain.Month.June, 11, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1955, 1955, PublicDomain.Month.September, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1956, 1956, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1956, 1956, PublicDomain.Month.September, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1957, 1957, PublicDomain.Month.April, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1957, 1957, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1974, 1974, PublicDomain.Month.July, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1974, 1974, PublicDomain.Month.October, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1975, 1975, PublicDomain.Month.April, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1975, 1975, PublicDomain.Month.August, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1985, 1985, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1985, 1985, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1986, 1986, PublicDomain.Month.May, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1986, 1986, PublicDomain.Month.September, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1987, 1987, PublicDomain.Month.April, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1987, 1987, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1988, 1988, PublicDomain.Month.April, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1988, 1988, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1989, 1989, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1989, 1989, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1990, 1990, PublicDomain.Month.March, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1990, 1990, PublicDomain.Month.August, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1991, 1991, PublicDomain.Month.March, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1991, 1991, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1992, 1992, PublicDomain.Month.March, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1992, 1992, PublicDomain.Month.September, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1993, 1993, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1993, 1993, PublicDomain.Month.September, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1994, 1994, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1994, 1994, PublicDomain.Month.August, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1995, 1995, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1995, 1995, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1996, 1996, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1996, 1996, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1997, 1997, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1997, 1997, PublicDomain.Month.September, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1998, 1998, PublicDomain.Month.March, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1998, 1998, PublicDomain.Month.September, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1999, 1999, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1999, 1999, PublicDomain.Month.September, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2000, 2000, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2000, 2000, PublicDomain.Month.October, 6, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2001, 2001, PublicDomain.Month.April, 9, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2001, 2001, PublicDomain.Month.September, 24, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2002, 2002, PublicDomain.Month.March, 29, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2002, 2002, PublicDomain.Month.October, 7, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2003, 2003, PublicDomain.Month.March, 28, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2003, 2003, PublicDomain.Month.October, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2004, 2004, PublicDomain.Month.April, 7, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2004, 2004, PublicDomain.Month.September, 22, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2005, 2005, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2005, 2005, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2006, 2010, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2006, 2006, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2007, 2007, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2008, 2008, PublicDomain.Month.October, 5, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2009, 2009, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2010, 2010, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2011, 2011, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2011, 2011, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2012, 2015, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2012, 2012, PublicDomain.Month.September, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2013, 2013, PublicDomain.Month.September, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2014, 2014, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2015, 2015, PublicDomain.Month.September, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2016, 2016, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2016, 2016, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2017, 2021, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2017, 2017, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2018, 2018, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2019, 2019, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2020, 2020, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2021, 2021, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2022, 2022, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2022, 2022, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2023, 2032, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2023, 2023, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2024, 2024, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2025, 2025, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2026, 2026, PublicDomain.Month.September, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2027, 2027, PublicDomain.Month.October, 10, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2028, 2028, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2029, 2029, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2030, 2030, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2031, 2031, PublicDomain.Month.September, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2032, 2032, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2033, 2033, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2033, 2033, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2034, 2037, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2034, 2034, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2035, 2035, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2036, 2036, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2037, 2037, PublicDomain.Month.September, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaThimbu_294()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Thimbu", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Thimphu", new TimeSpan(215160000000), "-", "LMT", 1947, PublicDomain.Month.August, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Thimbu"), new PublicDomain.TzDatabase.TzZone("Asia/Thimphu", new TimeSpan(198000000000), "-", "IST", 1987, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Thimphu", new TimeSpan(216000000000), "-", "BTT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bhutan Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaThimphu_295()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Thimphu", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Thimphu", new TimeSpan(215160000000), "-", "LMT", 1947, PublicDomain.Month.August, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Thimbu"), new PublicDomain.TzDatabase.TzZone("Asia/Thimphu", new TimeSpan(198000000000), "-", "IST", 1987, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Thimphu", new TimeSpan(216000000000), "-", "BTT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bhutan Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaTokyo_296()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Tokyo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Tokyo", new TimeSpan(335390000000), "-", "LMT", 1887, PublicDomain.Month.December, 31, null, new TimeSpan(540000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tokyo", new TimeSpan(324000000000), "-", "JST", 1896, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tokyo", new TimeSpan(324000000000), "-", "CJT", 1938, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tokyo", new TimeSpan(324000000000), "Japan", "J%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Japan", 1948, 1948, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Japan", 1948, 1951, PublicDomain.Month.September, 8, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Japan", 1949, 1949, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Japan", 1950, 1951, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaUjung_Pandang_297()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Ujung_Pandang", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Makassar", new TimeSpan(286560000000), "-", "LMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Makassar", new TimeSpan(286560000000), "-", "MMT", 1932, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Macassar MT"), new PublicDomain.TzDatabase.TzZone("Asia/Makassar", new TimeSpan(288000000000), "-", "CIT", 1942, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Makassar", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.September, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Makassar", new TimeSpan(288000000000), "-", "CIT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaUlaanbaatar_298()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Ulaanbaatar", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Ulaanbaatar", new TimeSpan(256520000000), "-", "LMT", 1905, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ulaanbaatar", new TimeSpan(252000000000), "-", "ULAT", 1978, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Ulaanbaatar Time"), new PublicDomain.TzDatabase.TzZone("Asia/Ulaanbaatar", new TimeSpan(288000000000), "Mongol", "ULA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mongol", 1983, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1983, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1984, 1998, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1985, 1998, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2001, 2001, PublicDomain.Month.April, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2001, 2006, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2002, 2006, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaUlan_Bator_299()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Ulan_Bator", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Ulaanbaatar", new TimeSpan(256520000000), "-", "LMT", 1905, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Ulaanbaatar", new TimeSpan(252000000000), "-", "ULAT", 1978, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Ulaanbaatar Time"), new PublicDomain.TzDatabase.TzZone("Asia/Ulaanbaatar", new TimeSpan(288000000000), "Mongol", "ULA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mongol", 1983, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1983, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1984, 1998, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 1985, 1998, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2001, 2001, PublicDomain.Month.April, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2001, 2006, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mongol", 2002, 2006, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaUrumqi_300()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Urumqi", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Urumqi", new TimeSpan(210200000000), "-", "LMT", 1928, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Urumchi"), new PublicDomain.TzDatabase.TzZone("Asia/Urumqi", new TimeSpan(216000000000), "-", "URUT", 1980, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Urumqi Time"), new PublicDomain.TzDatabase.TzZone("Asia/Urumqi", new TimeSpan(288000000000), "PRC", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1986, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1991, PublicDomain.Month.September, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("PRC", 1987, 1991, PublicDomain.Month.April, 10, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaVientiane_301()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Vientiane", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Vientiane", new TimeSpan(246240000000), "-", "LMT", 1906, PublicDomain.Month.June, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Viangchan"), new PublicDomain.TzDatabase.TzZone("Asia/Vientiane", new TimeSpan(255800000000), "-", "SMT", 1911, PublicDomain.Month.March, 11, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, "Saigon MT?"), new PublicDomain.TzDatabase.TzZone("Asia/Vientiane", new TimeSpan(252000000000), "-", "ICT", 1912, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Vientiane", new TimeSpan(288000000000), "-", "ICT", 1931, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Vientiane", new TimeSpan(252000000000), "-", "ICT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaVladivostok_302()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Vladivostok", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Vladivostok", new TimeSpan(316640000000), "-", "LMT", 1922, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Vladivostok", new TimeSpan(324000000000), "-", "VLAT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Vladivostok Time"), new PublicDomain.TzDatabase.TzZone("Asia/Vladivostok", new TimeSpan(360000000000), "Russia", "VLA%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Vladivostok", new TimeSpan(324000000000), "Russia", "VLA%sST", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Vladivostok", new TimeSpan(360000000000), "Russia", "VLA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaYakutsk_303()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Yakutsk", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Yakutsk", new TimeSpan(311200000000), "-", "LMT", 1919, PublicDomain.Month.December, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Yakutsk", new TimeSpan(288000000000), "-", "YAKT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Yakutsk Time"), new PublicDomain.TzDatabase.TzZone("Asia/Yakutsk", new TimeSpan(324000000000), "Russia", "YAK%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Yakutsk", new TimeSpan(288000000000), "Russia", "YAK%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Yakutsk", new TimeSpan(324000000000), "Russia", "YAK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaYekaterinburg_304()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Yekaterinburg", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Yekaterinburg", new TimeSpan(145440000000), "-", "LMT", 1919, PublicDomain.Month.July, 15, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Yekaterinburg", new TimeSpan(144000000000), "-", "SVET", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Sverdlovsk Time"), new PublicDomain.TzDatabase.TzZone("Asia/Yekaterinburg", new TimeSpan(180000000000), "Russia", "SVE%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Yekaterinburg", new TimeSpan(144000000000), "Russia", "SVE%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Yekaterinburg", new TimeSpan(180000000000), "Russia", "YEK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Yekaterinburg Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAsiaYerevan_305()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Asia/Yerevan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Yerevan", new TimeSpan(106800000000), "-", "LMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Yerevan", new TimeSpan(108000000000), "-", "YERT", 1957, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Yerevan Time"), new PublicDomain.TzDatabase.TzZone("Asia/Yerevan", new TimeSpan(144000000000), "RussiaAsia", "YER%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Yerevan", new TimeSpan(108000000000), "1:00", "YERST", 1991, PublicDomain.Month.September, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Yerevan", new TimeSpan(108000000000), "RussiaAsia", "AM%sT", 1995, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Yerevan", new TimeSpan(144000000000), "-", "AMT", 1997, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Yerevan", new TimeSpan(144000000000), "RussiaAsia", "AM%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("RussiaAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticAzores_306()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/Azores", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/Azores", new TimeSpan(-61600000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Ponta Delgada"), new PublicDomain.TzDatabase.TzZone("Atlantic/Azores", new TimeSpan(-68720000000), "-", "HMT", 1911, PublicDomain.Month.May, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Horta Mean Time"), new PublicDomain.TzDatabase.TzZone("Atlantic/Azores", new TimeSpan(-72000000000), "Port", "AZO%sT", 1966, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, "Azores Time"), new PublicDomain.TzDatabase.TzZone("Atlantic/Azores", new TimeSpan(-36000000000), "Port", "AZO%sT", 1983, PublicDomain.Month.September, 25, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Azores", new TimeSpan(-36000000000), "W-Eur", "AZO%sT", 1992, PublicDomain.Month.September, 27, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Azores", TimeSpan.Zero, "EU", "WE%sT", 1993, PublicDomain.Month.March, 28, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Azores", new TimeSpan(-36000000000), "EU", "AZO%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Port", 1916, 1916, PublicDomain.Month.June, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1916, 1916, PublicDomain.Month.November, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1917, 1917, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1917, 1921, PublicDomain.Month.October, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1918, 1918, PublicDomain.Month.March, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1919, 1919, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1920, 1920, PublicDomain.Month.February, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1921, 1921, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1924, 1924, PublicDomain.Month.April, 16, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1924, 1924, PublicDomain.Month.October, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1926, 1929, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1929, 1929, PublicDomain.Month.April, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1931, 1931, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1931, 1932, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1932, 1932, PublicDomain.Month.April, 2, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1934, 1934, PublicDomain.Month.April, 7, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1934, 1938, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1935, 1935, PublicDomain.Month.March, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1936, 1936, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1937, 1937, PublicDomain.Month.April, 3, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1938, 1938, PublicDomain.Month.March, 26, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1939, 1939, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1939, 1939, PublicDomain.Month.November, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1940, 1940, PublicDomain.Month.February, 24, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1940, 1941, PublicDomain.Month.October, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1941, 1941, PublicDomain.Month.April, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1945, PublicDomain.Month.March, 8, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1942, PublicDomain.Month.April, 25, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1942, PublicDomain.Month.August, 15, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1945, PublicDomain.Month.October, 24, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1943, 1943, PublicDomain.Month.April, 17, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Port", 1943, 1945, PublicDomain.Month.August, 25, DayOfWeek.Saturday, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1944, 1945, PublicDomain.Month.April, 21, DayOfWeek.Saturday, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Port", 1946, 1946, PublicDomain.Month.April, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1946, 1946, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1947, 1949, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1947, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1951, 1965, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1951, 1965, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1977, 1977, PublicDomain.Month.March, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1977, 1977, PublicDomain.Month.September, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1978, 1979, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1979, 1982, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1980, 1980, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1981, 1982, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1983, 1983, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticBermuda_307()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/Bermuda", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/Bermuda", new TimeSpan(-155440000000), "-", "LMT", 1930, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, "Hamilton"), new PublicDomain.TzDatabase.TzZone("Atlantic/Bermuda", new TimeSpan(-144000000000), "-", "AST", 1974, PublicDomain.Month.April, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Bermuda", new TimeSpan(-144000000000), "Bahamas", "A%sT", 1976, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Bermuda", new TimeSpan(-144000000000), "US", "A%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Bahamas", 1964, 1975, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Bahamas", 1964, 1975, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticCanary_308()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/Canary", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/Canary", new TimeSpan(-36960000000), "-", "LMT", 1922, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Las Palmas de Gran C."), new PublicDomain.TzDatabase.TzZone("Atlantic/Canary", new TimeSpan(-36000000000), "-", "CANT", 1946, PublicDomain.Month.September, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, "Canaries Time"), new PublicDomain.TzDatabase.TzZone("Atlantic/Canary", TimeSpan.Zero, "-", "WET", 1980, PublicDomain.Month.April, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Canary", TimeSpan.Zero, "1:00", "WEST", 1980, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Canary", TimeSpan.Zero, "EU", "WE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticCape_Verde_309()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/Cape_Verde", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/Cape_Verde", new TimeSpan(-56440000000), "-", "LMT", 1907, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Praia"), new PublicDomain.TzDatabase.TzZone("Atlantic/Cape_Verde", new TimeSpan(-72000000000), "-", "CVT", 1942, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Cape_Verde", new TimeSpan(-72000000000), "1:00", "CVST", 1945, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Cape_Verde", new TimeSpan(-72000000000), "-", "CVT", 1975, PublicDomain.Month.November, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Cape_Verde", new TimeSpan(-36000000000), "-", "CVT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticFaeroe_310()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/Faeroe", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/Faroe", new TimeSpan(-16240000000), "-", "LMT", 1908, PublicDomain.Month.January, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Torshavn"), new PublicDomain.TzDatabase.TzZone("Atlantic/Faroe", TimeSpan.Zero, "-", "WET", 1981, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Faroe", TimeSpan.Zero, "EU", "WE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticFaroe_311()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/Faroe", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/Faroe", new TimeSpan(-16240000000), "-", "LMT", 1908, PublicDomain.Month.January, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Torshavn"), new PublicDomain.TzDatabase.TzZone("Atlantic/Faroe", TimeSpan.Zero, "-", "WET", 1981, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Faroe", TimeSpan.Zero, "EU", "WE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticJan_Mayen_312()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/Jan_Mayen", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(25800000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "Norway", "CE%sT", 1940, PublicDomain.Month.August, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "Norway", "CE%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Norway", 1916, 1916, PublicDomain.Month.May, 22, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Norway", 1916, 1916, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1945, 1945, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1959, 1964, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Norway", 1959, 1965, PublicDomain.Month.September, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1965, 1965, PublicDomain.Month.April, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticMadeira_313()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/Madeira", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/Madeira", new TimeSpan(-40560000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Funchal"), new PublicDomain.TzDatabase.TzZone("Atlantic/Madeira", new TimeSpan(-40560000000), "-", "FMT", 1911, PublicDomain.Month.May, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Funchal Mean Time"), new PublicDomain.TzDatabase.TzZone("Atlantic/Madeira", new TimeSpan(-36000000000), "Port", "MAD%sT", 1966, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, "Madeira Time"), new PublicDomain.TzDatabase.TzZone("Atlantic/Madeira", TimeSpan.Zero, "Port", "WE%sT", 1983, PublicDomain.Month.September, 25, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Madeira", TimeSpan.Zero, "EU", "WE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Port", 1916, 1916, PublicDomain.Month.June, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1916, 1916, PublicDomain.Month.November, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1917, 1917, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1917, 1921, PublicDomain.Month.October, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1918, 1918, PublicDomain.Month.March, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1919, 1919, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1920, 1920, PublicDomain.Month.February, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1921, 1921, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1924, 1924, PublicDomain.Month.April, 16, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1924, 1924, PublicDomain.Month.October, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1926, 1929, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1929, 1929, PublicDomain.Month.April, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1931, 1931, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1931, 1932, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1932, 1932, PublicDomain.Month.April, 2, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1934, 1934, PublicDomain.Month.April, 7, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1934, 1938, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1935, 1935, PublicDomain.Month.March, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1936, 1936, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1937, 1937, PublicDomain.Month.April, 3, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1938, 1938, PublicDomain.Month.March, 26, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1939, 1939, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1939, 1939, PublicDomain.Month.November, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1940, 1940, PublicDomain.Month.February, 24, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1940, 1941, PublicDomain.Month.October, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1941, 1941, PublicDomain.Month.April, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1945, PublicDomain.Month.March, 8, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1942, PublicDomain.Month.April, 25, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1942, PublicDomain.Month.August, 15, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1945, PublicDomain.Month.October, 24, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1943, 1943, PublicDomain.Month.April, 17, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Port", 1943, 1945, PublicDomain.Month.August, 25, DayOfWeek.Saturday, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1944, 1945, PublicDomain.Month.April, 21, DayOfWeek.Saturday, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Port", 1946, 1946, PublicDomain.Month.April, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1946, 1946, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1947, 1949, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1947, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1951, 1965, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1951, 1965, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1977, 1977, PublicDomain.Month.March, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1977, 1977, PublicDomain.Month.September, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1978, 1979, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1979, 1982, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1980, 1980, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1981, 1982, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1983, 1983, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticReykjavik_314()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/Reykjavik", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/Reykjavik", new TimeSpan(-52440000000), "-", "LMT", 1837, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Reykjavik", new TimeSpan(-52680000000), "-", "RMT", 1908, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Reykjavik Mean Time?"), new PublicDomain.TzDatabase.TzZone("Atlantic/Reykjavik", new TimeSpan(-36000000000), "Iceland", "IS%sT", 1968, PublicDomain.Month.April, 7, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Reykjavik", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Iceland", 1917, 1918, PublicDomain.Month.February, 19, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1917, 1917, PublicDomain.Month.October, 21, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1918, 1918, PublicDomain.Month.November, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1939, 1939, PublicDomain.Month.April, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1939, 1939, PublicDomain.Month.November, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1940, 1940, PublicDomain.Month.February, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1940, 1940, PublicDomain.Month.November, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1941, 1941, PublicDomain.Month.March, 2, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1941, 1941, PublicDomain.Month.November, 2, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1942, 1942, PublicDomain.Month.March, 8, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1942, 1942, PublicDomain.Month.October, 25, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1943, 1946, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1943, 1948, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1947, 1967, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1950, 1966, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1967, 1967, PublicDomain.Month.October, 29, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticSouth_Georgia_315()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/South_Georgia", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/South_Georgia", new TimeSpan(-87680000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Grytviken"), new PublicDomain.TzDatabase.TzZone("Atlantic/South_Georgia", new TimeSpan(-72000000000), "-", "GST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "South Georgia Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticSt_Helena_316()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/St_Helena", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/St_Helena", new TimeSpan(-13680000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Jamestown"), new PublicDomain.TzDatabase.TzZone("Atlantic/St_Helena", new TimeSpan(-13680000000), "-", "JMT", 1951, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Jamestown Mean Time"), new PublicDomain.TzDatabase.TzZone("Atlantic/St_Helena", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAtlanticStanley_317()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Atlantic/Stanley", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/Stanley", new TimeSpan(-138840000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Stanley", new TimeSpan(-138840000000), "-", "SMT", 1912, PublicDomain.Month.March, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Stanley Mean Time"), new PublicDomain.TzDatabase.TzZone("Atlantic/Stanley", new TimeSpan(-144000000000), "Falk", "FK%sT", 1983, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Falkland Is Time"), new PublicDomain.TzDatabase.TzZone("Atlantic/Stanley", new TimeSpan(-108000000000), "Falk", "FK%sT", 1985, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Stanley", new TimeSpan(-144000000000), "Falk", "FK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Falk", 1937, 1938, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Falk", 1938, 1942, PublicDomain.Month.March, 19, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Falk", 1939, 1939, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Falk", 1940, 1942, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Falk", 1943, 1943, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Falk", 1983, 1983, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Falk", 1984, 1985, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Falk", 1984, 1984, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Falk", 1985, 2000, PublicDomain.Month.September, 9, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Falk", 1986, 2000, PublicDomain.Month.April, 16, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Falk", 2001, 2147483647, PublicDomain.Month.April, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Falk", 2001, 2147483647, PublicDomain.Month.September, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaACT_318()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/ACT", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(362920000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(360000000000), "Aus", "EST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(360000000000), "AN", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1972, 1972, PublicDomain.Month.February, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1973, 1981, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1982, 1982, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1983, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1987, 1999, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1990, 1995, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1996, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2001, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaAdelaide_319()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Adelaide", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Adelaide", new TimeSpan(332600000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Adelaide", new TimeSpan(324000000000), "-", "CST", 1899, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Adelaide", new TimeSpan(342000000000), "Aus", "CST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Adelaide", new TimeSpan(342000000000), "AS", "CST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1972, 1972, PublicDomain.Month.February, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1973, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1987, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1990, 1990, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1991, 1991, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1992, 1992, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1993, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1994, 1994, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1995, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaBrisbane_320()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Brisbane", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Brisbane", new TimeSpan(367280000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Brisbane", new TimeSpan(360000000000), "Aus", "EST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Brisbane", new TimeSpan(360000000000), "AQ", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1971, 1971, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1972, 1972, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1989, 1991, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1990, 1992, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaBroken_Hill_321()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Broken_Hill", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(339480000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(360000000000), "-", "EST", 1896, PublicDomain.Month.August, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(324000000000), "-", "CST", 1899, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(342000000000), "Aus", "CST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(342000000000), "AN", "CST", 2000, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(342000000000), "AS", "CST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1972, 1972, PublicDomain.Month.February, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1972, 1972, PublicDomain.Month.February, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1973, 1981, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1973, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1982, 1982, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1983, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1987, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1987, 1999, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1990, 1995, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1990, 1990, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1991, 1991, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1992, 1992, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1993, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1994, 1994, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1995, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1996, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2001, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaCanberra_322()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Canberra", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(362920000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(360000000000), "Aus", "EST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(360000000000), "AN", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1972, 1972, PublicDomain.Month.February, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1973, 1981, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1982, 1982, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1983, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1987, 1999, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1990, 1995, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1996, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2001, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaCurrie_323()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Currie", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Currie", new TimeSpan(345280000000), "-", "LMT", 1895, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Currie", new TimeSpan(360000000000), "-", "EST", 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Currie", new TimeSpan(360000000000), "1:00", "EST", 1917, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Currie", new TimeSpan(360000000000), "Aus", "EST", 1971, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Currie", new TimeSpan(360000000000), "AT", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1967, 1967, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1968, 1968, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1968, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1969, 1971, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1972, 1972, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1973, 1981, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1982, 1983, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1984, 1986, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1986, 1986, PublicDomain.Month.October, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1987, 1990, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1987, 1987, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1988, 1990, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1991, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1991, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2001, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaDarwin_324()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Darwin", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Darwin", new TimeSpan(314000000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Darwin", new TimeSpan(324000000000), "-", "CST", 1899, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Darwin", new TimeSpan(342000000000), "Aus", "CST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaEucla_325()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Eucla", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Eucla", new TimeSpan(309280000000), "-", "LMT", 1895, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Eucla", new TimeSpan(315000000000), "Aus", "CWST", 1943, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Eucla", new TimeSpan(315000000000), "AW", "CWST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1974, 1974, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1975, 1975, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1983, 1983, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1984, 1984, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1991, 1991, PublicDomain.Month.November, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1992, 1992, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 2006, 2006, PublicDomain.Month.December, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 2007, 2009, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 2007, 2008, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaHobart_326()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Hobart", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Hobart", new TimeSpan(353560000000), "-", "LMT", 1895, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Hobart", new TimeSpan(360000000000), "-", "EST", 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Hobart", new TimeSpan(360000000000), "1:00", "EST", 1917, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Hobart", new TimeSpan(360000000000), "Aus", "EST", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Hobart", new TimeSpan(360000000000), "AT", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1967, 1967, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1968, 1968, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1968, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1969, 1971, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1972, 1972, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1973, 1981, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1982, 1983, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1984, 1986, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1986, 1986, PublicDomain.Month.October, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1987, 1990, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1987, 1987, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1988, 1990, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1991, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1991, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2001, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaLHI_327()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/LHI", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Lord_Howe", new TimeSpan(381800000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Lord_Howe", new TimeSpan(360000000000), "-", "EST", 1981, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Lord_Howe", new TimeSpan(378000000000), "LH", "LHST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("LH", 1981, 1984, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1982, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1985, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1987, 1999, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1990, 1995, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1996, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2001, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaLindeman_328()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Lindeman", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Lindeman", new TimeSpan(357560000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Lindeman", new TimeSpan(360000000000), "Aus", "EST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Lindeman", new TimeSpan(360000000000), "AQ", "EST", 1992, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Lindeman", new TimeSpan(360000000000), "Holiday", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1971, 1971, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1972, 1972, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1989, 1991, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1990, 1992, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Holiday", 1992, 1993, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Holiday", 1993, 1994, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaLord_Howe_329()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Lord_Howe", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Lord_Howe", new TimeSpan(381800000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Lord_Howe", new TimeSpan(360000000000), "-", "EST", 1981, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Lord_Howe", new TimeSpan(378000000000), "LH", "LHST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("LH", 1981, 1984, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1982, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1985, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1987, 1999, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1990, 1995, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 1996, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2001, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("LH", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaMelbourne_330()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Melbourne", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Melbourne", new TimeSpan(347920000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Melbourne", new TimeSpan(360000000000), "Aus", "EST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Melbourne", new TimeSpan(360000000000), "AV", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1972, 1972, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1973, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1986, 1990, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1986, 1987, PublicDomain.Month.October, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1988, 1999, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1991, 1994, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1995, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2001, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaNorth_331()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/North", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Darwin", new TimeSpan(314000000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Darwin", new TimeSpan(324000000000), "-", "CST", 1899, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Darwin", new TimeSpan(342000000000), "Aus", "CST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaNSW_332()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/NSW", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(362920000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(360000000000), "Aus", "EST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(360000000000), "AN", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1972, 1972, PublicDomain.Month.February, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1973, 1981, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1982, 1982, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1983, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1987, 1999, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1990, 1995, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1996, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2001, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaPerth_333()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Perth", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Perth", new TimeSpan(278040000000), "-", "LMT", 1895, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Perth", new TimeSpan(288000000000), "Aus", "WST", 1943, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Perth", new TimeSpan(288000000000), "AW", "WST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1974, 1974, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1975, 1975, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1983, 1983, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1984, 1984, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1991, 1991, PublicDomain.Month.November, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1992, 1992, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 2006, 2006, PublicDomain.Month.December, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 2007, 2009, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 2007, 2008, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaQueensland_334()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Queensland", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Brisbane", new TimeSpan(367280000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Brisbane", new TimeSpan(360000000000), "Aus", "EST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Brisbane", new TimeSpan(360000000000), "AQ", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1971, 1971, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1972, 1972, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1989, 1991, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AQ", 1990, 1992, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaSouth_335()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/South", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Adelaide", new TimeSpan(332600000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Adelaide", new TimeSpan(324000000000), "-", "CST", 1899, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Adelaide", new TimeSpan(342000000000), "Aus", "CST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Adelaide", new TimeSpan(342000000000), "AS", "CST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1972, 1972, PublicDomain.Month.February, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1973, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1987, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1990, 1990, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1991, 1991, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1992, 1992, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1993, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1994, 1994, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1995, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaSydney_336()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Sydney", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(362920000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(360000000000), "Aus", "EST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Sydney", new TimeSpan(360000000000), "AN", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1972, 1972, PublicDomain.Month.February, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1973, 1981, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1982, 1982, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1983, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1987, 1999, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1990, 1995, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1996, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2001, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaTasmania_337()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Tasmania", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Hobart", new TimeSpan(353560000000), "-", "LMT", 1895, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Hobart", new TimeSpan(360000000000), "-", "EST", 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Hobart", new TimeSpan(360000000000), "1:00", "EST", 1917, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Hobart", new TimeSpan(360000000000), "Aus", "EST", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Hobart", new TimeSpan(360000000000), "AT", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1967, 1967, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1968, 1968, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1968, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1969, 1971, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1972, 1972, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1973, 1981, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1982, 1983, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1984, 1986, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1986, 1986, PublicDomain.Month.October, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1987, 1990, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1987, 1987, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1988, 1990, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1991, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 1991, 1999, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2001, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AT", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaVictoria_338()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Victoria", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Melbourne", new TimeSpan(347920000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Melbourne", new TimeSpan(360000000000), "Aus", "EST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Melbourne", new TimeSpan(360000000000), "AV", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1972, 1972, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1973, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1986, 1990, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1986, 1987, PublicDomain.Month.October, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1988, 1999, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1991, 1994, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 1995, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2001, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AV", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaWest_339()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/West", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Perth", new TimeSpan(278040000000), "-", "LMT", 1895, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Perth", new TimeSpan(288000000000), "Aus", "WST", 1943, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Perth", new TimeSpan(288000000000), "AW", "WST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1974, 1974, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1975, 1975, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1983, 1983, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1984, 1984, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1991, 1991, PublicDomain.Month.November, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 1992, 1992, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 2006, 2006, PublicDomain.Month.December, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AW", 2007, 2009, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AW", 2007, 2008, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesAustraliaYancowinna_340()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Australia/Yancowinna", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(339480000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(360000000000), "-", "EST", 1896, PublicDomain.Month.August, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(324000000000), "-", "CST", 1899, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(342000000000), "Aus", "CST", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(342000000000), "AN", "CST", 2000, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Australia/Broken_Hill", new TimeSpan(342000000000), "AS", "CST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1917, 1917, PublicDomain.Month.March, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1942, 1942, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1944, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Aus", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1971, 1985, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1972, 1972, PublicDomain.Month.February, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1972, 1972, PublicDomain.Month.February, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1973, 1981, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1973, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1982, 1982, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1983, 1985, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1989, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1986, 1986, PublicDomain.Month.October, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1987, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1987, 1999, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1990, 1995, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1990, 1990, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1991, 1991, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1992, 1992, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1993, 1993, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1994, 1994, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 1995, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 1996, 2005, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2000, 2000, PublicDomain.Month.August, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2001, 2007, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2006, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2007, 2007, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("AN", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null), new PublicDomain.TzDatabase.TzRule("AS", 2008, 2147483647, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesBrazilAcre_341()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Brazil/Acre", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-162720000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-180000000000), "Brazil", "AC%sT", 1988, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-180000000000), "-", "ACT", 2008, PublicDomain.Month.June, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Rio_Branco", new TimeSpan(-144000000000), "-", "AMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesBrazilDeNoronha_342()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Brazil/DeNoronha", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-77800000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "Brazil", "FN%sT", 1990, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "-", "FNT", 1999, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "Brazil", "FN%sT", 2000, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "-", "FNT", 2001, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "Brazil", "FN%sT", 2002, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Noronha", new TimeSpan(-72000000000), "-", "FNT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesBrazilEast_343()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Brazil/East", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Sao_Paulo", new TimeSpan(-111880000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Sao_Paulo", new TimeSpan(-108000000000), "Brazil", "BR%sT", 1963, PublicDomain.Month.October, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Sao_Paulo", new TimeSpan(-108000000000), "1:00", "BRST", 1964, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Sao_Paulo", new TimeSpan(-108000000000), "Brazil", "BR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesBrazilWest_344()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Brazil/West", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Manaus", new TimeSpan(-144040000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Manaus", new TimeSpan(-144000000000), "Brazil", "AM%sT", 1988, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Manaus", new TimeSpan(-144000000000), "-", "AMT", 1993, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Manaus", new TimeSpan(-144000000000), "Brazil", "AM%sT", 1994, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Manaus", new TimeSpan(-144000000000), "-", "AMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Brazil", 1931, 1931, PublicDomain.Month.October, 3, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1933, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1932, 1932, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1949, 1952, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1950, 1950, PublicDomain.Month.April, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1951, 1952, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1953, 1953, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1963, 1963, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1964, 1964, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1965, 1965, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1968, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1966, 1967, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1985, 1985, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1986, 1986, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.February, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1987, 1987, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.February, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1988, 1988, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.January, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1989, 1989, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1990, 1990, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.February, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1991, 1991, PublicDomain.Month.October, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.February, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1992, 1992, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1993, PublicDomain.Month.January, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1993, 1995, PublicDomain.Month.October, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1994, 1995, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.February, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1997, 1997, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1998, 1998, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 1999, 1999, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2000, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2000, 2001, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2001, 2006, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2002, 2002, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2003, 2003, PublicDomain.Month.October, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2004, 2004, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2005, 2005, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2006, 2006, PublicDomain.Month.November, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2007, PublicDomain.Month.February, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2007, 2147483647, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Brazil", 2008, 2147483647, PublicDomain.Month.February, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCanadaAtlantic_345()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Canada/Atlantic", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-152640000000), "-", "LMT", 1902, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Halifax", "A%sT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Canada", "A%sT", 1919, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Halifax", "A%sT", 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Canada", "A%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Halifax", "A%sT", 1974, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Halifax", new TimeSpan(-144000000000), "Canada", "A%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Halifax", 1916, 1916, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1920, 1920, PublicDomain.Month.May, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1920, 1920, PublicDomain.Month.August, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1921, 1921, PublicDomain.Month.May, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1921, 1922, PublicDomain.Month.September, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1922, 1922, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1923, 1925, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1923, 1923, PublicDomain.Month.September, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1924, 1924, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1925, 1925, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1926, 1926, PublicDomain.Month.May, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1926, 1926, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1927, 1927, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1927, 1927, PublicDomain.Month.September, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1928, 1931, PublicDomain.Month.May, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1928, 1928, PublicDomain.Month.September, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1929, 1929, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1930, 1930, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1931, 1932, PublicDomain.Month.September, 24, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1932, 1932, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1933, 1933, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1933, 1933, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1934, 1934, PublicDomain.Month.May, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1934, 1934, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1935, 1935, PublicDomain.Month.June, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1935, 1935, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1936, 1936, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1936, 1936, PublicDomain.Month.September, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1937, 1938, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1937, 1941, PublicDomain.Month.September, 24, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1939, 1939, PublicDomain.Month.May, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1940, 1941, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1946, 1949, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1946, 1949, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1951, 1954, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1951, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1956, 1959, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1956, 1959, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1962, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Halifax", 1962, 1973, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCanadaCentral_346()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Canada/Central", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Winnipeg", new TimeSpan(-233160000000), "-", "LMT", 1887, PublicDomain.Month.July, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Winnipeg", new TimeSpan(-216000000000), "Winn", "C%sT", 2006, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Winnipeg", new TimeSpan(-216000000000), "Canada", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Winn", 1916, 1916, PublicDomain.Month.April, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1916, 1916, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1937, 1937, PublicDomain.Month.May, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1937, 1937, PublicDomain.Month.September, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Winn", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Winn", 1945, 1945, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1946, 1946, PublicDomain.Month.May, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1946, 1946, PublicDomain.Month.October, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1947, 1949, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1947, 1949, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1950, 1950, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1950, 1950, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1951, 1960, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1951, 1958, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1959, 1959, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1960, 1960, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1963, 1963, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1963, 1963, PublicDomain.Month.September, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Winn", 1966, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1966, 2005, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Winn", 1987, 2005, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCanadaEastern_347()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Canada/Eastern", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-190520000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-180000000000), "Canada", "E%sT", 1919, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-180000000000), "Toronto", "E%sT", 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-180000000000), "Canada", "E%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-180000000000), "Toronto", "E%sT", 1974, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Toronto", new TimeSpan(-180000000000), "Canada", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(846000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1919, 1919, PublicDomain.Month.October, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1920, 1920, PublicDomain.Month.May, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1920, 1920, PublicDomain.Month.September, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1921, 1921, PublicDomain.Month.May, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1921, 1921, PublicDomain.Month.September, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1922, 1923, PublicDomain.Month.May, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1922, 1926, PublicDomain.Month.September, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1924, 1927, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1927, 1932, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1928, 1931, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1932, 1932, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1933, 1940, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1933, 1933, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1934, 1939, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Toronto", 1945, 1946, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1946, 1946, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1947, 1949, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1947, 1948, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1949, 1949, PublicDomain.Month.November, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1950, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1950, 1950, PublicDomain.Month.November, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1951, 1956, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Toronto", 1957, 1973, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCanadaEastSaskatchewan_348()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Canada/East-Saskatchewan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Regina", new TimeSpan(-251160000000), "-", "LMT", 1905, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Regina", new TimeSpan(-252000000000), "Regina", "M%sT", 1960, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Regina", new TimeSpan(-216000000000), "-", "CST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Regina", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1930, 1934, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1930, 1934, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1937, 1941, PublicDomain.Month.April, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1937, 1937, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1938, 1938, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1939, 1941, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Regina", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Regina", 1945, 1945, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1946, 1946, PublicDomain.Month.April, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1946, 1946, PublicDomain.Month.October, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1947, 1957, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1947, 1957, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1959, 1959, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1959, 1959, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCanadaMountain_349()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Canada/Mountain", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Edmonton", new TimeSpan(-272320000000), "-", "LMT", 1906, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Edmonton", new TimeSpan(-252000000000), "Edm", "M%sT", 1987, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Edmonton", new TimeSpan(-252000000000), "Canada", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Edm", 1918, 1919, PublicDomain.Month.April, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1919, 1919, PublicDomain.Month.May, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1920, 1923, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1920, 1920, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1921, 1923, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Edm", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1945, 1945, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1947, 1947, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1947, 1947, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1967, 1967, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1967, 1967, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1969, 1969, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1969, 1969, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Edm", 1972, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Edm", 1972, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCanadaNewfoundland_350()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Canada/Newfoundland", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126520000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126520000000), "StJohns", "N%sT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126520000000), "Canada", "N%sT", 1919, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126520000000), "StJohns", "N%sT", 1935, PublicDomain.Month.March, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126000000000), "StJohns", "N%sT", 1942, PublicDomain.Month.May, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126000000000), "Canada", "N%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/St_Johns", new TimeSpan(-126000000000), "StJohns", "N%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("StJohns", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1919, 1919, PublicDomain.Month.May, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1919, 1919, PublicDomain.Month.August, 12, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1920, 1935, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1920, 1935, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1936, 1941, PublicDomain.Month.May, 9, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1936, 1941, PublicDomain.Month.October, 2, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1946, 1950, PublicDomain.Month.May, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1946, 1950, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1951, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1951, 1959, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1960, 1986, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1987, 1987, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1987, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1988, 1988, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("StJohns", 1989, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("StJohns", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCanadaPacific_351()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Canada/Pacific", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Vancouver", new TimeSpan(-295480000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Vancouver", new TimeSpan(-288000000000), "Vanc", "P%sT", 1987, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Vancouver", new TimeSpan(-288000000000), "Canada", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Vanc", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Vanc", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1946, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1946, 1946, PublicDomain.Month.October, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1947, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Vanc", 1962, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCanadaSaskatchewan_352()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Canada/Saskatchewan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Regina", new TimeSpan(-251160000000), "-", "LMT", 1905, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Regina", new TimeSpan(-252000000000), "Regina", "M%sT", 1960, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Regina", new TimeSpan(-216000000000), "-", "CST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Regina", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1930, 1934, PublicDomain.Month.May, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1930, 1934, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1937, 1941, PublicDomain.Month.April, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1937, 1937, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1938, 1938, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1939, 1941, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Regina", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("Regina", 1945, 1945, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1946, 1946, PublicDomain.Month.April, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1946, 1946, PublicDomain.Month.October, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1947, 1957, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1947, 1957, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Regina", 1959, 1959, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Regina", 1959, 1959, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCanadaYukon_353()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Canada/Yukon", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Whitehorse", new TimeSpan(-324120000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Whitehorse", new TimeSpan(-324000000000), "NT_YK", "Y%sT", 1966, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Whitehorse", new TimeSpan(-288000000000), "NT_YK", "P%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Whitehorse", new TimeSpan(-288000000000), "Canada", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1918, 1918, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1918, 1918, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.May, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1919, 1919, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("NT_YK", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1965, 1965, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1974, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1980, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NT_YK", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Canada", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCET_354()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("CET", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("CET", new TimeSpan(36000000000), "C-Eur", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesChileContinental_355()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Chile/Continental", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-169660000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-169660000000), "-", "SMT", 1910, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Santiago Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-180000000000), "-", "CLT", 1916, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Chile Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-169660000000), "-", "SMT", 1918, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Santiago Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-144000000000), "-", "CLT", 1919, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Chile Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-169660000000), "-", "SMT", 1927, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Santiago Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-180000000000), "Chile", "CL%sT", 1947, PublicDomain.Month.May, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Chile Time"), new PublicDomain.TzDatabase.TzZone("America/Santiago", new TimeSpan(-144000000000), "Chile", "CL%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Chile", 1927, 1932, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1928, 1932, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1942, 1942, PublicDomain.Month.June, 1, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1942, 1942, PublicDomain.Month.August, 1, null, new TimeSpan(180000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1946, 1946, PublicDomain.Month.July, 15, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1946, 1946, PublicDomain.Month.September, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1947, 1947, PublicDomain.Month.April, 1, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1968, 1968, PublicDomain.Month.November, 3, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1969, 1969, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1969, 1969, PublicDomain.Month.November, 23, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1970, 1970, PublicDomain.Month.March, 29, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1970, 1972, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1971, 1971, PublicDomain.Month.March, 14, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1972, 1986, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1973, 1973, PublicDomain.Month.September, 30, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1974, 1987, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1987, 1987, PublicDomain.Month.April, 12, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1988, 1989, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1988, 1988, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1989, 1989, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1990, 1990, PublicDomain.Month.March, 18, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1990, 1990, PublicDomain.Month.September, 16, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1991, 1996, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1991, 1997, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1997, 1997, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1998, 1998, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1998, 1998, PublicDomain.Month.September, 27, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1999, 1999, PublicDomain.Month.April, 4, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1999, 2147483647, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 2000, 2007, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 2008, 2008, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 2009, 2147483647, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesChileEasterIsland_356()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Chile/EasterIsland", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Easter", new TimeSpan(-262640000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Easter", new TimeSpan(-262480000000), "-", "EMT", 1932, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Easter Mean Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Easter", new TimeSpan(-252000000000), "Chile", "EAS%sT", 1982, PublicDomain.Month.March, 13, null, new TimeSpan(756000000000), TzDatabase.TimeModifier.LocalWallTime, "Easter I Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Easter", new TimeSpan(-216000000000), "Chile", "EAS%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Chile", 1927, 1932, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1928, 1932, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1942, 1942, PublicDomain.Month.June, 1, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1942, 1942, PublicDomain.Month.August, 1, null, new TimeSpan(180000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1946, 1946, PublicDomain.Month.July, 15, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1946, 1946, PublicDomain.Month.September, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1947, 1947, PublicDomain.Month.April, 1, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1968, 1968, PublicDomain.Month.November, 3, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1969, 1969, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1969, 1969, PublicDomain.Month.November, 23, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1970, 1970, PublicDomain.Month.March, 29, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1970, 1972, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1971, 1971, PublicDomain.Month.March, 14, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1972, 1986, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1973, 1973, PublicDomain.Month.September, 30, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1974, 1987, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1987, 1987, PublicDomain.Month.April, 12, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1988, 1989, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1988, 1988, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1989, 1989, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1990, 1990, PublicDomain.Month.March, 18, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1990, 1990, PublicDomain.Month.September, 16, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1991, 1996, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1991, 1997, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1997, 1997, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1998, 1998, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1998, 1998, PublicDomain.Month.September, 27, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1999, 1999, PublicDomain.Month.April, 4, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1999, 2147483647, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 2000, 2007, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 2008, 2008, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 2009, 2147483647, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCST6CDT_357()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("CST6CDT", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("CST6CDT", new TimeSpan(-216000000000), "US", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesCuba_358()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Cuba", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Havana", new TimeSpan(-197680000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Havana", new TimeSpan(-197760000000), "-", "HMT", 1925, PublicDomain.Month.July, 19, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, "Havana MT"), new PublicDomain.TzDatabase.TzZone("America/Havana", new TimeSpan(-180000000000), "Cuba", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Cuba", 1928, 1928, PublicDomain.Month.June, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1928, 1928, PublicDomain.Month.October, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1940, 1942, PublicDomain.Month.June, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1940, 1942, PublicDomain.Month.September, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1945, 1946, PublicDomain.Month.June, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1945, 1946, PublicDomain.Month.September, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1965, 1965, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1965, 1965, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1966, 1966, PublicDomain.Month.May, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1966, 1966, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1967, 1967, PublicDomain.Month.April, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1967, 1968, PublicDomain.Month.September, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1968, 1968, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1969, 1977, PublicDomain.Month.April, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1969, 1971, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1972, 1974, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1975, 1977, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1978, 1978, PublicDomain.Month.May, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1978, 1990, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1979, 1980, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1981, 1985, PublicDomain.Month.May, 5, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1986, 1989, PublicDomain.Month.March, 14, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1990, 1997, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1991, 1995, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1996, 1996, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1997, 1997, PublicDomain.Month.October, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1998, 1999, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 1998, 2003, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 2000, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 2006, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Cuba", 2007, 2007, PublicDomain.Month.March, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Cuba", 2008, 2147483647, PublicDomain.Month.March, 15, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEET_359()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("EET", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("EET", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEgypt_360()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Egypt", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Cairo", new TimeSpan(75000000000), "-", "LMT", 1900, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Cairo", new TimeSpan(72000000000), "Egypt", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Egypt", 1940, 1940, PublicDomain.Month.July, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1940, 1940, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1941, 1941, PublicDomain.Month.April, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1941, 1941, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1942, 1944, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1942, 1942, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1943, 1945, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1945, 1945, PublicDomain.Month.April, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1957, 1957, PublicDomain.Month.May, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1957, 1958, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1958, 1958, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1959, 1981, PublicDomain.Month.May, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1959, 1965, PublicDomain.Month.September, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1966, 1994, PublicDomain.Month.October, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1982, 1982, PublicDomain.Month.July, 25, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1983, 1983, PublicDomain.Month.July, 12, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1984, 1988, PublicDomain.Month.May, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1989, 1989, PublicDomain.Month.May, 6, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1990, 1994, PublicDomain.Month.May, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1995, 2147483647, PublicDomain.Month.April, -1, DayOfWeek.Friday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Egypt", 1995, 2005, PublicDomain.Month.September, -1, DayOfWeek.Thursday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 2006, 2006, PublicDomain.Month.September, 21, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 2007, 2007, PublicDomain.Month.September, 1, DayOfWeek.Thursday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Egypt", 2008, 2147483647, PublicDomain.Month.August, -1, DayOfWeek.Thursday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEire_361()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Eire", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Dublin", new TimeSpan(-15000000000), "-", "LMT", 1880, PublicDomain.Month.August, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", new TimeSpan(-15210000000), "-", "DMT", 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", new TimeSpan(-15210000000), "1:00", "IST", 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "GB-Eire", "%s", 1921, PublicDomain.Month.December, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "GB-Eire", "GMT/IST", 1940, PublicDomain.Month.February, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "1:00", "IST", 1946, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "-", "GMT", 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "1:00", "IST", 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "-", "GMT", 1948, PublicDomain.Month.April, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "GB-Eire", "GMT/IST", 1968, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", new TimeSpan(36000000000), "-", "IST", 1971, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "GB-Eire", "GMT/IST", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "EU", "GMT/IST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.March, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.March, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1923, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1924, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1924, 1924, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1926, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1927, 1927, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1928, 1929, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1930, 1930, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1931, 1932, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1933, 1933, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1934, 1934, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1935, 1935, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1936, 1937, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1938, 1938, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.November, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1940, 1940, PublicDomain.Month.February, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1941, PublicDomain.Month.May, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1943, PublicDomain.Month.August, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1942, 1944, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1944, 1944, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.April, 2, DayOfWeek.Monday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.July, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1946, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1946, 1946, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.April, 14, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.October, 21, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1953, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1960, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1954, 1954, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1955, 1956, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1957, 1957, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1958, 1959, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1960, 1960, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1963, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1968, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1964, 1967, PublicDomain.Month.March, 19, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1968, 1968, PublicDomain.Month.February, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1989, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1990, 1995, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEST_362()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("EST", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("EST", new TimeSpan(-180000000000), "-", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEST5EDT_363()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("EST5EDT", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("EST5EDT", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT_364()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT0_365()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+0", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT1_366()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+1", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+1", new TimeSpan(-36000000000), "-", "GMT+1", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT10_367()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+10", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+10", new TimeSpan(-360000000000), "-", "GMT+10", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT11_368()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+11", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+11", new TimeSpan(-396000000000), "-", "GMT+11", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT12_369()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+12", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+12", new TimeSpan(-432000000000), "-", "GMT+12", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT2_370()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+2", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+2", new TimeSpan(-72000000000), "-", "GMT+2", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT3_371()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+3", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+3", new TimeSpan(-108000000000), "-", "GMT+3", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT4_372()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+4", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+4", new TimeSpan(-144000000000), "-", "GMT+4", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT5_373()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+5", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+5", new TimeSpan(-180000000000), "-", "GMT+5", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT6_374()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+6", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+6", new TimeSpan(-216000000000), "-", "GMT+6", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT7_375()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+7", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+7", new TimeSpan(-252000000000), "-", "GMT+7", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT8_376()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+8", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+8", new TimeSpan(-288000000000), "-", "GMT+8", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT9_377()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT+9", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT+9", new TimeSpan(-324000000000), "-", "GMT+9", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT0_378()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT0", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT0_379()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-0", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT1_380()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-1", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-1", new TimeSpan(36000000000), "-", "GMT-1", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT10_381()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-10", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-10", new TimeSpan(360000000000), "-", "GMT-10", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT11_382()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-11", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-11", new TimeSpan(396000000000), "-", "GMT-11", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT12_383()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-12", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-12", new TimeSpan(432000000000), "-", "GMT-12", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT13_384()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-13", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-13", new TimeSpan(468000000000), "-", "GMT-13", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT14_385()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-14", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-14", new TimeSpan(504000000000), "-", "GMT-14", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "14 hours ahead of GMT") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT2_386()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-2", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-2", new TimeSpan(72000000000), "-", "GMT-2", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT3_387()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-3", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-3", new TimeSpan(108000000000), "-", "GMT-3", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT4_388()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-4", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-4", new TimeSpan(144000000000), "-", "GMT-4", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT5_389()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-5", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-5", new TimeSpan(180000000000), "-", "GMT-5", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT6_390()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-6", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-6", new TimeSpan(216000000000), "-", "GMT-6", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT7_391()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-7", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-7", new TimeSpan(252000000000), "-", "GMT-7", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT8_392()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-8", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-8", new TimeSpan(288000000000), "-", "GMT-8", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGMT9_393()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/GMT-9", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT-9", new TimeSpan(324000000000), "-", "GMT-9", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcGreenwich_394()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/Greenwich", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcUCT_395()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/UCT", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/UCT", TimeSpan.Zero, "-", "UCT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcUniversal_396()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/Universal", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/UTC", TimeSpan.Zero, "-", "UTC", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcUTC_397()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/UTC", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/UTC", TimeSpan.Zero, "-", "UTC", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEtcZulu_398()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Etc/Zulu", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/UTC", TimeSpan.Zero, "-", "UTC", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeAmsterdam_399()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Amsterdam", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Amsterdam", new TimeSpan(11720000000), "-", "LMT", 1835, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Amsterdam", new TimeSpan(11720000000), "Neth", "%s", 1937, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Amsterdam", new TimeSpan(12000000000), "Neth", "NE%sT", 1940, PublicDomain.Month.May, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Dutch Time"), new PublicDomain.TzDatabase.TzZone("Europe/Amsterdam", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Amsterdam", new TimeSpan(36000000000), "Neth", "CE%sT", 1977, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Amsterdam", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Neth", 1916, 1916, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "NST", "Netherlands Summer Time"), new PublicDomain.TzDatabase.TzRule("Neth", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "AMT", "Amsterdam Mean Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Neth", 1917, 1917, PublicDomain.Month.April, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "NST", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Neth", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "AMT", null), new PublicDomain.TzDatabase.TzRule("Neth", 1918, 1921, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "NST", null), new PublicDomain.TzDatabase.TzRule("Neth", 1918, 1921, PublicDomain.Month.September, -1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "AMT", null), new PublicDomain.TzDatabase.TzRule("Neth", 1922, 1922, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "NST", null), new PublicDomain.TzDatabase.TzRule("Neth", 1922, 1936, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "AMT", null), new PublicDomain.TzDatabase.TzRule("Neth", 1923, 1923, PublicDomain.Month.June, 1, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "NST", null), new PublicDomain.TzDatabase.TzRule("Neth", 1924, 1924, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "NST", null), new PublicDomain.TzDatabase.TzRule("Neth", 1925, 1925, PublicDomain.Month.June, 1, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "NST", null), new PublicDomain.TzDatabase.TzRule("Neth", 1926, 1931, PublicDomain.Month.May, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "NST", null), new PublicDomain.TzDatabase.TzRule("Neth", 1932, 1932, PublicDomain.Month.May, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "NST", null), new PublicDomain.TzDatabase.TzRule("Neth", 1933, 1936, PublicDomain.Month.May, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "NST", null), new PublicDomain.TzDatabase.TzRule("Neth", 1937, 1937, PublicDomain.Month.May, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "NST", null), new PublicDomain.TzDatabase.TzRule("Neth", 1937, 1937, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Neth", 1937, 1939, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Neth", 1938, 1939, PublicDomain.Month.May, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Neth", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Neth", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeAndorra_400()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Andorra", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Andorra", new TimeSpan(3640000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Andorra", TimeSpan.Zero, "-", "WET", 1946, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Andorra", new TimeSpan(36000000000), "-", "CET", 1985, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Andorra", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeAthens_401()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Athens", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Athens", new TimeSpan(56920000000), "-", "LMT", 1895, PublicDomain.Month.September, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Athens", new TimeSpan(56920000000), "-", "AMT", 1916, PublicDomain.Month.July, 28, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, "Athens MT"), new PublicDomain.TzDatabase.TzZone("Europe/Athens", new TimeSpan(72000000000), "Greece", "EE%sT", 1941, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Athens", new TimeSpan(36000000000), "Greece", "CE%sT", 1944, PublicDomain.Month.April, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Athens", new TimeSpan(72000000000), "Greece", "EE%sT", 1981, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Athens", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Greece", 1932, 1932, PublicDomain.Month.July, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Greece", 1932, 1932, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Greece", 1941, 1941, PublicDomain.Month.April, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Greece", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Greece", 1943, 1943, PublicDomain.Month.March, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Greece", 1943, 1943, PublicDomain.Month.October, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Greece", 1952, 1952, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Greece", 1952, 1952, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Greece", 1975, 1975, PublicDomain.Month.April, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Greece", 1975, 1975, PublicDomain.Month.November, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Greece", 1976, 1976, PublicDomain.Month.April, 11, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Greece", 1976, 1976, PublicDomain.Month.October, 10, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Greece", 1977, 1978, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Greece", 1977, 1977, PublicDomain.Month.September, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Greece", 1978, 1978, PublicDomain.Month.September, 24, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Greece", 1979, 1979, PublicDomain.Month.April, 1, null, new TimeSpan(324000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Greece", 1979, 1979, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Greece", 1980, 1980, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Greece", 1980, 1980, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeBelfast_402()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Belfast", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(-750000000), "-", "LMT", 1847, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1968, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(36000000000), "-", "BST", 1971, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "EU", "GMT/BST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.March, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.March, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1923, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1924, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1924, 1924, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1926, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1927, 1927, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1928, 1929, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1930, 1930, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1931, 1932, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1933, 1933, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1934, 1934, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1935, 1935, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1936, 1937, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1938, 1938, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.November, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1940, 1940, PublicDomain.Month.February, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1941, PublicDomain.Month.May, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1943, PublicDomain.Month.August, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1942, 1944, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1944, 1944, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.April, 2, DayOfWeek.Monday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.July, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1946, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1946, 1946, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.April, 14, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.October, 21, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1953, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1960, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1954, 1954, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1955, 1956, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1957, 1957, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1958, 1959, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1960, 1960, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1963, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1968, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1964, 1967, PublicDomain.Month.March, 19, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1968, 1968, PublicDomain.Month.February, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1989, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1990, 1995, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeBelgrade_403()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Belgrade", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(49200000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1941, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1945, PublicDomain.Month.May, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "1:00", "CEST", 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1982, PublicDomain.Month.November, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeBerlin_404()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Berlin", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Berlin", new TimeSpan(32080000000), "-", "LMT", 1893, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Berlin", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.May, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Berlin", new TimeSpan(36000000000), "SovietZone", "CE%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Berlin", new TimeSpan(36000000000), "Germany", "CE%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Berlin", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("SovietZone", 1945, 1945, PublicDomain.Month.May, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("SovietZone", 1945, 1945, PublicDomain.Month.September, 24, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("SovietZone", 1945, 1945, PublicDomain.Month.November, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Germany", 1946, 1946, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Germany", 1946, 1946, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Germany", 1947, 1947, PublicDomain.Month.April, 6, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Germany", 1947, 1947, PublicDomain.Month.May, 11, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Germany", 1947, 1947, PublicDomain.Month.June, 29, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Germany", 1947, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Germany", 1948, 1948, PublicDomain.Month.April, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Germany", 1949, 1949, PublicDomain.Month.April, 10, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeBratislava_405()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Bratislava", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Prague", new TimeSpan(34640000000), "-", "LMT", 1850, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Prague", new TimeSpan(34640000000), "-", "PMT", 1891, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Prague Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Prague", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Prague", new TimeSpan(36000000000), "Czech", "CE%sT", 1979, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Prague", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Czech", 1945, 1945, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Czech", 1945, 1945, PublicDomain.Month.November, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Czech", 1946, 1946, PublicDomain.Month.May, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Czech", 1946, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Czech", 1947, 1947, PublicDomain.Month.April, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Czech", 1948, 1948, PublicDomain.Month.April, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Czech", 1949, 1949, PublicDomain.Month.April, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeBrussels_406()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Brussels", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Brussels", new TimeSpan(10500000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Brussels", new TimeSpan(10500000000), "-", "BMT", 1892, PublicDomain.Month.May, 1, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, "Brussels MT"), new PublicDomain.TzDatabase.TzZone("Europe/Brussels", TimeSpan.Zero, "-", "WET", 1914, PublicDomain.Month.November, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Brussels", new TimeSpan(36000000000), "-", "CET", 1916, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Brussels", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1918, PublicDomain.Month.November, 11, null, new TimeSpan(396000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Brussels", TimeSpan.Zero, "Belgium", "WE%sT", 1940, PublicDomain.Month.May, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Brussels", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Brussels", new TimeSpan(36000000000), "Belgium", "CE%sT", 1977, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Brussels", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1918, 1918, PublicDomain.Month.March, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1918, 1919, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1919, 1919, PublicDomain.Month.March, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1920, 1920, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1920, 1920, PublicDomain.Month.October, 23, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1921, 1921, PublicDomain.Month.March, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1921, 1921, PublicDomain.Month.October, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1922, 1922, PublicDomain.Month.March, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1922, 1927, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1923, 1923, PublicDomain.Month.April, 21, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1924, 1924, PublicDomain.Month.March, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1925, 1925, PublicDomain.Month.April, 4, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1928, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1929, 1929, PublicDomain.Month.April, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1930, 1930, PublicDomain.Month.April, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1931, 1931, PublicDomain.Month.April, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1932, 1932, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1933, 1933, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1934, 1934, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1935, 1935, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1936, 1936, PublicDomain.Month.April, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1937, 1937, PublicDomain.Month.April, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1938, 1938, PublicDomain.Month.March, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1939, 1939, PublicDomain.Month.April, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1939, 1939, PublicDomain.Month.November, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1940, 1940, PublicDomain.Month.February, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1944, 1944, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1946, 1946, PublicDomain.Month.May, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1946, 1946, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeBucharest_407()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Bucharest", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Bucharest", new TimeSpan(62640000000), "-", "LMT", 1891, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Bucharest", new TimeSpan(62640000000), "-", "BMT", 1931, PublicDomain.Month.July, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bucharest MT"), new PublicDomain.TzDatabase.TzZone("Europe/Bucharest", new TimeSpan(72000000000), "Romania", "EE%sT", 1981, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Bucharest", new TimeSpan(72000000000), "C-Eur", "EE%sT", 1991, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Bucharest", new TimeSpan(72000000000), "Romania", "EE%sT", 1994, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Bucharest", new TimeSpan(72000000000), "E-Eur", "EE%sT", 1997, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Bucharest", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1932, 1932, PublicDomain.Month.May, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1932, 1939, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1933, 1939, PublicDomain.Month.April, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1979, 1979, PublicDomain.Month.May, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1979, 1979, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1980, 1980, PublicDomain.Month.April, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1980, 1980, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1991, 1993, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1991, 1993, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeBudapest_408()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Budapest", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Budapest", new TimeSpan(45800000000), "-", "LMT", 1890, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Budapest", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Budapest", new TimeSpan(36000000000), "Hungary", "CE%sT", 1941, PublicDomain.Month.April, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Budapest", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Budapest", new TimeSpan(36000000000), "Hungary", "CE%sT", 1980, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Budapest", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1918, 1918, PublicDomain.Month.April, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1918, 1918, PublicDomain.Month.September, 29, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1919, 1919, PublicDomain.Month.April, 15, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1919, 1919, PublicDomain.Month.September, 15, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1920, 1920, PublicDomain.Month.April, 5, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1920, 1920, PublicDomain.Month.September, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1945, 1945, PublicDomain.Month.May, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1945, 1945, PublicDomain.Month.November, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1946, 1946, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1946, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1947, 1949, PublicDomain.Month.April, 4, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1950, 1950, PublicDomain.Month.April, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1950, 1950, PublicDomain.Month.October, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1954, 1955, PublicDomain.Month.May, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1954, 1955, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1956, 1956, PublicDomain.Month.June, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1956, 1956, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1957, 1957, PublicDomain.Month.June, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1957, 1957, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Hungary", 1980, 1980, PublicDomain.Month.April, 6, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeChisinau_409()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Chisinau", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(69200000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(69000000000), "-", "CMT", 1918, PublicDomain.Month.February, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Chisinau MT"), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(62640000000), "-", "BMT", 1931, PublicDomain.Month.July, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bucharest MT"), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "Romania", "EE%sT", 1940, PublicDomain.Month.August, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "1:00", "EEST", 1941, PublicDomain.Month.July, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.August, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1990, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(108000000000), "-", "MSK", 1990, PublicDomain.Month.May, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "-", "EET", 1991, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "Russia", "EE%sT", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "E-Eur", "EE%sT", 1997, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1932, 1932, PublicDomain.Month.May, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1932, 1939, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1933, 1939, PublicDomain.Month.April, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1979, 1979, PublicDomain.Month.May, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1979, 1979, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1980, 1980, PublicDomain.Month.April, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1980, 1980, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1991, 1993, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1991, 1993, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeCopenhagen_410()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Copenhagen", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Copenhagen", new TimeSpan(30200000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Copenhagen", new TimeSpan(30200000000), "-", "CMT", 1894, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Copenhagen MT"), new PublicDomain.TzDatabase.TzZone("Europe/Copenhagen", new TimeSpan(36000000000), "Denmark", "CE%sT", 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Copenhagen", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Copenhagen", new TimeSpan(36000000000), "Denmark", "CE%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Copenhagen", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1916, 1916, PublicDomain.Month.May, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1916, 1916, PublicDomain.Month.September, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1940, 1940, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1945, 1945, PublicDomain.Month.August, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1946, 1946, PublicDomain.Month.May, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1946, 1946, PublicDomain.Month.September, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1947, 1947, PublicDomain.Month.May, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1948, 1948, PublicDomain.Month.May, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Denmark", 1948, 1948, PublicDomain.Month.August, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeDublin_411()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Dublin", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Dublin", new TimeSpan(-15000000000), "-", "LMT", 1880, PublicDomain.Month.August, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", new TimeSpan(-15210000000), "-", "DMT", 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", new TimeSpan(-15210000000), "1:00", "IST", 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "GB-Eire", "%s", 1921, PublicDomain.Month.December, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "GB-Eire", "GMT/IST", 1940, PublicDomain.Month.February, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "1:00", "IST", 1946, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "-", "GMT", 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "1:00", "IST", 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "-", "GMT", 1948, PublicDomain.Month.April, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "GB-Eire", "GMT/IST", 1968, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", new TimeSpan(36000000000), "-", "IST", 1971, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "GB-Eire", "GMT/IST", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Dublin", TimeSpan.Zero, "EU", "GMT/IST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.March, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.March, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1923, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1924, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1924, 1924, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1926, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1927, 1927, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1928, 1929, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1930, 1930, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1931, 1932, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1933, 1933, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1934, 1934, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1935, 1935, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1936, 1937, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1938, 1938, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.November, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1940, 1940, PublicDomain.Month.February, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1941, PublicDomain.Month.May, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1943, PublicDomain.Month.August, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1942, 1944, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1944, 1944, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.April, 2, DayOfWeek.Monday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.July, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1946, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1946, 1946, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.April, 14, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.October, 21, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1953, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1960, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1954, 1954, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1955, 1956, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1957, 1957, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1958, 1959, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1960, 1960, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1963, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1968, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1964, 1967, PublicDomain.Month.March, 19, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1968, 1968, PublicDomain.Month.February, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1989, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1990, 1995, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeGibraltar_412()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Gibraltar", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Gibraltar", new TimeSpan(-12840000000), "-", "LMT", 1880, PublicDomain.Month.August, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Gibraltar", TimeSpan.Zero, "GB-Eire", "%s", 1957, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Gibraltar", new TimeSpan(36000000000), "-", "CET", 1982, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Gibraltar", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.March, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.March, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1923, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1924, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1924, 1924, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1926, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1927, 1927, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1928, 1929, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1930, 1930, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1931, 1932, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1933, 1933, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1934, 1934, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1935, 1935, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1936, 1937, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1938, 1938, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.November, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1940, 1940, PublicDomain.Month.February, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1941, PublicDomain.Month.May, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1943, PublicDomain.Month.August, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1942, 1944, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1944, 1944, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.April, 2, DayOfWeek.Monday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.July, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1946, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1946, 1946, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.April, 14, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.October, 21, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1953, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1960, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1954, 1954, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1955, 1956, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1957, 1957, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1958, 1959, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1960, 1960, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1963, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1968, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1964, 1967, PublicDomain.Month.March, 19, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1968, 1968, PublicDomain.Month.February, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1989, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1990, 1995, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeGuernsey_413()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Guernsey", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(-750000000), "-", "LMT", 1847, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1968, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(36000000000), "-", "BST", 1971, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "EU", "GMT/BST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.March, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.March, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1923, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1924, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1924, 1924, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1926, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1927, 1927, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1928, 1929, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1930, 1930, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1931, 1932, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1933, 1933, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1934, 1934, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1935, 1935, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1936, 1937, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1938, 1938, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.November, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1940, 1940, PublicDomain.Month.February, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1941, PublicDomain.Month.May, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1943, PublicDomain.Month.August, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1942, 1944, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1944, 1944, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.April, 2, DayOfWeek.Monday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.July, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1946, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1946, 1946, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.April, 14, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.October, 21, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1953, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1960, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1954, 1954, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1955, 1956, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1957, 1957, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1958, 1959, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1960, 1960, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1963, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1968, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1964, 1967, PublicDomain.Month.March, 19, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1968, 1968, PublicDomain.Month.February, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1989, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1990, 1995, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeHelsinki_414()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Helsinki", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Helsinki", new TimeSpan(59920000000), "-", "LMT", 1878, PublicDomain.Month.May, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Helsinki", new TimeSpan(59920000000), "-", "HMT", 1921, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Helsinki Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Helsinki", new TimeSpan(72000000000), "Finland", "EE%sT", 1981, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Helsinki", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Finland", 1942, 1942, PublicDomain.Month.April, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Finland", 1942, 1942, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeIsle_of_Man_415()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Isle_of_Man", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(-750000000), "-", "LMT", 1847, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1968, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(36000000000), "-", "BST", 1971, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "EU", "GMT/BST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.March, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.March, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1923, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1924, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1924, 1924, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1926, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1927, 1927, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1928, 1929, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1930, 1930, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1931, 1932, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1933, 1933, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1934, 1934, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1935, 1935, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1936, 1937, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1938, 1938, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.November, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1940, 1940, PublicDomain.Month.February, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1941, PublicDomain.Month.May, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1943, PublicDomain.Month.August, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1942, 1944, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1944, 1944, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.April, 2, DayOfWeek.Monday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.July, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1946, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1946, 1946, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.April, 14, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.October, 21, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1953, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1960, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1954, 1954, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1955, 1956, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1957, 1957, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1958, 1959, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1960, 1960, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1963, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1968, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1964, 1967, PublicDomain.Month.March, 19, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1968, 1968, PublicDomain.Month.February, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1989, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1990, 1995, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeIstanbul_416()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Istanbul", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(69520000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(70160000000), "-", "IMT", 1910, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Istanbul Mean Time?"), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(72000000000), "Turkey", "EE%sT", 1978, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(108000000000), "Turkey", "TR%sT", 1985, PublicDomain.Month.April, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Turkey Time"), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(72000000000), "Turkey", "EE%sT", 2007, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Turkey", 1916, 1916, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1920, 1920, PublicDomain.Month.March, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1920, 1920, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1921, 1921, PublicDomain.Month.April, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1921, 1921, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1922, 1922, PublicDomain.Month.March, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1922, 1922, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1924, 1924, PublicDomain.Month.May, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1924, 1925, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1925, 1925, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1940, 1940, PublicDomain.Month.June, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1940, 1940, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1940, 1940, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1941, 1941, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1942, 1942, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1942, 1942, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1945, 1945, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1945, 1945, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1946, 1946, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1947, 1948, PublicDomain.Month.April, 16, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1947, 1950, PublicDomain.Month.October, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1949, 1949, PublicDomain.Month.April, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1950, 1950, PublicDomain.Month.April, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1951, 1951, PublicDomain.Month.April, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1951, 1951, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1962, 1962, PublicDomain.Month.July, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1962, 1962, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1964, 1964, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1964, 1964, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1970, 1972, PublicDomain.Month.May, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1970, 1972, PublicDomain.Month.October, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1973, 1973, PublicDomain.Month.June, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1973, 1973, PublicDomain.Month.November, 4, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1974, 1974, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1974, 1974, PublicDomain.Month.November, 3, null, new TimeSpan(180000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1975, 1975, PublicDomain.Month.March, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1975, 1976, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1976, 1976, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1977, 1978, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1977, 1977, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1979, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1979, 1982, PublicDomain.Month.October, 11, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1981, 1982, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1983, 1983, PublicDomain.Month.July, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1983, 1983, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1985, 1985, PublicDomain.Month.April, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1985, 1985, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1986, 1990, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1986, 1990, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1991, 2006, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1991, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1996, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeJersey_417()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Jersey", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(-750000000), "-", "LMT", 1847, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1968, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(36000000000), "-", "BST", 1971, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "EU", "GMT/BST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.March, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.March, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1923, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1924, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1924, 1924, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1926, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1927, 1927, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1928, 1929, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1930, 1930, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1931, 1932, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1933, 1933, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1934, 1934, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1935, 1935, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1936, 1937, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1938, 1938, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.November, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1940, 1940, PublicDomain.Month.February, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1941, PublicDomain.Month.May, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1943, PublicDomain.Month.August, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1942, 1944, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1944, 1944, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.April, 2, DayOfWeek.Monday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.July, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1946, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1946, 1946, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.April, 14, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.October, 21, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1953, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1960, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1954, 1954, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1955, 1956, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1957, 1957, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1958, 1959, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1960, 1960, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1963, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1968, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1964, 1967, PublicDomain.Month.March, 19, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1968, 1968, PublicDomain.Month.February, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1989, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1990, 1995, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeKaliningrad_418()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Kaliningrad", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Kaliningrad", new TimeSpan(49200000000), "-", "LMT", 1893, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kaliningrad", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kaliningrad", new TimeSpan(72000000000), "Poland", "CE%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kaliningrad", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kaliningrad", new TimeSpan(72000000000), "Russia", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Poland", 1918, 1919, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1919, 1919, PublicDomain.Month.April, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1944, 1944, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1944, 1944, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1945, 1945, PublicDomain.Month.April, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1945, 1945, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1946, 1946, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1946, 1946, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1947, 1947, PublicDomain.Month.May, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1947, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1948, 1948, PublicDomain.Month.April, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1949, 1949, PublicDomain.Month.April, 10, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1957, 1957, PublicDomain.Month.June, 2, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1957, 1958, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1958, 1958, PublicDomain.Month.March, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1959, 1959, PublicDomain.Month.May, 31, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1959, 1961, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1960, 1960, PublicDomain.Month.April, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1961, 1964, PublicDomain.Month.May, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1962, 1964, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeKiev_419()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Kiev", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Kiev", new TimeSpan(73240000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kiev", new TimeSpan(73240000000), "-", "KMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kiev Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Kiev", new TimeSpan(72000000000), "-", "EET", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kiev", new TimeSpan(108000000000), "-", "MSK", 1941, PublicDomain.Month.September, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kiev", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1943, PublicDomain.Month.November, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kiev", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1990, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kiev", new TimeSpan(108000000000), "-", "MSK", 1990, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kiev", new TimeSpan(72000000000), "-", "EET", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kiev", new TimeSpan(72000000000), "E-Eur", "EE%sT", 1995, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Kiev", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeLisbon_420()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Lisbon", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", new TimeSpan(-21920000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", new TimeSpan(-21920000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Lisbon Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", TimeSpan.Zero, "Port", "WE%sT", 1966, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", new TimeSpan(36000000000), "-", "CET", 1976, PublicDomain.Month.September, 26, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", TimeSpan.Zero, "Port", "WE%sT", 1983, PublicDomain.Month.September, 25, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", TimeSpan.Zero, "W-Eur", "WE%sT", 1992, PublicDomain.Month.September, 27, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", new TimeSpan(36000000000), "EU", "CE%sT", 1996, PublicDomain.Month.March, 31, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", TimeSpan.Zero, "EU", "WE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Port", 1916, 1916, PublicDomain.Month.June, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1916, 1916, PublicDomain.Month.November, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1917, 1917, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1917, 1921, PublicDomain.Month.October, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1918, 1918, PublicDomain.Month.March, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1919, 1919, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1920, 1920, PublicDomain.Month.February, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1921, 1921, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1924, 1924, PublicDomain.Month.April, 16, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1924, 1924, PublicDomain.Month.October, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1926, 1929, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1929, 1929, PublicDomain.Month.April, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1931, 1931, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1931, 1932, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1932, 1932, PublicDomain.Month.April, 2, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1934, 1934, PublicDomain.Month.April, 7, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1934, 1938, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1935, 1935, PublicDomain.Month.March, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1936, 1936, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1937, 1937, PublicDomain.Month.April, 3, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1938, 1938, PublicDomain.Month.March, 26, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1939, 1939, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1939, 1939, PublicDomain.Month.November, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1940, 1940, PublicDomain.Month.February, 24, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1940, 1941, PublicDomain.Month.October, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1941, 1941, PublicDomain.Month.April, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1945, PublicDomain.Month.March, 8, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1942, PublicDomain.Month.April, 25, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1942, PublicDomain.Month.August, 15, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1945, PublicDomain.Month.October, 24, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1943, 1943, PublicDomain.Month.April, 17, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Port", 1943, 1945, PublicDomain.Month.August, 25, DayOfWeek.Saturday, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1944, 1945, PublicDomain.Month.April, 21, DayOfWeek.Saturday, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Port", 1946, 1946, PublicDomain.Month.April, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1946, 1946, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1947, 1949, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1947, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1951, 1965, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1951, 1965, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1977, 1977, PublicDomain.Month.March, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1977, 1977, PublicDomain.Month.September, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1978, 1979, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1979, 1982, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1980, 1980, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1981, 1982, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1983, 1983, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeLjubljana_421()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Ljubljana", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(49200000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1941, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1945, PublicDomain.Month.May, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "1:00", "CEST", 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1982, PublicDomain.Month.November, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeLondon_422()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/London", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(-750000000), "-", "LMT", 1847, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1968, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(36000000000), "-", "BST", 1971, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "EU", "GMT/BST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.March, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.March, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1923, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1924, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1924, 1924, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1926, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1927, 1927, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1928, 1929, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1930, 1930, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1931, 1932, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1933, 1933, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1934, 1934, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1935, 1935, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1936, 1937, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1938, 1938, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.November, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1940, 1940, PublicDomain.Month.February, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1941, PublicDomain.Month.May, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1943, PublicDomain.Month.August, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1942, 1944, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1944, 1944, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.April, 2, DayOfWeek.Monday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.July, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1946, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1946, 1946, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.April, 14, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.October, 21, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1953, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1960, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1954, 1954, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1955, 1956, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1957, 1957, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1958, 1959, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1960, 1960, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1963, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1968, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1964, 1967, PublicDomain.Month.March, 19, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1968, 1968, PublicDomain.Month.February, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1989, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1990, 1995, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeLuxembourg_423()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Luxembourg", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Luxembourg", new TimeSpan(14760000000), "-", "LMT", 1904, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Luxembourg", new TimeSpan(36000000000), "Lux", "CE%sT", 1918, PublicDomain.Month.November, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Luxembourg", TimeSpan.Zero, "Lux", "WE%sT", 1929, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Luxembourg", TimeSpan.Zero, "Belgium", "WE%sT", 1940, PublicDomain.Month.May, 14, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Luxembourg", new TimeSpan(36000000000), "C-Eur", "WE%sT", 1944, PublicDomain.Month.September, 18, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Luxembourg", new TimeSpan(36000000000), "Belgium", "CE%sT", 1977, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Luxembourg", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1916, 1916, PublicDomain.Month.May, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1917, 1917, PublicDomain.Month.April, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1918, 1918, PublicDomain.Month.March, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1918, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1918, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1918, 1919, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lux", 1919, 1919, PublicDomain.Month.March, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1919, 1919, PublicDomain.Month.March, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1919, 1919, PublicDomain.Month.October, 5, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lux", 1920, 1920, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1920, 1920, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1920, 1920, PublicDomain.Month.October, 23, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lux", 1920, 1920, PublicDomain.Month.October, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lux", 1921, 1921, PublicDomain.Month.March, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1921, 1921, PublicDomain.Month.March, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1921, 1921, PublicDomain.Month.October, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lux", 1921, 1921, PublicDomain.Month.October, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lux", 1922, 1922, PublicDomain.Month.March, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1922, 1922, PublicDomain.Month.March, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1922, 1927, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lux", 1922, 1922, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lux", 1923, 1923, PublicDomain.Month.April, 21, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1923, 1923, PublicDomain.Month.April, 21, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1923, 1923, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lux", 1924, 1924, PublicDomain.Month.March, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1924, 1924, PublicDomain.Month.March, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1924, 1928, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1925, 1925, PublicDomain.Month.April, 4, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1925, 1925, PublicDomain.Month.April, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Lux", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1928, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Lux", 1929, 1929, PublicDomain.Month.April, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1929, 1929, PublicDomain.Month.April, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1930, 1930, PublicDomain.Month.April, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1931, 1931, PublicDomain.Month.April, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1932, 1932, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1933, 1933, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1934, 1934, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1935, 1935, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1936, 1936, PublicDomain.Month.April, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1937, 1937, PublicDomain.Month.April, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1938, 1938, PublicDomain.Month.March, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1939, 1939, PublicDomain.Month.April, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1939, 1939, PublicDomain.Month.November, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1940, 1940, PublicDomain.Month.February, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1944, 1944, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1946, 1946, PublicDomain.Month.May, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Belgium", 1946, 1946, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeMadrid_424()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Madrid", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Madrid", new TimeSpan(-8840000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Madrid", TimeSpan.Zero, "Spain", "WE%sT", 1946, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Madrid", new TimeSpan(36000000000), "Spain", "CE%sT", 1979, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Madrid", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Spain", 1917, 1917, PublicDomain.Month.May, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1917, 1919, PublicDomain.Month.October, 6, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1918, 1918, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1919, 1919, PublicDomain.Month.April, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1924, 1924, PublicDomain.Month.April, 16, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1924, 1924, PublicDomain.Month.October, 4, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1926, 1929, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1929, 1929, PublicDomain.Month.April, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1937, 1937, PublicDomain.Month.May, 22, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1937, 1939, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1938, 1938, PublicDomain.Month.March, 22, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1939, 1939, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1940, 1940, PublicDomain.Month.March, 16, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1942, 1942, PublicDomain.Month.May, 2, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Spain", 1942, 1942, PublicDomain.Month.September, 1, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1943, 1946, PublicDomain.Month.April, 13, DayOfWeek.Saturday, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Spain", 1943, 1943, PublicDomain.Month.October, 3, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1944, 1944, PublicDomain.Month.October, 10, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1946, 1946, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1949, 1949, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1949, 1949, PublicDomain.Month.September, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1974, 1975, PublicDomain.Month.April, 13, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1974, 1975, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1976, 1976, PublicDomain.Month.March, 27, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Spain", 1976, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1977, 1978, PublicDomain.Month.April, 2, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Spain", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeMalta_425()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Malta", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Malta", new TimeSpan(34840000000), "-", "LMT", 1893, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, "Valletta"), new PublicDomain.TzDatabase.TzZone("Europe/Malta", new TimeSpan(36000000000), "Italy", "CE%sT", 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Malta", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Malta", new TimeSpan(36000000000), "Italy", "CE%sT", 1973, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Malta", new TimeSpan(36000000000), "Malta", "CE%sT", 1981, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Malta", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1916, 1916, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1917, 1917, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1917, 1917, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1918, 1918, PublicDomain.Month.March, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1918, 1919, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1919, 1919, PublicDomain.Month.March, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1920, 1920, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1920, 1920, PublicDomain.Month.September, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1940, 1940, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1944, 1944, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1945, 1945, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1946, 1946, PublicDomain.Month.March, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1946, 1946, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1947, 1947, PublicDomain.Month.March, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1947, 1947, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1948, 1948, PublicDomain.Month.February, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1948, 1948, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1966, 1968, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1966, 1969, PublicDomain.Month.September, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1969, 1969, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1970, 1970, PublicDomain.Month.May, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1970, 1970, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1971, 1972, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1971, 1971, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1972, 1972, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Malta", 1973, 1973, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1973, 1973, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Malta", 1973, 1973, PublicDomain.Month.September, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1973, 1974, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Malta", 1974, 1974, PublicDomain.Month.April, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1974, 1974, PublicDomain.Month.May, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Malta", 1974, 1974, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Malta", 1975, 1979, PublicDomain.Month.April, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1975, 1975, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Malta", 1975, 1980, PublicDomain.Month.September, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1975, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1976, 1976, PublicDomain.Month.May, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1977, 1979, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1979, 1979, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Malta", 1980, 1980, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeMariehamn_426()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Mariehamn", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Helsinki", new TimeSpan(59920000000), "-", "LMT", 1878, PublicDomain.Month.May, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Helsinki", new TimeSpan(59920000000), "-", "HMT", 1921, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Helsinki Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Helsinki", new TimeSpan(72000000000), "Finland", "EE%sT", 1981, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Helsinki", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Finland", 1942, 1942, PublicDomain.Month.April, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Finland", 1942, 1942, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeMinsk_427()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Minsk", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(66160000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(66000000000), "-", "MMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Minsk Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(72000000000), "-", "EET", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(108000000000), "-", "MSK", 1941, PublicDomain.Month.June, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.July, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1990, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(108000000000), "-", "MSK", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(72000000000), "1:00", "EEST", 1991, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(72000000000), "-", "EET", 1992, PublicDomain.Month.March, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(72000000000), "1:00", "EEST", 1992, PublicDomain.Month.September, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Minsk", new TimeSpan(72000000000), "Russia", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeMonaco_428()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Monaco", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Monaco", new TimeSpan(17720000000), "-", "LMT", 1891, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Monaco", new TimeSpan(5610000000), "-", "PMT", 1911, PublicDomain.Month.March, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Paris Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Monaco", TimeSpan.Zero, "France", "WE%sT", 1945, PublicDomain.Month.September, 16, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Monaco", new TimeSpan(36000000000), "France", "CE%sT", 1977, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Monaco", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("France", 1916, 1916, PublicDomain.Month.June, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1916, 1919, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1917, 1917, PublicDomain.Month.March, 24, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1918, 1918, PublicDomain.Month.March, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1919, 1919, PublicDomain.Month.March, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1920, 1920, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1920, 1920, PublicDomain.Month.October, 23, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1921, 1921, PublicDomain.Month.March, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1921, 1921, PublicDomain.Month.October, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1922, 1922, PublicDomain.Month.March, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1922, 1938, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1923, 1923, PublicDomain.Month.May, 26, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1924, 1924, PublicDomain.Month.March, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1925, 1925, PublicDomain.Month.April, 4, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1929, 1929, PublicDomain.Month.April, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1930, 1930, PublicDomain.Month.April, 12, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1931, 1931, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1932, 1932, PublicDomain.Month.April, 2, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1933, 1933, PublicDomain.Month.March, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1934, 1934, PublicDomain.Month.April, 7, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1935, 1935, PublicDomain.Month.March, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1936, 1936, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1937, 1937, PublicDomain.Month.April, 3, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1938, 1938, PublicDomain.Month.March, 26, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1939, 1939, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1939, 1939, PublicDomain.Month.November, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1940, 1940, PublicDomain.Month.February, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1941, 1941, PublicDomain.Month.May, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("France", 1941, 1941, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1942, 1942, PublicDomain.Month.March, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("France", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("France", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1944, 1944, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("France", 1944, 1944, PublicDomain.Month.October, 8, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("France", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1976, 1976, PublicDomain.Month.March, 28, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1976, 1976, PublicDomain.Month.September, 26, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeMoscow_429()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Moscow", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(90200000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(90000000000), "-", "MMT", 1916, PublicDomain.Month.July, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Moscow Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(90480000000), "Russia", "%s", 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1922, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(72000000000), "-", "EET", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(72000000000), "Russia", "EE%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(108000000000), "Russia", "MSK/MSD", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeNicosia_430()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Nicosia", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Nicosia", new TimeSpan(80080000000), "-", "LMT", 1921, PublicDomain.Month.November, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Nicosia", new TimeSpan(72000000000), "Cyprus", "EE%sT", 1998, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Nicosia", new TimeSpan(72000000000), "EUAsia", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Cyprus", 1975, 1975, PublicDomain.Month.April, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1975, 1975, PublicDomain.Month.October, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1976, 1976, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1976, 1976, PublicDomain.Month.October, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1977, 1977, PublicDomain.Month.September, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1978, 1978, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1979, 1997, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EUAsia", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cyprus", 1981, 1998, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EUAsia", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EUAsia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeOslo_431()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Oslo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(25800000000), "-", "LMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "Norway", "CE%sT", 1940, PublicDomain.Month.August, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "Norway", "CE%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Oslo", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Norway", 1916, 1916, PublicDomain.Month.May, 22, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Norway", 1916, 1916, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1945, 1945, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1959, 1964, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Norway", 1959, 1965, PublicDomain.Month.September, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Norway", 1965, 1965, PublicDomain.Month.April, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeParis_432()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Paris", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Paris", new TimeSpan(5610000000), "-", "LMT", 1891, PublicDomain.Month.March, 15, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Paris", new TimeSpan(5610000000), "-", "PMT", 1911, PublicDomain.Month.March, 11, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, "Paris MT"), new PublicDomain.TzDatabase.TzZone("Europe/Paris", TimeSpan.Zero, "France", "WE%sT", 1940, PublicDomain.Month.June, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Paris", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.August, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Paris", TimeSpan.Zero, "France", "WE%sT", 1945, PublicDomain.Month.September, 16, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Paris", new TimeSpan(36000000000), "France", "CE%sT", 1977, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Paris", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1916, 1916, PublicDomain.Month.June, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1916, 1919, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1917, 1917, PublicDomain.Month.March, 24, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1918, 1918, PublicDomain.Month.March, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1919, 1919, PublicDomain.Month.March, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1920, 1920, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1920, 1920, PublicDomain.Month.October, 23, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1921, 1921, PublicDomain.Month.March, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1921, 1921, PublicDomain.Month.October, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1922, 1922, PublicDomain.Month.March, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1922, 1938, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1923, 1923, PublicDomain.Month.May, 26, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1924, 1924, PublicDomain.Month.March, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1925, 1925, PublicDomain.Month.April, 4, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1929, 1929, PublicDomain.Month.April, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1930, 1930, PublicDomain.Month.April, 12, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1931, 1931, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1932, 1932, PublicDomain.Month.April, 2, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1933, 1933, PublicDomain.Month.March, 25, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1934, 1934, PublicDomain.Month.April, 7, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1935, 1935, PublicDomain.Month.March, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1936, 1936, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1937, 1937, PublicDomain.Month.April, 3, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1938, 1938, PublicDomain.Month.March, 26, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1939, 1939, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1939, 1939, PublicDomain.Month.November, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1940, 1940, PublicDomain.Month.February, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1941, 1941, PublicDomain.Month.May, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("France", 1941, 1941, PublicDomain.Month.October, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1942, 1942, PublicDomain.Month.March, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1944, 1944, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1944, 1944, PublicDomain.Month.October, 8, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("France", 1976, 1976, PublicDomain.Month.March, 28, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("France", 1976, 1976, PublicDomain.Month.September, 26, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropePodgorica_433()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Podgorica", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(49200000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1941, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1945, PublicDomain.Month.May, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "1:00", "CEST", 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1982, PublicDomain.Month.November, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropePrague_434()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Prague", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Prague", new TimeSpan(34640000000), "-", "LMT", 1850, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Prague", new TimeSpan(34640000000), "-", "PMT", 1891, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Prague Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Prague", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Prague", new TimeSpan(36000000000), "Czech", "CE%sT", 1979, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Prague", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Czech", 1945, 1945, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Czech", 1945, 1945, PublicDomain.Month.November, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Czech", 1946, 1946, PublicDomain.Month.May, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Czech", 1946, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Czech", 1947, 1947, PublicDomain.Month.April, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Czech", 1948, 1948, PublicDomain.Month.April, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Czech", 1949, 1949, PublicDomain.Month.April, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeRiga_435()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Riga", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(57840000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(57840000000), "-", "RMT", 1918, PublicDomain.Month.April, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, "Riga Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(57840000000), "1:00", "LST", 1918, PublicDomain.Month.September, 16, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, "Latvian Summer"), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(57840000000), "-", "RMT", 1919, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(57840000000), "1:00", "LST", 1919, PublicDomain.Month.May, 22, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(57840000000), "-", "RMT", 1926, PublicDomain.Month.May, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(72000000000), "-", "EET", 1940, PublicDomain.Month.August, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(108000000000), "-", "MSK", 1941, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.October, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1989, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(72000000000), "1:00", "EEST", 1989, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(72000000000), "Latvia", "EE%sT", 1997, PublicDomain.Month.January, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(72000000000), "EU", "EE%sT", 2000, PublicDomain.Month.February, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(72000000000), "-", "EET", 2001, PublicDomain.Month.January, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Riga", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Latvia", 1989, 1996, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Latvia", 1989, 1996, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeRome_436()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Rome", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(29960000000), "-", "LMT", 1866, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(29960000000), "-", "RMT", 1893, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, "Rome Mean"), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "Italy", "CE%sT", 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "Italy", "CE%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1916, 1916, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1917, 1917, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1917, 1917, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1918, 1918, PublicDomain.Month.March, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1918, 1919, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1919, 1919, PublicDomain.Month.March, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1920, 1920, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1920, 1920, PublicDomain.Month.September, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1940, 1940, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1944, 1944, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1945, 1945, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1946, 1946, PublicDomain.Month.March, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1946, 1946, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1947, 1947, PublicDomain.Month.March, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1947, 1947, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1948, 1948, PublicDomain.Month.February, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1948, 1948, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1966, 1968, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1966, 1969, PublicDomain.Month.September, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1969, 1969, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1970, 1970, PublicDomain.Month.May, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1970, 1970, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1971, 1972, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1971, 1971, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1972, 1972, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1973, 1973, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1973, 1974, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1974, 1974, PublicDomain.Month.May, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1975, 1975, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1975, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1976, 1976, PublicDomain.Month.May, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1977, 1979, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1979, 1979, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeSamara_437()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Samara", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Samara", new TimeSpan(120360000000), "-", "LMT", 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Samara", new TimeSpan(108000000000), "-", "SAMT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Samara", new TimeSpan(144000000000), "-", "SAMT", 1935, PublicDomain.Month.January, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Samara", new TimeSpan(144000000000), "Russia", "KUY%sT", 1989, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, "Kuybyshev"), new PublicDomain.TzDatabase.TzZone("Europe/Samara", new TimeSpan(108000000000), "Russia", "KUY%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Samara", new TimeSpan(72000000000), "Russia", "KUY%sT", 1991, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Samara", new TimeSpan(108000000000), "-", "KUYT", 1991, PublicDomain.Month.October, 20, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Samara", new TimeSpan(144000000000), "Russia", "SAM%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Samara Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeSan_Marino_438()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/San_Marino", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(29960000000), "-", "LMT", 1866, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(29960000000), "-", "RMT", 1893, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, "Rome Mean"), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "Italy", "CE%sT", 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "Italy", "CE%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1916, 1916, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1917, 1917, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1917, 1917, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1918, 1918, PublicDomain.Month.March, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1918, 1919, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1919, 1919, PublicDomain.Month.March, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1920, 1920, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1920, 1920, PublicDomain.Month.September, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1940, 1940, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1944, 1944, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1945, 1945, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1946, 1946, PublicDomain.Month.March, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1946, 1946, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1947, 1947, PublicDomain.Month.March, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1947, 1947, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1948, 1948, PublicDomain.Month.February, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1948, 1948, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1966, 1968, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1966, 1969, PublicDomain.Month.September, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1969, 1969, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1970, 1970, PublicDomain.Month.May, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1970, 1970, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1971, 1972, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1971, 1971, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1972, 1972, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1973, 1973, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1973, 1974, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1974, 1974, PublicDomain.Month.May, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1975, 1975, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1975, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1976, 1976, PublicDomain.Month.May, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1977, 1979, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1979, 1979, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeSarajevo_439()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Sarajevo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(49200000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1941, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1945, PublicDomain.Month.May, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "1:00", "CEST", 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1982, PublicDomain.Month.November, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeSimferopol_440()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Simferopol", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(81840000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(81600000000), "-", "SMT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Simferopol Mean T"), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(72000000000), "-", "EET", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(108000000000), "-", "MSK", 1941, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.April, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1990, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(108000000000), "-", "MSK", 1990, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(72000000000), "-", "EET", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(72000000000), "E-Eur", "EE%sT", 1994, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(108000000000), "E-Eur", "MSK/MSD", 1996, PublicDomain.Month.March, 31, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(108000000000), "1:00", "MSD", 1996, PublicDomain.Month.October, 27, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1997, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(108000000000), "-", "MSK", 1997, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Simferopol", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeSkopje_441()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Skopje", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(49200000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1941, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1945, PublicDomain.Month.May, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "1:00", "CEST", 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1982, PublicDomain.Month.November, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeSofia_442()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Sofia", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Sofia", new TimeSpan(55960000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Sofia", new TimeSpan(70160000000), "-", "IMT", 1894, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Istanbul MT?"), new PublicDomain.TzDatabase.TzZone("Europe/Sofia", new TimeSpan(72000000000), "-", "EET", 1942, PublicDomain.Month.November, 2, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Sofia", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Sofia", new TimeSpan(36000000000), "-", "CET", 1945, PublicDomain.Month.April, 2, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Sofia", new TimeSpan(72000000000), "-", "EET", 1979, PublicDomain.Month.March, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Sofia", new TimeSpan(72000000000), "Bulg", "EE%sT", 1982, PublicDomain.Month.September, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Sofia", new TimeSpan(72000000000), "C-Eur", "EE%sT", 1991, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Sofia", new TimeSpan(72000000000), "E-Eur", "EE%sT", 1997, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Sofia", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Bulg", 1979, 1979, PublicDomain.Month.March, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Bulg", 1979, 1979, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Bulg", 1980, 1982, PublicDomain.Month.April, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Bulg", 1980, 1980, PublicDomain.Month.September, 29, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Bulg", 1981, 1981, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeStockholm_443()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Stockholm", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Stockholm", new TimeSpan(43320000000), "-", "LMT", 1879, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Stockholm", new TimeSpan(36140000000), "-", "SET", 1900, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Swedish Time"), new PublicDomain.TzDatabase.TzZone("Europe/Stockholm", new TimeSpan(36000000000), "-", "CET", 1916, PublicDomain.Month.May, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Stockholm", new TimeSpan(36000000000), "1:00", "CEST", 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Stockholm", new TimeSpan(36000000000), "-", "CET", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Stockholm", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeTallinn_444()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Tallinn", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(59400000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(59400000000), "-", "TMT", 1918, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tallinn Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1919, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(59400000000), "-", "TMT", 1921, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(72000000000), "-", "EET", 1940, PublicDomain.Month.August, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(108000000000), "-", "MSK", 1941, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1989, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(72000000000), "1:00", "EEST", 1989, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(72000000000), "C-Eur", "EE%sT", 1998, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(72000000000), "EU", "EE%sT", 1999, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(72000000000), "-", "EET", 2002, PublicDomain.Month.February, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tallinn", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeTirane_445()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Tirane", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Tirane", new TimeSpan(47600000000), "-", "LMT", 1914, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tirane", new TimeSpan(36000000000), "-", "CET", 1940, PublicDomain.Month.June, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tirane", new TimeSpan(36000000000), "Albania", "CE%sT", 1984, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Tirane", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Albania", 1940, 1940, PublicDomain.Month.June, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1943, 1943, PublicDomain.Month.April, 10, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1974, 1974, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1974, 1974, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1975, 1975, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1975, 1975, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1976, 1976, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1976, 1976, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1977, 1977, PublicDomain.Month.May, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1977, 1977, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1978, 1978, PublicDomain.Month.May, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1979, 1979, PublicDomain.Month.May, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1979, 1979, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1980, 1980, PublicDomain.Month.May, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1980, 1980, PublicDomain.Month.October, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1981, 1981, PublicDomain.Month.April, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1981, 1981, PublicDomain.Month.September, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1982, 1982, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1982, 1982, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1983, 1983, PublicDomain.Month.April, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Albania", 1983, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Albania", 1984, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeTiraspol_446()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Tiraspol", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(69200000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(69000000000), "-", "CMT", 1918, PublicDomain.Month.February, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Chisinau MT"), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(62640000000), "-", "BMT", 1931, PublicDomain.Month.July, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bucharest MT"), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "Romania", "EE%sT", 1940, PublicDomain.Month.August, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "1:00", "EEST", 1941, PublicDomain.Month.July, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.August, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1990, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(108000000000), "-", "MSK", 1990, PublicDomain.Month.May, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "-", "EET", 1991, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "Russia", "EE%sT", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "E-Eur", "EE%sT", 1997, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Chisinau", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1932, 1932, PublicDomain.Month.May, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1932, 1939, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1933, 1939, PublicDomain.Month.April, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1979, 1979, PublicDomain.Month.May, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1979, 1979, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Romania", 1980, 1980, PublicDomain.Month.April, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1980, 1980, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1991, 1993, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Romania", 1991, 1993, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeUzhgorod_447()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Uzhgorod", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(53520000000), "-", "LMT", 1890, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(36000000000), "-", "CET", 1940, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(36000000000), "1:00", "CEST", 1944, PublicDomain.Month.October, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(36000000000), "-", "CET", 1945, PublicDomain.Month.June, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1990, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(108000000000), "-", "MSK", 1990, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(36000000000), "-", "CET", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(72000000000), "-", "EET", 1992, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(72000000000), "E-Eur", "EE%sT", 1995, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Uzhgorod", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeVaduz_448()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Vaduz", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Vaduz", new TimeSpan(22840000000), "-", "LMT", 1894, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vaduz", new TimeSpan(36000000000), "-", "CET", 1981, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vaduz", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeVatican_449()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Vatican", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(29960000000), "-", "LMT", 1866, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(29960000000), "-", "RMT", 1893, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, "Rome Mean"), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "Italy", "CE%sT", 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "Italy", "CE%sT", 1980, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Rome", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1916, 1916, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1917, 1917, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1917, 1917, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1918, 1918, PublicDomain.Month.March, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1918, 1919, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1919, 1919, PublicDomain.Month.March, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1920, 1920, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1920, 1920, PublicDomain.Month.September, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1940, 1940, PublicDomain.Month.June, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1944, 1944, PublicDomain.Month.September, 17, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1945, 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1945, 1945, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1946, 1946, PublicDomain.Month.March, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1946, 1946, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1947, 1947, PublicDomain.Month.March, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1947, 1947, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1948, 1948, PublicDomain.Month.February, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1948, 1948, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1966, 1968, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1966, 1969, PublicDomain.Month.September, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1969, 1969, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1970, 1970, PublicDomain.Month.May, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1970, 1970, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1971, 1972, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1971, 1971, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1972, 1972, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1973, 1973, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1973, 1974, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1974, 1974, PublicDomain.Month.May, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1975, 1975, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1975, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1976, 1976, PublicDomain.Month.May, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Italy", 1977, 1979, PublicDomain.Month.May, 22, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Italy", 1979, 1979, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeVienna_450()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Vienna", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Vienna", new TimeSpan(39200000000), "-", "LMT", 1893, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vienna", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vienna", new TimeSpan(36000000000), "Austria", "CE%sT", 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vienna", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vienna", new TimeSpan(36000000000), "1:00", "CEST", 1945, PublicDomain.Month.April, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vienna", new TimeSpan(36000000000), "-", "CET", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vienna", new TimeSpan(36000000000), "Austria", "CE%sT", 1981, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vienna", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Austria", 1920, 1920, PublicDomain.Month.April, 5, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Austria", 1920, 1920, PublicDomain.Month.September, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Austria", 1946, 1946, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Austria", 1946, 1948, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Austria", 1947, 1947, PublicDomain.Month.April, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Austria", 1948, 1948, PublicDomain.Month.April, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Austria", 1980, 1980, PublicDomain.Month.April, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Austria", 1980, 1980, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeVilnius_451()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Vilnius", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(60760000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(50400000000), "-", "WMT", 1917, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Warsaw Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(57360000000), "-", "KMT", 1919, PublicDomain.Month.October, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kaunas Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(36000000000), "-", "CET", 1920, PublicDomain.Month.July, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(72000000000), "-", "EET", 1920, PublicDomain.Month.October, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(36000000000), "-", "CET", 1940, PublicDomain.Month.August, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(108000000000), "-", "MSK", 1941, PublicDomain.Month.June, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.August, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(72000000000), "1:00", "EEST", 1991, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(72000000000), "C-Eur", "EE%sT", 1998, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(72000000000), "-", "EET", 1998, PublicDomain.Month.March, 29, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(36000000000), "EU", "CE%sT", 1999, PublicDomain.Month.October, 31, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(72000000000), "-", "EET", 2003, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Vilnius", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeVolgograd_452()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Volgograd", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Volgograd", new TimeSpan(106600000000), "-", "LMT", 1920, PublicDomain.Month.January, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Volgograd", new TimeSpan(108000000000), "-", "TSAT", 1925, PublicDomain.Month.April, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tsaritsyn Time"), new PublicDomain.TzDatabase.TzZone("Europe/Volgograd", new TimeSpan(108000000000), "-", "STAT", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Stalingrad Time"), new PublicDomain.TzDatabase.TzZone("Europe/Volgograd", new TimeSpan(144000000000), "-", "STAT", 1961, PublicDomain.Month.November, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Volgograd", new TimeSpan(144000000000), "Russia", "VOL%sT", 1989, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, "Volgograd T"), new PublicDomain.TzDatabase.TzZone("Europe/Volgograd", new TimeSpan(108000000000), "Russia", "VOL%sT", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Volgograd", new TimeSpan(144000000000), "-", "VOLT", 1992, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Volgograd", new TimeSpan(108000000000), "Russia", "VOL%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeWarsaw_453()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Warsaw", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(50400000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(50400000000), "-", "WMT", 1915, PublicDomain.Month.August, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Warsaw Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1918, PublicDomain.Month.September, 16, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(72000000000), "Poland", "EE%sT", 1922, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "Poland", "CE%sT", 1940, PublicDomain.Month.June, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "Poland", "CE%sT", 1977, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "W-Eur", "CE%sT", 1988, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1918, 1919, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1919, 1919, PublicDomain.Month.April, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1944, 1944, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1944, 1944, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1945, 1945, PublicDomain.Month.April, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1945, 1945, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1946, 1946, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1946, 1946, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1947, 1947, PublicDomain.Month.May, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1947, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1948, 1948, PublicDomain.Month.April, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1949, 1949, PublicDomain.Month.April, 10, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1957, 1957, PublicDomain.Month.June, 2, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1957, 1958, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1958, 1958, PublicDomain.Month.March, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1959, 1959, PublicDomain.Month.May, 31, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1959, 1961, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1960, 1960, PublicDomain.Month.April, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1961, 1964, PublicDomain.Month.May, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1962, 1964, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeZagreb_454()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Zagreb", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(49200000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1941, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1945, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1945, PublicDomain.Month.May, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "1:00", "CEST", 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "-", "CET", 1982, PublicDomain.Month.November, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Belgrade", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeZaporozhye_455()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Zaporozhye", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Zaporozhye", new TimeSpan(84400000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Zaporozhye", new TimeSpan(84000000000), "-", "CUT", 1924, PublicDomain.Month.May, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Central Ukraine T"), new PublicDomain.TzDatabase.TzZone("Europe/Zaporozhye", new TimeSpan(72000000000), "-", "EET", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Zaporozhye", new TimeSpan(108000000000), "-", "MSK", 1941, PublicDomain.Month.August, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Zaporozhye", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1943, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Zaporozhye", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Zaporozhye", new TimeSpan(72000000000), "E-Eur", "EE%sT", 1995, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Zaporozhye", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("E-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesEuropeZurich_456()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Europe/Zurich", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Zurich", new TimeSpan(20480000000), "-", "LMT", 1848, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Zurich", new TimeSpan(17840000000), "-", "BMT", 1894, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bern Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Zurich", new TimeSpan(36000000000), "Swiss", "CE%sT", 1981, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Zurich", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Swiss", 1940, 1940, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Swiss", 1940, 1940, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Swiss", 1941, 1942, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Swiss", 1941, 1942, PublicDomain.Month.October, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesGB_457()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("GB", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(-750000000), "-", "LMT", 1847, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1968, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(36000000000), "-", "BST", 1971, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "EU", "GMT/BST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.March, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.March, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1923, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1924, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1924, 1924, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1926, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1927, 1927, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1928, 1929, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1930, 1930, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1931, 1932, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1933, 1933, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1934, 1934, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1935, 1935, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1936, 1937, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1938, 1938, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.November, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1940, 1940, PublicDomain.Month.February, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1941, PublicDomain.Month.May, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1943, PublicDomain.Month.August, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1942, 1944, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1944, 1944, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.April, 2, DayOfWeek.Monday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.July, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1946, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1946, 1946, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.April, 14, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.October, 21, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1953, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1960, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1954, 1954, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1955, 1956, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1957, 1957, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1958, 1959, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1960, 1960, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1963, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1968, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1964, 1967, PublicDomain.Month.March, 19, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1968, 1968, PublicDomain.Month.February, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1989, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1990, 1995, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesGBEire_458()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("GB-Eire", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(-750000000), "-", "LMT", 1847, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1968, PublicDomain.Month.October, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", new TimeSpan(36000000000), "-", "BST", 1971, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "GB-Eire", "%s", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/London", TimeSpan.Zero, "EU", "GMT/BST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.April, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1917, 1917, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.March, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1918, 1918, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.March, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1919, 1919, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.March, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1920, 1920, PublicDomain.Month.October, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1921, 1921, PublicDomain.Month.October, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.March, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1922, 1922, PublicDomain.Month.October, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1923, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1923, 1924, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1924, 1924, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1926, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1925, 1938, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1927, 1927, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1928, 1929, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1930, 1930, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1931, 1932, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1933, 1933, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1934, 1934, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1935, 1935, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1936, 1937, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1938, 1938, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1939, 1939, PublicDomain.Month.November, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1940, 1940, PublicDomain.Month.February, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1941, PublicDomain.Month.May, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1941, 1943, PublicDomain.Month.August, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1942, 1944, PublicDomain.Month.April, 2, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1944, 1944, PublicDomain.Month.September, 16, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.April, 2, DayOfWeek.Monday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1945, PublicDomain.Month.July, 9, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1945, 1946, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1946, 1946, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.March, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "BDST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.August, 10, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1947, 1947, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1948, 1948, PublicDomain.Month.October, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.April, 14, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1950, 1952, PublicDomain.Month.October, 21, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1953, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1953, 1960, PublicDomain.Month.October, 2, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1954, 1954, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1955, 1956, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1957, 1957, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1958, 1959, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1960, 1960, PublicDomain.Month.April, 9, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1963, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1961, 1968, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1964, 1967, PublicDomain.Month.March, 19, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1968, 1968, PublicDomain.Month.February, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.March, 16, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1972, 1980, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1995, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "BST", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1981, 1989, PublicDomain.Month.October, 23, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("GB-Eire", 1990, 1995, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "GMT", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesGMT_459()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("GMT", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesGMT0_460()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("GMT+0", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesGMT0_461()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("GMT0", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesGMT0_462()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("GMT-0", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesGreenwich_463()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Greenwich", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/GMT", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesHongkong_464()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Hongkong", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Hong_Kong", new TimeSpan(273960000000), "-", "LMT", 1904, PublicDomain.Month.October, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Hong_Kong", new TimeSpan(288000000000), "HK", "HK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("HK", 1946, 1946, PublicDomain.Month.April, 20, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1946, 1946, PublicDomain.Month.December, 1, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1947, 1947, PublicDomain.Month.April, 13, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1947, 1947, PublicDomain.Month.December, 30, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1948, 1948, PublicDomain.Month.May, 2, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1948, 1952, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1949, 1953, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1953, 1953, PublicDomain.Month.November, 1, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1954, 1964, PublicDomain.Month.March, 18, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1954, 1954, PublicDomain.Month.October, 31, null, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1955, 1964, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1965, 1977, PublicDomain.Month.April, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1965, 1977, PublicDomain.Month.October, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("HK", 1979, 1980, PublicDomain.Month.May, 8, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("HK", 1979, 1980, PublicDomain.Month.October, 16, DayOfWeek.Sunday, new TimeSpan(126000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesHST_465()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("HST", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("HST", new TimeSpan(-360000000000), "-", "HST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIceland_466()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Iceland", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Atlantic/Reykjavik", new TimeSpan(-52440000000), "-", "LMT", 1837, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Reykjavik", new TimeSpan(-52680000000), "-", "RMT", 1908, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Reykjavik Mean Time?"), new PublicDomain.TzDatabase.TzZone("Atlantic/Reykjavik", new TimeSpan(-36000000000), "Iceland", "IS%sT", 1968, PublicDomain.Month.April, 7, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Atlantic/Reykjavik", TimeSpan.Zero, "-", "GMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Iceland", 1917, 1918, PublicDomain.Month.February, 19, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1917, 1917, PublicDomain.Month.October, 21, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1918, 1918, PublicDomain.Month.November, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1939, 1939, PublicDomain.Month.April, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1939, 1939, PublicDomain.Month.November, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1940, 1940, PublicDomain.Month.February, 25, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1940, 1940, PublicDomain.Month.November, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1941, 1941, PublicDomain.Month.March, 2, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1941, 1941, PublicDomain.Month.November, 2, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1942, 1942, PublicDomain.Month.March, 8, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1942, 1942, PublicDomain.Month.October, 25, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1943, 1946, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1943, 1948, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1947, 1967, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1949, 1949, PublicDomain.Month.October, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1950, 1966, PublicDomain.Month.October, 22, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Iceland", 1967, 1967, PublicDomain.Month.October, 29, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianAntananarivo_467()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Antananarivo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Antananarivo", new TimeSpan(114040000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Indian/Antananarivo", new TimeSpan(108000000000), "-", "EAT", 1954, PublicDomain.Month.February, 27, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Indian/Antananarivo", new TimeSpan(108000000000), "1:00", "EAST", 1954, PublicDomain.Month.May, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Indian/Antananarivo", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianChagos_468()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Chagos", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Chagos", new TimeSpan(173800000000), "-", "LMT", 1907, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Indian/Chagos", new TimeSpan(180000000000), "-", "IOT", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "BIOT Time"), new PublicDomain.TzDatabase.TzZone("Indian/Chagos", new TimeSpan(216000000000), "-", "IOT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianChristmas_469()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Christmas", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Christmas", new TimeSpan(253720000000), "-", "LMT", 1895, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Indian/Christmas", new TimeSpan(252000000000), "-", "CXT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Christmas Island Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianCocos_470()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Cocos", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Cocos", new TimeSpan(232600000000), "-", "LMT", 1900, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Indian/Cocos", new TimeSpan(234000000000), "-", "CCT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Cocos Islands Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianComoro_471()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Comoro", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Comoro", new TimeSpan(103840000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Moroni, Gran Comoro"), new PublicDomain.TzDatabase.TzZone("Indian/Comoro", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianKerguelen_472()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Kerguelen", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Kerguelen", TimeSpan.Zero, "-", "zzz", 1950, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Port-aux-Francais"), new PublicDomain.TzDatabase.TzZone("Indian/Kerguelen", new TimeSpan(180000000000), "-", "TFT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "ISO code TF Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianMahe_473()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Mahe", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Mahe", new TimeSpan(133080000000), "-", "LMT", 1906, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Victoria"), new PublicDomain.TzDatabase.TzZone("Indian/Mahe", new TimeSpan(144000000000), "-", "SCT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Seychelles Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianMaldives_474()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Maldives", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Maldives", new TimeSpan(176400000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Male"), new PublicDomain.TzDatabase.TzZone("Indian/Maldives", new TimeSpan(176400000000), "-", "MMT", 1960, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Male Mean Time"), new PublicDomain.TzDatabase.TzZone("Indian/Maldives", new TimeSpan(180000000000), "-", "MVT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Maldives Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianMauritius_475()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Mauritius", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Mauritius", new TimeSpan(138000000000), "-", "LMT", 1907, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Port Louis"), new PublicDomain.TzDatabase.TzZone("Indian/Mauritius", new TimeSpan(144000000000), "Mauritius", "MU%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Mauritius Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mauritius", 1982, 1982, PublicDomain.Month.October, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mauritius", 1983, 1983, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Mauritius", 2008, 2008, PublicDomain.Month.October, 26, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Mauritius", 2009, 2009, PublicDomain.Month.March, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianMayotte_476()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Mayotte", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Mayotte", new TimeSpan(108560000000), "-", "LMT", 1911, PublicDomain.Month.July, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Mamoutzou"), new PublicDomain.TzDatabase.TzZone("Indian/Mayotte", new TimeSpan(108000000000), "-", "EAT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIndianReunion_477()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Indian/Reunion", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Indian/Reunion", new TimeSpan(133120000000), "-", "LMT", 1911, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Saint-Denis"), new PublicDomain.TzDatabase.TzZone("Indian/Reunion", new TimeSpan(144000000000), "-", "RET", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Reunion Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIran_478()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Iran", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Tehran", new TimeSpan(123440000000), "-", "LMT", 1916, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tehran", new TimeSpan(123440000000), "-", "TMT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tehran Mean Time"), new PublicDomain.TzDatabase.TzZone("Asia/Tehran", new TimeSpan(126000000000), "-", "IRST", 1977, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tehran", new TimeSpan(144000000000), "Iran", "IR%sT", 1979, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tehran", new TimeSpan(126000000000), "Iran", "IR%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Iran", 1978, 1980, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 1978, 1978, PublicDomain.Month.October, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 1979, 1979, PublicDomain.Month.September, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 1980, 1980, PublicDomain.Month.September, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 1991, 1991, PublicDomain.Month.May, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 1991, 1995, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 1992, 1995, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 1996, 1996, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 1996, 1996, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 1997, 1999, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 1997, 1999, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2000, 2000, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2000, 2000, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2001, 2003, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2001, 2003, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2004, 2004, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2004, 2004, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2005, 2005, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2005, 2005, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2008, 2008, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2008, 2008, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2009, 2011, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2009, 2011, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2012, 2012, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2012, 2012, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2013, 2015, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2013, 2015, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2016, 2016, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2016, 2016, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2017, 2019, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2017, 2019, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2020, 2020, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2020, 2020, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2021, 2023, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2021, 2023, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2024, 2024, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2024, 2024, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2025, 2027, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2025, 2027, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2028, 2029, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2028, 2029, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2030, 2031, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2030, 2031, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2032, 2033, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2032, 2033, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2034, 2035, PublicDomain.Month.March, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2034, 2035, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Iran", 2036, 2037, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Iran", 2036, 2037, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesIsrael_479()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Israel", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Jerusalem", new TimeSpan(84560000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Jerusalem", new TimeSpan(84400000000), "-", "JMT", 1918, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Jerusalem Mean Time?"), new PublicDomain.TzDatabase.TzZone("Asia/Jerusalem", new TimeSpan(72000000000), "Zion", "I%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Zion", 1940, 1940, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1942, 1944, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1943, 1943, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1944, 1944, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1945, 1945, PublicDomain.Month.April, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1945, 1945, PublicDomain.Month.November, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1946, 1946, PublicDomain.Month.April, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1946, 1946, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1948, PublicDomain.Month.May, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "DD", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1948, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1948, 1949, PublicDomain.Month.November, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1949, 1949, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1950, 1950, PublicDomain.Month.April, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1950, 1950, PublicDomain.Month.September, 15, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1951, 1951, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1951, 1951, PublicDomain.Month.November, 11, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1952, 1952, PublicDomain.Month.April, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1952, 1952, PublicDomain.Month.October, 19, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1953, 1953, PublicDomain.Month.April, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1953, 1953, PublicDomain.Month.September, 13, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1954, 1954, PublicDomain.Month.June, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1954, 1954, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1955, 1955, PublicDomain.Month.June, 11, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1955, 1955, PublicDomain.Month.September, 11, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1956, 1956, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1956, 1956, PublicDomain.Month.September, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1957, 1957, PublicDomain.Month.April, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1957, 1957, PublicDomain.Month.September, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1974, 1974, PublicDomain.Month.July, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1974, 1974, PublicDomain.Month.October, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1975, 1975, PublicDomain.Month.April, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1975, 1975, PublicDomain.Month.August, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1985, 1985, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1985, 1985, PublicDomain.Month.September, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1986, 1986, PublicDomain.Month.May, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1986, 1986, PublicDomain.Month.September, 7, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1987, 1987, PublicDomain.Month.April, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1987, 1987, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1988, 1988, PublicDomain.Month.April, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1988, 1988, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1989, 1989, PublicDomain.Month.April, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1989, 1989, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1990, 1990, PublicDomain.Month.March, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1990, 1990, PublicDomain.Month.August, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1991, 1991, PublicDomain.Month.March, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1991, 1991, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1992, 1992, PublicDomain.Month.March, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1992, 1992, PublicDomain.Month.September, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1993, 1993, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1993, 1993, PublicDomain.Month.September, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1994, 1994, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1994, 1994, PublicDomain.Month.August, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1995, 1995, PublicDomain.Month.March, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1995, 1995, PublicDomain.Month.September, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1996, 1996, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1996, 1996, PublicDomain.Month.September, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1997, 1997, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1997, 1997, PublicDomain.Month.September, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1998, 1998, PublicDomain.Month.March, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1998, 1998, PublicDomain.Month.September, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 1999, 1999, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 1999, 1999, PublicDomain.Month.September, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2000, 2000, PublicDomain.Month.April, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2000, 2000, PublicDomain.Month.October, 6, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2001, 2001, PublicDomain.Month.April, 9, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2001, 2001, PublicDomain.Month.September, 24, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2002, 2002, PublicDomain.Month.March, 29, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2002, 2002, PublicDomain.Month.October, 7, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2003, 2003, PublicDomain.Month.March, 28, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2003, 2003, PublicDomain.Month.October, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2004, 2004, PublicDomain.Month.April, 7, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2004, 2004, PublicDomain.Month.September, 22, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2005, 2005, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2005, 2005, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2006, 2010, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2006, 2006, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2007, 2007, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2008, 2008, PublicDomain.Month.October, 5, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2009, 2009, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2010, 2010, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2011, 2011, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2011, 2011, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2012, 2015, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2012, 2012, PublicDomain.Month.September, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2013, 2013, PublicDomain.Month.September, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2014, 2014, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2015, 2015, PublicDomain.Month.September, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2016, 2016, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2016, 2016, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2017, 2021, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2017, 2017, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2018, 2018, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2019, 2019, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2020, 2020, PublicDomain.Month.September, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2021, 2021, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2022, 2022, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2022, 2022, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2023, 2032, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2023, 2023, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2024, 2024, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2025, 2025, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2026, 2026, PublicDomain.Month.September, 20, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2027, 2027, PublicDomain.Month.October, 10, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2028, 2028, PublicDomain.Month.September, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2029, 2029, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2030, 2030, PublicDomain.Month.October, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2031, 2031, PublicDomain.Month.September, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2032, 2032, PublicDomain.Month.September, 12, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2033, 2033, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2033, 2033, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2034, 2037, PublicDomain.Month.March, 26, DayOfWeek.Friday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Zion", 2034, 2034, PublicDomain.Month.September, 17, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2035, 2035, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2036, 2036, PublicDomain.Month.September, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Zion", 2037, 2037, PublicDomain.Month.September, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesJamaica_480()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Jamaica", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Jamaica", new TimeSpan(-184320000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kingston"), new PublicDomain.TzDatabase.TzZone("America/Jamaica", new TimeSpan(-184320000000), "-", "KMT", 1912, PublicDomain.Month.February, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kingston Mean Time"), new PublicDomain.TzDatabase.TzZone("America/Jamaica", new TimeSpan(-180000000000), "-", "EST", 1974, PublicDomain.Month.April, 28, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Jamaica", new TimeSpan(-180000000000), "US", "E%sT", 1984, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Jamaica", new TimeSpan(-180000000000), "-", "EST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesJapan_481()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Japan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Tokyo", new TimeSpan(335390000000), "-", "LMT", 1887, PublicDomain.Month.December, 31, null, new TimeSpan(540000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tokyo", new TimeSpan(324000000000), "-", "JST", 1896, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tokyo", new TimeSpan(324000000000), "-", "CJT", 1938, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Tokyo", new TimeSpan(324000000000), "Japan", "J%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Japan", 1948, 1948, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Japan", 1948, 1951, PublicDomain.Month.September, 8, DayOfWeek.Saturday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Japan", 1949, 1949, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Japan", 1950, 1951, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesKwajalein_482()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Kwajalein", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Kwajalein", new TimeSpan(401600000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Kwajalein", new TimeSpan(396000000000), "-", "MHT", 1969, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Kwajalein", new TimeSpan(-432000000000), "-", "KWAT", 1993, PublicDomain.Month.August, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kwajalein Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Kwajalein", new TimeSpan(432000000000), "-", "MHT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesLibya_483()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Libya", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(31640000000), "-", "LMT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(36000000000), "Libya", "CE%sT", 1959, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(72000000000), "-", "EET", 1982, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(36000000000), "Libya", "CE%sT", 1990, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(72000000000), "-", "EET", 1996, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(36000000000), "-", "CET", 1997, PublicDomain.Month.April, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(36000000000), "1:00", "CEST", 1997, PublicDomain.Month.October, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Africa/Tripoli", new TimeSpan(72000000000), "-", "EET", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("Libya", 1951, 1951, PublicDomain.Month.October, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1952, 1952, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Libya", 1953, 1953, PublicDomain.Month.October, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1954, 1954, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Libya", 1955, 1955, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1956, 1956, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Libya", 1982, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1982, 1985, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Libya", 1985, 1985, PublicDomain.Month.April, 6, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1986, 1986, PublicDomain.Month.April, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1986, 1986, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Libya", 1987, 1989, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Libya", 1987, 1989, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesMET_484()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("MET", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("MET", new TimeSpan(36000000000), "C-Eur", "ME%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesMexicoBajaNorte_485()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Mexico/BajaNorte", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-280840000000), "-", "LMT", 1922, PublicDomain.Month.January, 1, null, new TimeSpan(7160000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-252000000000), "-", "MST", 1924, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1927, PublicDomain.Month.June, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-252000000000), "-", "MST", 1930, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1931, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PDT", 1931, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1942, PublicDomain.Month.April, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PWT", 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PPT", 1945, PublicDomain.Month.November, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Peace"), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1948, PublicDomain.Month.April, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "1:00", "PDT", 1949, PublicDomain.Month.January, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1954, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "CA", "P%sT", 1961, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "-", "PST", 1976, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "US", "P%sT", 1996, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "Mexico", "P%sT", 2001, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "US", "P%sT", 2002, PublicDomain.Month.February, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Tijuana", new TimeSpan(-288000000000), "Mexico", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1949, 1949, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1962, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesMexicoBajaSur_486()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Mexico/BajaSur", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-255400000000), "-", "LMT", 1921, PublicDomain.Month.December, 31, null, new TimeSpan(860600000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-252000000000), "-", "MST", 1927, PublicDomain.Month.June, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-216000000000), "-", "CST", 1930, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-252000000000), "-", "MST", 1931, PublicDomain.Month.May, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-216000000000), "-", "CST", 1931, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-252000000000), "-", "MST", 1932, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-216000000000), "-", "CST", 1942, PublicDomain.Month.April, 24, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-252000000000), "-", "MST", 1949, PublicDomain.Month.January, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-288000000000), "-", "PST", 1970, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mazatlan", new TimeSpan(-252000000000), "Mexico", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesMexicoGeneral_487()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Mexico/General", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-237960000000), "-", "LMT", 1922, PublicDomain.Month.January, 1, null, new TimeSpan(14040000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-252000000000), "-", "MST", 1927, PublicDomain.Month.June, 10, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-216000000000), "-", "CST", 1930, PublicDomain.Month.November, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-252000000000), "-", "MST", 1931, PublicDomain.Month.May, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-216000000000), "-", "CST", 1931, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-252000000000), "-", "MST", 1932, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-216000000000), "Mexico", "C%sT", 2001, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-216000000000), "-", "CST", 2002, PublicDomain.Month.February, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Mexico_City", new TimeSpan(-216000000000), "Mexico", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.February, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1939, 1939, PublicDomain.Month.June, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1940, 1940, PublicDomain.Month.December, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1941, 1941, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1943, 1943, PublicDomain.Month.December, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("Mexico", 1944, 1944, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.February, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1950, 1950, PublicDomain.Month.July, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 1996, 2000, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.May, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2001, 2001, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Mexico", 2002, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesMideastRiyadh87_488()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Mideast/Riyadh87", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Riyadh87", new TimeSpan(112240000000), "-", "zzz", 1987, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh87", new TimeSpan(112240000000), "sol87", "zzz", 1988, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh87", new TimeSpan(112240000000), "-", "zzz", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 1, null, new TimeSpan(434000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 2, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 3, null, new TimeSpan(434550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 4, null, new TimeSpan(434850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 5, null, new TimeSpan(435100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 6, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 7, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 8, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 9, null, new TimeSpan(436150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 10, null, new TimeSpan(436400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 11, null, new TimeSpan(436650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 12, null, new TimeSpan(436900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 13, null, new TimeSpan(437100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 14, null, new TimeSpan(437350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 15, null, new TimeSpan(437550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 16, null, new TimeSpan(437750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 17, null, new TimeSpan(437950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 18, null, new TimeSpan(438150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 19, null, new TimeSpan(438350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 20, null, new TimeSpan(438550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 21, null, new TimeSpan(438700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 22, null, new TimeSpan(438900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 23, null, new TimeSpan(439050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 24, null, new TimeSpan(439200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 25, null, new TimeSpan(439350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 26, null, new TimeSpan(439500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 27, null, new TimeSpan(439600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 28, null, new TimeSpan(439750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 29, null, new TimeSpan(439850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 30, null, new TimeSpan(439950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.January, 31, null, new TimeSpan(440050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 1, null, new TimeSpan(440150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 2, null, new TimeSpan(440200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 3, null, new TimeSpan(440300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 4, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 5, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 6, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 7, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 8, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 9, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 10, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 11, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 12, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 13, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 14, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 15, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 16, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 17, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 18, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 19, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 20, null, new TimeSpan(440300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 21, null, new TimeSpan(440250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 22, null, new TimeSpan(440150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 23, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 24, null, new TimeSpan(440000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 25, null, new TimeSpan(439900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 26, null, new TimeSpan(439800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 27, null, new TimeSpan(439700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.February, 28, null, new TimeSpan(439600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 1, null, new TimeSpan(439500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 2, null, new TimeSpan(439400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 3, null, new TimeSpan(439250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 4, null, new TimeSpan(439150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 5, null, new TimeSpan(439000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 6, null, new TimeSpan(438850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 7, null, new TimeSpan(438750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 8, null, new TimeSpan(438600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 9, null, new TimeSpan(438450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 10, null, new TimeSpan(438300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 11, null, new TimeSpan(438150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 12, null, new TimeSpan(437950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 13, null, new TimeSpan(437800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 14, null, new TimeSpan(437650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 15, null, new TimeSpan(437500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 16, null, new TimeSpan(437300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 17, null, new TimeSpan(437150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 18, null, new TimeSpan(436950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 19, null, new TimeSpan(436800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 20, null, new TimeSpan(436600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 21, null, new TimeSpan(436450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 22, null, new TimeSpan(436250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 23, null, new TimeSpan(436100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 24, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 25, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 26, null, new TimeSpan(435550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 27, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 28, null, new TimeSpan(435150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 29, null, new TimeSpan(435000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 30, null, new TimeSpan(434800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.March, 31, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 1, null, new TimeSpan(434450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 2, null, new TimeSpan(434250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 3, null, new TimeSpan(434100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 4, null, new TimeSpan(433900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 5, null, new TimeSpan(433750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 6, null, new TimeSpan(433550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 7, null, new TimeSpan(433400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 8, null, new TimeSpan(433250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 9, null, new TimeSpan(433050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 10, null, new TimeSpan(432900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 11, null, new TimeSpan(432750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 12, null, new TimeSpan(432550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 13, null, new TimeSpan(432400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 14, null, new TimeSpan(432250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 15, null, new TimeSpan(432100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 16, null, new TimeSpan(431950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(50000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 17, null, new TimeSpan(431850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 18, null, new TimeSpan(431700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 19, null, new TimeSpan(431550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 20, null, new TimeSpan(431450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 21, null, new TimeSpan(431300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 22, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 23, null, new TimeSpan(431050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 24, null, new TimeSpan(430950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 25, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 26, null, new TimeSpan(430750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 27, null, new TimeSpan(430650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 28, null, new TimeSpan(430550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 29, null, new TimeSpan(430450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.April, 30, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 1, null, new TimeSpan(430300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 2, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 3, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 4, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 5, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 6, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 7, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 8, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 9, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 10, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 11, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 12, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 13, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 14, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 15, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 16, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 17, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 18, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 19, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 20, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 21, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 22, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 23, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 24, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 25, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 26, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 27, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 28, null, new TimeSpan(430300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 29, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 30, null, new TimeSpan(430450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.May, 31, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 1, null, new TimeSpan(430600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 2, null, new TimeSpan(430700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 3, null, new TimeSpan(430800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 4, null, new TimeSpan(430900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 5, null, new TimeSpan(431000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 6, null, new TimeSpan(431100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 7, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 8, null, new TimeSpan(431300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 9, null, new TimeSpan(431450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 10, null, new TimeSpan(431550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 11, null, new TimeSpan(431700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 12, null, new TimeSpan(431800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 13, null, new TimeSpan(431900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 14, null, new TimeSpan(432050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-50000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 15, null, new TimeSpan(432150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 16, null, new TimeSpan(432300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 17, null, new TimeSpan(432450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 18, null, new TimeSpan(432550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 19, null, new TimeSpan(432700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 20, null, new TimeSpan(432800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 21, null, new TimeSpan(432950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 22, null, new TimeSpan(433100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 23, null, new TimeSpan(433200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 24, null, new TimeSpan(433350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 25, null, new TimeSpan(433450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 26, null, new TimeSpan(433600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 27, null, new TimeSpan(433700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 28, null, new TimeSpan(433850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 29, null, new TimeSpan(433950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.June, 30, null, new TimeSpan(434100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 1, null, new TimeSpan(434200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 2, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 3, null, new TimeSpan(434450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 4, null, new TimeSpan(434550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 5, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 6, null, new TimeSpan(434750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 7, null, new TimeSpan(434850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 8, null, new TimeSpan(434950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 9, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 10, null, new TimeSpan(435150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 11, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 12, null, new TimeSpan(435300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 13, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 14, null, new TimeSpan(435450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 15, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 16, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 17, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 18, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 19, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 20, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 21, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 22, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 23, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 24, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 25, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 26, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 27, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 28, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 29, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 30, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.July, 31, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 1, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 2, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 3, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 4, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 5, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 6, null, new TimeSpan(435550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 7, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 8, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 9, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 10, null, new TimeSpan(435250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 11, null, new TimeSpan(435150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 12, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 13, null, new TimeSpan(434950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 14, null, new TimeSpan(434850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 15, null, new TimeSpan(434750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 16, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 17, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 18, null, new TimeSpan(434400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 19, null, new TimeSpan(434250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 20, null, new TimeSpan(434100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 21, null, new TimeSpan(433950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 22, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 23, null, new TimeSpan(433650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 24, null, new TimeSpan(433500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 25, null, new TimeSpan(433350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 26, null, new TimeSpan(433200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 27, null, new TimeSpan(433000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 28, null, new TimeSpan(432850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 29, null, new TimeSpan(432650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 30, null, new TimeSpan(432500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.August, 31, null, new TimeSpan(432300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 1, null, new TimeSpan(432100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 2, null, new TimeSpan(431900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 3, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 4, null, new TimeSpan(431550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 5, null, new TimeSpan(431350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 6, null, new TimeSpan(431150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 7, null, new TimeSpan(430950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 8, null, new TimeSpan(430750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 9, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 10, null, new TimeSpan(430300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 11, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 12, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 13, null, new TimeSpan(429700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 14, null, new TimeSpan(429450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 15, null, new TimeSpan(429250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 16, null, new TimeSpan(429050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 17, null, new TimeSpan(428850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 18, null, new TimeSpan(428600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 19, null, new TimeSpan(428400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 20, null, new TimeSpan(428200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 21, null, new TimeSpan(427950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 22, null, new TimeSpan(427750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 23, null, new TimeSpan(427550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 24, null, new TimeSpan(427350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 25, null, new TimeSpan(427150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 26, null, new TimeSpan(426950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 27, null, new TimeSpan(426700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 28, null, new TimeSpan(426500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 29, null, new TimeSpan(426300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.September, 30, null, new TimeSpan(426100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 1, null, new TimeSpan(425900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 2, null, new TimeSpan(425750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 3, null, new TimeSpan(425550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 4, null, new TimeSpan(425350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 5, null, new TimeSpan(425150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 6, null, new TimeSpan(425000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 7, null, new TimeSpan(424800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 8, null, new TimeSpan(424650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 9, null, new TimeSpan(424450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 10, null, new TimeSpan(424300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 11, null, new TimeSpan(424150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 12, null, new TimeSpan(424000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 13, null, new TimeSpan(423850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 14, null, new TimeSpan(423700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 15, null, new TimeSpan(423550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 16, null, new TimeSpan(423450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 17, null, new TimeSpan(423300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 18, null, new TimeSpan(423200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 19, null, new TimeSpan(423050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 20, null, new TimeSpan(422950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 21, null, new TimeSpan(422850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 22, null, new TimeSpan(422750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 23, null, new TimeSpan(422650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 24, null, new TimeSpan(422600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 25, null, new TimeSpan(422500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 26, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 27, null, new TimeSpan(422350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 28, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 29, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 30, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.October, 31, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 1, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 2, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 3, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 4, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 5, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 6, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 7, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 8, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 9, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 10, null, new TimeSpan(422350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 11, null, new TimeSpan(422400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 12, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 13, null, new TimeSpan(422550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 14, null, new TimeSpan(422600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 15, null, new TimeSpan(422700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 16, null, new TimeSpan(422800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 17, null, new TimeSpan(422900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 18, null, new TimeSpan(423050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 19, null, new TimeSpan(423150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 20, null, new TimeSpan(423300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 21, null, new TimeSpan(423450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 22, null, new TimeSpan(423600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 23, null, new TimeSpan(423750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 24, null, new TimeSpan(423900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 25, null, new TimeSpan(424100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 26, null, new TimeSpan(424300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 27, null, new TimeSpan(424450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 28, null, new TimeSpan(424650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 29, null, new TimeSpan(424850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.November, 30, null, new TimeSpan(425100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 1, null, new TimeSpan(425300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 2, null, new TimeSpan(425500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 3, null, new TimeSpan(425750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 4, null, new TimeSpan(426000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 5, null, new TimeSpan(426250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 6, null, new TimeSpan(426500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 7, null, new TimeSpan(426750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 8, null, new TimeSpan(427000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 9, null, new TimeSpan(427250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 10, null, new TimeSpan(427500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 11, null, new TimeSpan(427800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 12, null, new TimeSpan(428050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 13, null, new TimeSpan(428350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 14, null, new TimeSpan(428650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 15, null, new TimeSpan(428900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 16, null, new TimeSpan(429200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 17, null, new TimeSpan(429500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 18, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 19, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 20, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 21, null, new TimeSpan(430700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 22, null, new TimeSpan(431000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 23, null, new TimeSpan(431300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 24, null, new TimeSpan(431600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 25, null, new TimeSpan(431900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 26, null, new TimeSpan(432200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 27, null, new TimeSpan(432450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 28, null, new TimeSpan(432750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 29, null, new TimeSpan(433050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 30, null, new TimeSpan(433350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol87", 1987, 1987, PublicDomain.Month.December, 31, null, new TimeSpan(433650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1650000000), "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesMideastRiyadh88_489()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Mideast/Riyadh88", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Riyadh88", new TimeSpan(112240000000), "-", "zzz", 1988, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh88", new TimeSpan(112240000000), "sol88", "zzz", 1989, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh88", new TimeSpan(112240000000), "-", "zzz", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 1, null, new TimeSpan(433950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 2, null, new TimeSpan(434200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 3, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 4, null, new TimeSpan(434800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 5, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 6, null, new TimeSpan(435300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 7, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 8, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 9, null, new TimeSpan(436100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 10, null, new TimeSpan(436350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 11, null, new TimeSpan(436600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 12, null, new TimeSpan(436850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 13, null, new TimeSpan(437050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 14, null, new TimeSpan(437300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 15, null, new TimeSpan(437500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 16, null, new TimeSpan(437700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 17, null, new TimeSpan(437900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 18, null, new TimeSpan(438100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 19, null, new TimeSpan(438300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 20, null, new TimeSpan(438500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 21, null, new TimeSpan(438650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 22, null, new TimeSpan(438850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 23, null, new TimeSpan(439000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 24, null, new TimeSpan(439150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 25, null, new TimeSpan(439300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 26, null, new TimeSpan(439450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 27, null, new TimeSpan(439600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 28, null, new TimeSpan(439700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 29, null, new TimeSpan(439800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 30, null, new TimeSpan(439900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.January, 31, null, new TimeSpan(440000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 1, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 2, null, new TimeSpan(440200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 3, null, new TimeSpan(440250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 4, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 5, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 6, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 7, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 8, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 9, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 10, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 11, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 12, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 13, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 14, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 15, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 16, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 17, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 18, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 19, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 20, null, new TimeSpan(440300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 21, null, new TimeSpan(440250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 22, null, new TimeSpan(440200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 23, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 24, null, new TimeSpan(440000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 25, null, new TimeSpan(439950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 26, null, new TimeSpan(439850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 27, null, new TimeSpan(439750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 28, null, new TimeSpan(439650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.February, 29, null, new TimeSpan(439500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 1, null, new TimeSpan(439400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 2, null, new TimeSpan(439300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 3, null, new TimeSpan(439150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 4, null, new TimeSpan(439050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 5, null, new TimeSpan(438900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 6, null, new TimeSpan(438750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 7, null, new TimeSpan(438600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 8, null, new TimeSpan(438450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 9, null, new TimeSpan(438300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 10, null, new TimeSpan(438150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 11, null, new TimeSpan(438000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 12, null, new TimeSpan(437850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 13, null, new TimeSpan(437700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 14, null, new TimeSpan(437500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 15, null, new TimeSpan(437350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 16, null, new TimeSpan(437200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 17, null, new TimeSpan(437000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 18, null, new TimeSpan(436850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 19, null, new TimeSpan(436650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 20, null, new TimeSpan(436500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 21, null, new TimeSpan(436300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 22, null, new TimeSpan(436100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 23, null, new TimeSpan(435950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 24, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 25, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 26, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 27, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 28, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 29, null, new TimeSpan(434850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 30, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.March, 31, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 1, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 2, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 3, null, new TimeSpan(433950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 4, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 5, null, new TimeSpan(433600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 6, null, new TimeSpan(433450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 7, null, new TimeSpan(433250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 8, null, new TimeSpan(433100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 9, null, new TimeSpan(432950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 10, null, new TimeSpan(432750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 11, null, new TimeSpan(432600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 12, null, new TimeSpan(432450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 13, null, new TimeSpan(432300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 14, null, new TimeSpan(432150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 15, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 16, null, new TimeSpan(431850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 17, null, new TimeSpan(431700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 18, null, new TimeSpan(431600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 19, null, new TimeSpan(431450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 20, null, new TimeSpan(431350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 21, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 22, null, new TimeSpan(431100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 23, null, new TimeSpan(430950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 24, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 25, null, new TimeSpan(430750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 26, null, new TimeSpan(430650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 27, null, new TimeSpan(430550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 28, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 29, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.April, 30, null, new TimeSpan(430300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 1, null, new TimeSpan(430250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 2, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 3, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 4, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 5, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 6, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 7, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 8, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 9, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 10, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 11, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 12, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 13, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 14, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 15, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 16, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 17, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 18, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 19, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 20, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 21, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 22, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 23, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 24, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 25, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 26, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 27, null, new TimeSpan(430250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 28, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 29, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 30, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.May, 31, null, new TimeSpan(430600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 1, null, new TimeSpan(430700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 2, null, new TimeSpan(430750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 3, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 4, null, new TimeSpan(430950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 5, null, new TimeSpan(431100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 6, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 7, null, new TimeSpan(431300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 8, null, new TimeSpan(431400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 9, null, new TimeSpan(431550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 10, null, new TimeSpan(431650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 11, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 12, null, new TimeSpan(431900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 13, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 14, null, new TimeSpan(432150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 15, null, new TimeSpan(432250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 16, null, new TimeSpan(432400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 17, null, new TimeSpan(432550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 18, null, new TimeSpan(432650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 19, null, new TimeSpan(432800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 20, null, new TimeSpan(432900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 21, null, new TimeSpan(433050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 22, null, new TimeSpan(433200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 23, null, new TimeSpan(433300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 24, null, new TimeSpan(433450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 25, null, new TimeSpan(433550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 26, null, new TimeSpan(433700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 27, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 28, null, new TimeSpan(433950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 29, null, new TimeSpan(434050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.June, 30, null, new TimeSpan(434200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 1, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 2, null, new TimeSpan(434400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 3, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 4, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 5, null, new TimeSpan(434750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 6, null, new TimeSpan(434850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 7, null, new TimeSpan(434950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 8, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 9, null, new TimeSpan(435100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 10, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 11, null, new TimeSpan(435300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 12, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 13, null, new TimeSpan(435450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 14, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 15, null, new TimeSpan(435550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 16, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 17, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 18, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 19, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 20, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 21, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 22, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 23, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 24, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 25, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 26, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 27, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 28, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 29, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 30, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.July, 31, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 1, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 2, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 3, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 4, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 5, null, new TimeSpan(435550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 6, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 7, null, new TimeSpan(435450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 8, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 9, null, new TimeSpan(435250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 10, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 11, null, new TimeSpan(435100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 12, null, new TimeSpan(435000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 13, null, new TimeSpan(434900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 14, null, new TimeSpan(434750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 15, null, new TimeSpan(434650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 16, null, new TimeSpan(434550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 17, null, new TimeSpan(434400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 18, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 19, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 20, null, new TimeSpan(434000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 21, null, new TimeSpan(433850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 22, null, new TimeSpan(433700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 23, null, new TimeSpan(433550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 24, null, new TimeSpan(433400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 25, null, new TimeSpan(433200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 26, null, new TimeSpan(433050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 27, null, new TimeSpan(432900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 28, null, new TimeSpan(432700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 29, null, new TimeSpan(432500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 30, null, new TimeSpan(432350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.August, 31, null, new TimeSpan(432150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 1, null, new TimeSpan(431950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(50000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 2, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 3, null, new TimeSpan(431600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 4, null, new TimeSpan(431400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 5, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 6, null, new TimeSpan(431000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 7, null, new TimeSpan(430800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 8, null, new TimeSpan(430550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 9, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 10, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 11, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 12, null, new TimeSpan(429750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 13, null, new TimeSpan(429500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 14, null, new TimeSpan(429300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 15, null, new TimeSpan(429100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 16, null, new TimeSpan(428900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 17, null, new TimeSpan(428650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 18, null, new TimeSpan(428450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 19, null, new TimeSpan(428250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 20, null, new TimeSpan(428050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 21, null, new TimeSpan(427800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 22, null, new TimeSpan(427600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 23, null, new TimeSpan(427400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 24, null, new TimeSpan(427200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 25, null, new TimeSpan(427000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 26, null, new TimeSpan(426750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 27, null, new TimeSpan(426550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 28, null, new TimeSpan(426350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 29, null, new TimeSpan(426150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.September, 30, null, new TimeSpan(425950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 1, null, new TimeSpan(425750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 2, null, new TimeSpan(425600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 3, null, new TimeSpan(425400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 4, null, new TimeSpan(425200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 5, null, new TimeSpan(425050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 6, null, new TimeSpan(424850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 7, null, new TimeSpan(424700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 8, null, new TimeSpan(424500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 9, null, new TimeSpan(424350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 10, null, new TimeSpan(424200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 11, null, new TimeSpan(424050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 12, null, new TimeSpan(423900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 13, null, new TimeSpan(423750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 14, null, new TimeSpan(423600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 15, null, new TimeSpan(423450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 16, null, new TimeSpan(423350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 17, null, new TimeSpan(423200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 18, null, new TimeSpan(423100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 19, null, new TimeSpan(423000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 20, null, new TimeSpan(422850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 21, null, new TimeSpan(422800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 22, null, new TimeSpan(422700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 23, null, new TimeSpan(422600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 24, null, new TimeSpan(422500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 25, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 26, null, new TimeSpan(422400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 27, null, new TimeSpan(422350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 28, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 29, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 30, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.October, 31, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 1, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 2, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 3, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 4, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 5, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 6, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 7, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 8, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 9, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 10, null, new TimeSpan(422400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 11, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 12, null, new TimeSpan(422500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 13, null, new TimeSpan(422600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 14, null, new TimeSpan(422700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 15, null, new TimeSpan(422800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 16, null, new TimeSpan(422900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 17, null, new TimeSpan(423000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 18, null, new TimeSpan(423150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 19, null, new TimeSpan(423250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 20, null, new TimeSpan(423400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 21, null, new TimeSpan(423550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 22, null, new TimeSpan(423700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 23, null, new TimeSpan(423900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 24, null, new TimeSpan(424050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 25, null, new TimeSpan(424250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 26, null, new TimeSpan(424400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 27, null, new TimeSpan(424600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 28, null, new TimeSpan(424800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 29, null, new TimeSpan(425050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.November, 30, null, new TimeSpan(425250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 1, null, new TimeSpan(425450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 2, null, new TimeSpan(425700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 3, null, new TimeSpan(425950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 4, null, new TimeSpan(426150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 5, null, new TimeSpan(426400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 6, null, new TimeSpan(426650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 7, null, new TimeSpan(426950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 8, null, new TimeSpan(427200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 9, null, new TimeSpan(427450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 10, null, new TimeSpan(427750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 11, null, new TimeSpan(428000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 12, null, new TimeSpan(428300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 13, null, new TimeSpan(428550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 14, null, new TimeSpan(428850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 15, null, new TimeSpan(429150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 16, null, new TimeSpan(429450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 17, null, new TimeSpan(429750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 18, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 19, null, new TimeSpan(430300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 20, null, new TimeSpan(430600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 21, null, new TimeSpan(430900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 22, null, new TimeSpan(431200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 23, null, new TimeSpan(431500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 24, null, new TimeSpan(431800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 25, null, new TimeSpan(432100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 26, null, new TimeSpan(432400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 27, null, new TimeSpan(432700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 28, null, new TimeSpan(433000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 29, null, new TimeSpan(433300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 30, null, new TimeSpan(433550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol88", 1988, 1988, PublicDomain.Month.December, 31, null, new TimeSpan(433850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1850000000), "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesMideastRiyadh89_490()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Mideast/Riyadh89", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Riyadh89", new TimeSpan(112240000000), "-", "zzz", 1989, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh89", new TimeSpan(112240000000), "sol89", "zzz", 1990, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Riyadh89", new TimeSpan(112240000000), "-", "zzz", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 1, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 2, null, new TimeSpan(434450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 3, null, new TimeSpan(434700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 4, null, new TimeSpan(435000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 5, null, new TimeSpan(435250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 6, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 7, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 8, null, new TimeSpan(436050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 9, null, new TimeSpan(436300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 10, null, new TimeSpan(436550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 11, null, new TimeSpan(436750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 12, null, new TimeSpan(437000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 13, null, new TimeSpan(437250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 14, null, new TimeSpan(437450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 15, null, new TimeSpan(437650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 16, null, new TimeSpan(437850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 17, null, new TimeSpan(438050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 18, null, new TimeSpan(438250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 19, null, new TimeSpan(438450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 20, null, new TimeSpan(438650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 21, null, new TimeSpan(438800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 22, null, new TimeSpan(438950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 23, null, new TimeSpan(439150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 24, null, new TimeSpan(439300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 25, null, new TimeSpan(439400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 26, null, new TimeSpan(439550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 27, null, new TimeSpan(439700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 28, null, new TimeSpan(439800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 29, null, new TimeSpan(439900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 30, null, new TimeSpan(440000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.January, 31, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 1, null, new TimeSpan(440200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 2, null, new TimeSpan(440250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 3, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 4, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 5, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 6, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 7, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 8, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 9, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 10, null, new TimeSpan(440600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 11, null, new TimeSpan(440600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 12, null, new TimeSpan(440600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 13, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 14, null, new TimeSpan(440550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 15, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 16, null, new TimeSpan(440500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 17, null, new TimeSpan(440450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 18, null, new TimeSpan(440400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 19, null, new TimeSpan(440350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 20, null, new TimeSpan(440300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 21, null, new TimeSpan(440200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 22, null, new TimeSpan(440150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 23, null, new TimeSpan(440050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-8050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 24, null, new TimeSpan(439950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 25, null, new TimeSpan(439850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 26, null, new TimeSpan(439750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 27, null, new TimeSpan(439650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.February, 28, null, new TimeSpan(439550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 1, null, new TimeSpan(439450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 2, null, new TimeSpan(439300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 3, null, new TimeSpan(439200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 4, null, new TimeSpan(439050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-7050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 5, null, new TimeSpan(438950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 6, null, new TimeSpan(438800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 7, null, new TimeSpan(438650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 8, null, new TimeSpan(438500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 9, null, new TimeSpan(438350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 10, null, new TimeSpan(438200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 11, null, new TimeSpan(438050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-6050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 12, null, new TimeSpan(437900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 13, null, new TimeSpan(437700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 14, null, new TimeSpan(437550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 15, null, new TimeSpan(437400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 16, null, new TimeSpan(437200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 17, null, new TimeSpan(437050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-5050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 18, null, new TimeSpan(436850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 19, null, new TimeSpan(436700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 20, null, new TimeSpan(436500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 21, null, new TimeSpan(436350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 22, null, new TimeSpan(436150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-4150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 23, null, new TimeSpan(435950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 24, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 25, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 26, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 27, null, new TimeSpan(435250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 28, null, new TimeSpan(435050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 29, null, new TimeSpan(434900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 30, null, new TimeSpan(434700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.March, 31, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 1, null, new TimeSpan(434350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 2, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 3, null, new TimeSpan(434000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 4, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 5, null, new TimeSpan(433650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 6, null, new TimeSpan(433450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 7, null, new TimeSpan(433300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 8, null, new TimeSpan(433100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 9, null, new TimeSpan(432950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 10, null, new TimeSpan(432800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 11, null, new TimeSpan(432650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 12, null, new TimeSpan(432500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 13, null, new TimeSpan(432350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 14, null, new TimeSpan(432200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 15, null, new TimeSpan(432050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-50000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 16, null, new TimeSpan(431900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 17, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 18, null, new TimeSpan(431600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 19, null, new TimeSpan(431500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 20, null, new TimeSpan(431350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 21, null, new TimeSpan(431250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 22, null, new TimeSpan(431100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 23, null, new TimeSpan(431000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 24, null, new TimeSpan(430900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 25, null, new TimeSpan(430800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 26, null, new TimeSpan(430700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 27, null, new TimeSpan(430600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 28, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 29, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.April, 30, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 1, null, new TimeSpan(430250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 2, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 3, null, new TimeSpan(430100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 4, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 5, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 6, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 7, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 8, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 9, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 10, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 11, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 12, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 13, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 14, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 15, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 16, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 17, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 18, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 19, null, new TimeSpan(429850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 20, null, new TimeSpan(429900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 21, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 22, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 23, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 24, null, new TimeSpan(430050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 25, null, new TimeSpan(430150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 26, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 27, null, new TimeSpan(430250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 28, null, new TimeSpan(430350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 29, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 30, null, new TimeSpan(430500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.May, 31, null, new TimeSpan(430550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 1, null, new TimeSpan(430650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 2, null, new TimeSpan(430750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 3, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 4, null, new TimeSpan(430950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 5, null, new TimeSpan(431050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 6, null, new TimeSpan(431150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 7, null, new TimeSpan(431250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 8, null, new TimeSpan(431400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 9, null, new TimeSpan(431500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 10, null, new TimeSpan(431600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 11, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 12, null, new TimeSpan(431850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 13, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 14, null, new TimeSpan(432100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 15, null, new TimeSpan(432250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 16, null, new TimeSpan(432350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 17, null, new TimeSpan(432500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 18, null, new TimeSpan(432650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 19, null, new TimeSpan(432750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 20, null, new TimeSpan(432900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 21, null, new TimeSpan(433000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 22, null, new TimeSpan(433150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 23, null, new TimeSpan(433300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 24, null, new TimeSpan(433400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 25, null, new TimeSpan(433550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 26, null, new TimeSpan(433650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 27, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 28, null, new TimeSpan(433900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 29, null, new TimeSpan(434050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.June, 30, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 1, null, new TimeSpan(434250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 2, null, new TimeSpan(434400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 3, null, new TimeSpan(434500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 4, null, new TimeSpan(434600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 5, null, new TimeSpan(434700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 6, null, new TimeSpan(434800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 7, null, new TimeSpan(434900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 8, null, new TimeSpan(435000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 9, null, new TimeSpan(435100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 10, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 11, null, new TimeSpan(435250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 12, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 13, null, new TimeSpan(435400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 14, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 15, null, new TimeSpan(435550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 16, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 17, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 18, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 19, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 20, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 21, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 22, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 23, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 24, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 25, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 26, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 27, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 28, null, new TimeSpan(435900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 29, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 30, null, new TimeSpan(435850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.July, 31, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 1, null, new TimeSpan(435800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 2, null, new TimeSpan(435750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 3, null, new TimeSpan(435700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 4, null, new TimeSpan(435650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 5, null, new TimeSpan(435600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 6, null, new TimeSpan(435500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 7, null, new TimeSpan(435450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 8, null, new TimeSpan(435350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 9, null, new TimeSpan(435300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 10, null, new TimeSpan(435200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 11, null, new TimeSpan(435100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 12, null, new TimeSpan(435000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-3000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 13, null, new TimeSpan(434900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 14, null, new TimeSpan(434800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 15, null, new TimeSpan(434700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 16, null, new TimeSpan(434550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 17, null, new TimeSpan(434450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 18, null, new TimeSpan(434300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 19, null, new TimeSpan(434150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 20, null, new TimeSpan(434050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 21, null, new TimeSpan(433900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 22, null, new TimeSpan(433750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 23, null, new TimeSpan(433600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 24, null, new TimeSpan(433400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 25, null, new TimeSpan(433250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 26, null, new TimeSpan(433100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 27, null, new TimeSpan(432900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 28, null, new TimeSpan(432750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 29, null, new TimeSpan(432550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 30, null, new TimeSpan(432400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.August, 31, null, new TimeSpan(432200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 1, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 2, null, new TimeSpan(431850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 3, null, new TimeSpan(431650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 4, null, new TimeSpan(431450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 5, null, new TimeSpan(431250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 6, null, new TimeSpan(431050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 7, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 8, null, new TimeSpan(430650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 9, null, new TimeSpan(430400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 10, null, new TimeSpan(430200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 11, null, new TimeSpan(430000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 12, null, new TimeSpan(429800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 13, null, new TimeSpan(429600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 14, null, new TimeSpan(429350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 15, null, new TimeSpan(429150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 16, null, new TimeSpan(428950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 17, null, new TimeSpan(428750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 18, null, new TimeSpan(428500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 19, null, new TimeSpan(428300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 20, null, new TimeSpan(428100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 21, null, new TimeSpan(427900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 22, null, new TimeSpan(427650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 23, null, new TimeSpan(427450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 24, null, new TimeSpan(427250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 25, null, new TimeSpan(427050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 26, null, new TimeSpan(426850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 27, null, new TimeSpan(426650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 28, null, new TimeSpan(426400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 29, null, new TimeSpan(426200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.September, 30, null, new TimeSpan(426000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 1, null, new TimeSpan(425850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 2, null, new TimeSpan(425650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 3, null, new TimeSpan(425450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 4, null, new TimeSpan(425250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 5, null, new TimeSpan(425100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 6, null, new TimeSpan(424900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 7, null, new TimeSpan(424700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 8, null, new TimeSpan(424550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 9, null, new TimeSpan(424400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 10, null, new TimeSpan(424200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 11, null, new TimeSpan(424050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 12, null, new TimeSpan(423900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 13, null, new TimeSpan(423750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 14, null, new TimeSpan(423600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 15, null, new TimeSpan(423500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 16, null, new TimeSpan(423350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 17, null, new TimeSpan(423200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 18, null, new TimeSpan(423100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 19, null, new TimeSpan(423000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 20, null, new TimeSpan(422900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 21, null, new TimeSpan(422800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 22, null, new TimeSpan(422700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 23, null, new TimeSpan(422600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 24, null, new TimeSpan(422500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 25, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 26, null, new TimeSpan(422400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 27, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 28, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 29, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 30, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.October, 31, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 1, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 2, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 3, null, new TimeSpan(422100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 4, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 5, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 6, null, new TimeSpan(422150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 7, null, new TimeSpan(422200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 8, null, new TimeSpan(422250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 9, null, new TimeSpan(422300000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9700000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 10, null, new TimeSpan(422350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 11, null, new TimeSpan(422400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 12, null, new TimeSpan(422450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 13, null, new TimeSpan(422550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 14, null, new TimeSpan(422650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 15, null, new TimeSpan(422750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 16, null, new TimeSpan(422850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 17, null, new TimeSpan(422950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(9050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 18, null, new TimeSpan(423100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 19, null, new TimeSpan(423200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 20, null, new TimeSpan(423350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 21, null, new TimeSpan(423500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 22, null, new TimeSpan(423650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 23, null, new TimeSpan(423850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 24, null, new TimeSpan(424000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(8000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 25, null, new TimeSpan(424200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 26, null, new TimeSpan(424400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 27, null, new TimeSpan(424550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 28, null, new TimeSpan(424750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 29, null, new TimeSpan(425000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(7000000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.November, 30, null, new TimeSpan(425200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6800000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 1, null, new TimeSpan(425400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 2, null, new TimeSpan(425650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 3, null, new TimeSpan(425900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(6100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 4, null, new TimeSpan(426150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 5, null, new TimeSpan(426350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 6, null, new TimeSpan(426600000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5400000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 7, null, new TimeSpan(426900000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(5100000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 8, null, new TimeSpan(427150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 9, null, new TimeSpan(427400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 10, null, new TimeSpan(427700000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4300000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 11, null, new TimeSpan(427950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(4050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 12, null, new TimeSpan(428250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 13, null, new TimeSpan(428500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 14, null, new TimeSpan(428800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(3200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 15, null, new TimeSpan(429100000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2900000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 16, null, new TimeSpan(429400000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2600000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 17, null, new TimeSpan(429650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 18, null, new TimeSpan(429950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(2050000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 19, null, new TimeSpan(430250000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1750000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 20, null, new TimeSpan(430550000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1450000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 21, null, new TimeSpan(430850000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(1150000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 22, null, new TimeSpan(431150000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(850000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 23, null, new TimeSpan(431450000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(550000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 24, null, new TimeSpan(431750000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(250000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 25, null, new TimeSpan(432050000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-50000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 26, null, new TimeSpan(432350000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-350000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 27, null, new TimeSpan(432650000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-650000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 28, null, new TimeSpan(432950000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-950000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 29, null, new TimeSpan(433200000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1200000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 30, null, new TimeSpan(433500000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1500000000), "-", null), new PublicDomain.TzDatabase.TzRule("sol89", 1989, 1989, PublicDomain.Month.December, 31, null, new TimeSpan(433800000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(-1800000000), "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesMST_491()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("MST", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("MST", new TimeSpan(-252000000000), "-", "MST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesMST7MDT_492()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("MST7MDT", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("MST7MDT", new TimeSpan(-252000000000), "US", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesNavajo_493()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Navajo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-251960000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(432040000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "Denver", "M%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "Denver", "M%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1920, 1921, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Denver", 1920, 1920, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1921, 1921, PublicDomain.Month.May, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1965, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Denver", 1965, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesNZ_494()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("NZ", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Auckland", new TimeSpan(419440000000), "-", "LMT", 1868, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Auckland", new TimeSpan(414000000000), "NZ", "NZ%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Auckland", new TimeSpan(432000000000), "NZ", "NZ%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NZ", 1927, 1927, PublicDomain.Month.November, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1928, 1928, PublicDomain.Month.March, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "M", null), new PublicDomain.TzDatabase.TzRule("NZ", 1928, 1933, PublicDomain.Month.October, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1929, 1933, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "M", null), new PublicDomain.TzDatabase.TzRule("NZ", 1934, 1940, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "M", null), new PublicDomain.TzDatabase.TzRule("NZ", 1934, 1940, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1946, 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1974, 1974, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZ", 1975, 1975, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1975, 1988, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZ", 1976, 1989, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1989, 1989, PublicDomain.Month.October, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZ", 1990, 2007, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1990, 2006, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZ", 2007, 2147483647, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZ", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesNZCHAT_495()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("NZ-CHAT", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Chatham", new TimeSpan(440280000000), "-", "LMT", 1957, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Chatham", new TimeSpan(459000000000), "Chatham", "CHA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Chatham", 1974, 1974, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1975, 1975, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1975, 1988, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1976, 1989, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1989, 1989, PublicDomain.Month.October, 8, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1990, 2007, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1990, 2006, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chatham", 2007, 2147483647, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chatham", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificApia_496()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Apia", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Apia", new TimeSpan(451840000000), "-", "LMT", 1879, PublicDomain.Month.July, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Apia", new TimeSpan(-412160000000), "-", "LMT", 1911, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Apia", new TimeSpan(-414000000000), "-", "SAMT", 1950, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Samoa Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Apia", new TimeSpan(-396000000000), "-", "WST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Samoa Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificAuckland_497()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Auckland", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Auckland", new TimeSpan(419440000000), "-", "LMT", 1868, PublicDomain.Month.November, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Auckland", new TimeSpan(414000000000), "NZ", "NZ%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Auckland", new TimeSpan(432000000000), "NZ", "NZ%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NZ", 1927, 1927, PublicDomain.Month.November, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1928, 1928, PublicDomain.Month.March, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "M", null), new PublicDomain.TzDatabase.TzRule("NZ", 1928, 1933, PublicDomain.Month.October, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1929, 1933, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "M", null), new PublicDomain.TzDatabase.TzRule("NZ", 1934, 1940, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "M", null), new PublicDomain.TzDatabase.TzRule("NZ", 1934, 1940, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1946, 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1974, 1974, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZ", 1975, 1975, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1975, 1988, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZ", 1976, 1989, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1989, 1989, PublicDomain.Month.October, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZ", 1990, 2007, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NZ", 1990, 2006, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZ", 2007, 2147483647, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NZ", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificChatham_498()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Chatham", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Chatham", new TimeSpan(440280000000), "-", "LMT", 1957, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Chatham", new TimeSpan(459000000000), "Chatham", "CHA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Chatham", 1974, 1974, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1975, 1975, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1975, 1988, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1976, 1989, PublicDomain.Month.March, 1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1989, 1989, PublicDomain.Month.October, 8, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1990, 2007, PublicDomain.Month.March, 15, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chatham", 1990, 2006, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chatham", 2007, 2147483647, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chatham", 2008, 2147483647, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(99000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificEaster_499()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Easter", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Easter", new TimeSpan(-262640000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Easter", new TimeSpan(-262480000000), "-", "EMT", 1932, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Easter Mean Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Easter", new TimeSpan(-252000000000), "Chile", "EAS%sT", 1982, PublicDomain.Month.March, 13, null, new TimeSpan(756000000000), TzDatabase.TimeModifier.LocalWallTime, "Easter I Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Easter", new TimeSpan(-216000000000), "Chile", "EAS%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Chile", 1927, 1932, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1928, 1932, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1942, 1942, PublicDomain.Month.June, 1, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1942, 1942, PublicDomain.Month.August, 1, null, new TimeSpan(180000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1946, 1946, PublicDomain.Month.July, 15, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1946, 1946, PublicDomain.Month.September, 1, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1947, 1947, PublicDomain.Month.April, 1, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1968, 1968, PublicDomain.Month.November, 3, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1969, 1969, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1969, 1969, PublicDomain.Month.November, 23, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1970, 1970, PublicDomain.Month.March, 29, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1970, 1972, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1971, 1971, PublicDomain.Month.March, 14, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1972, 1986, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1973, 1973, PublicDomain.Month.September, 30, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1974, 1987, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1987, 1987, PublicDomain.Month.April, 12, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1988, 1989, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1988, 1988, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1989, 1989, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1990, 1990, PublicDomain.Month.March, 18, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1990, 1990, PublicDomain.Month.September, 16, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1991, 1996, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1991, 1997, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1997, 1997, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1998, 1998, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1998, 1998, PublicDomain.Month.September, 27, null, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 1999, 1999, PublicDomain.Month.April, 4, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 1999, 2147483647, PublicDomain.Month.October, 9, DayOfWeek.Sunday, new TimeSpan(144000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Chile", 2000, 2007, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 2008, 2008, PublicDomain.Month.March, 30, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Chile", 2009, 2147483647, PublicDomain.Month.March, 9, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificEfate_500()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Efate", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Efate", new TimeSpan(403960000000), "-", "LMT", 1912, PublicDomain.Month.January, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Vila"), new PublicDomain.TzDatabase.TzZone("Pacific/Efate", new TimeSpan(396000000000), "Vanuatu", "VU%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Vanuatu Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Vanuatu", 1983, 1983, PublicDomain.Month.September, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Vanuatu", 1984, 1991, PublicDomain.Month.March, 23, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Vanuatu", 1984, 1984, PublicDomain.Month.October, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Vanuatu", 1985, 1991, PublicDomain.Month.September, 23, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Vanuatu", 1992, 1993, PublicDomain.Month.January, 23, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Vanuatu", 1992, 1992, PublicDomain.Month.October, 23, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificEnderbury_501()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Enderbury", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Enderbury", new TimeSpan(-410600000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Enderbury", new TimeSpan(-432000000000), "-", "PHOT", 1979, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Phoenix Is Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Enderbury", new TimeSpan(-396000000000), "-", "PHOT", 1995, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Enderbury", new TimeSpan(468000000000), "-", "PHOT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificFakaofo_502()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Fakaofo", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Fakaofo", new TimeSpan(-410960000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Fakaofo", new TimeSpan(-360000000000), "-", "TKT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tokelau Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificFiji_503()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Fiji", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Fiji", new TimeSpan(428200000000), "-", "LMT", 1915, PublicDomain.Month.October, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Suva"), new PublicDomain.TzDatabase.TzZone("Pacific/Fiji", new TimeSpan(432000000000), "Fiji", "FJ%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Fiji Time") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Fiji", 1998, 1999, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Fiji", 1999, 2000, PublicDomain.Month.February, -1, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificFunafuti_504()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Funafuti", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Funafuti", new TimeSpan(430120000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Funafuti", new TimeSpan(432000000000), "-", "TVT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tuvalu Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificGalapagos_505()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Galapagos", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Galapagos", new TimeSpan(-215040000000), "-", "LMT", 1931, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Puerto Baquerizo Moreno"), new PublicDomain.TzDatabase.TzZone("Pacific/Galapagos", new TimeSpan(-180000000000), "-", "ECT", 1986, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Galapagos", new TimeSpan(-216000000000), "-", "GALT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Galapagos Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificGambier_506()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Gambier", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Gambier", new TimeSpan(-323880000000), "-", "LMT", 1912, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Rikitea"), new PublicDomain.TzDatabase.TzZone("Pacific/Gambier", new TimeSpan(-324000000000), "-", "GAMT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Gambier Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificGuadalcanal_507()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Guadalcanal", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Guadalcanal", new TimeSpan(383880000000), "-", "LMT", 1912, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Honiara"), new PublicDomain.TzDatabase.TzZone("Pacific/Guadalcanal", new TimeSpan(396000000000), "-", "SBT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Solomon Is Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificGuam_508()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Guam", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Guam", new TimeSpan(-516600000000), "-", "LMT", 1844, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Guam", new TimeSpan(347400000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Agana"), new PublicDomain.TzDatabase.TzZone("Pacific/Guam", new TimeSpan(360000000000), "-", "GST", 2000, PublicDomain.Month.December, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Guam"), new PublicDomain.TzDatabase.TzZone("Pacific/Guam", new TimeSpan(360000000000), "-", "ChST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Chamorro Standard Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificHonolulu_509()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Honolulu", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Honolulu", new TimeSpan(-378860000000), "-", "LMT", 1900, PublicDomain.Month.January, 1, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Honolulu", new TimeSpan(-378000000000), "-", "HST", 1933, PublicDomain.Month.April, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Honolulu", new TimeSpan(-378000000000), "1:00", "HDT", 1933, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Honolulu", new TimeSpan(-378000000000), "US", "H%sT", 1947, PublicDomain.Month.June, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Honolulu", new TimeSpan(-360000000000), "-", "HST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificJohnston_510()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Johnston", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Johnston", new TimeSpan(-360000000000), "-", "HST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificKiritimati_511()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Kiritimati", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Kiritimati", new TimeSpan(-377600000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Kiritimati", new TimeSpan(-384000000000), "-", "LINT", 1979, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Line Is Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Kiritimati", new TimeSpan(-360000000000), "-", "LINT", 1995, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Kiritimati", new TimeSpan(504000000000), "-", "LINT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificKosrae_512()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Kosrae", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Kosrae", new TimeSpan(391160000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Kosrae", new TimeSpan(396000000000), "-", "KOST", 1969, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kosrae Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Kosrae", new TimeSpan(432000000000), "-", "KOST", 1999, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Kosrae", new TimeSpan(396000000000), "-", "KOST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificKwajalein_513()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Kwajalein", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Kwajalein", new TimeSpan(401600000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Kwajalein", new TimeSpan(396000000000), "-", "MHT", 1969, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Kwajalein", new TimeSpan(-432000000000), "-", "KWAT", 1993, PublicDomain.Month.August, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kwajalein Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Kwajalein", new TimeSpan(432000000000), "-", "MHT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificMajuro_514()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Majuro", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Majuro", new TimeSpan(410880000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Majuro", new TimeSpan(396000000000), "-", "MHT", 1969, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Marshall Islands Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Majuro", new TimeSpan(432000000000), "-", "MHT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificMarquesas_515()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Marquesas", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Marquesas", new TimeSpan(-334800000000), "-", "LMT", 1912, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Marquesas", new TimeSpan(-342000000000), "-", "MART", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Marquesas Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificMidway_516()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Midway", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Midway", new TimeSpan(-425680000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Midway", new TimeSpan(-396000000000), "-", "NST", 1956, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Midway", new TimeSpan(-396000000000), "1:00", "NDT", 1956, PublicDomain.Month.September, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Midway", new TimeSpan(-396000000000), "-", "NST", 1967, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "N=Nome"), new PublicDomain.TzDatabase.TzZone("Pacific/Midway", new TimeSpan(-396000000000), "-", "BST", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "B=Bering"), new PublicDomain.TzDatabase.TzZone("Pacific/Midway", new TimeSpan(-396000000000), "-", "SST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "S=Samoa") }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificNauru_517()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Nauru", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Nauru", new TimeSpan(400600000000), "-", "LMT", 1921, PublicDomain.Month.January, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Uaobe"), new PublicDomain.TzDatabase.TzZone("Pacific/Nauru", new TimeSpan(414000000000), "-", "NRT", 1942, PublicDomain.Month.March, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Nauru Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Nauru", new TimeSpan(324000000000), "-", "JST", 1944, PublicDomain.Month.August, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Nauru", new TimeSpan(414000000000), "-", "NRT", 1979, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Nauru", new TimeSpan(432000000000), "-", "NRT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificNiue_518()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Niue", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Niue", new TimeSpan(-407800000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Alofi"), new PublicDomain.TzDatabase.TzZone("Pacific/Niue", new TimeSpan(-408000000000), "-", "NUT", 1951, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Niue Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Niue", new TimeSpan(-414000000000), "-", "NUT", 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Niue", new TimeSpan(-396000000000), "-", "NUT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificNorfolk_519()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Norfolk", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Norfolk", new TimeSpan(403120000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kingston"), new PublicDomain.TzDatabase.TzZone("Pacific/Norfolk", new TimeSpan(403200000000), "-", "NMT", 1951, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Norfolk Mean Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Norfolk", new TimeSpan(414000000000), "-", "NFT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Norfolk Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificNoumea_520()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Noumea", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Noumea", new TimeSpan(399480000000), "-", "LMT", 1912, PublicDomain.Month.January, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Noumea", new TimeSpan(396000000000), "NC", "NC%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("NC", 1977, 1978, PublicDomain.Month.December, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("NC", 1978, 1979, PublicDomain.Month.February, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("NC", 1996, 1996, PublicDomain.Month.December, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("NC", 1997, 1997, PublicDomain.Month.March, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificPago_Pago_521()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Pago_Pago", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(454320000000), "-", "LMT", 1879, PublicDomain.Month.July, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-409680000000), "-", "LMT", 1911, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-414000000000), "-", "SAMT", 1950, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Samoa Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-396000000000), "-", "NST", 1967, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "N=Nome"), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-396000000000), "-", "BST", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "B=Bering"), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-396000000000), "-", "SST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "S=Samoa") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificPalau_522()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Palau", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Palau", new TimeSpan(322760000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Koror"), new PublicDomain.TzDatabase.TzZone("Pacific/Palau", new TimeSpan(324000000000), "-", "PWT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Palau Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificPitcairn_523()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Pitcairn", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Pitcairn", new TimeSpan(-312200000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Adamstown"), new PublicDomain.TzDatabase.TzZone("Pacific/Pitcairn", new TimeSpan(-306000000000), "-", "PNT", 1998, PublicDomain.Month.April, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Pitcairn", new TimeSpan(-288000000000), "-", "PST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Pitcairn Standard Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificPonape_524()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Ponape", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Ponape", new TimeSpan(379720000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Kolonia"), new PublicDomain.TzDatabase.TzZone("Pacific/Ponape", new TimeSpan(396000000000), "-", "PONT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Ponape Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificPort_Moresby_525()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Port_Moresby", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Port_Moresby", new TimeSpan(353200000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Port_Moresby", new TimeSpan(353120000000), "-", "PMMT", 1895, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Port Moresby Mean Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Port_Moresby", new TimeSpan(360000000000), "-", "PGT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Papua New Guinea Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificRarotonga_526()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Rarotonga", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Rarotonga", new TimeSpan(-383440000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Avarua"), new PublicDomain.TzDatabase.TzZone("Pacific/Rarotonga", new TimeSpan(-378000000000), "-", "CKT", 1978, PublicDomain.Month.November, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Cook Is Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Rarotonga", new TimeSpan(-360000000000), "Cook", "CK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Cook", 1978, 1978, PublicDomain.Month.November, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HS", null), new PublicDomain.TzDatabase.TzRule("Cook", 1979, 1991, PublicDomain.Month.March, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Cook", 1979, 1990, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(18000000000), "HS", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificSaipan_527()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Saipan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Saipan", new TimeSpan(-514200000000), "-", "LMT", 1844, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Saipan", new TimeSpan(349800000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Saipan", new TimeSpan(324000000000), "-", "MPT", 1969, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "N Mariana Is Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Saipan", new TimeSpan(360000000000), "-", "MPT", 2000, PublicDomain.Month.December, 23, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Saipan", new TimeSpan(360000000000), "-", "ChST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Chamorro Standard Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificSamoa_528()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Samoa", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(454320000000), "-", "LMT", 1879, PublicDomain.Month.July, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-409680000000), "-", "LMT", 1911, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-414000000000), "-", "SAMT", 1950, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Samoa Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-396000000000), "-", "NST", 1967, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "N=Nome"), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-396000000000), "-", "BST", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "B=Bering"), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-396000000000), "-", "SST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "S=Samoa") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificTahiti_529()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Tahiti", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Tahiti", new TimeSpan(-358960000000), "-", "LMT", 1912, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Papeete"), new PublicDomain.TzDatabase.TzZone("Pacific/Tahiti", new TimeSpan(-360000000000), "-", "TAHT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tahiti Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificTarawa_530()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Tarawa", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Tarawa", new TimeSpan(415240000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Bairiki"), new PublicDomain.TzDatabase.TzZone("Pacific/Tarawa", new TimeSpan(432000000000), "-", "GILT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Gilbert Is Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificTongatapu_531()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Tongatapu", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Tongatapu", new TimeSpan(443600000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Tongatapu", new TimeSpan(444000000000), "-", "TOT", 1941, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Tonga Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Tongatapu", new TimeSpan(468000000000), "-", "TOT", 1999, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Tongatapu", new TimeSpan(468000000000), "Tonga", "TO%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Tonga", 1999, 1999, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tonga", 2000, 2000, PublicDomain.Month.March, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Tonga", 2000, 2001, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Tonga", 2001, 2002, PublicDomain.Month.January, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null) }), false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificTruk_532()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Truk", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Truk", new TimeSpan(364280000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Truk", new TimeSpan(360000000000), "-", "TRUT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Truk Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificWake_533()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Wake", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Wake", new TimeSpan(399880000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Wake", new TimeSpan(432000000000), "-", "WAKT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Wake Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificWallis_534()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Wallis", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Wallis", new TimeSpan(441200000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Wallis", new TimeSpan(432000000000), "-", "WFT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Wallis & Futuna Time") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPacificYap_535()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Pacific/Yap", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Truk", new TimeSpan(364280000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Truk", new TimeSpan(360000000000), "-", "TRUT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Truk Time") }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPoland_536()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Poland", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(50400000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(50400000000), "-", "WMT", 1915, PublicDomain.Month.August, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Warsaw Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1918, PublicDomain.Month.September, 16, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(72000000000), "Poland", "EE%sT", 1922, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "Poland", "CE%sT", 1940, PublicDomain.Month.June, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "C-Eur", "CE%sT", 1944, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "Poland", "CE%sT", 1977, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "W-Eur", "CE%sT", 1988, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Warsaw", new TimeSpan(36000000000), "EU", "CE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.April, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1916, 1916, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.April, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1917, 1918, PublicDomain.Month.September, 15, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1918, 1919, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1919, 1919, PublicDomain.Month.April, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1940, 1940, PublicDomain.Month.April, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1942, 1942, PublicDomain.Month.November, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.March, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1943, 1943, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1945, PublicDomain.Month.April, 1, DayOfWeek.Monday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1944, 1944, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1944, 1944, PublicDomain.Month.October, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1944, 1944, PublicDomain.Month.October, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1945, 1945, PublicDomain.Month.April, 29, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1945, 1945, PublicDomain.Month.September, 16, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1945, 1945, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1946, 1946, PublicDomain.Month.April, 14, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1946, 1946, PublicDomain.Month.October, 7, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1947, 1947, PublicDomain.Month.May, 4, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1947, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1948, 1948, PublicDomain.Month.April, 18, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1949, 1949, PublicDomain.Month.April, 10, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1957, 1957, PublicDomain.Month.June, 2, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1957, 1958, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1958, 1958, PublicDomain.Month.March, 30, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1959, 1959, PublicDomain.Month.May, 31, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1959, 1961, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Poland", 1960, 1960, PublicDomain.Month.April, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1961, 1964, PublicDomain.Month.May, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Poland", 1962, 1964, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("C-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPortugal_537()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Portugal", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", new TimeSpan(-21920000000), "-", "LMT", 1884, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", new TimeSpan(-21920000000), "-", "LMT", 1912, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Lisbon Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", TimeSpan.Zero, "Port", "WE%sT", 1966, PublicDomain.Month.April, 3, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", new TimeSpan(36000000000), "-", "CET", 1976, PublicDomain.Month.September, 26, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", TimeSpan.Zero, "Port", "WE%sT", 1983, PublicDomain.Month.September, 25, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", TimeSpan.Zero, "W-Eur", "WE%sT", 1992, PublicDomain.Month.September, 27, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", new TimeSpan(36000000000), "EU", "CE%sT", 1996, PublicDomain.Month.March, 31, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Lisbon", TimeSpan.Zero, "EU", "WE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Port", 1916, 1916, PublicDomain.Month.June, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1916, 1916, PublicDomain.Month.November, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1917, 1917, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1917, 1921, PublicDomain.Month.October, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1918, 1918, PublicDomain.Month.March, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1919, 1919, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1920, 1920, PublicDomain.Month.February, 29, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1921, 1921, PublicDomain.Month.February, 28, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1924, 1924, PublicDomain.Month.April, 16, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1924, 1924, PublicDomain.Month.October, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1926, 1926, PublicDomain.Month.April, 17, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1926, 1929, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1927, 1927, PublicDomain.Month.April, 9, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1928, 1928, PublicDomain.Month.April, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1929, 1929, PublicDomain.Month.April, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1931, 1931, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1931, 1932, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1932, 1932, PublicDomain.Month.April, 2, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1934, 1934, PublicDomain.Month.April, 7, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1934, 1938, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1935, 1935, PublicDomain.Month.March, 30, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1936, 1936, PublicDomain.Month.April, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1937, 1937, PublicDomain.Month.April, 3, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1938, 1938, PublicDomain.Month.March, 26, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1939, 1939, PublicDomain.Month.April, 15, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1939, 1939, PublicDomain.Month.November, 18, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1940, 1940, PublicDomain.Month.February, 24, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1940, 1941, PublicDomain.Month.October, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1941, 1941, PublicDomain.Month.April, 5, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1945, PublicDomain.Month.March, 8, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1942, PublicDomain.Month.April, 25, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1942, PublicDomain.Month.August, 15, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1942, 1945, PublicDomain.Month.October, 24, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1943, 1943, PublicDomain.Month.April, 17, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Port", 1943, 1945, PublicDomain.Month.August, 25, DayOfWeek.Saturday, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1944, 1945, PublicDomain.Month.April, 21, DayOfWeek.Saturday, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(72000000000), "M", null), new PublicDomain.TzDatabase.TzRule("Port", 1946, 1946, PublicDomain.Month.April, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1946, 1946, PublicDomain.Month.October, 1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1947, 1949, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1947, 1949, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1951, 1965, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1951, 1965, PublicDomain.Month.October, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1977, 1977, PublicDomain.Month.March, 27, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1977, 1977, PublicDomain.Month.September, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1978, 1979, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1978, 1978, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1979, 1982, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Port", 1980, 1980, PublicDomain.Month.March, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1981, 1982, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Port", 1983, 1983, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("W-Eur", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPRC_538()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("PRC", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Shanghai", new TimeSpan(291520000000), "-", "LMT", 1928, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Shanghai", new TimeSpan(288000000000), "Shang", "C%sT", 1949, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Shanghai", new TimeSpan(288000000000), "PRC", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Shang", 1940, 1940, PublicDomain.Month.June, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Shang", 1940, 1941, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Shang", 1941, 1941, PublicDomain.Month.March, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1986, PublicDomain.Month.May, 4, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("PRC", 1986, 1991, PublicDomain.Month.September, 11, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("PRC", 1987, 1991, PublicDomain.Month.April, 10, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesPST8PDT_539()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("PST8PDT", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("PST8PDT", new TimeSpan(-288000000000), "US", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesROC_540()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("ROC", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Taipei", new TimeSpan(291600000000), "-", "LMT", 1896, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "or Taibei or T'ai-pei"), new PublicDomain.TzDatabase.TzZone("Asia/Taipei", new TimeSpan(288000000000), "Taiwan", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Taiwan", 1945, 1951, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1945, 1951, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1952, 1952, PublicDomain.Month.March, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1952, 1954, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1953, 1959, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1955, 1961, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1960, 1961, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1974, 1975, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1974, 1975, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1980, 1980, PublicDomain.Month.June, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Taiwan", 1980, 1980, PublicDomain.Month.September, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesROK_541()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("ROK", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(304720000000), "-", "LMT", 1890, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(306000000000), "-", "KST", 1904, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(324000000000), "-", "KST", 1928, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(306000000000), "-", "KST", 1932, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(324000000000), "-", "KST", 1954, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(288000000000), "ROK", "K%sT", 1961, PublicDomain.Month.August, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(306000000000), "-", "KST", 1968, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Seoul", new TimeSpan(324000000000), "ROK", "K%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("ROK", 1960, 1960, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("ROK", 1960, 1960, PublicDomain.Month.September, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("ROK", 1987, 1988, PublicDomain.Month.May, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("ROK", 1987, 1988, PublicDomain.Month.October, 8, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesSingapore_542()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Singapore", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(249250000000), "-", "LMT", 1901, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(249250000000), "-", "SMT", 1905, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Singapore M.T."), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(252000000000), "-", "MALT", 1933, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Malaya Time"), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(252000000000), "0:20", "MALST", 1936, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(264000000000), "-", "MALT", 1941, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(270000000000), "-", "MALT", 1942, PublicDomain.Month.February, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(324000000000), "-", "JST", 1945, PublicDomain.Month.September, 12, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(270000000000), "-", "MALT", 1965, PublicDomain.Month.August, 9, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "independence"), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(270000000000), "-", "SGT", 1982, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Singapore Time"), new PublicDomain.TzDatabase.TzZone("Asia/Singapore", new TimeSpan(288000000000), "-", "SGT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("0:20", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(12000000000), null, null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesTurkey_543()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Turkey", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(69520000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(70160000000), "-", "IMT", 1910, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Istanbul Mean Time?"), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(72000000000), "Turkey", "EE%sT", 1978, PublicDomain.Month.October, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(108000000000), "Turkey", "TR%sT", 1985, PublicDomain.Month.April, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Turkey Time"), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(72000000000), "Turkey", "EE%sT", 2007, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Istanbul", new TimeSpan(72000000000), "EU", "EE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Turkey", 1916, 1916, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1916, 1916, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1920, 1920, PublicDomain.Month.March, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1920, 1920, PublicDomain.Month.October, 25, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1921, 1921, PublicDomain.Month.April, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1921, 1921, PublicDomain.Month.October, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1922, 1922, PublicDomain.Month.March, 26, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1922, 1922, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1924, 1924, PublicDomain.Month.May, 13, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1924, 1925, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1925, 1925, PublicDomain.Month.May, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1940, 1940, PublicDomain.Month.June, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1940, 1940, PublicDomain.Month.October, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1940, 1940, PublicDomain.Month.December, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1941, 1941, PublicDomain.Month.September, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1942, 1942, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1942, 1942, PublicDomain.Month.November, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1945, 1945, PublicDomain.Month.April, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1945, 1945, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1946, 1946, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1946, 1946, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1947, 1948, PublicDomain.Month.April, 16, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1947, 1950, PublicDomain.Month.October, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1949, 1949, PublicDomain.Month.April, 10, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1950, 1950, PublicDomain.Month.April, 19, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1951, 1951, PublicDomain.Month.April, 22, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1951, 1951, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1962, 1962, PublicDomain.Month.July, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1962, 1962, PublicDomain.Month.October, 8, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1964, 1964, PublicDomain.Month.May, 15, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1964, 1964, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1970, 1972, PublicDomain.Month.May, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1970, 1972, PublicDomain.Month.October, 2, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1973, 1973, PublicDomain.Month.June, 3, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1973, 1973, PublicDomain.Month.November, 4, null, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1974, 1974, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1974, 1974, PublicDomain.Month.November, 3, null, new TimeSpan(180000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1975, 1975, PublicDomain.Month.March, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1975, 1976, PublicDomain.Month.October, -1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1976, 1976, PublicDomain.Month.June, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1977, 1978, PublicDomain.Month.April, 1, DayOfWeek.Sunday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1977, 1977, PublicDomain.Month.October, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1979, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1979, 1982, PublicDomain.Month.October, 11, DayOfWeek.Monday, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1981, 1982, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(108000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1983, 1983, PublicDomain.Month.July, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1983, 1983, PublicDomain.Month.October, 2, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1985, 1985, PublicDomain.Month.April, 20, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1985, 1985, PublicDomain.Month.September, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1986, 1990, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1986, 1990, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1991, 2006, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1991, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Turkey", 1996, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUCT_544()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("UCT", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/UCT", TimeSpan.Zero, "-", "UCT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUniversal_545()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Universal", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/UTC", TimeSpan.Zero, "-", "UTC", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSAlaska_546()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Alaska", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(504240000000), "-", "LMT", 1867, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-359760000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "-", "CAT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "US", "CAT/CAWT", 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "US", "CAT/CAPT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Peace"), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "-", "CAT", 1967, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "-", "AHST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-360000000000), "US", "AH%sT", 1983, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-324000000000), "US", "Y%sT", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Anchorage", new TimeSpan(-324000000000), "US", "AK%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSAleutian_547()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Aleutian", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(440010000000), "-", "LMT", 1867, PublicDomain.Month.October, 18, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-423980000000), "-", "LMT", 1900, PublicDomain.Month.August, 20, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "-", "NST", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "US", "N%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "-", "NST", 1967, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "-", "BST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-396000000000), "US", "B%sT", 1983, PublicDomain.Month.October, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-360000000000), "US", "AH%sT", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Adak", new TimeSpan(-360000000000), "US", "HA%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSArizona_548()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Arizona", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-268980000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(415020000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "US", "M%sT", 1944, PublicDomain.Month.January, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "-", "MST", 1944, PublicDomain.Month.April, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "US", "M%sT", 1944, PublicDomain.Month.October, 1, null, new TimeSpan(600000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "-", "MST", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "US", "M%sT", 1968, PublicDomain.Month.March, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Phoenix", new TimeSpan(-252000000000), "-", "MST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSCentral_549()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Central", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-210360000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(437640000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "US", "C%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "Chicago", "C%sT", 1936, PublicDomain.Month.March, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-180000000000), "-", "EST", 1936, PublicDomain.Month.November, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "Chicago", "C%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "Chicago", "C%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Chicago", new TimeSpan(-216000000000), "US", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1920, 1920, PublicDomain.Month.June, 13, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1920, 1921, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1921, 1921, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1922, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1922, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Chicago", 1955, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSEastern_550()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Eastern", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-177620000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(434380000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-180000000000), "US", "E%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-180000000000), "NYC", "E%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-180000000000), "US", "E%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-180000000000), "NYC", "E%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/New_York", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NYC", 1920, 1920, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NYC", 1920, 1920, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NYC", 1921, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("NYC", 1921, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("NYC", 1955, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSEastIndiana_551()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/East-Indiana", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-206780000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(441220000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "US", "C%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "Indianapolis", "C%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "US", "C%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "Indianapolis", "C%sT", 1955, PublicDomain.Month.April, 24, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 1957, PublicDomain.Month.September, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-216000000000), "-", "CST", 1958, PublicDomain.Month.April, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 1969, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "US", "E%sT", 1971, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Indianapolis", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1941, 1941, PublicDomain.Month.June, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1941, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Indianapolis", 1946, 1954, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSHawaii_552()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Hawaii", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Honolulu", new TimeSpan(-378860000000), "-", "LMT", 1900, PublicDomain.Month.January, 1, null, new TimeSpan(432000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Honolulu", new TimeSpan(-378000000000), "-", "HST", 1933, PublicDomain.Month.April, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Honolulu", new TimeSpan(-378000000000), "1:00", "HDT", 1933, PublicDomain.Month.May, 21, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Honolulu", new TimeSpan(-378000000000), "US", "H%sT", 1947, PublicDomain.Month.June, 8, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Honolulu", new TimeSpan(-360000000000), "-", "HST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("1:00", -2147483648, 2147483647, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), null, null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSIndianaStarke_553()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Indiana-Starke", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-207900000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(440100000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "US", "C%sT", 1947, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "Starke", "C%sT", 1962, PublicDomain.Month.April, 29, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-180000000000), "-", "EST", 1963, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "US", "C%sT", 1991, PublicDomain.Month.October, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-180000000000), "-", "EST", 2006, PublicDomain.Month.April, 2, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Indiana/Knox", new TimeSpan(-216000000000), "US", "C%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1947, 1961, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Starke", 1947, 1954, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1955, 1956, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1957, 1958, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Starke", 1959, 1961, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSMichigan_554()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Michigan", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-199310000000), "-", "LMT", 1905, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-216000000000), "-", "CST", 1915, PublicDomain.Month.May, 15, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "-", "EST", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "US", "E%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "Detroit", "E%sT", 1973, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "US", "E%sT", 1975, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "-", "EST", 1975, PublicDomain.Month.April, 27, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Detroit", new TimeSpan(-180000000000), "US", "E%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Detroit", 1948, 1948, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Detroit", 1948, 1948, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Detroit", 1967, 1967, PublicDomain.Month.June, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Detroit", 1967, 1967, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSMountain_555()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Mountain", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-251960000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(432040000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 1920, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "Denver", "M%sT", 1942, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "Denver", "M%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Denver", new TimeSpan(-252000000000), "US", "M%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1920, 1921, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Denver", 1920, 1920, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1921, 1921, PublicDomain.Month.May, 22, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("Denver", 1965, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("Denver", 1965, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSPacific_556()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Pacific", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-283780000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(436220000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-288000000000), "US", "P%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-288000000000), "CA", "P%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-288000000000), "US", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1949, 1949, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1962, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSPacificNew_557()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Pacific-New", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-283780000000), "-", "LMT", 1883, PublicDomain.Month.November, 18, null, new TimeSpan(436220000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-288000000000), "US", "P%sT", 1946, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-288000000000), "CA", "P%sT", 1967, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("America/Los_Angeles", new TimeSpan(-288000000000), "US", "P%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1918, 1919, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1942, 1942, PublicDomain.Month.February, 9, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "W", "War"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.August, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "P", "Peace"), new PublicDomain.TzDatabase.TzRule("US", 1945, 1945, PublicDomain.Month.September, 30, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1948, 1948, PublicDomain.Month.March, 14, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1949, 1949, PublicDomain.Month.January, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1966, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("CA", 1950, 1961, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("CA", 1962, 1966, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 1973, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1967, 2006, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null), new PublicDomain.TzDatabase.TzRule("US", 1974, 1974, PublicDomain.Month.January, 6, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1975, 1975, PublicDomain.Month.February, 23, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1976, 1986, PublicDomain.Month.April, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 1987, 2006, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.March, 8, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "D", null), new PublicDomain.TzDatabase.TzRule("US", 2007, 2147483647, PublicDomain.Month.November, 1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "S", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUSSamoa_558()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("US/Samoa", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(454320000000), "-", "LMT", 1879, PublicDomain.Month.July, 5, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-409680000000), "-", "LMT", 1911, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-414000000000), "-", "SAMT", 1950, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Samoa Time"), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-396000000000), "-", "NST", 1967, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "N=Nome"), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-396000000000), "-", "BST", 1983, PublicDomain.Month.November, 30, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "B=Bering"), new PublicDomain.TzDatabase.TzZone("Pacific/Pago_Pago", new TimeSpan(-396000000000), "-", "SST", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "S=Samoa") }), null, false);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesUTC_559()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("UTC", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/UTC", TimeSpan.Zero, "-", "UTC", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesWET_560()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("WET", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("WET", TimeSpan.Zero, "EU", "WE%sT", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("EU", 1977, 1980, PublicDomain.Month.April, 1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1977, 1977, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1978, 1978, PublicDomain.Month.October, 1, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1979, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("EU", 1981, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("EU", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(36000000000), TzDatabase.TimeModifier.UniversalTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesWSU_561()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("W-SU", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(90200000000), "-", "LMT", 1880, PublicDomain.Month.January, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(90000000000), "-", "MMT", 1916, PublicDomain.Month.July, 3, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, "Moscow Mean Time"), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(90480000000), "Russia", "%s", 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1922, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(72000000000), "-", "EET", 1930, PublicDomain.Month.June, 21, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(108000000000), "Russia", "MSK/MSD", 1991, PublicDomain.Month.March, 31, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(72000000000), "Russia", "EE%sT", 1992, PublicDomain.Month.January, 19, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, null), new PublicDomain.TzDatabase.TzZone("Europe/Moscow", new TimeSpan(108000000000), "Russia", "MSK/MSD", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), new List<PublicDomain.TzDatabase.TzRule>(new PublicDomain.TzDatabase.TzRule[] { new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.July, 1, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", "Moscow Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1917, 1917, PublicDomain.Month.December, 28, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "MMT", "Moscow Mean Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.May, 31, null, new TimeSpan(792000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", "Moscow Double Summer Time"), new PublicDomain.TzDatabase.TzRule("Russia", 1918, 1918, PublicDomain.Month.September, 16, null, new TimeSpan(36000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "MST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.May, 31, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "MDST", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.July, 1, null, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1919, 1919, PublicDomain.Month.August, 16, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.February, 14, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.March, 20, null, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(72000000000), "M", "Midsummer"), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.September, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1921, 1921, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1984, PublicDomain.Month.April, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1981, 1983, PublicDomain.Month.October, 1, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1984, 1991, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1985, 1991, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.March, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1992, 1992, PublicDomain.Month.September, -1, DayOfWeek.Saturday, new TimeSpan(828000000000), TzDatabase.TimeModifier.LocalWallTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 2147483647, PublicDomain.Month.March, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, new TimeSpan(36000000000), "S", null), new PublicDomain.TzDatabase.TzRule("Russia", 1993, 1995, PublicDomain.Month.September, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null), new PublicDomain.TzDatabase.TzRule("Russia", 1996, 2147483647, PublicDomain.Month.October, -1, DayOfWeek.Sunday, new TimeSpan(72000000000), TzDatabase.TimeModifier.LocalStandardTime, TimeSpan.Zero, "-", null) }), true);
        }

        private static PublicDomain.TzTimeZone.TzZoneInfo InitializeZonesZulu_562()
        {
            return new PublicDomain.TzTimeZone.TzZoneInfo("Zulu", new List<PublicDomain.TzDatabase.TzZone>(new PublicDomain.TzDatabase.TzZone[] { new PublicDomain.TzDatabase.TzZone("Etc/UTC", TimeSpan.Zero, "-", "UTC", 2147483647, PublicDomain.Month.December, 31, null, TimeSpan.Zero, TzDatabase.TimeModifier.LocalWallTime, null) }), null, true);
        }

        #endregion

        /// <summary>
        /// Represents a time zone, with all of its transitions and rules.
        /// </summary>
        [Serializable]
        public class TzZoneInfo
        {
            private ReadOnlyCollection<PublicDomain.TzDatabase.TzRule> m_rules;

            private ReadOnlyCollection<PublicDomain.TzDatabase.TzZone> m_zones;

            private string m_zoneName;

            private bool m_isLatitudeNorth = true;

            /// <summary>
            /// Initializes a new instance of the <see cref="PublicDomain.TzDatabase.TzZone"/> class.
            /// </summary>
            /// <param name="zoneName">Name of the zone.</param>
            public TzZoneInfo(string zoneName)
                : this(zoneName, null, null, true)
            {
            }

            /// <summary>
            /// Initializes a new instance of the <see cref="TzZoneInfo"/> class.
            /// </summary>
            /// <param name="zoneName">Name of the zone.</param>
            /// <param name="zones">The zones.</param>
            /// <param name="rules">The rules.</param>
            [Obsolete]
            public TzZoneInfo(string zoneName, List<PublicDomain.TzDatabase.TzZone> zones, List<PublicDomain.TzDatabase.TzRule> rules)
                : this(zoneName, zones, rules, true)
            {
            }

            /// <summary>
            /// Initializes a new instance of the <see cref="PublicDomain.TzDatabase.TzZone"/> class.
            /// </summary>
            /// <param name="zoneName">Name of the zone.</param>
            /// <param name="zones">The zones.</param>
            /// <param name="rules">The rules.</param>
            /// <param name="isLatitudeNorth">if set to <c>true</c> [is latitude north].</param>
            public TzZoneInfo(string zoneName, List<PublicDomain.TzDatabase.TzZone> zones, List<PublicDomain.TzDatabase.TzRule> rules, bool isLatitudeNorth)
            {
                if (zones == null)
                {
                    zones = new List<TzDatabase.TzZone>();
                }
                if (rules == null)
                {
                    rules = new List<TzDatabase.TzRule>();
                }
                m_zoneName = zoneName;
                m_zones = zones.AsReadOnly();
                m_rules = rules.AsReadOnly();
                m_isLatitudeNorth = isLatitudeNorth;
            }

            /// <summary>
            /// Gets the rules.
            /// </summary>
            /// <value>The rules.</value>
            public ReadOnlyCollection<PublicDomain.TzDatabase.TzRule> Rules
            {
                get
                {
                    return m_rules;
                }
            }

            /// <summary>
            /// Gets the zones.
            /// </summary>
            /// <value>The zones.</value>
            public ReadOnlyCollection<PublicDomain.TzDatabase.TzZone> Zones
            {
                get
                {
                    return m_zones;
                }
            }

            /// <summary>
            /// Gets the name of the zone.
            /// </summary>
            /// <value>The name of the zone.</value>
            public string ZoneName
            {
                get
                {
                    return m_zoneName;
                }
            }

            /// <summary>
            /// Returns a <see cref="T:System.String"></see> that represents the current <see cref="T:System.Object"></see>.
            /// </summary>
            /// <returns>
            /// A <see cref="T:System.String"></see> that represents the current <see cref="T:System.Object"></see>.
            /// </returns>
            public override string ToString()
            {
                return ZoneName;
            }

            /// <summary>
            /// Serves as a hash function for a particular type. <see cref="M:System.Object.GetHashCode"></see> is suitable for use in hashing algorithms and data structures like a hash table.
            /// </summary>
            /// <returns>
            /// A hash code for the current <see cref="T:System.Object"></see>.
            /// </returns>
            public override int GetHashCode()
            {
                return ZoneName.GetHashCode();
            }

            /// <summary>
            /// Gets a value indicating whether this instance is latitude north.
            /// </summary>
            /// <value>
            /// 	<c>true</c> if this instance is latitude north; otherwise, <c>false</c>.
            /// </value>
            public bool IsLatitudeNorth
            {
                get
                {
                    return m_isLatitudeNorth;
                }
            }

            /// <summary>
            /// Determines whether the specified <see cref="T:System.Object"></see> is equal to the current <see cref="T:System.Object"></see>.
            /// </summary>
            /// <param name="obj">The <see cref="T:System.Object"></see> to compare with the current <see cref="T:System.Object"></see>.</param>
            /// <returns>
            /// true if the specified <see cref="T:System.Object"></see> is equal to the current <see cref="T:System.Object"></see>; otherwise, false.
            /// </returns>
            public override bool Equals(object obj)
            {
                TzZoneInfo z = obj as TzZoneInfo;
                if (z != null && z.ZoneName.Equals(ZoneName))
                {
                    return true;
                }
                return base.Equals(obj);
            }

            /// <summary>
            /// Clones the specified zone name.
            /// </summary>
            /// <param name="zoneName">Name of the zone.</param>
            /// <returns></returns>
            public TzZoneInfo Clone(string zoneName)
            {
                TzZoneInfo result = (TzZoneInfo)MemberwiseClone();
                result.m_zoneName = zoneName;
                return result;
            }
        }

        /// <summary>
        /// Logical zone description taken from the zone tab file in the tz database.
        /// </summary>
        [Serializable]
        public class TzZoneDescription
        {
            private string m_twoLetterCode;
            private Iso6709 m_location;
            private string m_zoneName;
            private string m_comments;

            /// <summary>
            /// Initializes a new instance of the <see cref="TzZoneDescription"/> class.
            /// </summary>
            public TzZoneDescription()
            {
            }

            /// <summary>
            /// Initializes a new instance of the <see cref="TzZoneDescription"/> class.
            /// </summary>
            /// <param name="twoLetterCode">The two letter code.</param>
            /// <param name="location">The location.</param>
            /// <param name="zoneName">Name of the zone.</param>
            /// <param name="comments">The comments.</param>
            public TzZoneDescription(string twoLetterCode, Iso6709 location, string zoneName, string comments)
            {
                m_twoLetterCode = twoLetterCode;
                m_location = location;
                m_zoneName = zoneName;
                m_comments = comments;
            }

            /// <summary>
            /// 
            /// </summary>
            public string TwoLetterCode
            {
                get
                {
                    return m_twoLetterCode;
                }
            }

            /// <summary>
            /// 
            /// </summary>
            public Iso6709 Location
            {
                get
                {
                    return m_location;
                }
            }

            /// <summary>
            /// 
            /// </summary>
            public string ZoneName
            {
                get
                {
                    return m_zoneName;
                }
            }

            /// <summary>
            /// 
            /// </summary>
            public string Comments
            {
                get
                {
                    return m_comments;
                }
            }

            /// <summary>
            /// Returns the fully qualified type name of this instance.
            /// </summary>
            /// <returns>
            /// A <see cref="T:System.String"></see> containing a fully qualified type name.
            /// </returns>
            public override string ToString()
            {
                return string.Format("{0}\t{1}\t{2}\t{3}", TwoLetterCode, Location, ZoneName, Comments);
            }
        }
    }
}
